// convexify.c
//          
// Chop polygons into convex polygons. 
// Input shapes must not contain holes.

// Slices are most likely made where the polygon kinks in.

// v1.2
// A new command line option allows this program to chop
// polygons into parallelograms which can be
// exposed on the GPG, or on a UPG without Firebird.
// The fussy trickiness of generating gpf trapezoids
// is left to gpfout.c. To get good results when 
// generating parallelograms, first snap to the bss then
// trapify using 1nm :
//         snap a.gds as.gds 5 fix
//         convexify as.gds ast.gds 1 header trap
// where the flag "header" means "create a gds header" 
// and the flag "trap" means "make parallelograms".


#include "libgdsmin.h"
#include "polyq.h"
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

int bss, bss2;    // beam step size, in bits, and some fraction of that
int bssm;

#define VERSION 1.2
    
#define TOOCLOSE( a ) ((fabs((a)) < bss2))

#define PI    3.1415926
#define TWOPI 6.2831853
#define PI2   1.5707963
#define PI32  4.7123890

#define printv noop
#define ifv(a) if(0)

#define MAXMARRY 25000   

void marry_vertices( queue *tout );           // not used
double pointoline_dist( xy p, xy a, xy b );
void snap_to_grid( polygon *p );


BOOL ding=FALSE;
int total=0;
char ofilename[256], *ifilename;

void noop(){}

//-------------------------------------------------------------------------------------------------------------

BOOL
dequal( double a, double b )   // are two doubles roughly equal?
    {
    if ( a == b ) return( TRUE );

    if ( a == 0.0 )
        return( fabs(b) < 1.0e-7 );
    else if ( b == 0.0 )
        return( fabs(a) < 1.0e-7 );
    else
        return( fabs( (a - b)/a ) < 1.0e-7 );
    }

//---------------------------------------------------------------------------------------------------

BOOL
samepoint( xy a, xy b )
    {
    return( (abs(a.x - b.x) < bss) && (abs(a.y - b.y) < bss) );  
    }

BOOL
closepoint( xy a, xy b )
    {
    return( (abs(a.x - b.x) <= bssm) && (abs(a.y - b.y) <= bssm) );  
    }


BOOL
exact_samepoint( xy a, xy b )
    {
    return( (a.x == b.x) && (a.y == b.y) );
    }

//------------------------------------------------------------------------------------------------------

BOOL
between( int x, int a, int b )       // Is x between a and b?
    {
    if ( x == a ) return( TRUE );
    if ( x == b ) return( TRUE );

    if ( a < b )  
	return( x >= a && x <= b );   
    else
	return( x >= b && x <= a );
    }

//------------------------------------------------------------------------------------------------------

BOOL
near( double x, double y, xy v )
    {
    // x,y is near v if it is less than one beam step away

    static double vx, vy, fbss;
    vx = v.x;
    vy = v.y;
    fbss = bss;
    return( (fabs( vx - x ) < fbss) && (fabs( vy - y ) < fbss) );
    }

//---------------------------------------------------------------------------------------------------------

double 
angle( xy a, xy b )   // The positive angle 0..2pi ccw with point a at the origin
    {
    static int quad;
    static double theta;

    if ( samepoint( a, b ) ) return( -1 ); // invalid, but this will avoid 180 heading change

    if ( a.y == b.y )
	{
	if ( b.x > a.x ) 
	    { return( 0.0 ); }
	else
	    { return( PI ); }
	}
    else if ( a.x == b.x )
	{
	if ( b.y > a.y ) 
	    { return( PI2 ); }
	else
	    { return( PI32 ); }
	}

    if ( b.y > a.y && b.x > a.x ) 
	quad = 1;
    else if ( b.y > a.y && b.x < a.x )
	quad = 2;
    else if ( b.y < a.y && b.x < a.x )
	quad = 3;
    else if ( b.y < a.y && b.x > a.x )
	quad = 4;
    else
        bail( "logic error in angle" );

    theta = atan( fabs(b.y - a.y) / fabs( b.x - a.x ) );

    if ( quad == 1 )
	return( theta );
    else if ( quad == 2 )
	return( PI - theta );
    else if ( quad == 3 )
	return( PI + theta );
    else
	return( TWOPI - theta );

    }  // angle


double
heading_change( xy a, xy b, xy c )
    {
    static double t1, t2, hc;
    t1 = angle( a, b );
    t2 = angle( b, c );
    hc = t2 - t1;
    if ( hc > PI ) 
	hc = hc - TWOPI;
    if ( hc < -PI )
	hc = hc + TWOPI;
    if ( dequal( hc, PI ) ) hc = PI;
    if ( fabs(hc) > PI )
	{
	printf( "t1=%1.2lf  t2=%1.2lf  hc=%1.2lf\n", t1, t2, hc );
	bail( "logic error in heading change" );
	}
    return( hc );
    }

//-------------------------------------------------------------------------------------------------------------

BOOL
proche( double x, int i )   // la meme chose, mais seulement un nombre
    {
    static double di, fbss;
    di = i;
    fbss = bss;
    return( fabs( x - di ) < fbss );
    }

//-------------------------------------------------------------------------------------------------------------


BOOL
entre( double x, int a, int b )       // est-ce que x entre a et b?
    {
    static double fa, fb;

    if ( proche( x, a ) ) return( TRUE );
    if ( proche( x, b ) ) return( TRUE );

    fa = a;
    fb = b;

    if ( a < b )  
        return( x >= fa && x <= fb );   
    else
        return( x >= fb && x <= fa );
    }


//-------------------------------------------------------------------------------------------------------------

void
remove_duplicate_vertices( polygon *p )
    {
    static int n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    static BOOL escape;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
	{
	vi  = vertex( p, i );
	vi1 = vertex( p, i+1 );
	escape = FALSE;
	while ( ! escape && (i < nm1) && samepoint( vi, vi1 ) )
	    {
	    escape = delete_vertex( p, i+1 );
	    n = p->len;
	    nm1 = n - 1;
	    vi  = vertex( p, i );
	    if ( i == nm1 )
		vi1 = vertex( p, 0 );
	    else 
		vi1 = vertex( p, i+1 );
	    }
	i += 1;
	}

    if ( n < 3 ) return;

    v0   = vertex( p, 0 );

    vnm1 = vertex( p, nm1 );

    if ( samepoint( v0, vnm1 ) ) p->len -= 1;

    }

//-------------remove_duplicate_vertices----------------------------------------------------------------

void
remove_flagpoles( polygon *p, BOOL severe )
    {
    static int n, nm1, i, i1, i2;
    static xy  a, b, c;
    static double tooclose;
    
    if ( severe ) 
	{
	tooclose = bss;
	}
    else
	{
        tooclose = bss/4.0;
        if ( tooclose < 1.0 ) tooclose = 1.0;
	}

    tooclose = 1;  //??????? this seems to make a difference, preventing some gaps 

    if ( ! p ) return;
    if ( ! p->active ) return;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;
    i1  = (i + 1) % n;
    i2  = (i + 2) % n;

    while ( i < nm1 )
	{
	a = vertex( p, i  );
	b = vertex( p, i1 );
        c = vertex( p, i2 );

        if ( //dequal( PI, fabs(heading_change( a, b, c )) ) ||  // redundant
            ( pointoline_dist(a,b,c) < tooclose )            || 
	    ( pointoline_dist(c,a,b) < tooclose )                )
	    {
    	    delete_precious_vertex( p, i1 );
	    n = p->len;
	    nm1 = n - 1;
	    i1 = (i + 1) % n;
	    i2 = (i + 2) % n;
	    }
	else
	    {
	    i  = (i + 1) % n;
            i1 = (i + 1) % n;
            i2 = (i + 2) % n;
	    }
	}

    if ( n < 3 ) 
	{
	p->active = FALSE;
	return;
	}
    else 
	{
        a = vertex( p, nm1 );
        b = vertex( p, 0   );
        c = vertex( p, 1   );


        if  ( //dequal( PI, fabs(heading_change( a, b, c )) ) || 
            ( pointoline_dist(a,b,c) < tooclose )             || 
	    ( pointoline_dist(c,a,b) < tooclose )                 ) 
	    {
	    delete_precious_vertex( p, 0 );
	    }

        if ( p->len < 3 ) p->active = FALSE;
	}

    }

//--- end remove_flagpoles ------------------------------------------------------------------------------

void
remove_precise_duplicates( polygon *p )
    {
    static int n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    static BOOL escape;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
        {
        vi  = vertex( p, i );
        vi1 = vertex( p, i+1 );
        escape = FALSE;
        while ( ! escape && (i < nm1) && exact_samepoint( vi, vi1 ) )
            {
            escape = delete_vertex( p, i+1 );
            n = p->len;
            nm1 = n - 1;
            vi  = vertex( p, i );
            if ( i == nm1 )
                vi1 = vertex( p, 0 );
            else 
                vi1 = vertex( p, i+1 );
            }
        i += 1;
        }

    if ( n < 3 ) return;

    v0   = vertex( p, 0 );

    vnm1 = vertex( p, nm1 );

    if ( exact_samepoint( v0, vnm1 ) ) p->len -= 1;

    }

//-------------remove_precise_duplicates----------------------------------------------------------------

double
distance( xy p, xy q )
    {
    static double dx, dy, d;
    dx = p.x - q.x;
    dy = p.y - q.y;
    d = sqrt( dx * dx + dy * dy );
    return( d );
    }


double
pointoline_dist( xy p, xy a, xy b )
    {
    static double dx, dy, d, px, py, s, u, xx, yy;

    if ( samepoint( p, a ) )
        return( 0 );
    else if ( samepoint( p, b ) )
        return( 0 );
    else if ( samepoint( a, b ) ) 
        {
        dx = p.x - a.x;
        dy = p.y - a.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }
    else 
        {
        px = b.x - a.x;
        py = b.y - a.y;
        s = px * px + py * py;
        u = ((p.x - a.x) * px + (p.y - a.y) * py) / s;
        if ( u > 1.0 )
            u = 1.0;
        else if ( u < 0.0 )
            u = 0.0;
        xx = a.x + u * px;
        yy = a.y + u * py;
        dx = xx - p.x;
        dy = yy - p.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }

    }

//---------end pointoline_dist-----------------------------------------------------------------------------------

BOOL
colinear( xy a, xy b, xy c, int *middle )
    {
    // If the three points are colinear then return TRUE. The return parameter
    // "middle" will indicate which of the three points is in the middle.

    static BOOL first, second, third, vertical, horizontal, colinear;
    static double m;
    static double abc, bac, cab;

    first  = FALSE;
    second = FALSE;
    third  = FALSE;

    if ( (c.x == a.x) || (c.y == a.y) || (b.x == a.x) || (b.y == a.y) )
        {
        vertical   = (a.x == b.x) && (b.x == c.x); 
        horizontal = (a.y == b.y) && (b.y == c.y);
        colinear   = horizontal || vertical;
        if ( horizontal )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second ) first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        if ( vertical )
            {
            second = between( b.y, a.y, c.y );
            if ( ! second ) first = between( a.y, b.y, c.y );
            third = (! second) && (! first);
            }
        }
    else 
        {
        abc = pointoline_dist( a, b, c );
        bac = pointoline_dist( b, a, c );
        cab = pointoline_dist( c, a, b );

        colinear = TOOCLOSE( abc ) || TOOCLOSE( bac ) || TOOCLOSE( cab ); 

        if ( colinear )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second ) first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        }
         
    // It's very difficult to decide which point to delete. Usually, we should delete
    // the middle point. However, if two of the points are very close then we should
    // delete the endpoint, maybe.

    //                 It's more reliable to delete the middle point always,
    //                 but it leads to more nasty gaps.
    //*middle = 1;
    //return( colinear );

    if ( colinear )
        {
        if ( samepoint( a, b ) )
            { 
            *middle = 0;              // delete the first vertex
            //printf( "\ndelete a\n" );
            }
        else if ( samepoint( b, c ) )
            {            
            *middle = 2;              // delete the third vertex
            //printf( "\ndelete c\n" );
            }
        else        
            {
            if ( first ) 
                *middle = 0;
            else if ( second ) 
                *middle = 1;
            else if ( third )
                *middle = 2;
            }
        }
    

    return colinear;
    }

//---------end colinear------------------------------------------------------------------------------------------

void
remove_colinear_vertices( polygon *p, BOOL severe )
    {
    // Much harder than it might seem. Normally we would want to delete 
    // the vertex in the middle, but if the polygon doubles back on itself
    // then this can cause the polygon to become distorted. Therefore
    // we start by removing flag poles before looking at colinearity.
    

    static int i, n, nm1, nm2, j, k, middle;
    static BOOL escape;
    static xy vi, vj, vk;
    
    if ( ! p->active ) return;
    
    remove_flagpoles( p, severe );
    
    if ( ! severe ) return;

    i = 0;
    middle = -1;
    n = p->len;
    nm2 = n - 2;
    nm1 = n - 1;

    while ( i < nm2 )
        {
        j = i + 1;
        k = j + 1;
        escape = FALSE;

        while ( ! escape && (i < nm2) && colinear( vertex(p,i), vertex(p,j), vertex(p,k), &middle ) )
            {
            if ( middle == 1 )
                {
                escape = delete_vertex( p, j );   // if the vertex is precious then we should escape
                }
            else
                escape = TRUE;

            n = p->len;
            nm1 = n - 1;
            nm2 = n - 2;
            }    
        i = i + 1;
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;

    if ( n > 2 )
        {
        if ( colinear( vertex(p,nm2), vertex(p,nm1), vertex(p,0), &middle ) )
            {
            if ( middle == 1 ) delete_vertex( p, nm1 );
            }
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;
        
    if ( n > 2 )
        {
        if ( colinear( vertex(p,n-1), vertex(p,0), vertex(p,1), &middle ) )
            { 
            if ( middle == 1 ) delete_vertex( p, 0 );
            }
        }
    }

//-----------end remove_colinear_vertices------------------------------------------------------------------------

void
extent( polygon *p )
    {
    static xy v;
    static int n, i, minx, maxx, miny, maxy, b2;
    static BOOL init=FALSE;

    if ( ! init ) { init=TRUE; b2 = bss; }

    n = p->len;

    if ( n < 3 ) 
        {
        p->active = FALSE;
        p->ll.x = 0;
        p->ll.y = 0;
        p->ur.x = 0;
        p->ur.y = 0;
        return;
        }

    v = vertex( p, 0 );
    minx = v.x; 
    maxx = minx;
    miny = v.y;
    maxy = miny;

    for ( i=1; i < n; i++ )
        {
        v = vertex( p, i );
        if ( v.x < minx ) minx = v.x;
        if ( v.x > maxx ) maxx = v.x;
        if ( v.y < miny ) miny = v.y;
        if ( v.y > maxy ) maxy = v.y;
        }

    p->ll.x = minx - b2;
    p->ll.y = miny - b2;
    p->ur.x = maxx + b2;
    p->ur.y = maxy + b2;
    }

//----------end extent-------------------------------------------------------------------------------------------

polygon *
copy_from_gds_polygon( struct gds_itemtype *item )
    {
    // create a poly and then copy the item to it

    static int i;
    polygon *poly;

    poly = initialize_polygon( item->n, item->layer, item->dt, FALSE );

    for ( i=0; i < item->n; i++ )
        {
        append_polygon( poly, item->x[i], item->y[i] );
        }

    snap_to_grid( poly );
    extent( poly );
    return( poly );

    }  // copy_from_gds_polygon

//---------------------------------------------------------------------------------------------------------------

BOOL
on_segment( double x, double y, xy a1, xy a2, xy b1, xy b2 )
    {

    // Is the point x,y somewhere on both line segments a1-a2 and b1-b2 ?

    static BOOL cross;

    // If a1-a2 is the same as b1-b2 then they definitely cross, and the intersection
    // is definitely on the segment. If overlaps are bad, then we should think about
    // that elsewhere.

    if ( (samepoint(a1, b1) && samepoint(a2, b2)) || (samepoint(a1, b2) && samepoint(a2, b1)) ) return( TRUE );


    if ( a1.y == a2.y ) 
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else if ( b1.y == b2.y )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( a1.x == a2.x )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( b1.x == b2.x )
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else
        cross = entre( x, b1.x, b2.x ) && entre( y, b1.y, b2.y ) &&
                entre( x, a1.x, a2.x ) && entre( y, a1.y, a2.y );

    return( cross );

    }  // on_segment

//--------------------------------------------------------------------------------------------

BOOL
in_segment( double x, double y, xy a1, xy a2, xy b1, xy b2 )
    {

    // Is the point x,y somewhere on both line segments a1-a2 and b1-b2 ...
    // BUT not on the end points?

    static BOOL cross;
    static xy   v;

    if ( samepoint(a1, b1) || samepoint(a2, b2) || samepoint(a1, b2) || samepoint(a2, b1) ) return( FALSE );

    v.x = (int) round( x );
    v.y = (int) round( y );

    if ( samepoint(v, a1) || samepoint(v, a2) || samepoint(v, b1) || samepoint(v, b2) ) return( FALSE ); 

    if ( a1.y == a2.y ) 
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else if ( b1.y == b2.y )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( a1.x == a2.x )
        cross = entre( x, b1.x, b2.x ) && entre( y, a1.y, a2.y );
    else if ( b1.x == b2.x )
        cross = entre( x, a1.x, a2.x ) && entre( y, b1.y, b2.y );
    else
        cross = entre( x, b1.x, b2.x ) && entre( y, b1.y, b2.y ) &&
                entre( x, a1.x, a2.x ) && entre( y, a1.y, a2.y );

    return( cross );

    }  // in_segment

//--------------------------------------------------------------------------------------------

int
snap( int x )
    {
    if ( ! (x % bss) ) return( x );
    x = x / bss;
    x = x * bss;
    return( x );
    }

void
snapoint( xy *v ) 
    {
    v->x = snap( v->x );
    v->y = snap( v->y );
    }

void
snap_to_grid( polygon *p )
    {
    static int i;
    static xy v;
    if ( ! p || ! p->active ) return;
    for ( i=0; i < p->len; i++ )
	{
        v = vertex( p, i );
        snapoint( &v );
        set_vertex( p, i, v );
	}
    }

//--------------------------------------------------------------------------------------------

BOOL
not_crossing( polygon *p, int a, int b, int c, int d, xy *intersection, BOOL overlapsbad )
    {
    // Does the segment a-b cross the segment c-d?
    // Return True if they do not cross, and return the 
    // intersection point if they do. 
    
    // If the line a-b is the same as the line c-d then
    // these line segments are still "crossing", because
    // doughnut polygons must be eliminated. The program
    // darn.c should have eliminated these sort of holes, 
    // but we must make sure here.

    // If we are making a slice, then we should NOT consider
    // coincident lines to be crossing. It's ok to slice
    // along one of the sides. Hence the logical overlapsbad.

    static xy
        inter,
        a1, a2, b1, b2;

    static BOOL
        crossing,
        a_is_vertical,
        b_is_vertical,
        ma_equals_mb,
        ba_equals_bb;

    static double
        fb1x, fb1y,
        fb2x, fb2y,
        fa1x, fa1y,
        fa2x, fa2y,
        fx, fy,
        ma, mb, 
        ba, bb;

    static int
        ix, iy;

    a1 = vertex( p, a );
    a2 = vertex( p, b );
    b1 = vertex( p, c );
    b2 = vertex( p, d );

    intersection->x = 0;
    intersection->y = 0;

    // first check that the segments are anywhere near each other

    if ( a1.x < b1.x && a1.x < b2.x && a2.x < b1.x && a2.x < b2.x ) return( TRUE );
    if ( a1.x > b1.x && a1.x > b2.x && a2.x > b1.x && a2.x > b2.x ) return( TRUE );
    if ( a1.y < b1.y && a1.y < b2.y && a2.y < b1.y && a2.y < b2.y ) return( TRUE );
    if ( a1.y > b1.y && a1.y > b2.y && a2.y > b1.y && a2.y > b2.y ) return( TRUE );

    // overlaps are bad when we are looking for self-intersections or doughnut shapes.
    // overlaps are not bad when we are slicing.

    if ( ( samepoint( a1, b1 ) && samepoint( a2, b2 ) ) || ( samepoint( a1, b2 ) && samepoint( a2, b1 ) ) )
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        if ( overlapsbad )
            {
            //printf( "matching segment - probably a doughnut\n" );
            return( FALSE ); // they are crossing, probably a doughnut shape
            }
        else
            return( TRUE );  // they are the same line segment, but this is ok
        }


    if ( samepoint( a1, b1 ) || samepoint( a1, b2 ) )
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        if ( overlapsbad )
            {
            //printf( "lines touch at an end point\n" ); 
            return( FALSE ); // => crossing
            }
        else
            return( TRUE );
        }

    if ( samepoint( a2, b1 ) || samepoint( a2, b2 ) )  
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        if ( overlapsbad )
            {
            //printf( "lines touch at an end point\n" ); 
            return( FALSE );
            }
        else
            return( TRUE );
        }

    a_is_vertical = (a1.x == a2.x); 
    b_is_vertical = (b1.x == b2.x); 

    if ( a_is_vertical && b_is_vertical && ! overlapsbad ) return( TRUE );

    if ( a_is_vertical && b_is_vertical )  // and overlaps are bad...
        {
        //printf( "both line segments are vertical\n" ); 
        if ( a1.x == b1.x )
            {
            if ( between( a1.y, b1.y, b2.y ) )
                {
                //printf( "vertical segments overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else if ( between( a2.y, b1.y, b2.y ) )
                {
                //printf( "vertical segments overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else
                {
                return( TRUE );
                }
            }
        else
            {
            return( TRUE );
            }
        }

    if ( a_is_vertical && ! b_is_vertical )
        {
        //printf( "a is vertical but b is not\n" );

        fb1y = b1.y;
        fb2y = b2.y;
        fb1x = b1.x;
        fb2x = b2.x;

        mb = (fb1y - fb2y) / (fb1x - fb2x);
        bb = fb1y - mb * fb1x;

        fx = a1.x;
        fy = mb * fx + bb;

        // fx,fy are floats which have not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        crossing = on_segment( fx, fy, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            //printf( "segments a and b cross. a is vertical, b is not\n" );
            intersection->x = a1.x; //snap( a1.x );
            iy = round( fy );
            intersection->y = iy; //snap( iy );
            return( FALSE );
            }
        }
        

    if ( b_is_vertical && ! a_is_vertical )
        {
        //printf( "b is vertical but a is not\n" );

        fa1y = a1.y;
        fa2y = a2.y;
        fa1x = a1.x;
        fa2x = a2.x;

        ma = (fa1y - fa2y) / (fa1x - fa2x);
        ba = fa1y - ma * fa1x;

        fx = b1.x;
        fy = ma * fx + ba;

        // fx,fy are floats which have not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        crossing = on_segment( fx, fy, a1, a2, b1, b2 );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            //printf( "segments a and b cross. b is vertical, a is not\n" );
            intersection->x = b1.x; //snap( b1.x );
            iy = round( fy );
            intersection->y = iy; //snap( iy );
            return( FALSE );
            }
        }

    // neither segment is vertical
        
    fa1x = a1.x;  fa1y = a1.y;
    fa2x = a2.x;  fa2y = a2.y;
    fb1x = b1.x;  fb1y = b1.y;
    fb2x = b2.x;  fb2y = b2.y;

    ma = (fa1y - fa2y) / (fa1x - fa2x);
    mb = (fb1y - fb2y) / (fb1x - fb2x);

    bb = b1.y - mb * b1.x;
    ba = a1.y - ma * a1.x;

    ma_equals_mb = dequal( ma, mb );
    ba_equals_bb = dequal( ba, bb );

    if ( ma_equals_mb && ! ba_equals_bb )
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );      // they are not crossing
        }

    if ( ma_equals_mb )  // Coincident lines, but do the segments overlap?
        {                // We already considered the case where the segments
                         // are the same, so this is not a doughnut cut.
                         // Also, it is not a slice cut, since slice lines
                         // are always on-axis.

        //printf( "coincident lines\n" );

        if ( ma > 1.0 )
            {
            if ( between( a2.y, b1.y, b2.y ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.y, b1.y, b2.y ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            } 
        else
            {
            if ( between( a2.x, b1.x, b2.x ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.x, b1.x, b2.x ) )
                {
                //printf( "off-axis segments that overlap\n" );
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            }
        }
            
    // Finally, just regular lines not on-axis, not parallel

    fx = (bb - ba) / (ma - mb); 
    fy = ma * fx + ba; 

    crossing = on_segment( fx, fy, a1, a2, b1, b2 ); 

    if ( crossing )
        {
        //printf( "off axis segments that cross\n" );
        //printf( "a = %d,%d  %d,%d\n", a1.x, a1.y, a2.x, a2.y );
        //printf( "b = %d,%d  %d,%d\n", b1.x, b1.y, b2.x, b2.y );
        //printf( "intersection = %1.1lf, %1.1lf \n", fx, fy );
        ix = round( fx );
        intersection->x = ix; //snap( ix );
        iy = round( fy );
        intersection->y = iy; //snap( iy );
        return( FALSE );
        }
    else
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );
        }

    }

//---------end not_crossing -------------------------------------------------------------------------------------

BOOL
clockwise( polygon *p )
    {
    // Determine the sign of a polygon's curl by calculating the net change in heading.
    // Seems like a stupid way of calculating the curl, but it's actually a lot more reliable
    // than doing the cross-product. Why?

    static xy a, b, c;
    static int n, i;
    static double dt, hc;

    if ( ! p ) bail( "null polygon sent to clockwise" );
    if ( ! p->active ) bail( "inactive polygon sent to clockwise" );
    if ( p->len < 3 ) bail( "degenerate polygon sent to clockwise" );

    n = p->len;
    dt = 0.0;

    for ( i=0; i < n; i++ )
	{
	a = vertex( p, i );
	b = vertex( p, (i+1) % n );
	c = vertex( p, (i+2) % n );
        hc = heading_change( a, b, c );
        dt = dt + hc;
	}

    return( dt <= 0.0 );
    }


//-------clockwise-----------------------------------------------------------------------------------------------


 
double
circumference( polygon *p )
    {                                   
    static int    n;
    static double c, dx, dy;
    static int    i, j;
    static xy     vi, vj;

    n = p->len;
    c = 0.0;

    i = 0;
    while ( i < n )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
        dx = vi.x - vj.x;
        dy = vi.y - vj.y;
        c = c + sqrt( dx * dx + dy * dy );
        i++;
        }
    
    return( c );
    }
 

//-----ccw et al---------------------------------------------------------------------------------------

polygon_pair *
split_touchy_polygon( polygon *p, int v, int a, int b )
    {
    // Vertex v touches line segment p[a]--p[b]. Split p into two polygons. Let's be careful
    // to retain the original order so that both resulting polygons have the same
    // clock-wiseiness. This is quite tricky.

    static int     n, i;

    static double  change_in_area, area, 
                   area1, area2,
                   toosmall, toomuchange;

    static BOOL    p1_degenerate, p2_degenerate,
                   ccw_p,  
                   ccw_p1, 
                   ccw_p2, 
                   doughnut, init=FALSE;

    polygon *p1, *p2;
    polygon_pair *pp;

    xy pv;

    pp = initialize_polygon_pair();

    printv( "split touchy\n" );

    if ( ! init ) 
        { 
        init=TRUE; 
        toosmall = bss * bss * 0.8; 
        toomuchange = bss * bss * 4.0;
        }

    doughnut = FALSE;

    area = polygon_area( p, bss );

    if ( area < toosmall ) 
        {
	//printv( "the area %1.1lf is < %1.1lf\n", area, toosmall );
        pp->a = NULL;
        pp->b = NULL;
        p->active = FALSE;
        return( pp );
        }

    p1 = initialize_polygon( p->len, p->layer, p->dt, p->obese );
    p2 = initialize_polygon( p->len, p->layer, p->dt, p->obese );

    n = p->len; 
    i = v;

    while ( (i != b) && (i != a) )
        {
        append_polygon_xy( p1, vertex( p, i ) );
        i = (i + 1) % n;
        }
    append_polygon_xy( p1, vertex( p, i ) );

    i = v;
    while ( (i != b) && (i != a) )
        {
        append_polygon_xy( p2, vertex( p, i ) );
        i -= 1;
        if ( i == -1 ) i = n-1;
        }

    append_polygon_xy( p2, vertex( p, i ) ); 
    
    extent( p1 );
    extent( p2 );

    p1_degenerate = (p1->len < 3) || (polygon_area(p1, bss) < toosmall);
    p2_degenerate = (p2->len < 3) || (polygon_area(p2, bss) < toosmall);

    if ( p1_degenerate && p2_degenerate )
        {
        //printv( "p1 and p2 are both degenerate\n" );
        pp->a = NULL;
        pp->b = NULL;
        p->active = FALSE;
        return( pp );
        }
    else if ( p1_degenerate )
        {
        // sometimes causes a fill-in, so we reject them both
        pp->a = NULL;
        pp->b = NULL; //p2;
        return( pp );
        }
    else if ( p2_degenerate )
        {
        // sometimes causes a fill-in, so we reject them both
        pp->a = NULL; //p1;
        pp->b = NULL;
        return( pp );
        }

    area1 = polygon_area( p1, bss );
    area2 = polygon_area( p2, bss );

    ccw_p1 = TRUE;
    ccw_p2 = FALSE;
    if ( p1->active && p2->active )
	{
        ccw_p1 = ! clockwise( p1 );
        ccw_p2 = ! clockwise( p2 );
	}

    if ( area1 < toosmall && area2 < toosmall )
        {
        pp->a = NULL;
        pp->b = NULL;
        p->active = FALSE;
        change_in_area = 0.0;
        }
    else if ( area1 < toosmall ) 
        {
        pp->a = NULL;
        pp->b = p2;
        change_in_area = fabs(area - area1 - area2);  // is this right? or should it be zero?
        }
    else if ( area2 < toosmall )
        {
        pp->a = p1;
        pp->b = NULL;
        change_in_area = fabs(area - area1 - area2); // maybe it should be zero?;
        }
    else
        {
        pp->a = p1;
        pp->b = p2;
        change_in_area = fabs(area - area1 - area2);
        }

    if ( change_in_area > toomuchange )
        {
        printv( "large change in area - split rejected\n" );
	pp->a = NULL;
        pp->b = NULL;
        return( pp );
	}

    doughnut = (ccw_p1 == ccw_p2);

    ifv( doughnut ) printf( "p1 and p2 have the same curl ... ignored\n" );

    return( pp );

    }

//-------split_touchy_polygon ------------------------------------------------------------------------------------


polygon_pair *
untie_bow( polygon *p, xy ip, int i1, int i2, int j1, int j2 )
    {
    polygon_pair *pp;
    polygon *p1, *p2;
    static int n, i, t;
    
    pp = initialize_polygon_pair();

    p1 = initialize_polygon( p->len, p->layer, p->dt, p->obese );
    p2 = initialize_polygon( p->len, p->layer, p->dt, p->obese );

    n = p->len;

    if ( i2 < i1 )
        {
        t  = i1;
        i1 = i2;
        i2 = t;
        }

    if ( i2 == (n-1) )
        {
        t  = i1;
        i1 = i2;
        i2 = t;
        }

    if ( j2 < j1 )
        {
        t  = j1;
        j1 = j2;
        j2 = t;
        }

    if ( j2 == (n-1) )
        {
        t  = j1;
        j1 = j2;
        j2 = t;
        }

    i = i2;
    while ( (i != j1) && (i != j2) ) 
        {
        append_polygon_xy( p1, vertex( p, i ) );
        i = (i + 1) % n;
        }
    append_polygon_xy( p1, vertex( p, i ) );
    append_polygon_xy( p1, ip );

    i = j2;
    while ( (i != i1) && (i != i2) )
        {
        append_polygon_xy( p2, vertex( p, i ) );
        i = (i + 1) % n;
        }
    append_polygon_xy( p2, vertex( p, i ) );
    append_polygon_xy( p2, ip ); 

    //p1->layer = 7;
    //p2->layer = 7;

    extent( p1 );
    extent( p2 );
    pp->a = p1;
    pp->b = p2;
    return( pp );
    }

//------ untie_bow ------------------------------------------------------------------------------------------


BOOL
find_and_fix_self_crossing_polygon( polygon *p, queue *todo_list )
    {
    // If the polygon is self crossing, this function chops it into two polygons which
    // are pushed onto the todo_list queue. If one or both of those new polygons are
    // self-crossing, then they will be further chopped as the queue is processed.

    static BOOL   good, init=FALSE; 
    static int    n, i, i1, j, j1;
    static xy     ip, vi, vj, vi1, vj1;
    polygon_pair *pp;
    static double area, toosmall;


    if ( ! init ) {init=TRUE; toosmall=bss*bss; }
    if ( ! p->active ) return( TRUE );

    if ( polygon_area( p, bss ) < toosmall )
        {
	//printf( "polygon is too small, find/fix deletes it\n" );
        p->active = FALSE;
        return( TRUE );    
        }

    good = TRUE;
    n = p->len;
    if ( n < 4 ) return( FALSE );

    i = 0;
    while ( good && (i < (n-2)) )   // notice that the polygon will be bad only once,
        {                           // at which point it will be split and the pieces 
        j = i + 2;                  // appended to the todo_list.
        while ( good && (j < n))
            {
            j1 = (j+1) % n;
            i1 = i + 1;

            if ( j1 != i )
                good = not_crossing( p, i, i1, j, j1, &ip, TRUE );  // overlapsbad=true since doughnuts are bad

            //if ( ! good ) printf( "n: %d  i: %d   j: %d   ip: %d,%d \n", n, i, j, ip.x, ip.y );

            if ( ! good )  
                {
                vi  = vertex( p, i );
                vi1 = vertex( p, i1 );
                vj  = vertex( p, j );
                vj1 = vertex( p, j1 );

                if ( samepoint( ip, vi ) )
                    {
                    pp = split_touchy_polygon( p, i, j, j1 );
                    }
                else if ( samepoint( ip, vi1 ) )
                    { 
                    pp = split_touchy_polygon( p, i1, j, j1 );
                    }
                else if ( samepoint( ip, vj ) )
                    {
                    pp = split_touchy_polygon( p, j, i, i1 );
                    }
                else if ( samepoint( ip, vj1 ) ) 
                    {
                    pp = split_touchy_polygon( p, j1, i, i1 );
                    }
                else
                    {
                    pp = untie_bow( p, ip, i, i1, j, j1 );
                    }

                if ( ! pp->a && ! pp->b && p->active )  // maybe not a failure! maybe just tiny shapes
                    {                                   // maybe it really was not a bowtie
                    //print_polygon( p, 1000 );
                    //printf( "ip: %d, %d \n", ip.x, ip.y );
                    }
                else            // success
                    {
                    p->active = FALSE;
                    remove_flagpoles( pp->a, FALSE );
                    remove_flagpoles( pp->b, FALSE );
                    if ( pp->a ) append_queue( todo_list, pp->a );
                    if ( pp->b ) append_queue( todo_list, pp->b );
                    }
                }

            j += 1;
            }
        i += 1;
        }

    return( ! good );
    }

//-----------end find_and_fix_self_crossing_polygon ----------------------------------------------------------

BOOL
self_crossing_polygon( polygon *p )
    {
    static BOOL   good, init=FALSE;
    static int    n, i, i1, j, j1;
    static xy     ip;

    good = TRUE;
    n = p->len;
    if ( n < 4 ) return( FALSE );

    i = 0;
    while ( good && (i < (n-2)) ) 
        {                         
        j = i + 2;                
        while ( good && (j < n))
            {
            j1 = (j+1) % n;
            i1 = i + 1;

            if ( j1 != i )
                good = not_crossing( p, i, i1, j, j1, &ip, FALSE );  // overlapsbad=false

            j += 1;
            }
        i += 1;
        }

    return( ! good );
    }

//-----------end self-crossing polygon------------------------------------------------------------------------

void
copy_to_gds_polygon( polygon *poly, struct gds_itemtype *item )
    {
    // copy poly to item 
    // item has already been allocated 
    // note that we are not snapping here

    static int i;
    static int ix, iy;
    static xy  v;
    static long li;

    if ( item->n < poly->len ) bail( "attempt to copy large poly into small gds item" );

    for ( i=0; i < poly->len; i++ )
        {
        li = i;
        v = vertex( poly, li );
        item->x[i] = v.x;
        item->y[i] = v.y;
        }

    item->layer = poly->layer;
    item->dt    = poly->dt;
    item->n     = poly->len;
    
    }  // copy_to_gds_polygon

//-----------------------------------------------------------------------------------------------------


void write_trap_list( int fdo, queue *list, BOOL traps ) 
    {
   
    static polygon     *p;
    static polygon_ref *pr;
    static int          ilayer, idt, ilen;
    static BOOL         init=FALSE;
    static int          maxlen=0;
    static struct gds_itemtype *gdspoly;

    if ( ! init )
        {
        gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        init = TRUE;
        }
    
    pr = list->first;    
    
    while ( pr )
        {
        p = pr->p;
        if ( p->active && p->len != 0 )
            {
            if ( traps && p->len > 4 )
                {
                print_polygon( p, 1 );
                printf( "input  file: %s\n", ifilename );
                printf( "output file: %s\n", ofilename );
                bail( "trapezoid has more than 4 vertices" );
                }
            
            ilen = p->len;
            if ( ilen > maxlen )
                {
                gdspoly->x = (int *) malloc( ilen * sizeof( int ) ); 
                gdspoly->y = (int *) malloc( ilen * sizeof( int ) );
                maxlen = ilen;
                }
            gdspoly->n = ilen; 

            copy_to_gds_polygon( p, gdspoly );

            //if ( p->len > 4 ) p->layer = ERRORLAYER; 

            gds_write_boundary( fdo );
            ilayer = p->layer;
            idt    = p->dt;
            gds_write_layer(    fdo, ilayer );
            gds_write_datatype( fdo, idt    );
            gds_write_xy( fdo, gdspoly->x, gdspoly->y, gdspoly->n );
            gds_write_endel( fdo );      
            total++;

            // printf( "." ); fflush( stdout );
            }
        pr = pr->next;
        }

    // note that this function does not close the structure 

    } // write_trap_list

//---------------------------------------------------------------------------------------------------------------


void 
write_polygon( int fdo, polygon *p, BOOL traps ) 
    {
   
    static int    ilayer, idt, ilen;
    static BOOL   init=FALSE;
    static int    maxlen=0;
    static struct gds_itemtype *gdspoly;

    if ( ! init )
        {
        gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        init = TRUE;
        }
    
    if ( ! p ) return;

    if ( ! p->active ) return;
    
    if ( traps && p->len > 4 ) bail( "trapezoid has more than 4 vertices" );

    if ( p->len != 0 )
        {
        ilen = p->len;
        if ( ilen > maxlen )
            {
            gdspoly->x = (int *) malloc( ilen * sizeof( int ) ); 
            gdspoly->y = (int *) malloc( ilen * sizeof( int ) );
            maxlen = ilen;
            }
        gdspoly->n = ilen; 

        copy_to_gds_polygon( p, gdspoly );

        //if ( p->len > 4 ) p->layer = ERRORLAYER; 

        gds_write_boundary( fdo );
        ilayer = p->layer;
        idt    = p->dt;
        gds_write_layer(    fdo, ilayer );
        gds_write_datatype( fdo, idt    );
        gds_write_xy( fdo, gdspoly->x, gdspoly->y, gdspoly->n );
        gds_write_endel( fdo );      

        }

    // note that this function does not close the structure 

    } // write_polygon

//---------------------------------------------------------------------------------------------------------------

BOOL 
is_parallelogram_already( polygon *p )
    {

    // Are two sides parallel and on-axis?
    // Or maybe it's a triangle with one side on-axis.
    // Off-axis parallelograms - nope.
    
    static int  n;
    static BOOL yup, onaxis, horizontal, vertical,
                hor01, hor12, hor23, hor30,
                ver01, ver12, ver23, ver30; 
    static xy   v0, v1, v2, v3;

    n = p->len;

    if ( (n != 3) && (n != 4) ) return( FALSE );

    yup = FALSE;
    v0 = vertex( p, 0 );
    v1 = vertex( p, 1 );
    v2 = vertex( p, 2 );

    if ( n == 3 )
        {  
        onaxis = v0.y == v1.y;
        if ( ! onaxis ) onaxis = (v1.y == v2.y);
        if ( ! onaxis ) onaxis = (v2.y == v0.y); 
        if ( ! onaxis ) onaxis = (v0.x == v1.x);
        if ( ! onaxis ) onaxis = (v1.x == v2.x); 
        if ( ! onaxis ) onaxis = (v2.x == v0.x); 
        yup = onaxis;
        }
    else
        {
        v3 = vertex( p, 3 );
        hor01 = (v0.y == v1.y); 
        hor12 = (v1.y == v2.y); 
        hor23 = (v2.y == v3.y); 
        hor30 = (v3.y == v0.y); 
        horizontal = (hor01 && hor23) || (hor12 && hor30);
        vertical = FALSE;
        if ( ! horizontal )
            {
            ver01 = (v0.x == v1.x); 
            ver12 = (v1.x == v2.x); 
            ver23 = (v2.x == v3.x); 
            ver30 = (v3.x == v0.x); 
            vertical = (ver01 && ver23) || (ver12 && ver30);
            }
        yup = horizontal || vertical;
        }
    return yup;
    }

//-----------end is_parallelogram_already---------------------------------------------------------------------

BOOL
crossing_point( xy a1, xy a2, xy b1, xy b2, xy *intersection )
    {
    // Does the line a1-a2 cross the line b1-b2 ? Return True if they cross, and return the 
    // intersection point if they do. This is the same logic as not_crossing, but
    // we have to repackage the vertices first.

    static BOOL crossing;
    static BOOL init=FALSE;
    static polygon *p;

    if ( ! init ) { init=TRUE; p = initialize_polygon( 4, 0, 0, FALSE ); }

    zero_polygon( p );
    append_polygon_xy( p, a1 );
    append_polygon_xy( p, a2 );
    append_polygon_xy( p, b1 );
    append_polygon_xy( p, b2 );

    crossing = ! not_crossing( p, 0, 1, 2, 3, intersection, TRUE );

    return( crossing );
    }

//----------end crossing_point-----------------------------------------------------------------------------------------

BOOL
convex( polygon *p, int *turn, BOOL traps )
    {
    // to produce parallelograms instead of convex shapes, simply use traps=TRUE

    if ( traps )
        {
        *turn = 0;
        return( is_parallelogram_already( p ) );
        }

    // Is this polygon convex? Look at the heading change as we go around the
    // polygon. If the sign changes, then it is not convex.
    // If not convex, *turn will be returned with the index of a
    // kink, which is perhaps a good place to cut the polygon. 
    // In this version, we find the kink which is next to the longest
    // polygon segment, so we can avoid bologna slices.

    static xy a, b, c;
    static int n, i, s, ns, besti;
    static double hc, dx, dy, d, nd;
    static BOOL first, isconvex;

    if ( ! p ) bail( "null polygon sent to convex" );
    if ( ! p->active ) return( TRUE );
    if ( p->len < 3 ) bail( "degenerate polygon sent to convex" );

    isconvex = TRUE;
    *turn = 0;
    n = p->len;
    first = TRUE;

    for ( i=0; i < n; i++ )
	{
	a = vertex( p, i );
	b = vertex( p, (i+1) % n );
	c = vertex( p, (i+2) % n );
        hc = heading_change( a, b, c );
	if ( first )
	    {
	    first = FALSE;
	    s = hc >= 0.0 ? 1 : -1;
            dx = c.x - b.x;
	    dy = c.y - b.y;
	    d  = dx * dx + dy * dy;
	    besti = (i+1) % n;
	    }
	else
	    {
	    ns = hc >= 0.0 ? 1 : -1;
	    if ( ns != s ) 
		{
		isconvex = FALSE;
                dx = c.x - b.x;
                dy = c.y - b.y;
	        nd  = dx * dx + dy * dy;
                if ( nd > d )
		    {
	            besti = (i+1) % n;
		    d = nd;
		    }
		}
	    }
	}

    *turn = besti;
    return( isconvex );

    }

//----------end convex--------------------------------------------------------------------------------------------------

int
get_midpoint( polygon *p )
    {                                // which vertex is closest to the center?
    static int i, j;
    static double dmin, d, dx, dy;
    static dxy c, dv;
    static xy v;

    c.x = (p->ur.x + p->ll.x) / 2.0;
    c.y = (p->ur.y + p->ll.y) / 2.0;
    v = vertex( p, 0 );
    dv.x = v.x;
    dv.y = v.y;
    dx = dv.x - c.x;
    dy = dv.y - c.y;
    dmin = dx * dx + dy * dy;
    j = 0;
    for ( i=1; i < p->len; i++ )
	{
        v = vertex( p, i );
        dv.x = v.x;
        dv.y = v.y;
        dx = dv.x - c.x;
        dy = dv.y - c.y;
        d = dx * dx + dy * dy;
	if ( d < dmin )
	    {
	    dmin = d;
	    j = i;
	    }
	}
    return( j );
    }

//----------end get_midpoint--------------------------------------------------------------------------------------------

int
random_vertex_of( polygon *p, BOOL blobby, BOOL traps )   // return an unused vertex of p. if there are none, return -1.
    {                                                     // ok not really random - we first return the vertex
                                                          // where the polygon kinks. That's the most likely spot.
                                                          // But if blobby then we look for a vertex near the middle.

    static int i, j, left, n, this, pick;
    static BOOL is_convex;

    if ( ! p->active ) bail( "inactive poly sent to random_vertex generator" );

    if ( blobby ) 
	j = get_midpoint( p );
    else
        is_convex = convex( p, &j, traps );   // sends back a kink


    if ( ! p->hit[j] )          
	{                       
	p->hit[j] = TRUE;
	return( j );
	}

    n = p->len;
    left = 0;

    for ( j=0; j < n; j++ ) if ( ! p->hit[j] ) left++;
    if ( left == 0 ) return( -1 );

    pick = (left - 1) * random() / RAND_MAX;

    j    = -1;
    this = -1;

    while ( this < pick && j < n )
	{
	j++;
	if ( ! p->hit[j] ) this++;
	}

    if ( j >= 0 && j < n && ! p->hit[j] )
        {
        p->hit[j] = TRUE;
        return( j );
	}
    else
	{
        printf( "logic error in random_vertex_of\n" );
	return( -1 );
	}
    }

//---------

int
random_vertex2_of( polygon *p, BOOL traps )  // not really random, unless the kink was already hit
    {
    static int i, j, left, n, this, pick;
    static BOOL is_convex;

    /*
    if ( ! p->active ) return( -1 );
    i = 0;
    while ( i < p->len && p->hit2[i] ) i++;
    if ( i == p->len ) return( -1 );
    p->hit2[i] = TRUE;
    return( i );
    */

    if ( ! p->active ) bail( "inactive poly sent to random_vertex2 generator" );

    is_convex = convex( p, &j, traps );   // sends back a kink
    //if ( is_convex ) bail( "convex polygon sent to random_vertex2 generator" );

    if ( ! p->hit2[j] )          
	{                       
	p->hit2[j] = TRUE;
	return( j );
	}

    n = p->len;
    left = 0;

    for ( j=0; j < n; j++ ) if ( ! p->hit2[j] ) left++;
    if ( left == 0 ) return( -1 );

    pick = (left - 1) * random() / RAND_MAX;

    j    = -1;
    this = -1;

    while ( this < pick && j < n )
	{
	j++;
	if ( ! p->hit2[j] ) this++;
	}

    if ( j >= 0 && j < n && ! p->hit2[j] )
        {
        p->hit2[j] = TRUE;
        return( j );
	}
    else
	{
        printf( "logic error in random_vertex2_of\n" );
	return( -1 );
	}

    }

//---------------------------------------------------------------------------------------

BOOL
overlapping_range( polygon *p, polygon *q )
    {
    if ( p->ll.x == 0 && p->ll.y == 0 && p->ur.x == 0 && p->ur.y == 0 ) extent( p );
    if ( q->ll.x == 0 && q->ll.y == 0 && q->ur.x == 0 && q->ur.y == 0 ) extent( q );
    if ( p->ll.x < q->ll.x && p->ur.x < q->ll.x ) return( FALSE );
    if ( p->ur.x > q->ur.x && p->ll.x > q->ur.x ) return( FALSE );
    if ( p->ll.y < q->ll.y && p->ur.y < q->ll.y ) return( FALSE );
    if ( p->ur.y > q->ur.y && p->ll.y > q->ur.y ) return( FALSE );
    return( TRUE );
    }


//---------------------------------------------------------------------------------------

void
snap_to_feature( polygon *p, queue *list )
    {
    // Vertices which are not marked as precious, that is, new vertices of p, 
    // are made to snap to nearby vertices in list. 
    // This version looks through the entire list. This definitely needs
    // to be improved. We should look only in the local neighborhood.
    // Well it turns out that our overlaps were caused by darn.c, not 
    // by this program. So for now this function is turned off. We probably
    // should insert it in darn.c 

    return;

    static polygon_ref *pr;
    static polygon     *poly;
    static int          i, j;
    static xy           a, b;

    pr = list->first;

    while ( pr )                                      // for every poly in the list
	{
	poly = pr->p;
        if ( overlapping_range( poly, p ) )           // that overlaps with p
	    {
	    for ( i=0; i < p->len; i++ )              // and for every vertex in p
	        {
	        if ( ! p->precious[i] )               // that is new
		    {
		    for ( j=0; j < poly->len; j++ )   // compare it to every vertex in poly
		        {
		        a = vertex( p,    i );
		        b = vertex( poly, j );
		        if ( closepoint( a, b ) )     // and set them equal if they are close
			    {
			    assign_vertex( p, i, b ); 
			    p->precious[i] = TRUE;
			    }
		        }
		    }
	        }
            }

        pr = pr->next;
	}

    }

//----------------------------------------------------------------------------------------

BOOL
slice_once( polygon *p, queue *todo_list, BOOL traps )
    {
    static int    iii, n, i, j, j1, extrax, extray, minx, maxx, miny, maxy;
    static BOOL   init=FALSE, found_a_cut, found_another_cut, hit, outside;
    static xy     a, b, c, d, cutpoint;
    polygon       *p1, *p2, *p3, *p4;
    static double toosmall, toomuchange, rathersmall, dx, dy;
    static BOOL   ccw_p1, ccw_p2;
    static BOOL   ccw_p3, ccw_p4;
    static BOOL   ccw_p;  
    static BOOL   mostly_horizontal=TRUE, blobby;
    static double area=0;

    // cleaned up

    if ( ! p->active ) return( TRUE );
    if ( convex( p, &j, traps ) ) return( TRUE );

    if ( ! init ) 
        { 
        init=TRUE; 
        toosmall    = bss * bss * 4;      //???? works with annezone but not well tested
        toomuchange = 0.1;  // percent
        rathersmall = bss * bss * 4;
        }


    p1 = NULL;
    p2 = NULL;
    p3 = NULL;
    p4 = NULL;

    if ( p->len < 3 ) { p->active = FALSE; return TRUE; }
    remove_duplicate_vertices( p );
    remove_colinear_vertices(  p, FALSE ); 
    if ( convex( p, &j, traps ) ) return TRUE;
    if ( p && p->ur.x == 0 && p->ll.x == 0 && p->ur.y == 0 && p->ll.y == 0 ) extent( p );

    area = polygon_area( p, bss ); 

    if ( area < toosmall )
        {
        p->active = FALSE;
        return( TRUE );    
        }

    dx = (p->ur.x - p->ll.x);
    dy = (p->ur.y - p->ll.y);
    mostly_horizontal = dx > dy;
    blobby = (fabs( dx - dy ) / dx) < 0.1;  // if dx == dy within 20% then we should slice it in half first
                                            // This bool is sent to the vertex selection function.

    extrax = fabs(p->ur.x - p->ll.x) / 10.0;
    extray = fabs(p->ur.y - p->ll.y) / 10.0;

    minx = p->ll.x - extrax;
    miny = p->ll.y - extray;
    maxx = p->ur.x + extrax;
    maxy = p->ur.y + extray;

    n = p->len;
    found_a_cut = FALSE;
    found_another_cut = FALSE;

    unhit_all( p );

    i = random_vertex_of( p, blobby, traps );  // not random. 'i' will be a kink, most likely. 
                                               // if blobby, 'i' will be near the center.

    while ( i >= 0 && ! found_a_cut )
        {
        a = vertex( p, i );

        if ( mostly_horizontal ) // then try a vertical cut first
	    {
            printv( "[%d] trying vertical cut first\n", i );
            a.y = miny;
            b.x = a.x;
            b.y = maxy;       
	    }
	else                     // try a horizontal cut first
	    {
            printv( "[%d] trying horizontal cut first\n", i );
            b.x = maxx;
            b.y = a.y; 
            a.x = minx;
	    }
       
        j = (i + 1) % n;

        while ( (j != i) && ! found_a_cut )
            {
            j1 = (j + 1) % n;

            if ( (j != i) && (j1 != i) )
                {                
                c = vertex( p, j  ); 
                d = vertex( p, j1 );

                hit = crossing_point( a, b, c, d, &cutpoint );

                if ( hit )
                    {
                    printv( "hit found on %d,%d __ %d,%d at %d,%d \n", c.x,c.y,d.x,d.y, cutpoint.x, cutpoint.y );

                    p1 = initialize_polygon( p->max, p->layer, p->dt, p->obese );
                    p2 = initialize_polygon( p->max, p->layer, p->dt, p->obese );

                    iii = i;
                    while ( iii != j )
                        {
                        append_polygon_xy( p1, vertex( p, iii ) );
			p1->precious[iii] = TRUE;
                        iii = (iii + 1) % n;
                        }
                    append_polygon_xy( p1, vertex( p, j ) );
		    p1->precious[j] = TRUE;
                    append_polygon_xy( p1, cutpoint       );
                    append_polygon_xy( p2, cutpoint );
                    iii = j1;
                    while ( iii != i )
                        {
                        append_polygon_xy( p2, vertex( p, iii ) );
			p2->precious[iii] = TRUE;
                        iii = (iii + 1) % n;
                        }
                    append_polygon_xy( p2, vertex( p, i ) );
		    p2->precious[i] = TRUE;

                    remove_precise_duplicates( p1 );
                    remove_precise_duplicates( p2 );

                    remove_colinear_vertices( p1, FALSE ); // TRUE leads to 
                    remove_colinear_vertices( p2, FALSE ); // shape creep big time
                   
                    // Do not remove duplicate vertices until after we look
                    // in both directions.

                    ccw_p1 = TRUE;
                    ccw_p2 = FALSE;
                    if ( p1->active && p2->active )
			{
                        ccw_p1 = ! clockwise( p1 );
		        ccw_p2 = ! clockwise( p2 );
		        }

                    outside = (ccw_p1 != ccw_p2); 

                    // We should remove duplicate vertices here, but that might
                    // make it difficult to deal with little slivers. Perhaps
                    // we can create a version of remove_duplicate_vertices that
                    // does not consider "nearby" vertices to be duplicates.

		    ifv( ! outside ) printv( ".....looks ok, but now check crossings\n" );
                    if ( ! outside ) 
			{
			outside = self_crossing_polygon( p1 );
			ifv( outside ) printv( ".....p1 is self-crossing\n" );
			}
                    if ( ! outside ) 
			{
			outside = self_crossing_polygon( p2 ) ;
                        ifv( outside ) printv( ".....p2 is self-crossing\n" );
			}

                    if ( outside ) 
                        {
                        //if ( p1 ) free( p1 );
                        //if ( p2 ) free( p2 );
                        }
                    found_a_cut = ! outside;
                    }
                }
            j = (j + 1) % n;
            }
        i = random_vertex_of( p, FALSE, traps );
        }

    // If we did not find a cut then we should look for a cut in the other direction.
    // If we found a cut but the area of one of them is rather small, then
    // we should look for a cut in the other direction.

    if ( ! found_a_cut )
        {
        printv( "looking in the other direction\n" );
        found_another_cut = FALSE;

        i = random_vertex2_of( p, traps );

        while ( i >= 0 && ! found_another_cut )
            {
            a = vertex( p, i );

	    if ( ! mostly_horizontal ) // then try a vertical cut (second)
		{
		a.y = miny;
		b.x = a.x;
		b.y = maxy;       
		}
	    else                       // try a horizontal cut (second)
		{
		b.x = maxx;
		b.y = a.y; 
		a.x = minx;
		}

            j = (i + 1) % n;

            while ( (j != i) && ! found_another_cut )
                {
                j1 = (j + 1) % n;

                if ( (j != i) && (j1 != i) )
                    {                
                    c = vertex( p, j  ); 
                    d = vertex( p, j1 );

                    hit = crossing_point( a, b, c, d, &cutpoint );

                    if ( hit )
                        {
                        printv( "hit found on %d,%d __ %d,%d at %d,%d \n", c.x,c.y,d.x,d.y,cutpoint.x,cutpoint.y );
                        p3 = initialize_polygon( p->max, p->layer, p->dt, p->obese );
                        p4 = initialize_polygon( p->max, p->layer, p->dt, p->obese );

                        iii = i;
                        while ( iii != j )
                            {
                            append_polygon_xy( p3, vertex( p, iii ) );
			    p3->precious[iii] = TRUE;
                            iii = (iii + 1) % n;
                            }
                        append_polygon_xy( p3, vertex( p, j ) );
			p3->precious[j] = TRUE;
                        append_polygon_xy( p3, cutpoint       );

                        append_polygon_xy( p4, cutpoint );
                        iii = j1;
                        while ( iii != i )
                            {
                            append_polygon_xy( p4, vertex( p, iii ) );
			    p4->precious[iii] = TRUE;
                            iii = (iii + 1) % n;
                            }
                        append_polygon_xy( p4, vertex( p, i ) );
			p4->precious[i] = TRUE;

                        remove_precise_duplicates( p3 );
                        remove_precise_duplicates( p4 );

                        remove_colinear_vertices( p3, FALSE ); 
                        remove_colinear_vertices( p4, FALSE ); 

                        ccw_p3 = TRUE;
                        ccw_p4 = FALSE;
                        if ( p3->active && p4->active )
			    {
   			    ccw_p3 = ! clockwise( p3 );
                            ccw_p4 = ! clockwise( p4 );
			    }

                        outside = (ccw_p3 != ccw_p4);

                        ifv( ! outside ) printv( ".....looks ok, so now check for crossings\n" );
                        if ( ! outside ) 
			    {
			    outside = self_crossing_polygon( p3 );
                            ifv( outside ) printv( ".....p3 is self-crossing\n" );
			    }
                        if ( ! outside ) 
			    {
			    outside = self_crossing_polygon( p4 ) ;
			    }

                        if ( outside ) 
                            {
                            //if ( p3 ) free( p3 );
                            //if ( p4 ) free( p4 );
                            }
                        found_another_cut = ! outside;
                        }
                    }
                j = (j + 1) % n;
                }
            i = random_vertex2_of( p, traps );
            }
      
        } // end looking for a vertical cut



    if ( ! found_a_cut && found_another_cut )    // the first cut was not found but the second one is ok
        {
	snap_to_feature( p3, todo_list );
	snap_to_feature( p4, todo_list );
        append_queue( todo_list, p3 );
        append_queue( todo_list, p4 );
        //if ( p1 ) free( p1 );     // why does this cause
        //if ( p2 ) free( p2 );     // a seg fault?
        p->active = FALSE;
        return( TRUE );
        }
    else if ( found_a_cut && ! found_another_cut )
        {
	snap_to_feature( p1, todo_list );
	snap_to_feature( p2, todo_list );
        append_queue( todo_list, p1 );
        append_queue( todo_list, p2 );
        //if ( p3 ) free( p3 );
        //if ( p4 ) free( p4 );
        p->active = FALSE;
        return( TRUE );
	}
    else if ( ! found_a_cut && ! found_another_cut )  //  could not find a cut in either direction
        {
	//if ( p1 ) free( p1 );
        //if ( p2 ) free( p2 );
	//if ( p3 ) free( p3 );     // causes seg fault later... why?
        //if ( p4 ) free( p4 );     // causes seg fault later
        return( FALSE );
	}
    else
        {
	bail( "logic error in slice_once\n" );
	}


    return( found_a_cut || found_another_cut );
    
    }

//----- end slice_once ---------------------------------------------------------------------------------------

void
double_polygon( polygon_ref *pr )
    {
    static xy a, b, m;
    static long i, n1;
    polygon *p, *d;

    p = pr->p;
    if ( p->len < 3 || ! p->active ) return;
    n1 = p->len - 1;
    d = initialize_polygon( 2 * p->len, p->layer, p->dt, p->obese );

    for ( i=0; i < n1; i++ )
        {
        a = vertex( p, i   );
        b = vertex( p, i+1 );
        m.x = (a.x + b.x) / 2;
        m.y = (a.y + b.y) / 2;
        append_polygon_xy( d, a );
        append_polygon_xy( d, m );
        }
    a = vertex( p, n1 );
    b = vertex( p, 0  );
    m.x = (a.x + b.x) / 2;
    m.y = (a.y + b.y) / 2;
    append_polygon_xy( d, a );
    append_polygon_xy( d, m );
    remove_duplicate_vertices( d );
    pr->p = d;
    }

//---------- double_polygon---------------------------------------------------------------------------------

BOOL
slicer( polygon *poly, int ofile, BOOL traps )
    {
    // This controls the stack used for manual recursion
    // of the slicing function

    static polygon     *p, *d;
    static polygon_ref *pr;
    static BOOL         ok;
    static FILE        *fp;
    static int          j;
    static double       area, toosmall;
    queue              *todo_list;

    if ( convex( poly, &j, traps ) )
        {
        write_polygon( ofile, poly, traps );
        return( TRUE );
        }

    todo_list = initialize_queue();

    append_queue( todo_list, poly );

    pr = todo_list->first;

    while ( pr )
        {
        p = pr->p;
                
        if ( p && p->active )
            {
            ok = slice_once( p, todo_list, traps );
            if ( ! ok ) 
		{
                unhit_all( p );
                snap_to_grid( p );
                remove_colinear_vertices( p, TRUE );                       // TRUE => severe, remove more flagpoles
                find_and_fix_self_crossing_polygon( p, todo_list );        // really does help!
                ok = slice_once( p, todo_list, traps );
                if ( ! ok )
		    {
		    int pbss, pbss2;
		    pbss = bss;
		    pbss2 = bss2;
		    bss = 2 * bss;  // 1   ????????? highly dubious
                    bss2 = bss;
                    no_precious( p ); 
		    snap_to_grid( p );
                    remove_colinear_vertices( p, TRUE );                    
                    find_and_fix_self_crossing_polygon( p, todo_list );
		    ok = slice_once( p, todo_list, traps );
                    bss = pbss;
                    bss2 = pbss2;
		    }
                if ( ! ok )
		    {
		    area = polygon_area( p, bss );
                    toosmall = 6;
		    if ( area < toosmall ) 
			{
			p->active = FALSE;
			ok = TRUE;
			}
		    else
			{
  	                printf( "---------------------------convexify fail\n" );
                        fflush( stdout );
                        p->layer = 15;
			p->active = TRUE;
			append_queue( todo_list, p );
			write_trap_list( ofile, todo_list, traps );
			gds_write_endstr( ofile ); 
                        gds_write_endlib( ofile );    
                        close( ofile );
                        print_polygon( p, 1 );
                        printf( "\noutput file: %s\n", ofilename );
		        printf( "area = %lf   dx= %ld  dy= %ld  toosmall= %lf  bss= %d\n", area, (p->ur.x - p->ll.x), (p->ur.y - p->ll.y), toosmall, bss );
                        bail( "cannot fracture polygon" );
			}
		    }
		}
            }

        pr = pr->next;
        } 

    write_trap_list( ofile, todo_list, traps );
    
    return( TRUE );
    }

//------- end slicer ------------------------------------------------------------------------

BOOL
they_might_overlap( polygon *p, polygon *q )
    {
    // note that the extents were expanded by +- bss
    //
    if ( p->ur.x < q->ll.x ) return( FALSE );
    if ( p->ll.x > q->ur.x ) return( FALSE );
    if ( p->ur.y < q->ll.y ) return( FALSE );
    if ( p->ll.y > q->ur.y ) return( FALSE );
    return( TRUE );    
    }

//----------------------------------------------------------------------------------------------

void
marry_vertices( queue *tout )    // look for vertices which are closer than bss and
    {                            // set them to the same value so they will snap together
    polygon_ref *pr, *qr;        // Maybe we should call this the "healing" function.
    polygon     *p,  *q, *ptemp;
    xy           v,   w;
    xy           wa,  wb;
    int          i,   j;
    int          dx,  dy,  d, range;
    int          besti, bestj, btemp;
    BOOL         escape, skipi;
    double       cw, cv;


    // look for vertices near each other -----------------------

    //printf( "Snapping vertices...\n" ); fflush(stdout);

 
    pr = tout->first;
    while ( pr && ! pr->p->active ) pr = pr->next;

    while ( pr )
        {
        p = pr->p;
        qr = pr->next;
        while ( qr && ! qr->p->active ) qr = qr->next;
        range = 0;
        while ( qr && (range < MAXMARRY))
            {
            q = qr->p;
            range++;
            if ( they_might_overlap( p, q ) )
                {
                i = 0;
                while ( i < p->len )
                    {
                    v = vertex( p, i );
                    j = 0;
                    while ( j < q->len && i < p->len ) 
                        {
                        w = vertex( q, j );

                        dx = abs( v.x - w.x );
                        dy = abs( v.y - w.y );

                        if ( dx < bss && dy < bss ) // kiss the bride
                            {
                            besti = i;
                            bestj = j;
                            v = vertex( p, besti );
                            w = vertex( q, bestj );
                         
                            // If we have already moved one of these vertices
                            // then that one should stay put. If we have already
                            // moved both vertices, then we should do nothing.

                            if ( ! p->precious[besti] && ! q->precious[bestj] )
                                {
                                // should we snap to v or w? 
                                // let's see which one creates the largest circumference

                                assign_vertex( p, besti, w );
                                assign_vertex( q, bestj, w );
                                cw = circumference( p ) + circumference( q );
                                assign_vertex( p, besti, v );
                                assign_vertex( q, bestj, v );
                                cv = circumference( p ) + circumference( q );

                                if ( cw > cv ) 
                                    {
                                    assign_vertex( q, bestj, w );
                                    assign_vertex( p, besti, w );
                                    }

                                p->precious[besti] = TRUE;
                                q->precious[bestj] = TRUE;
                                }
                            else if ( p->precious[besti] )
                                {
                                assign_vertex( q, bestj, v ); 
                                q->precious[bestj] = TRUE;
                                }
                            else if ( q->precious[bestj] )
                                {
                                assign_vertex( p, besti, w );
                                p->precious[besti] = TRUE; 
                                }

                            j = j + 1;  

                            }
                        j++;
                        }
                    i++;
                    }
                }

            qr = qr->next;
            while ( qr && ! qr->p->active ) qr = qr->next;
            }

        pr = pr->next;
        while ( pr && ! pr->p->active ) pr = pr->next;
        }
    


    //---------------------
    // roll through the list again, 
    // eliminating duplicate and colinear vertices.

    pr = tout->last;
    while ( pr && ! pr->p->active ) pr = pr->prev;

    printv( "removing duplicate and colinear vertices...\n" ); fflush(stdout);

    while ( pr )
        {
        p = pr->p;
        remove_duplicate_vertices( p );
        remove_colinear_vertices(  p, FALSE ); 
        snap_to_grid( p );
        pr = pr->prev;
        while ( pr && ! pr->p->active ) pr = pr->prev;
        }
        
    }

//-----------------marry_vertices------------------------------------------------------------------

//==============================================================================================================

int main( int argc, char *argv[] )
    {

    pid_t     pid;

    char      
              pidfile[50];

    FILE     *pidf,
             *thecount;

    BOOL      noheader,
	      not_eof,
              good,
              traps;

    int       // bss and bss2 are global
              fd,
              n,
              fdo,
              numpolys,
              layer,
              dt,
	      i, j;

    double    bss_nm,
	      bss_um,
              dbu_uu,
              dbu_m,
              dbu_um,
              progress,
              was;

    unsigned short 
              count,           
              token;   

    polygon  *poly, *p;        

    polygon_ref *pr;

    struct gds_itemtype *item;

    queue    *tous;

    /*
    poly = initialize_polygon( 10, 0, 0, FALSE );
    append_polygon( poly, 1761795, 901575 );
    append_polygon( poly, 1762470, 900255 );
    append_polygon( poly, 1762470, 901575 );
    append_polygon( poly, 1761645, 901575 );
    append_polygon( poly, 1761000, 902835 );
    append_polygon( poly, 1761795, 901575 );
    extent( poly );
    if ( self_crossing_polygon( poly ) )
	printf( "\nNot good. Self-crossing.\n" );
    else
	printf( "\nGood. Not self-crossing.\n" );
    exit(0);
    */

    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "convexify v%1.1f\n", VERSION );
        exit(0);
        }

    if ( argc < 3 )
        {
        printf( "\nUse: convexify infile.gds outfile.gds bss_nm [noheader] [traps]\n\n" );
        exit(0);
        }

    pid = getpid();
    ifilename = argv[1]; 

    sprintf( ofilename, "%s", argv[2] );
    sscanf( argv[3], "%lf", &bss_nm );

    noheader = FALSE;
    traps    = FALSE;

    if ( argc > 4 )
        noheader = (strcmp(argv[4], "noheader") == 0) ||
                   (strcmp(argv[5], "noheader") == 0);

    if ( argc > 5 )
        traps = (strcmp(argv[5], "traps") == 0) ||
                (strcmp(argv[4], "traps") == 0);


    //------- end command parse --------------------------

    int   ch1, ch2;
    char *sn1, *sn2, sn[10]="";

    if ( strncmp( argv[1], "moslice", 7 ) != 0 ) 
        {
        if ( strstr( ifilename, "_part_" ) )
            {
            ch1 = (int) '_';
            sn1 = rindex( ifilename, ch1 );
            ch2 = (int) '.';
            sn2 = rindex( ifilename, ch2 ) - 1;
            if ( sn1 && sn2 )
                {
                strncpy( sn, sn1+1, sn2 - sn1 );
                strcat(  sn, " " );
                //printf( "starting convexify part %s    \n", sn );    
                }
            }
	}
    else
        {
        //printf( "convexify %s %s %1.2lf \n", ifilename, ofilename, bss_nm );
        }
        


    //------------------------------------------ open the input file

    fd = open( ifilename, 0 );
    if ( fd <= 0 ) bail( "unable to open input file" );
    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! item ) bail( "unable to allocate gds item" );

    //------------------------------------------ read the polygons and push them onto the queue tous

    tous = initialize_queue();
    numpolys = 0;
    progress = 0.0;
    not_eof = read( fd, &count, 2 ) > 0;

    while ( not_eof )
        {
        gds_swap2bytes( (BYTE *) &count );
        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );

            switch( token )
                {
                case 0x0305:  // UNITS
                          gds_read_units( fd, &dbu_um, &dbu_uu, &dbu_m, 0 );
                          dbu_uu = dbu_m * 1.0e6;
                          dbu_um = dbu_m * 1.0e6; 
                          bss_um = bss_nm / 1000.0;
                          bss = (int) round( bss_um / dbu_um );    // beam step in gds database units eg 0.01 / 0.001 or 6 / 0.01
                          if ( bss == 0 ) bss = 1;
                          bss2 = bss * 0.8;            // this is what we mean by "too close"
                          bssm = bss * 1.2;            // in the neighborhood
                          if ( bss2 == 0 ) bss2 = 1;
                          break;
                case 0x2B02:  // printf( "\nPROPATTR means obese \n" ); not used
                          SKIPOVER( fd, count );
                          break;
                case 0x0D02:  //printf( "LAYER \n" ) ;
                          gds_read_layer( fd, count, &item, 0 );
                          layer = item->layer;
                          break;
                case 0x0E02:  //printf( "DATATYPE \n" ) ;
                          gds_read_datatype( fd, count, &item, 0 );
                          dt = item->dt;
                          break;
                case 0x1003:  //printf( "XY \n" ) ;

                          gds_read_xy( fd, count, 0, &item, 0 );

                          item->layer = layer;
                          item->dt = dt;
                          poly = copy_from_gds_polygon( item );
                          append_queue( tous, poly );
                          numpolys++;
                          break;
                default:  
                          SKIPOVER( fd, count );
                }    
            }      
        not_eof = read( fd, &count, 2 ) > 0;
        }          

    close(fd);


    // look for vertices which are close to each other,
    // then set them equal so they will not snap apart

    // This was already done in the 'snap' program, so
    // you would think it is not necessary here.
    // But it does actually help close some gaps.
    
    // marry_vertices( tous );  and yet very rarely it causes a self-cross

    // Why is it better to fix self-crossing polygons here,
    // instead of in snap.c ? Maybe because darn.c then
    // has the task of removing holes a second time, before
    // we work on self-crossing polygons. Ok that makes sense.

    pr = tous->first;
    while ( pr )
        {
        p = pr->p;
        remove_duplicate_vertices( p );
        find_and_fix_self_crossing_polygon( p, tous );
        pr = pr->next;
        }
    

    // open output file --------------------------------------

    fdo = open( ofilename, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( fdo <= 0 ) ( "unable to open output gds file" );

    if ( ! noheader )
        {
        gds_create_lib_correctly( fdo, "traps", dbu_um, dbu_uu );
        gds_write_bgnstr( fdo );
        gds_write_strname( fdo, "traps" );
        }

    // munch on polygons and write output file -----------------------------

    /*
    write_trap_list( fdo, tous, traps );
    gds_write_endstr( fdo ); 
    gds_write_endlib( fdo );    
    close( fdo );
    printf( "\ndone, but not sliced\n" );
    exit(0);
    */

    /*
    printf( "\nWriting %s\n", ofilename );
    write_trap_list( fdo, tous, traps );
    bail( "abort for debugging" ); 
    */

    printv( "slicing polygons...\n" ); fflush(stdout);

    pr = tous->first;
    while ( pr )
        {
        poly = pr->p;

        slicer( poly, fdo, traps ); 

        pr = pr->next;
        }

    if ( FALSE ) // ! noheader )
	{
	gds_write_endstr( fdo ); 
	gds_write_endlib( fdo );    
	}

    close( fdo );

    printv( "\ndone.\n" );

    }


