// idarn.c
//                 Separate polygons into outlines and holes,
//                 then chop up the polygon by creating two cuts
//                 on each hole. Uses the integer library polyq.c

// 1.1   Reject shapes whose extent is less than bss in either x or y.
//       This seems to eliminate the rare crazy shapes that look like memory errors.
// 1.2   Empty pattern files handled gracefully, as before.
// 1.3   Set dbu_rescale to 1.0 in case the user needs sub-pixel placement
//       Scaling everything to 1 nm was always a lazy idea.
// 1.4   xy type now uses long int
// 1.5   Bowties are identified differently. Search for "change 1.5"

#include "libgdsmin.h"
#include "polyq.h"
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#define VERSION      1.5
#define SHOW_HOLES   0
#define SHOWPROGRESS 0
#define ERRORLAYER   300

#define NEAR( a, b ) ((abs((a.x)-(b.x)) < nearness) && (abs((a.y)-(b.y)) < nearness))
#define EQUAL( a, b ) ((abs((a)-(b)) < nearness))
#define CLOSE( a, b ) ((fabs((a)-(b)) < 3))  
#define TOOCLOSE( a ) ((abs((a)) < tooclose))
#define TOOMUCHANGE 1

char  *ifilename;
int    num_touchies=0;
int    num_fail=0;
int    recursion=0;
int    fdo;
int    nearness, presque, tooclose;
int    bss;

void write_queue( int fdo, queue *list ); 
void remove_colinear_vertices( polygon *p );       
BOOL crossing( xy a1, xy a2, xy b1, xy b2, xy *intersection );
BOOL not_crossing( polygon *p, int a, int b, int c, int d, xy *intersection, BOOL overlapsbad );
BOOL there_is_an_inside_point( polygon *h, polygon *p );
BOOL self_crossing( polygon *p );
BOOL all_hole_vertices_are_inside_poly( polygon *h, polygon *p );

//---------------------------------------------------------------------------------------------------

BOOL
SAMEPOINT( xy a, xy b )        
    {
    return( abs(a.x - b.x) < 2 && abs(a.y - b.y) < 2 );
    }

//---------------------------------------------------------------------------------------------------

int                            
decrement( int a, int b )
    {
    if ( a != 0 ) 
        return( a - 1 );
    else
        return( b - 1 );

    }

//---------------------------------------------------------------------------------------------------         

void                                      
nag( char *message, int *count )
    {
    *count += 1;
    if ( *count < 10 )
        {
        if ( *count == 1 ) printf( "\n\n" );
        printf( "WARNING: %s\n\n", message );
        }
    else if ( *count == 10 )
        {
        printf( "*** REPEATED WARNING MESSAGES SUPPRESSED\n\n" );
        }
    fflush( stdout );
    }

//----------------------------------------------------------------------------------------------------------------

void                                    
set_shape_extents( polygon *p )
    {
    static int  i;
    static xy   v;

    if ( p->len < 3 )
        {
        p->active = FALSE;
        return;
        }

    p->ll = vertex( p, 0 );
    p->ur = vertex( p, 0 );

    for ( i=1; i < p->len; i++ )
        {
        v = vertex( p, i );
        if ( v.x < p->ll.x ) p->ll.x = v.x;
        if ( v.y < p->ll.y ) p->ll.y = v.y;
        if ( v.x > p->ur.x ) p->ur.x = v.x;
        if ( v.y > p->ur.y ) p->ur.y = v.y;
        }

    return;
    }

//------------end set_shape_extents----------------------------------------------------------------

void                                  
check_extents( polygon *p )
    {
    static int nn=0;
    if ( (p->ll.x == 0) && (p->ll.y == 0) && (p->ur.x == 0) && (p-> ur.y == 0) ) 
        {
        nag( "polygon extents have not been set", &nn );
        set_shape_extents( p );
        }
    }

//---------------------------------------------------------------------------------------------------------------

BOOL                                                                   
dequal( double a, double b )   // are two doubles roughly equal?
    {
    if ( a == b ) return( TRUE );

    if ( a == 0.0 )
        return( fabs(b) < 1.0e-7 );
    else if ( b == 0.0 )
        return( fabs(a) < 1.0e-7 );
    else
        return( fabs( (a - b)/a ) < 1.0e-7 );

    }

//-----------------------------------------------------------------------------------------------------------------

BOOL                                                                                     
between( int x, int a, int b )                // Is x between a and b?
    {                                         // Hitting an end point means 'yes'.
    if ( EQUAL( x, a ) ) return( TRUE );
    if ( EQUAL( x, b ) ) return( TRUE );

    if ( a < b )  
        return( x >= a && x <= b );   
    else
        return( x >= b && x <= a );
    }

//-------------------------------------------------------------------------------------------------------------------

int                                            
point2line_dist( xy p, xy a, xy b )
    {
    // we might have to change this back to returning a double ??????
    // calculate the distance from the point p to line a-b

    static double dx, dy, d, px, py, s, u, xx, yy, da, db;

    if ( SAMEPOINT( p, a ) )
        return( 0 );
    else if ( SAMEPOINT( p, b ) )
        return( 0 );
    else if ( SAMEPOINT( a, b ) ) 
        {
        dx = p.x - a.x;
        dy = p.y - a.y;
        d = sqrt( dx * dx + dy * dy );
        return( (int) round(d) );
        }
    else 
        {
        px = b.x - a.x;
        py = b.y - a.y;
        s = px * px + py * py;
        u = ((p.x - a.x) * px + (p.y - a.y) * py) / s;
        if ( u > 1.0 )
            u = 1.0;
        else if ( u < 0.0 )
            u = 0.0;
        xx = a.x + u * px;
        yy = a.y + u * py;

        if ( between( xx, a.x, b.x ) && between( yy, a.y, b.y ) )
            {
            dx = xx - p.x;
            dy = yy - p.y;
            d = sqrt( dx * dx + dy * dy );
            }
        else
            {
            dx = p.x - a.x;
            dy = p.y - a.y;
            da = sqrt( dx * dx + dy * dy );
            dx = p.x - b.x;
            dy = p.y - b.y;
            db = sqrt( dx * dx + dy * dy );
            d = da;
            if ( db < da ) d = db;
            }

        return( (int) round(d) );
        }

    }

//---------end pointoline_dist------------------------------------------------------------------------------


// Point-in-polygon test:  Is the given point inside or outside the polygon?
//                         This is much more difficult than it seems.
//                         The winding test algorithm does not work 
//                         for complex doughnut-like polygons. This version
//                         counts crossings of a right-pointing ray.

BOOL                                                                                   
very_between( int x, int a, int b )                // Is x between a and b?
    {                                              // Hitting the end means 'no'.
    if ( EQUAL( x, a ) ) return( FALSE );
    if ( EQUAL( x, b ) ) return( FALSE );

    if ( a < b )  
        return( x >= a && x <= b );   
    else
        return( x >= b && x <= a );
    }


BOOL                                                                                     
point_near_line( xy p, xy a, xy b )                // is p near the line a-b?
    {
    return( point2line_dist( p, a, b ) < tooclose );
    }


BOOL                                                                                    
ontheright( xy point, xy v0, xy v1 )
    {
    // Is the line segment v0-v1 on the right of 'point'?

    static double v0y, v1y, v0x, v1x, m, b, xi;
    static double num, den, pointy;
    static int ixi, iv0y, iv1y;

    if ( SAMEPOINT( v0, v1 ) ) return( FALSE );


    if ( (v0.x > point.x) && (v1.x > point.x) )
        {
        iv0y = v0.y - point.y;
        iv1y = v1.y - point.y;

        if ( ((iv0y >= 0) && (iv1y < 0)) || ((iv0y < 0) && (iv1y >= 0)) ) 
            {
            return( TRUE );
            }
        }
    else if ( (v0.x > point.x) || (v1.x > point.x) )
        {
        if ( ( v0.x == v1.x ) )           // vertical - will never happen
            {
            return( very_between( point.y, v0.y, v1.y ) );
            }
        else if ( ( v0.y == v1.y ) )      // horizontal - always false
            {
            return( FALSE );
            }
        else
            {
	    v0y = v0.y;
	    v1y = v1.y;
	    v0x = v0.x;
            v1x = v1.x;
	    num = (v0y - v1y);
	    den = (v0x - v1x);
            m = num / den;
            b = v0y - m * v0x;
            pointy = point.y;
            xi = (pointy - b) / m;
            }

        ixi = (int) round( xi );
        if ( ixi > point.x )
            {
            if ( very_between( ixi, v0.x, v1.x ) ) 
                return( TRUE );
            else if ( EQUAL( ixi, v0.x ) && EQUAL( point.y, v0.y ) )     // if we hit a vertex, then we should count one
                return( TRUE );
	    else if ( EQUAL( ixi, v1.x ) && EQUAL( point.y, v1.y ) )
		return( TRUE );
            }

        }
    return( FALSE );
    }


BOOL                                                           
ontheleft( xy point, xy v0, xy v1 )
    {
    // Is the line segment v0-v1 on the left of 'point'?

    static double m, b, xi, num, den, pointy;
    static int ixi, iv0y, iv1y;

    if ( SAMEPOINT( v0, v1 ) ) return( FALSE );

    if ( (v0.x < point.x) && (v1.x < point.x) )
        {
        iv0y = v0.y - point.y;
        iv1y = v1.y - point.y;
        if ( ((iv0y >= 0) && (iv1y < 0)) || ((iv0y < 0) && (iv1y >= 0)) ) 
            {
            return( TRUE );
            }
        }
    else if ( (v0.x < point.x) || (v1.x < point.x) )
        {
        if ( ( v0.x == v1.x ) )           // vertical - will never happen
            {
            return( very_between( point.y, v0.y, v1.y ) );
            }
        else if ( ( v0.y == v1.y ) )      // horizontal - always false
            {
            return( FALSE );
            }
        else
            {
	    num = (v0.y - v1.y);
	    den = (v0.x - v1.x);
            m = num / den;
            b = v0.y - m * v0.x;
	    pointy = point.y;
            xi = (pointy - b) / m;
            }

        ixi = (int) round( xi );
        if ( ixi > point.x )
            {
            if ( very_between( ixi, v0.x, v1.x ) ) 
                return( TRUE );
            else if ( EQUAL( ixi, v0.x ) && EQUAL( point.y, v0.y ) )     // if we hit a vertex, then we should count one
                return( TRUE );
	    else if ( EQUAL( ixi, v1.x ) && EQUAL( point.y, v1.y ) )
		return( TRUE );
            }

        }
    return( FALSE );
    }


BOOL                                                  
onthebottom( xy point, xy v0, xy v1 )
    {
    // Is the line segment v0-v1 below 'point'?

    static double v0x, v0y, m, b, yi, num, den, pointx;
    static int iyi, iv0x, iv1x;


    if ( SAMEPOINT( v0, v1 ) ) return( FALSE );

    if ( (v0.y < point.y) && (v1.y < point.y) )
        {
        iv0x = v0.x - point.x;
        iv1x = v1.x - point.x;
        if ( ((iv0x >= 0) && (iv1x < 0)) || ((iv0x < 0) && (iv1x >= 0)) ) 
            {
            return( TRUE );
            }
        }
    else if ( (v0.y < point.y) || (v1.y < point.y) )
        {
        if ( ( v0.y == v1.y ) )   
            {
            return( very_between( point.x, v0.x, v1.x ) );
            }
        else if ( ( v0.x == v1.x ) )     
            {
            return( FALSE );
            }
        else
            {
	    v0y = v0.y;
	    v0x = v0.x;
	    num = (v0.y - v1.y);
	    den = (v0.x - v1.x);
            m = num / den;
            b = v0y - m * v0x;
	    pointx = point.x;
	    yi = m * pointx + b;
            }

        iyi = (int) round( yi );
        if ( iyi < point.y )
            {
            if ( very_between( iyi, v0.y, v1.y ) ) 
                return( TRUE );
            else if ( EQUAL( iyi, v0.y ) && EQUAL( point.x, v0.x ) )
                return( TRUE );
            else if ( EQUAL( iyi, v1.y ) && EQUAL( point.x, v1.x ) )
		return( TRUE );
            }

        }
    return( FALSE );
    }


BOOL                                                             
onthetop( xy point, xy v0, xy v1 )
    {
    // Is the line segment v0-v1 above 'point'?

    static double m, b, yi, num, den;
    static int iyi, iv0x, iv1x;


    if ( SAMEPOINT( v0, v1 ) ) return( FALSE );

    if ( (v0.y > point.y) && (v1.y > point.y) )
        {
        iv0x = v0.x - point.x;
        iv1x = v1.x - point.x;
        if ( ((iv0x >= 0) && (iv1x < 0)) || ((iv0x < 0) && (iv1x >= 0)) ) 
            {
            return( TRUE );
            }
        }
    else if ( (v0.y > point.y) || (v1.y > point.y) )
        {
        if ( ( v0.y == v1.y ) )   
            {
            return( very_between( point.x, v0.x, v1.x ) );
            }
        else if ( ( v0.x == v1.x ) )     
            {
            return( FALSE );
            }
        else
            {
	    num = (v0.y - v1.y);
	    den = (v0.x - v1.x);
            m = num / den;
            b = v0.y - m * v0.x;
	    yi = m * point.x + b;
            }

        iyi = (int) round( yi );
        if ( iyi > point.y )
            {
            if ( very_between( iyi, v0.y, v1.y ) ) 
                return( TRUE );
            else if ( EQUAL( iyi, v0.y ) && EQUAL( point.x, v0.x ) )
                return( TRUE );
            else if ( EQUAL( iyi, v1.y ) && EQUAL( point.x, v1.x ) )
		return( TRUE );
            }

        }
    return( FALSE );
    }

//-------------------------------------------------------------------------------------------

BOOL *                                                       
find_cutlines( polygon *p )
    {
    // Search for interior cuts in the polygon. These sides will be excluded from
    // determination of points being inside or outside. We recycle the array,
    // since there is no need to retain this after processing each polygon.

    // It seems unlikely that we will have cutlines after removing the holes, 
    // but it really does happen!

    static BOOL init=FALSE;
    static int len_cutline, i, j, im1, ii;
    static xy   v0, v1, vv0, vv1;
    static BOOL *cutline;

    if ( ! init )
        {
        init = TRUE;
        cutline = (BOOL *) malloc( p->len * 10  * sizeof(BOOL) );
        if ( ! cutline ) bail( "unable to allocate array for cutline" );
        len_cutline = p->len * 10;
        }

    if ( p->len > len_cutline )
        {
        free( cutline );
        cutline = (BOOL *) malloc( p->len * 10 * sizeof(BOOL) );
        if ( ! cutline ) bail( "unable to allocate array for cutline" );
        len_cutline = p->len * 10;
        }

    // Look for internal cuts
    // cutline[i] refers to the line segment i-1, i
    // cutline[0] refers to the line segment p->len-1, 0

    for ( j=0; j < p->len; j++ ) cutline[j] = FALSE;
    
    for ( i=0; i < p->len; i++ )
        {
        im1 = i - 1;
        if ( i == 0 )  im1 = p->len - 1;
        v0 = vertex( p, im1 );
        v1 = vertex( p, i   );
        for ( ii=(i+1); ii < p->len && ! cutline[i]; ii++ )
            {
            vv0 = vertex( p, ii-1 );
            vv1 = vertex( p, ii   );
            cutline[i] = (SAMEPOINT( v0, vv0 ) && SAMEPOINT( v1, vv1 )) || (SAMEPOINT( v0, vv1 ) && SAMEPOINT( v1, vv0 ));
            if ( ! cutline[i] )
                {
                cutline[i] = point_near_line( vv0, v0, v1 ) && point_near_line( vv1, v0, v1 );
                }
            if ( cutline[i] ) cutline[ii] = TRUE;
            }
        }

    //for ( j=0; j < p->len; j++ ) if ( cutline[j] ) printf( "\ncut line found \n" ); // yes, really 
    
    return( cutline );
    }

//-------------------------------------------------------------------------------------------------------------------

BOOL                                                                    
inside_point( xy point, polygon *p, BOOL no_wallflowers ) 
    {
    // Count the number of crossings by a +X ray from the point.
    // But this is not entirely reliable, so we also draw rays up, down, and left.
    // If three out of four say "interior" then ok, the point is inside.
    // Why is one ray not enough? Because our polygons might be VERY nasty.
    
    static int    i, i1;
    static xy     v0, v1;
    static int    crossing;
    static BOOL  *cutline;
    static BOOL   interior_left, interior_right, interior_top, interior_bottom, interior;
    static int    vote;

    if ( p->len < 3 ) return( FALSE );

    // Bad news: crossing an internal cut does not count. Shapes like doughnuts
    // will have line segments that are not "inside" on one side and "outside"
    // on the other. Those nasty sides must be excluded from the count. We search
    // for inside cuts by looking for identical or overlapping segments on 
    // the same polygon.

    // Another subtle point: If a +X ray from the point intersects precisely 
    // on a vertex, then how many crossings is that? Just one. The function
    // ontheright() must consider this carefully, even though the case
    // is very rare.

    cutline = find_cutlines( p ); 

    // Count the number of crossings with a left-pointing ray

    crossing = 0;

    for ( i=1; i < p->len; i++ )
        {
        v0 = vertex( p, i-1 );
        v1 = vertex( p, i   );
        if ( ! cutline[i] && ontheleft( point, v0, v1 ) ) 
            {
            //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", i-1, i, v0.x, v0.y, v1.x, v1.y );
            crossing++;
            }
        }

    v1 = vertex( p, p->len - 1 );
    v0 = vertex( p, 0 );
    if ( ! cutline[0] && ontheleft( point, v0, v1 ) ) 
        {
        //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", 0, (p->len - 1), v0.x, v0.y, v1.x, v1.y );
        crossing++;
        }

    //printf( "\nCROSSINGS = %d\n", crossing );

    interior_left = (crossing % 2); 

    // if the point is allegedly interior, check that the point is far enough from the edge

    if ( interior_left && no_wallflowers )
        {
        for ( i=0; i < p->len && interior_left; i++ )
            { 
            v0 = vertex( p, i );
            if ( SAMEPOINT( point, v0 ) ) interior_left = FALSE; 
            }

        for ( i=0; i < p->len && interior_left; i++ )
            {
            i1 = (i + 1) % p->len;
            v0 = vertex( p, i  );
            v1 = vertex( p, i1 );
            if ( between( point.x, v0.x, v1.x ) && between( point.y, v0.y, v1.y ) )
                if ( point2line_dist( point, v0, v1 ) < presque ) interior_left = FALSE; 
            }
        }

    //.....
    // Count the number of crossings with a right-pointing ray

    crossing = 0;

    for ( i=1; i < p->len; i++ )
        {
        v0 = vertex( p, i-1 );
        v1 = vertex( p, i   );
        if ( ! cutline[i] && ontheright( point, v0, v1 ) ) 
            {
            //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", i-1, i, v0.x, v0.y, v1.x, v1.y );
            crossing++;
            }
        }

    v1 = vertex( p, p->len - 1 );
    v0 = vertex( p, 0 );
    if ( ! cutline[0] && ontheright( point, v0, v1 ) ) 
        {
        //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", 0, (p->len - 1), v0.x, v0.y, v1.x, v1.y );
        crossing++;
        }

    //printf( "\nCROSSINGS = %d\n", crossing );

    interior_right = (crossing % 2); 

    // if the point is allegedly interior, check that the point is far enough from the edge

    if ( interior_right && no_wallflowers )
        {
        for ( i=0; i < p->len && interior_right; i++ )
            { 
            v0 = vertex( p, i );
            if ( SAMEPOINT( point, v0 ) ) interior_right = FALSE; 
            }

        for ( i=0; i < p->len && interior_right; i++ )
            {
            i1 = (i + 1) % p->len;
            v0 = vertex( p, i  );
            v1 = vertex( p, i1 );
            if ( between( point.x, v0.x, v1.x ) && between( point.y, v0.y, v1.y ) )
                if ( point2line_dist( point, v0, v1 ) < presque ) interior_right = FALSE; 
            }
        }

    //.....
    // Count the number of crossings with a upward-pointing ray

    crossing = 0;

    for ( i=1; i < p->len; i++ )
        {
        v0 = vertex( p, i-1 );
        v1 = vertex( p, i   );
        if ( ! cutline[i] && onthetop( point, v0, v1 ) ) 
            {
            //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", i-1, i, v0.x, v0.y, v1.x, v1.y );
            crossing++;
            }
        }

    v1 = vertex( p, p->len - 1 );
    v0 = vertex( p, 0 );
    if ( ! cutline[0] && onthetop( point, v0, v1 ) ) 
        {
        //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", 0, (p->len - 1), v0.x, v0.y, v1.x, v1.y );
        crossing++;
        }

    //printf( "\nCROSSINGS = %d\n", crossing );

    interior_top = (crossing % 2); 

    // if the point is allegedly interior, check that the point is far enough from the edge

    if ( interior_top && no_wallflowers )
        {
        for ( i=0; i < p->len && interior_top; i++ )
            { 
            v0 = vertex( p, i );
            if ( SAMEPOINT( point, v0 ) ) interior_top = FALSE; 
            }

        for ( i=0; i < p->len && interior_top; i++ )
            {
            i1 = (i + 1) % p->len;
            v0 = vertex( p, i  );
            v1 = vertex( p, i1 );
            if ( between( point.x, v0.x, v1.x ) && between( point.y, v0.y, v1.y ) )
                if ( point2line_dist( point, v0, v1 ) < presque ) interior_top = FALSE; 
            }
        }


    //.....
    // Count the number of crossings with a downward-pointing ray

    crossing = 0;

    for ( i=1; i < p->len; i++ )
        {
        v0 = vertex( p, i-1 );
        v1 = vertex( p, i   );
        if ( ! cutline[i] && onthebottom( point, v0, v1 ) ) 
            {
            //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", i-1, i, v0.x, v0.y, v1.x, v1.y );
            crossing++;
            }
        }

    v1 = vertex( p, p->len - 1 );
    v0 = vertex( p, 0 );
    if ( ! cutline[0] && onthebottom( point, v0, v1 ) ) 
        {
        //printf( "[%d-%d] %ld,%ld to %ld,%ld \n", 0, (p->len - 1), v0.x, v0.y, v1.x, v1.y );
        crossing++;
        }

    //printf( "\nCROSSINGS = %d\n", crossing );

    interior_bottom = (crossing % 2); 

    // if the point is allegedly interior, check that the point is far enough from the edge

    if ( interior_bottom && no_wallflowers )
        {
        for ( i=0; i < p->len && interior_bottom; i++ )
            { 
            v0 = vertex( p, i );
            if ( SAMEPOINT( point, v0 ) ) interior_bottom = FALSE; 
            }

        for ( i=0; i < p->len && interior_bottom; i++ )
            {
            i1 = (i + 1) % p->len;
            v0 = vertex( p, i  );
            v1 = vertex( p, i1 );
            if ( between( point.x, v0.x, v1.x ) && between( point.y, v0.y, v1.y ) )
                if ( point2line_dist( point, v0, v1 ) < presque ) interior_bottom = FALSE; 
            }
        }

    vote = 0;
    if ( interior_left   ) vote++;
    if ( interior_right  ) vote++;
    if ( interior_top    ) vote++;
    if ( interior_bottom ) vote++;
  
    interior = (vote > 2);  // 2 works well for the stanford zone plate

    return ( interior );
    }



//---------end inside_point--------------------------------------------------------------------------------------------

BOOL                                                    
they_touch( polygon *h, polygon *p )
    {
    static int ih, ip;
    static xy  vh, vp, vp1;

    for ( ih=0; ih < h->len; ih++ )
        {
	vh = vertex( h, ih );
        for ( ip=0; ip < p->len; ip++ )
	    {
	    vp = vertex( p, ip );
            if ( SAMEPOINT( vp, vh ) )
		{
		printf( "\nhole and polygon touch at vertex %ld, %ld\n", vp.x, vp.y );
		return( TRUE );
		}
	    }
	}

    for ( ih=0; ih < h->len; ih++ )
	{
	vh = vertex( h, ih );
	for( ip=0; ip < p->len; ip++ )
	    {
	    vp  = vertex( p, ip );
	    vp1 = vertex( p, (ip+1) % p->len );
	    if ( point2line_dist( vh, vp, vp1 ) < nearness )
		{
                if ( between( vh.x, vp.x, vp1.x ) && between( vh.y, vp.y, vp1.y ) )
		    {
		    printf( "\nhole vertex %ld, %ld touches the polygon\n", vh.x, vh.y );
		    return( TRUE );
		    }
		}
	    }
	}

    return( FALSE );

    }


//---------end they_touch -----------------------------------------------------------------------------------------------

BOOL                                                         
point_touches_polygon( xy v, polygon *p )
    {
    static xy vp, vp1;
    static int ip;

    for( ip=0; ip < p->len; ip++ )
        {
        vp  = vertex( p, ip );
        vp1 = vertex( p, (ip+1) % p->len );

        if ( SAMEPOINT( v, vp ) || SAMEPOINT( v, vp1 ) ) return( TRUE );

        if ( point2line_dist( v, vp, vp1 ) < nearness ) 
	    {
            if ( between( v.x, vp.x, vp1.x ) && between( v.y, vp.y, vp1.y ) )
		{
		//printf( "\n......... point %ld, %ld touches the polygon\n", v.x, v.y );
		return( TRUE );
		}
	    }
	}
    return( FALSE );
    }

//---------end point_touches_polygon ------------------------------------------------------------------------------------

BOOL                                                                      
on_segment( xy p, xy a1, xy a2, xy b1, xy b2, BOOL overlapsbad )
    {

    // Is the point p somewhere on both line segments a1-a2 and b1-b2 ?

    static BOOL cross;

    if ( ! overlapsbad )   // p is not on a segment if the intersection is an endpoint, but yes if complete overlap
        {
        if ( (NEAR( a1, b1 ) && NEAR( a2, b2 )) || (NEAR( a1, b2 ) && NEAR( a2, b1 )) ) return( TRUE );
        if ( NEAR( p, a1 ) || NEAR( p, a2 ) || NEAR( p, b1 ) || NEAR( p, b2 ) ) return( FALSE );
        }

    if ( NEAR( p, a1 ) && (NEAR( p, b1 ) || NEAR( p, b2 )))
        return( TRUE );

    if ( NEAR( p, a2 ) && (NEAR( p, b1 ) || NEAR( p, b2 )))
        return( TRUE );

    if ( EQUAL(a1.y, a2.y) ) 
        cross = between( p.x, a1.x, a2.x ) && between( p.y, b1.y, b2.y );
    else if ( EQUAL(b1.y, b2.y) )
        cross = between( p.x, b1.x, b2.x ) && between( p.y, a1.y, a2.y );
    else if ( EQUAL(a1.x, a2.x) )
        cross = between( p.x, b1.x, b2.x ) && between( p.y, a1.y, a2.y );
    else if ( EQUAL(b1.x, b2.x) )
        cross = between( p.x, a1.x, a2.x ) && between( p.y, b1.y, b2.y );
    else
        cross = between( p.x, b1.x, b2.x ) && between( p.y, b1.y, b2.y ) &&
                between( p.x, a1.x, a2.x ) && between( p.y, a1.y, a2.y );

    return( cross );

    }  // on_segment

//----------------------------------------------------------------------------------------------


void                                                                     
read_gds_units( char *ifilename, double *dbu_uu, double *dbu_m )   
    {
    // previously named "count_polygons"

    double
        dbu_um;

    int i, fd;

    unsigned short
        count,           
        token;           

    fd = open( ifilename, 0 );
    if ( fd <= 0 ) bail( "idarn: unable to open input file" );

    while ( read( fd, &count, 2 ) )
        {
        gds_swap2bytes( (BYTE *) &count );

        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );
            
            if ( token == 0x0305 )  // UNITS
                {
                gds_read_units( fd, &dbu_um, dbu_uu, dbu_m, 0 );
                close( fd );
                return;   
                }
            else
                {
                SKIPOVER( fd, count );
                }
            }
        }
    bail( "cannot find gds units - invalid gds file" );
    } 

//-----------------end read_gds_units----------------------------------------------------------------------------------


void                                                    
remove_duplicate_vertices( polygon *p )
    {
    static int  n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
     
    if ( ! p->active ) return;

    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
        {
        vi  = vertex( p, i );
        vi1 = vertex( p, i+1 );
        while ( (i < nm1) && SAMEPOINT( vi, vi1 ) )
            {
            delete_vertex( p, i+1 );
            n = p->len;
            nm1 = n - 1;
            vi  = vertex( p, i );
            if ( i == nm1 )
                vi1 = vertex( p, 0 );
            else 
                vi1 = vertex( p, i+1 );
            }
        i += 1;
        }
    
    n = p->len;
    nm1 = n - 1;

    if ( n > 2 ) 
        {
        v0   = vertex( p, 0 );
        vnm1 = vertex( p, nm1 );
        if ( SAMEPOINT( v0, vnm1 ) ) delete_vertex( p, nm1 ); 
        }

    if ( n < 3 ) p->active = FALSE;

    }

//---------------------------------------------------------------------------------------------------------------------

void                                  
snap_to_grid( polygon *p )
    {
    static int i; 
    static xy v;
    
    for ( i=0; i < p->len; i++ )
        {
        v = vertex( p, i );
        //v.x = (int) bss * floor( v.x / bss );   
        //v.y = (int) bss * floor( v.y / bss );
        v.x = bss * ( v.x / bss );   
        v.y = bss * ( v.y / bss );
        assign_vertex( p, i, v );
        }

    remove_duplicate_vertices( p );

    }

//----------------------------------------------------------------------------------------------------------------------

BOOL
degenerate( polygon *poly )
    {
    static BOOL degen;
    set_shape_extents( poly );
    degen = (abs(poly->ll.x - poly->ur.x) < bss) || (abs(poly->ll.y - poly->ur.y) < bss);
    return( degen );
    }

//----------------------------------------------------------------------------------------------------------------------

void                                                                                             
copy_from_gds_polygon( struct gds_itemtype *item, polygon *poly, double dbu_rescale )
    {
    // copy item to poly
    // poly has already been initialized

    static int i, ix, iy;
    static double x, y;

    // the append function checks for overflow

    if ( dbu_rescale != 1.0 )
	{
	for ( i=0; i < item->n; i++ )
	    {
	    x = item->x[i];
	    y = item->y[i];
	    x = x * dbu_rescale;
	    y = y * dbu_rescale;
	    ix = (int) round( x );
	    iy = (int) round( y );
	    append_polygon( poly, ix, iy );
	    }
        }
    else
	{
	for ( i=0; i < item->n; i++ )
	    {
	    append_polygon( poly, item->x[i], item->y[i] );
	    }
	}
    
    //set_shape_extents( poly );  done in 'degenerate' now

    if ( degenerate( poly ) ) { poly->len = 0; poly->active = FALSE; }

    //snap_to_grid( poly );             // causes problems
    remove_colinear_vertices( poly );

    }  // copy_from_gds_polygon


//-----------------------------------------------------------------------------------------------------

BOOL                                                                                            
not_crossing( polygon *p, int a, int b, int c, int d, xy *intersection, BOOL overlapsbad )
    {
    // Does the segment a-b cross the segment c-d?
    // Return True if they do not cross, and return the 
    // intersection point if they do. 

    // If overlapsbad then we mark overlapping sides as crossing.

    static xy
        inter,
        a1, a2, b1, b2;

    static BOOL
        crossing,
        a_is_vertical,
        b_is_vertical,
        ma_equals_mb,
        ba_equals_bb;

    static double
        ma, mb, ba, bb, num, den,
	a1x, a1y, b1y, b1x, fb;    

    if ( p->len < 3 )
        {
        p->active = FALSE;
        intersection->x = a1.x;
        intersection->y = a1.y;
        return( TRUE );
        }
        
    a1 = vertex( p, a );
    a2 = vertex( p, b );
    b1 = vertex( p, c );
    b2 = vertex( p, d );

    intersection->x = 0;
    intersection->y = 0;

    // Let's do a quick check of the range first, to save a lot of time

    if ( a1.x > b1.x && a1.x > b2.x && a2.x > b1.x && a2.x > b2.x ) return( TRUE );
    if ( a1.x < b1.x && a1.x < b2.x && a2.x < b1.x && a2.x < b2.x ) return( TRUE );
    if ( a1.y > b1.y && a1.y > b2.y && a2.y > b1.y && a2.y > b2.y ) return( TRUE );
    if ( a1.y < b1.y && a1.y < b2.y && a2.y < b1.y && a2.y < b2.y ) return( TRUE );

    if ( ! overlapsbad )  // even if overlaps are ok, this case is too much
        if ( ( SAMEPOINT( a1, b1 ) && SAMEPOINT( a2, b2 ) ) ||
             ( SAMEPOINT( a1, b2 ) && SAMEPOINT( a2, b1 ) ) )    return( FALSE );

    if ( SAMEPOINT( a1, b1 ) ) 
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( FALSE );
        }

    if ( SAMEPOINT( a1, b2 ) ) 
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( FALSE );
        }

    if ( SAMEPOINT( a2, b1 ) ) 
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( FALSE );
        }

    if ( SAMEPOINT( a2, b2 ) ) 
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( FALSE );
        }

    a_is_vertical = EQUAL(a1.x, a2.x); 
    b_is_vertical = EQUAL(b1.x, b2.x); 

    if ( a_is_vertical && b_is_vertical && ! overlapsbad ) return( TRUE );

    if ( a_is_vertical && b_is_vertical )
        {
        //printf( "both line segments are vertical\n" ); 
        if ( EQUAL( a1.x, b1.x ) )
            {
            if ( between( a1.y, b1.y, b2.y ) )
                {
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else if ( between( a2.y, b1.y, b2.y ) )
                {
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else
                {
                return( TRUE );
                }
            }
        else
            {
            return( TRUE );
            }
        }

    if ( a_is_vertical && ! b_is_vertical )
        {
        //printf( "a is vertical but b is not\n" );
        intersection->x = a1.x;
        num = (b1.y - b2.y); 
	den = (b1.x - b2.x);
        mb  = num / den;
	a1x = a1.x;
	b1y = b1.y;
	b1x = b1.x;
	fb = mb * a1x + b1y - mb * b1x;
        intersection->y = (int) round( fb );

        // y is now an integer, but has not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        inter.x = intersection->x;
        inter.y = intersection->y;

        crossing = on_segment( inter, a1, a2, b1, b2, overlapsbad );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            return( FALSE );
            }
        }
        
    if ( b_is_vertical && ! a_is_vertical )
        {
        //printf( "b is vertical but a is not\n" );
        intersection->x = b1.x;
        num = (a1.y - a2.y);
	den = (a1.x - a2.x);
        ma = num / den;
	b1x = b1.x;
	a1y = a1.y;
	a1x = a1.x;
        fb = ma * b1x + a1y - ma * a1x;
        intersection->y = (int) round( fb );
        inter.x = intersection->x;
        inter.y = intersection->y;
        
        crossing = on_segment( inter, a1, a2, b1, b2, overlapsbad );

        if ( ! crossing )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( TRUE );
            }
        else
            {
            return( FALSE );
            }
        }
        
    num = (a1.y - a2.y);
    den = (a1.x - a2.x);
    ma = num / den;
    num = (b1.y - b2.y);
    den = (b1.x - b2.x);
    mb = num / den;

    b1x = b1.x;
    b1y = b1.y;
    a1x = a1.x;
    a1y = a1.y;
    
    bb = b1y - mb * b1x;
    ba = a1y - ma * a1x; 

    ma_equals_mb = dequal( ma, mb );
    ba_equals_bb = dequal( ba, bb );

    if ( ma_equals_mb && ! ba_equals_bb )
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );
        }

    if ( ma_equals_mb )  // coincident lines, but do the segments overlap?
        {
        //printf( "coincident lines\n" );

        if ( ma > 1.0 )
            {
            if ( between( a2.y, b1.y, b2.y ) )
                {
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.y, b1.y, b2.y ) )
                {
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            } 
        else
            {
            if ( between( a2.x, b1.x, b2.x ) )
                {
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( FALSE );
                }
            else if ( between( a1.x, b1.x, b2.x ) )
                {
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( FALSE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( TRUE );
                }
            }
        }
            
    // Finally, just regular lines not on-axis, not parallel

    fb = (bb - ba) / (ma - mb); 
    inter.x = (int) round( fb );
    fb = ma * fb + ba; 
    inter.y = (int) round( fb );

    crossing = on_segment( inter, a1, a2, b1, b2, overlapsbad );    

    if ( crossing )
        {
        intersection->x = inter.x;
        intersection->y = inter.y;
        return( FALSE );
        }
    else
        {
        intersection->x = 0;
        intersection->y = 0;
        return( TRUE );
        }

    }

//------end not_crossing-------------------------------------------------------------------------------

double                                         
distance_between( xy a, xy b )
    {
    static double dx, dy;
    dx = a.x - b.x;
    dy = a.y - b.y;
    return( sqrt( dx * dx + dy * dy ) );
    }

//-----------------------------------------------------------------------------------------------------

BOOL                                                               
crossing( xy a1, xy a2, xy b1, xy b2, xy *intersection )
    {
    // Similar to the previous function not_crossing, but here we have two 
    // line segments instead of parts of a polygon. Also, overlapsbad = TRUE
    // so we will find intersections with coincident lines.

    static xy inter;

    static BOOL
        cross,
        overlapsbad=TRUE,
        a_is_vertical,
        b_is_vertical,
        ma_equals_mb,
        ba_equals_bb;

    static double
        ma, mb, ba, bb,
	num, den,
	a1y, a1x, b1y, b1x, fb;
        
    intersection->x = 0;
    intersection->y = 0;


    if ( ( SAMEPOINT( a1, b1 ) && SAMEPOINT( a2, b2 ) ) || ( SAMEPOINT( a1, b2 ) && SAMEPOINT( a2, b1 ) ) )    
        {
        return( TRUE );
        }

    // First do a quick check of range
    // before we get all picky 

    if ( a1.x > b1.x && a1.x > b2.x && a2.x > b1.x && a2.x > b2.x ) return( FALSE );
    if ( a1.x < b1.x && a1.x < b2.x && a2.x < b1.x && a2.x < b2.x ) return( FALSE );
    if ( a1.y > b1.y && a1.y > b2.y && a2.y > b1.y && a2.y > b2.y ) return( FALSE );
    if ( a1.y < b1.y && a1.y < b2.y && a2.y < b1.y && a2.y < b2.y ) return( FALSE );

    if ( SAMEPOINT( a1, b1 ) ) 
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( TRUE );
        }

    if ( SAMEPOINT( a1, b2 ) ) 
        {
        intersection->x = a1.x;
        intersection->y = a1.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( TRUE );
        }

    if ( SAMEPOINT( a2, b1 ) ) 
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( TRUE );
        }

    if ( SAMEPOINT( a2, b2 ) ) 
        {
        intersection->x = a2.x;
        intersection->y = a2.y;
        //printf( "line segments share a point\n" );
        if ( overlapsbad ) return( TRUE );
        }

    a_is_vertical = EQUAL(a1.x, a2.x); 
    b_is_vertical = EQUAL(b1.x, b2.x); 

    if ( a_is_vertical && b_is_vertical )
        {
        //printf( "both line segments are vertical\n" ); 

        if ( EQUAL( a1.x, b1.x ) )
            {
            if ( between( a1.y, b1.y, b2.y ) )
                {
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( TRUE );
                }
            else if ( between( a2.y, b1.y, b2.y ) )
                {
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( TRUE );
                }
            else
                {
                return( FALSE );
                }
            }
        else
            {
            return( FALSE );
            }
        }

    if ( a_is_vertical && ! b_is_vertical )
        {
        //printf( "a is vertical but b is not\n" );
        intersection->x = a1.x;
        num = (b1.y - b2.y);
	den = (b1.x - b2.x);
        mb = num / den;
        a1x = a1.x;
        b1y = b1.y;
        b1x = b1.x;
	fb  = mb * a1x + b1y - mb * b1x;
        intersection->y = (int) round( fb );

        // y is now an integer, but has not snapped to the bss grid.
        // Is that what we want? Yes, probably.

        inter.x = intersection->x;
        inter.y = intersection->y;

        cross = on_segment( inter, a1, a2, b1, b2, overlapsbad );

        if ( ! cross )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( FALSE );
            }
        else
            {
            return( TRUE );
            }
        }
        
    if ( b_is_vertical && ! a_is_vertical )
        {
        //printf( "b is vertical but a is not\n" );
        intersection->x = b1.x;
	num = (a1.y - a2.y);
	den = (a1.x - a2.x);
        ma  = num / den;
	b1x = b1.x;
	a1y = a1.y;
	a1x = a1.x;
	fb = ma * b1x + a1y - ma * a1x;
        intersection->y = (int) round( fb );
        inter.x = intersection->x;
        inter.y = intersection->y;
        
        cross = on_segment( inter, a1, a2, b1, b2, overlapsbad );

        if ( ! cross )
            {
            intersection->x = 0;
            intersection->y = 0;
            return( FALSE );
            }
        else
            {
            return( TRUE );
            }
        }
        
    num = (a1.y - a2.y);
    den = (a1.x - a2.x);
    ma  = num / den;
    num = (b1.y - b2.y);  
    den = (b1.x - b2.x);
    mb  = num / den;

    b1y = b1.y;
    b1x = b1.x;
    a1y = a1.y;
    a1x = a1.x;
    bb = b1y - mb * b1x;
    ba = a1y - ma * a1x;

    ma_equals_mb = dequal( ma, mb );
    ba_equals_bb = dequal( ba, bb );

    if ( ma_equals_mb && ! ba_equals_bb )
        {
        intersection->x = 0;
        intersection->y = 0;
        return( FALSE );
        }

    if ( ma_equals_mb )  // coincident lines, but do the segments overlap?
        {
        //printf( "coincident lines\n" );

        if ( ma > 1.0 )
            {
            if ( between( a2.y, b1.y, b2.y ) )
                {
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( TRUE );
                }
            else if ( between( a1.y, b1.y, b2.y ) )
                {
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( TRUE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( FALSE );
                }
            } 
        else
            {
            if ( between( a2.x, b1.x, b2.x ) )
                {
                intersection->x = a2.x;
                intersection->y = a2.y;
                return( TRUE );
                }
            else if ( between( a1.x, b1.x, b2.x ) )
                {
                intersection->x = a1.x;
                intersection->y = a1.y;
                return( TRUE );
                }
            else
                {
                intersection->x = 0;
                intersection->y = 0;
                return( FALSE );
                }
            }
        }
            
    // Finally, just regular lines not on-axis, not parallel

    fb = (bb - ba) / (ma - mb); 
    inter.x = (int) round( fb );
    fb = ma * fb + ba; 
    inter.y = (int) round( fb );

    cross = on_segment( inter, a1, a2, b1, b2, overlapsbad );    

    if ( cross )
        {
        intersection->x = inter.x;
        intersection->y = inter.y;
        return( TRUE );
        }
    else
        {
        intersection->x = 0;
        intersection->y = 0;
        return( FALSE );
        }

    }


//-------end crossing-----------------------------------------------------------------------------------


BOOL                                                       
self_crossing_polygon( polygon *p, BOOL overlapsbad )
    {
    static BOOL good;
    static int  n, i, j;
    static xy   ip;

    good = TRUE;
    n = p->len;

    if ( n < 4 ) return( FALSE );

    i = 0;
    while ( good && (i < (n-2)) )
        {
        j = i + 2;
        while ( good && (j < (n-1)) )
            {
            good = not_crossing( p, i, i+1, j, j+1, &ip, overlapsbad );
            j += 1;
            }
        if ( good && (i > 0) )
            {                  
            good = not_crossing( p, i, i+1, n-1, 0, &ip, overlapsbad );
            }
        i += 1;
        }

    // if not good then we should see if this thing is a doughnut

    return( ! good );
    }

//---------end self_crossing_polygon-----------------------------------------------------------------------------------

BOOL                                                   
colinear( xy a, xy b, xy c, int *middle )
    {
    // If the three points are colinear then return TRUE. The return parameter
    // "middle" will indicate which of the three points is in the middle.

    static BOOL first, second, third, vertical, horizontal, colinear;
    
    static int abc, bac, cab;

    first  = FALSE;
    second = FALSE;
    third  = FALSE;

    if ( EQUAL(c.x, a.x) || EQUAL(c.y, a.y) || EQUAL(b.x, a.x) || EQUAL(b.y, a.y) )
        {
        vertical   = EQUAL(a.x, b.x) && EQUAL(b.x, c.x); 
        horizontal = EQUAL(a.y, b.y) && EQUAL(b.y, c.y);
        colinear   = horizontal || vertical;
        if ( horizontal )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second )
                first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        if ( vertical )
            {
            second = between( b.y, a.y, c.y );
            if ( ! second )
                first = between( a.y, b.y, c.y );
            third = (! second) && (! first);
            }
        }
    else
        {
        abc = point2line_dist( a, b, c );
        bac = point2line_dist( b, a, c );
        cab = point2line_dist( c, a, b );

        colinear = TOOCLOSE( abc ) || TOOCLOSE( bac ) || TOOCLOSE( cab ); 

        if ( colinear )
            {
            second = between( b.x, a.x, c.x );
            if ( ! second )
                first = between( a.x, b.x, c.x );
            third = (! second) && (! first);
            }
        }
         
    // It's very difficult to decide which point to delete. Usually, we should delete
    // the middle point. However, if two of the points are very close then we should
    // delete the endpoint, maybe.

    *middle = -1;

    if ( colinear )
        {
        if ( SAMEPOINT( a, b ) )
            { 
            *middle = 0;              // delete the first vertex
            //printf( "\ndelete a\n" );
            }
        else if ( SAMEPOINT( b, c ) )
            {            
            *middle = 2;              // delete the third vertex
            //printf( "\ndelete c\n" );
            }
        else        
            {
            if ( first ) 
                *middle = 0;
            else if ( second ) 
                *middle = 1;
            else if ( third )
                *middle = 2;
            }
        }

    return colinear;
    }

//---------end colinear------------------------------------------------------------------------------------------------

void                                                   
remove_colinear_vertices( polygon *p )
    {
    // Much harder than it might seem. Normally we would want to delete 
    // the vertex in the middle, but if the polygon doubles back on itself
    // then this can cause the polygon to become distorted. This is the
    // newer version that checks for dangling tails, and includes the call
    // to remove_duplicate_vertices.
    
    static int  i, n, nm1, nm2, j, k, previous;
    static int  middle;
    static BOOL escape;
    static xy   v0, v1, v2;
    
    if ( ! p->active ) return;

    previous = p->len;

    i = 0;
    middle = -1;
    n = p->len;
    nm2 = n - 2;
    nm1 = n - 1;

    while ( i < nm2 )
        {
        j = i + 1;
        k = j + 1;
        escape = FALSE;
        while ( ! escape && (i < nm2) && colinear( vertex(p,i), vertex(p,j), vertex(p,k), &middle ) )
            {
            if ( middle == 1 )
                delete_vertex( p, j );
            else
                {
                escape = TRUE;
                if ( SAMEPOINT( vertex(p,i), vertex(p,k) ) ) delete_vertex( p, j );
                }
            n = p->len;
            nm1 = n - 1;
            nm2 = n - 2;
            }    
        i = i + 1;
        }   

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;

    if ( n > 2 )
        {
        v2 = vertex( p, nm2 );
        v1 = vertex( p, nm1 );
        v0 = vertex( p, 0   );
        if ( colinear( v2, v1, v0, &middle ) )
            {
            if ( middle == 1 )          
                delete_vertex( p, nm1 );
            else if ( SAMEPOINT( v0, v2 ) )
                delete_vertex( p, nm1 );
            }
        }
    
    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;
        
    if ( n > 2 )
        {
        if ( colinear( vertex(p,n-1), vertex(p,0), vertex(p,1), &middle ) )
                delete_vertex( p, 0 );
        }
    
    // check for dangling stubs - could be generalized to delete longer tails,
    // but does that ever happen?
    
    n = p->len;
    nm1 = n - 1;
    if ( n > 2 )
        {
        v0 = vertex( p, 1 );
        v1 = vertex( p, nm1 );
        if ( SAMEPOINT( v0, v1 ) ) delete_vertex( p, 0 );
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;
    if ( n > 3 )
        {
        v0 = vertex( p, 0 );
        v2 = vertex( p, nm2 );
        if ( SAMEPOINT( v0, v2 ) ) delete_vertex( p, nm1 );
        }

    n = p->len;
    nm1 = n - 1;
    nm2 = n - 2;
    if ( n > 3 )
        {
        v0 = vertex( p, 1 );
        v2 = vertex( p, nm1 );
        if ( SAMEPOINT( v0, v2 ) ) delete_vertex( p, 0 );
        }
    
    remove_duplicate_vertices( p );

    if ( p->len < previous )
        {
        recursion++;
        remove_colinear_vertices( p );
        }
    else
        {
        //printf( "removed all colinear vertices after %d calls\n", recursion );
        recursion = 0;
        return;
        }
    }

//-----------end remove_colinear_vertices------------------------------------------------------------------------------


void                                                                     
copy_to_gds_polygon( polygon *poly, struct gds_itemtype *item )
    {
    // copy poly to item 
    // item has already been allocated 

    static int i;
    static xy  v;

    if ( ! poly->active || poly->len < 3 ) return;

    snap_to_grid( poly );
    remove_colinear_vertices( poly );

    if ( ! poly->active || poly->len < 3 ) return;

    if ( item->n < poly->len ) bail( "attempt to copy large poly into small gds item" );

    for ( i=0; i < poly->len; i++ )
        {
        v = vertex( poly, i );
        item->x[i] = (int) floor( v.x );
        item->y[i] = (int) floor( v.y );
        }

    item->layer = poly->layer;
    item->dt    = poly->dt;
    item->n     = poly->len;
    
    }  // copy_to_gds_polygon


//----------------------------------------------------------------------------------------------

polygon *                                 
double_polygon( polygon *p )                        // unused 
    {
    static xy  a, b, m;
    static int i, n1;
    polygon   *d;

    if ( p->len < 3 || ! p->active ) return( p );
    n1 = p->len - 1;
    d = initialize_polygon( 2 * p->len, p->layer, p->dt, p->obese );

    for ( i=0; i < n1; i++ )
        {
        a = vertex( p, i   );
        b = vertex( p, i+1 );
        m.x = (a.x + b.x) / 2;
        m.y = (a.y + b.y) / 2;
        append_polygon_xy( d, a );
        append_polygon_xy( d, m );
        }
    a = vertex( p, n1 );
    b = vertex( p, 0  );
    m.x = (a.x + b.x) / 2;
    m.y = (a.y + b.y) / 2;
    append_polygon_xy( d, a );
    append_polygon_xy( d, m );
    remove_duplicate_vertices( d );
    return( d );
    }

//-------------------------------------------------------------------------------------

polygon_pair                                                            
split_touchy_polygon( polygon *p, int v, int a, int b )
    {

    // Vertex v touches line segment p[a]--p[b]. Split p into two polygons. 
    // One will be the outside (mostly) and the other will be the hole (hopefully).
    // Elsewhere, in the calling function, these will go back onto the queue
    // for further splitting. After all, we might have multiple holes, or
    // lakes on islands. It's gastly what people will do.

    static int      n, i, start, finish;
    static BOOL     init=FALSE;
    polygon        *p1, *p2;
    static polygon *empty;
    polygon_pair    p1p2;
    

    if ( ! init )
        {
        init = TRUE;
        empty = initialize_polygon( 0, 0, 0, 0 );
        }

    n = p->len; 
    if ( n < 3 ) // could it be that splitting touchy polygons causes rare fill-in problems?
        {
        p1p2.a = p;
        p1p2.b = empty;
        return( p1p2 );
        }

    if ( a == b ) bail( "degenerate segment sent to split_touchy_polygon" );

    //printf( "P:\n" );
    //print_polygon( p, 1 );

    p1 = initialize_polygon( p->len + 2, p->layer, p->dt, p->obese );
    p2 = initialize_polygon( p->len + 2, p->layer, p->dt, p->obese );

    p1p2.a = p1;
    p1p2.b = p2;

    i = v;

    while ( (i != b) && (i != a) )
        {
        append_polygon_xy( p1, vertex( p, i ) );
        i = (i + 1) % n;
        }
    
    append_polygon_xy( p1, vertex( p, i ) );

    start = v - 1;

    if ( i == a )
        finish = b;
    else
        finish = a;

    i = start;
    if ( i == -1 ) i = n-1;
    while ( i != finish )
        {
        append_polygon_xy( p2, vertex( p, i ) );
        i -= 1;
        if ( i == -1 ) i = n-1;
        }

    append_polygon_xy( p2, vertex( p, i ) ); 
    append_polygon_xy( p2, vertex( p, v ) );

    remove_colinear_vertices( p1 );
    remove_colinear_vertices( p2 );
    
    if ( p2->len > 2 )  // mark the doughnut hole. but what if we are wrong?
        {
	if ( there_is_an_inside_point( p1, p2 ) )  
	//if ( all_hole_vertices_are_inside_poly( p1, p2 ) )
            {
            p1->layer = 9999; 
            if ( p2->layer < 9999 ) p2->layer = p2->layer + 10000;
            }
	else if ( there_is_an_inside_point( p2, p1 ) )
	//else if ( all_hole_vertices_are_inside_poly( p2, p1 ) )
            {
            p2->layer = 9999;              // bogus layer marks a hole
            if ( p1->layer < 9999 ) p1->layer = p1->layer + 10000;
            }
        //else
        //    printf( "\n------------ This is not a doughnut --------------\n" );
        }

    p1p2.a = p1;
    p1p2.b = p2;

    return( p1p2 );
    }

//---------end split_touchy_polygon------------------------------------------------------------------------------------


polygon_pair                                                             
untie_bow( polygon *p, xy ip, int i1, int i2, int j1, int j2 )
    {

    // Polygon p crosses itself on segments i1-i2 and j1-j2.
    // The intersection point is ip. Split p into two parts, then 
    // send back p1 and p2.

    // It might not be a bowtie, really. See below- it might
    // be a very weird doughnut.

    static int n, i, t;

    polygon *p1, *p2;

    polygon_pair p1p2;

    p1 = initialize_polygon( p->len, p->layer, p->dt, p->obese );
    p2 = initialize_polygon( p->len, p->layer, p->dt, p->obese );

    n = p->len;

    if ( i2 < i1 )
        {
        t  = i1;
        i1 = i2;
        i2 = t;
        }

    if ( i2 == (n-1) )
        {
        t  = i1;
        i1 = i2;
        i2 = t;
        }

    if ( j2 < j1 )
        {
        t  = j1;
        j1 = j2;
        j2 = t;
        }

    if ( j2 == (n-1) )
        {
        t  = j1;
        j1 = j2;
        j2 = t;
        }

    i = i2;
    while ( (i != j1) && (i != j2) ) 
        {
        append_polygon_xy( p1, vertex( p, i ) );
        i += 1;
        if ( i == n ) i = 0;
        }
    append_polygon_xy( p1, vertex( p, i ) );
    append_polygon_xy( p1, ip );

    i = j2;
    while ( (i != i1) && (i != i2) )
        {
        append_polygon_xy( p2, vertex( p, i ) );
        i += 1;
        if ( i == n ) i = 0;
        }
    append_polygon_xy( p2, vertex( p, i ) );
    append_polygon_xy( p2, ip ); 

    // but... this might not be a real bowtie. It might be
    // one side lying on top of another side, without any
    // identical vertices. criminy. So now we have to
    // look at the two polygons to see if one is inside 
    // the other.
    
    if ( p2->len > 2 )  // mark the doughnut hole if it's a hole
        {
	//if ( there_is_an_inside_point( p1, p2 ) )                 // this is dubious. mark change 1.5 
	if ( all_hole_vertices_are_inside_poly( p1, p2 ) )
            {
            p1->layer = 9999; 
            if ( p2->layer < 9999 ) p2->layer = p2->layer + 10000;
            }
	//else if ( there_is_an_inside_point( p2, p1 ) )
	else if ( all_hole_vertices_are_inside_poly( p2, p1 ) )
            {
            p2->layer = 9999;              // bogus layer marks a hole
            if ( p1->layer < 9999 ) p1->layer = p1->layer + 10000;
            }
        //else
        //    printf( "\n------------ This is not a doughnut --------------\n" );
        }


    p1p2.a = p1;
    p1p2.b = p2;
    return( p1p2 );

    }

//-----end untie_bow---------------------------------------------------------------------------------------------------


polygon_pair                                    
fix_bowties( polygon *p )
    {

    static BOOL     good, init=FALSE;
    static int      n, i, j, i1, i2, j1, j2;
    static xy       ip, vi, vj, vi1, vj1;
    
    static polygon *empty;
    polygon_pair    p1p2;

    if ( ! init )
        {
        init = TRUE;
        empty = initialize_polygon( 0, 0, 0, 0 );
        }

    if ( ! p->active )
        {
        p1p2.a = p;
        p1p2.b = empty;  
        return( p1p2 );
        }

    remove_colinear_vertices( p );

    good = TRUE;
    n = p->len;
    if ( n < 4 ) 
        {
        p1p2.a = p;
        p1p2.b = empty;
        return( p1p2 );
        }


    i = 0;
    while ( good && (i < (n-2)) )
        {
        j = i + 2;
        while ( good && (j < (n-1)))
            {
            good = not_crossing( p, i, i+1, j, j+1, &ip, TRUE );

	    if ( ! good )
		{
		vi  = vertex( p, i );
		vi1 = vertex( p, i+1 );
		vj  = vertex( p, j );
		vj1 = vertex( p, j+1 );
		good = ! SAMEPOINT( ip, vi ) && ! SAMEPOINT( ip, vj ) && ! SAMEPOINT( ip, vi1 ) && ! SAMEPOINT( ip, vj1 );
		}

            if ( ! good )
                {
                i1 = i;
                i2 = i+1;
                j1 = j;
                j2 = j+1;
                }
            j += 1;
            }
        if ( good && (i > 0) )
            {                  
            good = not_crossing( p, i, i+1, n-1, 0, &ip, TRUE );

	    if ( ! good )
		{
		vi  = vertex( p, i );
		vi1 = vertex( p, i+1 );
		vj  = vertex( p, j );
		vj1 = vertex( p, j+1 );
		good = ! SAMEPOINT( ip, vi ) && ! SAMEPOINT( ip, vj ) && ! SAMEPOINT( ip, vi1 ) && ! SAMEPOINT( ip, vj1 );
		}

            if ( ! good )
                {
                i1 = i;
                i2 = i+1;
                j1 = n-1;
                j2 = 0;
                }
            }
        i += 1;
        }

    //printf( "\nintersection: %g, %g\n", ip.x, ip.y );

    if ( ! good )
	{
        printf( "\n....................................Bow tie found. [%d-%d] crosses [%d-%d]\n", i1,i2,j1,j2 );
        p1p2 = untie_bow( p, ip, i1, i2, j1, j2 );
        if ( p1p2.b->len < 3 ) printf( "....................................the split was not successful\n" );
        return( p1p2 );
        }
    else
        {
        p1p2.a = p;
        p1p2.b = empty;  
        }
    
    return( p1p2 );
    }

//---------------------------------------------------------------------------------------------------------------------

polygon_pair                                           
look_for_self_crossing_polygon( polygon *p )
    {

    static BOOL     good, init=FALSE;
    static int      n, i, j, i1, i2, j1, j2;
    static xy       ip, vi, vj;
    
    static polygon *empty;
    polygon_pair    p1p2;

    if ( ! init )
        {
        init = TRUE;
        empty = initialize_polygon( 0, 0, 0, 0 );
        }

    if ( ! p->active )
        {
        p1p2.a = p;
        p1p2.b = empty;  
        return( p1p2 );
        }

    remove_colinear_vertices( p );

    good = TRUE;
    n = p->len;
    if ( n < 4 ) 
        {
        p1p2.a = p;
        p1p2.b = empty;
        return( p1p2 );
        }

    // If the polygon has two identical vertices then it's either a bowtie or a hole.
    // For non-crazy shapes we could look at the curl of the polygons. But some shapes
    // are nuts, so instead we try to see if one part is inside the other.

    good = TRUE;
    i = 0;
    n = p->len;
    while ( good && (i < n) )
	{
        vi = vertex( p, i % n );
	j = i + 1;
        while ( good && (j < n) )  
	    {
	    vj = vertex( p, j % n );
	    good = ! SAMEPOINT( vi, vj );
            j++;
	    }
        i++;        
	}

    i--;
    j--;

    if ( ! good ) 
	{
	p1p2 = split_touchy_polygon( p, i, j, (j+1) % n );
	return( p1p2 );
	}


    // Next we look for vertices that touch sides. These will not always be bowties.

    i = 0;
    while ( good && (i < (n-2)) )
        {
        j = i + 2;
        while ( good && (j < (n-1)))
            {
            good = not_crossing( p, i, i+1, j, j+1, &ip, TRUE );
            if ( ! good )
                {
                i1 = i;
                i2 = i+1;
                j1 = j;
                j2 = j+1;
                }
            j += 1;
            }
        if ( good && (i > 0) )
            {                  
            good = not_crossing( p, i, i+1, n-1, 0, &ip, TRUE );
            if ( ! good )
                {
                i1 = i;
                i2 = i+1;
                j1 = n-1;
                j2 = 0;
                }
            }
        i += 1;
        }

    //printf( "\nintersection: %g, %g\n", ip.x, ip.y );

    if ( ! good )
	{
	  //printf( "\n....................................Bow tie found. [%d-%d] crosses [%d-%d]\n", i1,i2,j1,j2 );
        p1p2 = untie_bow( p, ip, i1, i2, j1, j2 );
        if ( p1p2.b->len < 3 ) printf( "....................................the split was not successful\n" );
        return( p1p2 );
        }
    else
        {
        p1p2.a = p;
        p1p2.b = empty;  
        }
    
    return( p1p2 );
    }

//----- end look_for_self_crossing_polygon-----------------------------------------------------------------------------

BOOL                                                
ccw( polygon *p, double *area )
    {
    // returns truth of <counter clockwise>
    // Determine the clock wisdom of a polygon
    // by calculating the area. This will be
    // invalid for self-crossing polygons or
    // polygons with holes.

    static int    n;
    static double c, vix, viy, vjx, vjy;
    static int    i, j;
    static xy     vi, vj;

    n = p->len;
    c = 0.0;
    for ( i=0; i < n; i++ )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
	vix = vi.x;
	viy = vi.y;
	vjx = vj.x;
	vjy = vj.y;
        c = c + vix * vjy - vjx * viy;
        }
    *area = fabs( c / 2.0 );
    return( c > 0.0 );

    } // end ccw

//----------------------------------------------------------------------------------------------------------------

BOOL                                            
self_crossing( polygon *p )
    {
    static BOOL good, overlapsbad=TRUE;
    static int  n, i, j;
    static xy   ip;
    
    good = TRUE;
    n = p->len;

    if ( n < 4 ) return( FALSE );
    
    i = 0;
    while ( good && (i < (n-2)) )
        {
        j = i + 2;
        while ( good && (j < (n-1)) )
            {
            good = not_crossing( p, i, i+1, j, j+1, &ip, overlapsbad );
            j += 1;
            }
        if ( good && (i > 0) )
            {                  
            good = not_crossing( p, i, i+1, n-1, 0, &ip, overlapsbad );
            }
        i += 1;
        }

    return( ! good );
    }

//---------end self_crossing-----------------------------------------------------------------------------------

BOOL                                                    
neg_index_change( int a, int b, int n )
    {
    return ( ! ( b == ((a+1) % n)) );
    }


//--------------------------------------------------------------------------------------------------------------

void                                                                               
split_up( queue *q, polygon *outside, polygon *hole, 
          int hole_vertex_h, int out_vertex_h1, int out_vertex_h2, xy newhv,
          int hole_vertex_v, int out_vertex_v1, int out_vertex_v2, xy newvv, 
          queue **gob1, queue **gob2  )
    {
    // Split a polygon by making two cuts at the given hole: one horizontal and one vertical, 
    // extending from the hole to the outer polygon. After creating two new polygons,
    // assign the remaining holes to the appropriate polygon. We need 'q' so that we
    // have a list of all the holes. Output two new polygon gobs, each of which may
    // contain additional holes. It's likely that only one gob will still contain holes,
    // but we must not presume.

    static int          lh;
    static polygon     *p, *p1, *p2;
    static polygon_ref *pr;
    static int          oi, polylen, i, newhvi, newvvi;
    static BOOL         found;
    static BOOL         wallflowers=TRUE;
    static int          nnag=0;
    static xy           v1, v2, hvh, hvv, v;
           polygon     *newout;
           polygon_pair p1p2;

    if ( ! outside->active ) return;
 
    hvh = vertex( hole, hole_vertex_h );
    hvv = vertex( hole, hole_vertex_v );

    if ( ! CLOSE( hvh.y, newhv.y ) )
        {
        printf( "\nBad cut from %ld, %ld to %ld, %ld", hvh.x, hvh.y, newhv.x, newhv.y ); 
        write_queue( fdo, q );
        bail( "horizontal cut line is off-axis" );
        }
    if ( ! CLOSE( hvv.x, newvv.x ) )
        {
        printf( "\nBad cut from %ld, %ld to %ld, %ld", hvv.x, hvv.y, newvv.x, newvv.y ); 
        write_queue( fdo, q );
        bail( "vertical cut line is off-axis" );
        }

    // We have already used number_of_holes_in() to check the clock wisdom.
    // Everything should be ccw.

    if ( outside->len < 3 ) bail( "degenerate polygon in split_up" );
    if ( hole->len    < 3 ) bail( "degenerate hole in split_up" );

    if ( hole_vertex_h == hole_vertex_v ) bail( "we cannot use the same hole vertex for both cuts" );

    if ( neg_index_change( out_vertex_h1, out_vertex_h2, outside->len ) ) bail( "h2 index preceeds h1" );
    if ( neg_index_change( out_vertex_v1, out_vertex_v2, outside->len ) ) bail( "v2 index preceeds v1" );
 
    // Create a new outside polygon containing newvv and newhv.
    // We know that newhv goes between v1 and v2, and that newvv goes between h1 and h2.
    // But if h1==v1 and h2==v2 then it's pretty confusing. So instead of trying 
    // to keep track of who goes where, we start by inserting the new vertices
    // newhv and newvv into the outside polygon.
 
    newout = initialize_polygon( outside->len + 4, outside->layer, outside->dt, FALSE );

    if ( out_vertex_v1 == out_vertex_h1 )
        {
        if ( out_vertex_v2 != out_vertex_h2 ) bail( "why is v1=h1 but v2 <> h2 ?" );      
        v1 = vertex( outside, out_vertex_v1);
        v2 = vertex( outside, out_vertex_v2);

        append_polygon_xy( newout, v1 );

        if ( distance_between( v1, newhv ) < distance_between( v1, newvv ) )
            {
            append_polygon_xy( newout, newhv );
            newhvi = newout->len - 1;
            append_polygon_xy( newout, newvv );
            newvvi = newout->len - 1;
            }
        else 
            {
            append_polygon_xy( newout, newvv );  
            newvvi = newout->len - 1;
            append_polygon_xy( newout, newhv );  
            newhvi = newout->len - 1;
            }
        append_polygon_xy( newout, v2 ); 
        oi = out_vertex_v2;
        while ( oi != out_vertex_v1 )
            {
            oi = (oi + 1) % outside->len;
            if ( oi != out_vertex_v1 ) append_polygon_xy( newout, vertex( outside, oi )); 
            }        
        }
    else 
        {
        oi = out_vertex_h1;
        append_polygon_xy( newout, vertex( outside, oi ) ); 
        append_polygon_xy( newout, newhv ); 
        newhvi = newout->len - 1;
        oi = out_vertex_h2;
        append_polygon_xy( newout, vertex( outside, oi ) ); 
        while ( oi != out_vertex_v1 )
            {
            oi = (oi + 1) % outside->len;
            append_polygon_xy( newout, vertex( outside, oi ) ); 
            }
        append_polygon_xy( newout, newvv ); 
        newvvi = newout->len - 1;
        oi = out_vertex_v2;
        append_polygon_xy( newout, vertex( outside, oi ) ); 
        while ( oi != out_vertex_h1 )
            {
            oi = (oi + 1) % outside->len;
            if ( oi != out_vertex_h1 ) append_polygon_xy( newout, vertex( outside, oi ) ); 
            }
        }

    *gob1 = initialize_queue();
    *gob2 = initialize_queue();

    polylen = (newout->len + hole->len) * 2;

    p1 = initialize_polygon( polylen, outside->layer, outside->dt, FALSE );
    p2 = initialize_polygon( polylen, outside->layer, outside->dt, FALSE );

    oi = newvvi;
    append_polygon_xy( p1, vertex( newout, oi ) );
    while ( oi != newhvi )
        {
        oi = (oi + 1) % newout->len;
        append_polygon_xy( p1, vertex( newout, oi ) );
        }

    oi = hole_vertex_h;
    append_polygon_xy( p1, vertex( hole, oi ) );
    while ( oi != hole_vertex_v )
        {
        oi = decrement( oi, hole->len );
        append_polygon_xy( p1, vertex( hole, oi ) );
        }

    //.........

    oi = newhvi;
    append_polygon_xy( p2, vertex( newout, oi ) );
    while ( oi != newvvi )
        {
        oi = (oi + 1) % newout->len;
        append_polygon_xy( p2, vertex( newout, oi ));
        }
    oi = hole_vertex_v;
    append_polygon_xy( p2, vertex( hole, oi ) );
    while ( oi != hole_vertex_h )
        {
        oi = decrement( oi, hole->len );
        append_polygon_xy( p2, vertex( hole, oi ) );
        }

    set_shape_extents( p1 );
    set_shape_extents( p2 );

    remove_colinear_vertices(  p1 );
    remove_duplicate_vertices( p1 );
    remove_colinear_vertices(  p2 );
    remove_duplicate_vertices( p2 );

    // Check for self-crossings, which would be bad.
    // If we split the polygon then we must also separate the gobs.
    // But we cannot separate the gobs here because we do not have the qq.
    // The gobs will be separated in cut_one

    
    p1p2 = look_for_self_crossing_polygon( p1 );
    if ( p1p2.b->len > 2 )  
        {
	//nag( "split_up p1 is self crossing", &nnag );
        set_shape_extents( p1p2.a );
        set_shape_extents( p1p2.b );
        append_queue( *gob1, p1p2.a );
        append_queue( *gob1, p1p2.b );
        p1->active = FALSE;  
        }
    else
        append_queue( *gob1, p1 );

    p1p2 = look_for_self_crossing_polygon( p2 );
    if ( p1p2.b->len > 2 )
        {
        set_shape_extents( p1p2.a );
        set_shape_extents( p1p2.b );
        //nag( "split_up p2 is self crossing", &nnag );
        append_queue( *gob2, p1p2.a );
        append_queue( *gob2, p1p2.b );
        p2->active = FALSE; 
        }
    else
        append_queue( *gob2, p2 );

    hole->active = FALSE;
    outside->active = FALSE;
    
    // Assign the leftover holes to the appropriate polygons
    // This is strange. Why do we have to look at p1 and p2 
    // even if they are inactive? We had to disable the 'active'
    // check in the polyq vertex() function just to make this work.
    
    pr = q->first;    
    while ( pr )
        {
        p = pr->p;
        if ( p->active )
	    {
            if ( (p->layer == 9999) && (p != hole) )
                {
                if ( inside_point( vertex( p, 0 ), p1, wallflowers ) ) 
                    {
                    append_queue( *gob1, p );  
                    if ( inside_point( vertex( p, 0 ), p2, wallflowers ) ) 
                        {
                        p2->layer = 2;
                        append_queue( *gob1, p2 );
                        write_queue( fdo, *gob1 );
                        //print_polygon( p, 1000 );
                        v = vertex( p, 0 );
                        printf( "\nSomewhere around %ld, %ld \n", v.x, v.y );
                        nag( "a leftover hole seems to be inside both split_up polygons", &lh );
                        }
                    }
                else if ( inside_point( vertex( p, 0 ), p2, wallflowers ) )
    		    {
                    append_queue( *gob2, p );
    		    }
                else
                    {
                    found = FALSE;
                    for ( i = 1; i < p->len && ! found; i++ )
                        {
                        found = inside_point( vertex( p, i ), p1, wallflowers );
                        if ( found ) 
                            append_queue( *gob1, p );
                        else
                            {
                            found = inside_point( vertex( p, i ), p2, wallflowers );
                            if ( found ) append_queue( *gob2, p );
                            }
                        }
                    if ( ! found ) 
                        {
                        printf( "\nHole:\n" );
                        print_polygon( p, 1000 );
                        //printf( "\nPolygons:\n" );
                        //print_polygon( p1, 1000 );
                        //print_polygon( p2, 1000 );
                        nag( "a leftover hole does not appear to be inside either split_up part", &nnag );
                        }
                    }
                }
            }
        pr = pr->next;
        }
    free( newout );
    }

//------------end split_up-------------------------------------------------------------------------------------

int                                        
number_of_holes_in( queue *gob )
    {
    static polygon_ref *pr;
    static polygon     *p;
    static int          n;
    static double       area;

    if ( ! gob ) return( 0 );
    pr = gob->first;
    n = 0;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) ) n++;
        if ( ! ccw( p, &area ) ) reverse_polygon( p );
        pr = pr->next;
        }
    return( n );
    }
    


BOOL                                                                                 
it_does_not_hit_holes_in( queue *gob, polygon *hole, xy hv, int hvi, xy pv ) 
    {
    // Does the line hv-pv hit any of the holes in gob?
    // hv is a vertex on the hole. pv is a point on the outer polygon.
    // We exclude hv itself from consideration, since of course
    // it hits the hole. If pv is on a hole, that would be strange.
    // And yet it happens- the hole might butt up against the polygon.
    // We take care of this special case in find_hor_cut_betwen 
    // and find_ver_cut_between.

    static BOOL         hit;
    static polygon_ref *pr;
    static polygon     *p;
    static int          i, ip1;
    static xy           a, b, inter;
    
    
    // First check out the target hole itself.
    // Does the line intersect it?

    hit = FALSE;
    for ( i=0; (i < hole->len) && ! hit; i++ )
        {
        ip1 = (i + 1) % hole->len;
        if ( ! (i == hvi || ip1 == hvi) )   // exclude the starting vertex   
            {
            a = vertex( hole, i   );
            b = vertex( hole, ip1 );
            hit = crossing( a, b, hv, pv, &inter );
            }
        }

    // now consider all the other holes

    pr = gob->first;
    while ( pr && ! hit )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) && (p != hole) ) // for each hole
            {
            for ( i=0; (i < p->len) && ! hit; i++ )
                {
                ip1 = (i + 1) % p->len;
                a = vertex( p, i   );
                b = vertex( p, ip1 );
                hit = crossing( a, b, hv, pv, &inter );
                }
            }
        pr = pr->next;
        }

    return( ! hit );
    }

//------------end it_does_not_hit_holes_in---------------------------------------------------------------------

void                                               
write_queue( int fdo, queue *list ) 
    {
    static polygon             *p;
    static polygon_ref         *pr;
    static int                  ilayer, idt, ilen;
    static struct gds_itemtype *gdspoly;
    static BOOL                 init=FALSE;
    static int                  maxlen=0;
    static int                 *nullx, *nully;


    if ( list->len == 0 ) return;
    if ( ! init )
        {
        gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        nullx = (int *) malloc( sizeof( int ) ); 
        nully = (int *) malloc( sizeof( int ) );
        if ( ! gdspoly || ! nullx || ! nully ) bail( "unable to allocate memory in write_queue" );
        nullx[0] = 0;
        nully[0] = 0; 
        init = TRUE;
        }
    
    pr = list->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active )
            {
            ilen = p->len;
            if ( ilen > maxlen )
                {
                gdspoly->x = (int *) malloc( ilen * sizeof( int ) ); 
                gdspoly->y = (int *) malloc( ilen * sizeof( int ) );
                if ( ! gdspoly->x || ! gdspoly->y ) bail( "unable to allocate gdspoly in write_queue" );
                maxlen = ilen;
                }
            gdspoly->n = ilen; 

            copy_to_gds_polygon( p, gdspoly );

            gds_write_boundary( fdo );
            if ( p->layer == 9999 )
                ilayer = ERRORLAYER;
            else if ( p->layer > 9999 )
                ilayer = p->layer - 10000;
            else
                ilayer = p->layer;
            idt    = p->dt;

            if ( ilayer < 0 ) bail( "attempted to write a negative layer" );
            if ( idt < 0 )    bail( "attempted to write a negative datatype" );

            if ( ilen > 2 )
                {
                gds_write_layer(    fdo, ilayer );
                gds_write_datatype( fdo, idt    );
                gds_write_xy( fdo, gdspoly->x, gdspoly->y, gdspoly->n );
                gds_write_endel( fdo );
                }
            }
        pr = pr->next;
        }

    // note that this function does not close the gds structure 

    } // write_queue

//-----------------------------------------------------------------------------------------------


BOOL                                                                             
find_hor_cut_between( polygon *outside, polygon *hole, queue *gob, 
                      int *hole_vertex_h, int *out_vertex_h1, 
                      int *out_vertex_h2, xy  *newv )
    {
    static int    i, hvil, hvir, bestv1, bestv2;
    static xy     hv, ray, a, b, inter, bestinter;
    static int    minx, maxx, shortest, length;
    static BOOL   init;


    if ( ! outside->active ) bail( "why is the polygon not active in find_hor_cut?" );
    if ( ! hole->active )    bail( "why is the hole not active in find_hor_cut?" );

    // Start with the left-most vertex of the hole, since this seems most likely.
    // As long as we are looking for the left one, we can find the right one also.

    hv = vertex( hole, 0 );
    minx = hv.x;
    maxx = minx;
    hvil = 0;
    hvir = 0;

    for ( i=1; i < hole->len; i++ )
        {
        hv = vertex( hole, i );
        if ( hv.x < minx )
            {
            minx = hv.x;
            hvil = i; 
            }
        if ( hv.x > maxx )
            {
            maxx = hv.x;
            hvir = i; 
            }
        }

    hv = vertex( hole, hvil );

    // Now hv is the most likely vertex on the hole, and hvil is the vertex number.
    // Find the shortest horizontal line between this vertex and the outer polygon,
    // while avoiding any intersections with any of the holes. We need the
    // gob so that we can check for intersections with other holes.
    
    check_extents( outside );
    check_extents( hole );

    ray.x = outside->ll.x;      // a ray pointing left, horizontally
    ray.y = hv.y;

    // if the hole touches the polygon exterior

    if ( distance_between( hv, ray ) < nearness ) // then the hole touches the polygon!
        {
        //printf( "\n................A hole touches a polygon.\n" );

        *hole_vertex_h = hvil; 

        // loop around the outside polygon, finding the vertex which touches the hole

        init = TRUE;
        for ( i=0; i < outside->len; i++ )
            {
            a = vertex( outside, i );
            b = vertex( outside, (i+1) % outside->len );
            if ( init )
                {
                init = FALSE;
                shortest = point2line_dist( hv, a, b );
                bestv1 = i;
                bestv2 = (i + 1) % outside->len;
                bestinter.x = hv.x;
                bestinter.y = hv.y;
                }
            else
                {
                length = point2line_dist( hv, a, b );
                if ( length < shortest )
                    {
                    shortest = length;
                    bestv1 = i;
                    bestv2 = (i+1) % outside->len;
                    bestinter.x = hv.x;
                    bestinter.y = hv.y;
                    }
                }
            }        

        *out_vertex_h1 = bestv1;
        *out_vertex_h2 = bestv2;
        newv->x        = bestinter.x;
        newv->y        = bestinter.y;

        return( TRUE );
        }

    // And now, for a normal hole which does not touch the polygon...

    init = TRUE;

    for ( i=0; i < outside->len; i++ )
        {
        a = vertex( outside, i   );
        b = vertex( outside, (i+1) % outside->len );
        if ( crossing( a, b, hv, ray, &inter ) )
            {
            if ( it_does_not_hit_holes_in( gob, hole, hv, hvil, inter ) )
                {
                if ( init )
                    {
                    init = FALSE;
                    bestv1 = i;
                    bestv2 = (i+1) % outside->len;
                    bestinter.x = inter.x;
                    bestinter.y = inter.y;
                    shortest = abs( inter.x - hv.x ); 
                    }
                else
                    {
                    length = abs( inter.x - hv.x );
                    if ( length < shortest )
                        {
                        shortest = length;
                        bestv1 = i;
                        bestv2 = (i+1) % outside->len;
                        bestinter.x = inter.x;
                        bestinter.y = inter.y;
                        }
                    }
                }
            }
        }


    if ( ! init ) 
        {
        //printf( "\nThe best cut is to the left, from the hole at (%ld,%ld) to the polygon at (%ld,%ld)\n",
	//         hv.x, hv.y, bestinter.x, bestinter.y );

        *hole_vertex_h = hvil; 
        *out_vertex_h1 = bestv1;
        *out_vertex_h2 = bestv2;
        newv->x        = bestinter.x;
        newv->y        = bestinter.y;

        return( TRUE );
        }

    // Let's look for a right-pointing cut .......................

    hv = vertex( hole, hvir );
    ray.x = outside->ur.x;      // a ray pointing right, horizontally
    ray.y = hv.y;
    init = TRUE;

    for ( i=0; i < outside->len; i++ )
        {
        a = vertex( outside, i   );
        b = vertex( outside, (i+1) % outside->len );
        if ( crossing( a, b, hv, ray, &inter ) )
            {
            if ( it_does_not_hit_holes_in( gob, hole, hv, hvir, inter ) )
                {
                if ( init )
                    {
                    init = FALSE;
                    bestv1 = i;
                    bestv2 = (i+1) % outside->len;
                    bestinter.x = inter.x;
                    bestinter.y = inter.y;
                    shortest = abs( inter.x - hv.x ); 
                    }
                else
                    {
                    length = abs( inter.x - hv.x );
                    if ( length < shortest )
                        {
                        shortest = length;
                        bestv1 = i;
                        bestv2 = (i+1) % outside->len;
                        bestinter.x = inter.x;
                        bestinter.y = inter.y;
                        }
                    }
                }
            }
        }


    //printf( "\nThe best cut is to the right, from the hole at (%ld,%ld) to the polygon at (%ld,%ld)\n",
    //        hv.x, hv.y, bestinter.x, bestinter.y );

    *hole_vertex_h = hvir; 
    *out_vertex_h1 = bestv1;
    *out_vertex_h2 = bestv2;
    newv->x        = bestinter.x;
    newv->y        = bestinter.y;

    return( ! init );
    }
 
//------------end find_hor_cut_between-------------------------------------------------------------------------


BOOL                                                                         
find_ver_cut_between( polygon *outside, polygon *hole, queue *gob,
                      int *hole_vertex_v, int *out_vertex_v1, 
                      int *out_vertex_v2, xy  *newv,
                      int exclusion )
    {
    // same as the previous function, but here we exclude the hole vertex used for the horizontal cut

    static int    i, hvit, hvib, bestv1, bestv2, thisone, startone;
    static xy     hv, ray, a, b, inter, bestinter;
    static int    miny, maxy, shortest, length;
    static BOOL   init, first;


    if ( ! outside->active ) bail( "why is the polygon not active in find_ver_cut?" );
    if ( ! hole->active )    bail( "why is the hole not active in find_ver_cut?" );

    // Start with the top-most vertex of the hole, since this seems most likely.
    // As long as we are looking for the top one, we can find the lowest one also.

    hv = vertex( hole, 0 );
    miny = hv.y;
    maxy = miny;
    hvit = 0;
    hvib = 0;

    for ( i=1; i < hole->len; i++ )
        {
        hv = vertex( hole, i );
        if ( hv.y < miny )
            {
            miny = hv.y;
            hvib = i; 
            }
        if ( hv.y > maxy )
            {
            maxy = hv.y;
            hvit = i; 
            }
        }


    //if ( hvit == exclusion ) hvit = (hvit + 1) % hole->len;  // maybe we should do a more exhaustive search?
    //if ( hvib == exclusion ) hvib = (hvib + 1) % hole->len;  // yes indeed. that's what we are doing now.
    //if ( hvit == exclusion || hvib == exclusion ) bail( "What kind of hole is this? Degenerate?" );

    first    = TRUE;
    thisone  = hvit;
    startone = hvit;
    if ( thisone == exclusion ) 
        {
        thisone = (thisone + 1) % hole->len;
        startone = thisone;
        }
                                             // we ought to do this jazz for the horizontal cuts also
    while ( first || thisone != startone )
        {
        first = FALSE;
        hvit = thisone;

        hv = vertex( hole, hvit );

        // Now hv is the most likely vertex on the hole, and hvil is the vertex number.
        // Find the shortest vertical line between this vertex and the outer polygon,
        // while avoiding any intersections with any of the holes. We need the
        // gob so that we can check for intersections with other holes.
        
        check_extents( outside );
        check_extents( hole );

        ray.x = hv.x;
        ray.y = outside->ur.y;      // a ray pointing up from the hole


        //  if the hole touches the polygon exterior

        if ( distance_between( hv, ray ) < nearness ) // then the hole touches the polygon!
            {
            // printf( "\nA hole touches a polygon.\n" );

            *hole_vertex_v = hvit; 

            // loop around the outside polygon, finding the vertex which touches the hole

            init = TRUE;
            for ( i=0; i < outside->len; i++ )
                {
                a = vertex( outside, i );
                b = vertex( outside, (i+1) % outside->len );
                if ( init )
                    {
                    init = FALSE;
                    shortest = point2line_dist( hv, a, b );
                    bestv1 = i;
                    bestv2 = (i + 1) % outside->len;
                    bestinter.x = hv.x;
                    bestinter.y = hv.y;
                    }
                else
                    {
                    length = point2line_dist( hv, a, b );
                    if ( length < shortest )
                        {
                        shortest = length;
                        bestv1 = i;
                        bestv2 = (i+1) % outside->len;
                        bestinter.x = hv.x;
                        bestinter.y = hv.y;
                        }
                    }
                }        

            *out_vertex_v1 = bestv1;
            *out_vertex_v2 = bestv2;
            newv->x        = bestinter.x;
            newv->y        = bestinter.y;

            return( TRUE );
            }


        // And now, for a normal hole which does not touch the polygon...
        //.................

        init = TRUE;

        for ( i=0; i < outside->len; i++ )
            {
            a = vertex( outside, i   );
            b = vertex( outside, (i+1) % outside->len );
            if ( crossing( a, b, hv, ray, &inter ) )
                {
                if ( it_does_not_hit_holes_in( gob, hole, hv, hvit, inter ) )
                    {
                    if ( init )
                        {
                        init = FALSE;
                        bestv1 = i;
                        bestv2 = (i+1) % outside->len;
                        bestinter.x = inter.x;
                        bestinter.y = inter.y;
                        shortest = abs( inter.y - hv.y ); 
                        }
                    else
                        {
                        length = abs( inter.y - hv.y );
                        if ( length < shortest )
                            {
                            shortest = length;
                            bestv1 = i;
                            bestv2 = (i+1) % outside->len;
                            bestinter.x = inter.x;
                            bestinter.y = inter.y;
                            }
                        }
                    }
                }
            }

        //if ( init ) printf( "unable to find an upward pointing line from the hole to the polygon\n" );

        if ( ! init ) 
            {
            //printf( "\nThe best cut is upward, from the hole at (%ld,%ld) to the polygon at (%ld,%ld)\n",
            //        hv.x, hv.y, bestinter.x, bestinter.y );

            *hole_vertex_v = hvit; 
            *out_vertex_v1 = bestv1;
            *out_vertex_v2 = bestv2;
            newv->x        = bestinter.x;
            newv->y        = bestinter.y;

            return( TRUE );
            }

        thisone = (thisone + 1) % hole->len;
        if ( thisone == exclusion ) thisone = (thisone + 1) % hole->len;

        }  // end while thisone != startone



    // Let's look for a downward-pointing cut .......................

    first    = TRUE;
    thisone  = hvib;
    startone = hvib;
    if ( thisone == exclusion ) 
        {
        thisone = (thisone + 1) % hole->len;
        startone = thisone;
        }

    while ( first || thisone != startone )
        {
        first = FALSE;
        hvib = thisone;

        hv = vertex( hole, hvib );
        
        ray.x = hv.x;
        ray.y = outside->ll.y;      // a ray pointing down
        init = TRUE;

        for ( i=0; i < outside->len; i++ )
            {
            a = vertex( outside, i   );
            b = vertex( outside, (i+1) % outside->len );
            if ( crossing( a, b, hv, ray, &inter ) )
                {
                if ( it_does_not_hit_holes_in( gob, hole, hv, hvib, inter ) )
                    {
                    if ( init )
                        {
                        init = FALSE;
                        bestv1 = i;
                        bestv2 = (i+1) % outside->len;
                        bestinter.x = inter.x;
                        bestinter.y = inter.y;
                        shortest = abs( inter.y - hv.y ); 
                        }
                    else
                        {
                        length = abs( inter.y - hv.y );
                        if ( length < shortest )
                            {
                            shortest = length;
                            bestv1 = i;
                            bestv2 = (i+1) % outside->len;
                            bestinter.x = inter.x;
                            bestinter.y = inter.y;
                            }
                        }
                    }
                }
            }

        //printf( "\nThe best cut is downward, from the hole at (%ld,%ld) to the polygon at (%ld,%ld)\n",
        //        hv.x, hv.y, bestinter.x, bestinter.y );

        *hole_vertex_v = hvib; 
        *out_vertex_v1 = bestv1;
        *out_vertex_v2 = bestv2;
        newv->x        = bestinter.x;
        newv->y        = bestinter.y;

        // init true => nothing found => failure 

        if ( ! init ) return( TRUE );

        thisone = (thisone + 1) % hole->len;
        if ( thisone == exclusion ) thisone = (thisone + 1) % hole->len;

        }  // end while thisone != startone

    return( FALSE );  // omg can you imagine failure after all that? and yet it happens. 
    }
 
//------------end find_ver_cut_between-------------------------------------------------------------------------

BOOL                                                    
there_is_a_polygon_in( queue *gob )
    {
    static BOOL         oui;
    static polygon_ref *pr;
    static polygon     *p;

    if ( ! gob ) return( FALSE );

    oui = FALSE;
    pr = gob->first;

    while ( ! oui && pr )
        {
        p = pr->p;
        if ( p->active ) oui = (p->layer != 9999);
        pr = pr->next;
        }

    return( oui );
    }


BOOL                                          
there_are_holes_in( queue *gob )
    {
    static BOOL         oui;
    static polygon_ref *pr;
    static polygon     *p;

    if ( ! gob ) return( FALSE );

    oui = FALSE;
    pr = gob->first;

    while ( ! oui && pr )
        {
        p = pr->p;
        if ( p->active ) oui = (p->layer == 9999);
        pr = pr->next;
        }

    return( oui );
    }


BOOL                                                  
there_are_only_holes_in( queue *gob )
    {
    static BOOL         no_outer_polys;
    static polygon_ref *pr;
    static polygon     *p;

    if ( ! gob ) return( FALSE );

    no_outer_polys = TRUE;
    pr = gob->first;

    while ( no_outer_polys && pr )
        {
        p = pr->p;
        if ( p->active ) no_outer_polys = ! (p->layer != 9999);
        pr = pr->next;
        }

    return( no_outer_polys );
    }

//-------------------------------------------------------------------------------------------------------------

void                                               
concatenate( queue *q, queue *list )
    {
    polygon_ref *pr;
    pr = q->first;
    while ( pr )
        {
        if ( pr->p->active ) 
            {
            append_queue( list, pr->p );
            }
        pr = pr->next;
        }
    }

//------------end concatenate------------------------------------------------------------------------------------

polygon *                                          
left_hole_in( queue *gob )
    {
    static polygon *p, *theone;
    static polygon_ref *pr;

    static int  minx, maxy;
    static BOOL init;

    theone = NULL;
    if ( ! there_are_holes_in( gob ) ) bail( "there are no holes in this polygon, so I can't find the left-most" );
    init = TRUE;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) && ! p->hit[0] )
            {
            check_extents( p );
            if ( init )
                {
                init = FALSE;
                minx = p->ll.x; 
                theone = p;
                }
            if ( p->ll.x < minx ) 
                { 
                minx = p->ll.x; 
                theone = p;
                } 
            }
        pr = pr->next;
        }
    if ( ! theone ) bail( "unable to find any active holes in polygon (searching for left-most)" );
 
    // Look for holes with the same ll.x but higher ur.y. Holes near the UL corner are 
    // more likely to have two cuts to the outer polygon

    maxy = theone->ur.y;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) )
            {
            if ( EQUAL( p->ll.x, minx ) && (p->ur.y > maxy) ) 
                { 
                maxy = p->ur.y; 
                theone = p;
                } 
            }
        pr = pr->next;
        }
    

    return( theone );
    }

//------------end left_hole_in-----------------------------------------------------------------------------



polygon *                                          
lowest_hole_in( queue *gob )
    {
    static polygon *p, *theone;
    static polygon_ref *pr;
    static int  maxx, miny;
    static BOOL init;

    theone = NULL;
    if ( ! there_are_holes_in( gob ) ) bail( "there are no holes in this polygon, so I can't find the lowest" );
    init = TRUE;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) && ! p->hit[0] )
            {
            check_extents( p );
            if ( init )
                {
                init = FALSE;
                miny = p->ll.y;
                theone = p;
                }
            if ( p->ll.y < miny ) 
                { 
                miny = p->ll.y; 
                theone = p;
                } 
            }
        pr = pr->next;
        }
    if ( ! theone ) bail( "unable to find any active holes in polygon (searching for lowest)" );

    // Look for holes with the same ll.y but higher ur.x. Holes near the LR corner are 
    // more likely to have two cuts to the outer polygon

    maxx = theone->ur.x;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) )
            {
            if ( EQUAL( p->ll.y, miny ) && (p->ur.x > maxx) ) 
                { 
                maxx = p->ur.x; 
                theone = p;
                } 
            }
        pr = pr->next;
        }

    return( theone );
    }

//------------end lowest_hole_in-----------------------------------------------------------------------------

polygon *                                             
highest_hole_in( queue *gob )
    {
    static polygon *p, *theone;
    static polygon_ref *pr;

    static int  maxy;
    static BOOL init;

    theone = NULL;
    if ( ! there_are_holes_in( gob ) ) bail( "there are no holes in this polygon, so I can't find the highest" );
    init = TRUE;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) &&  ! p->hit[0] )
            {
            check_extents( p );
            if ( init )
                {
                init = FALSE;
                maxy = p->ur.y;
                theone = p;
                }
            if ( p->ur.y > maxy ) 
                { 
                maxy = p->ur.y; 
                theone = p;
                } 
            }
        pr = pr->next;
        }
    if ( ! theone ) bail( "unable to find any active holes in polygon (searching for highest)" );
    return( theone );
    }

//------------end highest_hole_in-------------------------------------------------------------------------------

polygon *                                            
right_hole_in( queue *gob )
    {
    static polygon *p, *theone;
    static polygon_ref *pr;

    static int  maxx, maxy;
    static BOOL init;

    theone = NULL;
    if ( ! there_are_holes_in( gob ) ) bail( "there are no holes in this polygon, so I can't find the right-most" );
    init = TRUE;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) && ! p->hit[0] )   // exclude previously found left-most hole
            {
            check_extents( p );
            if ( init )
                {
                init = FALSE;
                maxx = p->ur.x; 
                theone = p;
                }
            if ( p->ur.x > maxx ) 
                { 
                maxx = p->ur.x; 
                theone = p;
                } 
            }
        pr = pr->next;
        }
    if ( ! theone ) bail( "unable to find any active holes in polygon (searching for right-most)" );

    // Look for holes with the same ur.x but higher ur.y. Holes near the UR corner are 
    // more likely to have two cuts to the outer polygon

    maxy = theone->ur.y;
    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && (p->layer == 9999) )
            {
            if ( EQUAL( p->ur.x, maxx ) && (p->ur.y > maxy) ) 
                { 
                maxy = p->ur.y; 
                theone = p;
                } 
            }
        pr = pr->next;
        }

    return( theone );
    }

//------------end right_hole_in----------------------------------------------------------------------------

void
separate_gobs( queue_queue *qq, queue *gob ); // forward

void                                                                                 
cut_one( queue *q, queue_queue *qq, queue *list, BOOL *success, BOOL random )
    {
    // Search for two cuts between a hole and the outer polygon of 'q'.
    // If we can find two cuts from one hole then we separate the
    // polygon 'q' into two parts. If a part still has holes, then
    // it will be pushed onto 'qq'. If not, it will go onto the 'list'.

    static polygon_ref *pr, *firstholer;
    static polygon     *p, *outside, *hole, *left, *right, *high, *low;
    static BOOL         keep_looking, found;
    static int          hole_vertex_h, out_vertex_h1, out_vertex_h2;
    static int          hole_vertex_v, out_vertex_v1, out_vertex_v2;
    queue              *gob1, *gob2;
    static int          nnag=0, nholes;
    static xy           newhv, newvv;

    
    keep_looking = TRUE;
    firstholer   = NULL;
    hole         = NULL;
    outside      = NULL;
    *success     = FALSE;

    pr = q->first;
    while ( pr && keep_looking )
        {
        p = pr->p;
        if ( p->active )
            {
            if ( ! outside && p->layer > 9999 )  outside = p;
            if ( ! firstholer && p->layer == 9999 ) firstholer = pr;
            keep_looking = (! outside) || (! firstholer);
            }
        pr = pr->next;
        }

    pr = q->first;
    while ( pr )
        {
        p = pr->p;
        unhit_all( p );
        pr = pr->next;
        }

    if ( ! outside )    bail( "cut_one was not given a polygon with holes" );

    if ( ! firstholer ) 
        {
        nag( "cut_one did not find holes in the given polygon", &nnag );
        if ( there_are_holes_in( q ) )
            nag( "and yet there are holes in the polygon\n", &nnag );
        else
            nag( "so where did they go?\n", &nnag );
        *success = TRUE;
        concatenate( q, list );
        return;
        }

    if ( there_are_only_holes_in( q ) )
        {
        nag( "a gob contains nothing but holes", &nnag );
        *success = TRUE;
        return;
        }

    nholes = number_of_holes_in( q );

    // paranoid check that there is only one outer polygon - delete later

    while ( pr )
        {
        p = pr->p;
        if ( p->active && p->layer > 9999 ) bail( "more than one outer polygon appears in this gob and why is that?" );
        pr = pr->next;
        }

    // search for two cuts between a hole and the outer polygon.
    // start with the hole in the upper-left corner. 
    // if that does not work, then look in the other corners

    found = FALSE;
    left = left_hole_in( q );

    if ( ! random )
        {
        //printf( "\nleft? " ); fflush(stdout);
        hole = left;
        found =              find_hor_cut_between( outside, hole, q, &hole_vertex_h, &out_vertex_h1, &out_vertex_h2, &newhv );
        if ( found ) found = find_ver_cut_between( outside, hole, q, &hole_vertex_v, &out_vertex_v1, &out_vertex_v2, &newvv,
                             /* but the hole vertex cannot be... */   hole_vertex_h );
        }

    if ( ! found && nholes > 1 && ! random )
        {
        //printf( "right? " ); fflush(stdout);
        right = right_hole_in( q );  
        hole = right;
        hole->hit[0] = TRUE;
        found =              find_hor_cut_between( outside, hole, q, &hole_vertex_h, &out_vertex_h1, &out_vertex_h2, &newhv );
        if ( found ) found = find_ver_cut_between( outside, hole, q, &hole_vertex_v, &out_vertex_v1, &out_vertex_v2, &newvv,
                             /* but the hole vertex cannot be... */   hole_vertex_h );
        }
    
    if ( ! found && nholes > 2 && ! random )
        {
        hole->hit[0] = TRUE;
        //printf( "bottom? " ); fflush(stdout);
        low = lowest_hole_in( q );
        hole = low;
        found =              find_hor_cut_between( outside, hole, q, &hole_vertex_h, &out_vertex_h1, &out_vertex_h2, &newhv );
        if ( found ) found = find_ver_cut_between( outside, hole, q, &hole_vertex_v, &out_vertex_v1, &out_vertex_v2, &newvv,
                             /* but the hole vertex cannot be... */   hole_vertex_h );
        }
    
    if ( ! found && nholes > 3 && ! random )
        {
        hole->hit[0] = TRUE;
        //printf( "top? " ); fflush(stdout);
        high = highest_hole_in( q );
        hole = high;
        found =              find_hor_cut_between( outside, hole, q, &hole_vertex_h, &out_vertex_h1, &out_vertex_h2, &newhv );
        if ( found ) found = find_ver_cut_between( outside, hole, q, &hole_vertex_v, &out_vertex_v1, &out_vertex_v2, &newvv,
                             /* but the hole vertex cannot be... */   hole_vertex_h );
        }

    if ( ! found || random )
        {
        //printf( "\nlooking at all the holes... " ); fflush(stdout);
        pr = q->first;
        while ( pr && ! found )
            {
            hole = pr->p;
            if ( hole->active && hole->layer == 9999 && ! hole->hit[0] )
                {
                hole->hit[0] = TRUE;
                found = find_hor_cut_between( outside, hole, q, &hole_vertex_h, &out_vertex_h1, &out_vertex_h2, &newhv );
                if ( found ) 
                    {
                    found = find_ver_cut_between( outside, hole, q, &hole_vertex_v, &out_vertex_v1, &out_vertex_v2, &newvv,
                            /* but the hole vertex cannot be... */   hole_vertex_h );
                    }
                }
            //if ( ! found ) { printf( "no. " ); fflush(stdout); }
            pr = pr->next;
            }
        }

    if ( ! found ) 
        {
        write_queue( fdo, q );                                    // show us the gob that failed
        write_queue( fdo, list );
        nag( "unable to find two cuts from any corner hole", &nnag );   
        *success = FALSE;
        return;
        }

    split_up( q, outside, hole, hole_vertex_h, out_vertex_h1, out_vertex_h2, newhv,
                                hole_vertex_v, out_vertex_v1, out_vertex_v2, newvv, &gob1, &gob2  );


    outside->active = FALSE;
    hole->active = FALSE;

    // We have to separate the gobs, in case the splitting process 
    // created self-crossing polygons. So tricky!!

    if ( there_are_holes_in( gob1 ) )
        {
        //printf( "gob1 goes back onto the qq \n" );
        separate_gobs( qq, gob1 );
        }
    else
        {
        //printf( "gob1 is hole-free, pushed onto list \n" );
        if ( gob1 && gob1->first ) concatenate( gob1, list );
        }

    if ( there_are_holes_in( gob2 ) )
        {
        //printf( "gob2 goes back onto the qq \n" );
        separate_gobs( qq, gob2 );
        }
    else
        {
        //printf( "gob2 is hole-free, pushed onto list \n" );
        if ( gob2 && gob2->first ) concatenate( gob2, list );
        } 

    *success = TRUE;
    }

//-------------end cut_one----------------------------------------------------------------------------------------

BOOL                                                                                                   
there_is_an_inside_point( polygon *h, polygon *p )   // is the hole h inside the polygon p ?
    {                                                // maybe we should require that all points of h be inside p
    static BOOL yup;
    static int  i;

    yup = FALSE;
    for ( i=0; i < h->len && ! yup; i++ )
        {
        yup = inside_point( vertex( h, i ), p, TRUE );  // normally true wallflowers do not count as inside
        }
    return( yup );
    }


BOOL                                                                         
all_hole_vertices_are_inside_poly( polygon *h, polygon *p )
    {                                                
    static BOOL outside;
    static int  i;

    outside = FALSE;
    for ( i=0; i < h->len && ! outside; i++ )
        {
        outside = ! inside_point( vertex( h, i ), p, FALSE );
        if ( outside ) outside = ! point_touches_polygon( vertex( h, i ), p );
        }
    return( ! outside );
    }


//----------------------------------------------------------------------------------------------- 

void                                                         
separate_gobs( queue_queue *qq, queue *gob )
    {
    // For each outer polygon in gob, determine which holes are inside, and then
    // put these separated gobs onto qq.

    polygon_ref     *pr, *hr;
    polygon         *p,  *h;
    queue           *littlegob;
    static xy        v;


    static double    area;
    static int       nnag=0;
    static char      message[256];
    static int       mtop;


    //printf( "\nseparating gobs...\n" );

    // If there are no holes, then simply copy the polygon into qq

    if ( ! there_are_holes_in( gob ) )
        {
        append_queue_queue( qq, gob );
        return;
        }

    if ( there_are_only_holes_in( gob ) ) return;

    // mark all of the polygons as not hit

    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        unhit_all( p );
        pr = pr->next;
        }

    // For each outer polygon (with holes) find the associated holes and
    // create a little gob. Then push this little gob onto the qq.

    pr = gob->first;
    while ( pr )
        {
        p = pr->p;
        if ( p->active && p->layer != 9999 ) // this is an outside polygon
            {
            p->hit[0] = TRUE;
            littlegob = initialize_queue();
            append_queue( littlegob, p );
            hr = gob->first;
            while ( hr )
                {
                h = hr->p;
                if ( h->active && (h->layer == 9999) )
                    {
                    if ( there_is_an_inside_point( h, p ) ) 
		    //if ( all_hole_vertices_are_inside_poly( h, p ) )
                        {
                        if ( h->hit[0] ) nag( "Drat! A hole seems to be inside more than one polygon.", &mtop );
                        h->hit[0] = TRUE;
                        append_queue( littlegob, h );
                        }
                    }
                hr = hr->next;
                }
            append_queue_queue( qq, littlegob );
            }
        else if ( p->active && (p->layer < 9999) )    // we have a polygon without holes. Push it onto the qq.
            {
            p->hit[0] = TRUE;
            //printf( "\nI found a polygon without holes. Woo Hoo.\n" );
            littlegob = initialize_queue();
            append_queue( littlegob, p );
            append_queue_queue( qq, littlegob );   
            } 
        pr = pr->next;
        }

    // Check for leftovers. Leftovers are normally great, but in this case leftovers are bad.
    
    pr = gob->first;
    while ( pr )  
        {
        p = pr->p;
        if ( p->active && (! p->hit[0]) )
            {
            area = polygon_area( p, bss ); // ccw( p, &area );
            v = vertex( p, 0 );
            if ( area < 250.0 && p->layer == 9999 ) // no problem - tiny hole inside tiny sliver
	        {
                sprintf( message, "tiny hole at %ld %ld", v.x, v.y );
                nag( message, &nnag );
                p->active = FALSE;
                }
            else
                {
		p->active = FALSE;
                //write_queue( fdo, gob );   why are we writing the gob? that doesn't make any sense.
                if ( p->layer == 9999 )
                    printf( "we did not hit the hole at %ld %ld\n", v.x, v.y );
                else
                    printf( "we did not hit the polygon at %ld %ld\n", v.x, v.y );

                printf( "area = %1.3lf \n", area );
		
                }
            }
        pr = pr->next;
        }

    }

//-----------------end separate_gobs-----------------------------------------------------------------------------

BOOL                                                            
eat_doughnut( queue *gob, queue *list )
    {
    // The gob starts as a set of polygons and holes. This function will cut them into multiple polygons.

    // layer >= 10000 means this is a polygon with holes. Subtract 10000 to get the actual layer.
    // layer ==  9999 means this is a hole
    
    queue_queue    *gobq;
    queue_ref      *qr;
    queue *q;
    static BOOL     success;
    polygon_ref    *hr;
    polygon        *h;

    if ( ! there_are_holes_in( gob ) ) 
        {
        return( TRUE );
        }

    if ( there_are_only_holes_in( gob ) ) return( TRUE );

    gobq = initialize_queue_queue();

    // If gob contains several polygons then we need to separate them first.

    separate_gobs( gobq, gob );

    qr = gobq->first;

    while ( qr )
        {
        if ( qr->active )
            {
            q = qr->q;
            
            if ( there_are_holes_in( q ) ) 
                {
		cut_one( q, gobq, list, &success, FALSE );   // usually false = not random
                
                if ( success )
                    qr->active = FALSE;
                else
                    {
                    //print_polygon( q->first->p, 1000 );
                    printf( "\narea = %lf\n", polygon_area( q->first->p, bss ) );
                    printf( "unable to cut out a hole in %s\n", ifilename );
                    hr = q->first;
                    while ( FALSE ) //  hr )
			{
			h = hr->p;
			if ( h->layer == 9999 ) print_polygon( h, 1000 );
			hr = hr->next;
			}
                    printf( "unable to cut out a hole\n" );
                    return( FALSE );
                    }

                }
            else
                {
                concatenate( q, list );
                }
            }
        qr = qr->next;
        }

    return( TRUE );

    }  // end eat_doughnut

//---------------------------------------------------------------------------------------------------------------------

int_list *                                                  
parse_layers( char *layers )
    {
    // Create a list of layers to include.
    // Layer -1 implies "all".

    int_list *list;
    char     *p, *q;
    int       len, num, i, layer;

    q = layers;
    len = strlen( q );
    num = 0;
    for ( i=0; i<len; i++ ) 
        {
        if ( *q == ',' ) num++;
        q++;
        }
    num++;
    
    list = initialize_int_list( num );

    p = strtok( layers, "," );
    while ( p != NULL ) 
        {
        sscanf( p, "%d", &layer );
        append_int_list( list, layer );
        p = strtok( NULL, "," );
        }

    return( list );

    }  

//-------------------end parse_layers-------------------------------------------------------------------


//=====================================================================================================================

int main( int argc, char *argv[] )
    {
    char 
         *ofilename,
 	 *rotname,
         *rotempname,
 	 *tempfile1,
	 *tempfile2,
	 *tempfile3,
	 *tempfile4,
          command[256];

    double
	  angle, 
          bss_nm,
          dbu_rescale,
          dbu_uu,
          dbu_m,
	  dbu_um;

    int   layer,
          dt,
          fd, 
          n, 
	  i,
          minusone = -1,
          len;

    BOOL  allayers,
          not_eof,
          noheader,
          nofooter,
          we_must_snap;

    int_list
         *layers;

    unsigned short 
          count,           
          token;           

    struct gds_itemtype 
	 *item,
         *nullitem;

    polygon 
         *p, *g,
	 *poly;

    xy    v;

    polygon_pair 
          p1p2;

    polygon_ref 
         *pr, *gr;

    queue *list, 
          *gob, 
          *dblist, 
          *nice,
          *naughty;

    pid_t  pid;

    pid = getpid();


    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "darn v%1.1f\n", VERSION );
        exit(0);
        }
    
    if ( argc < 2 )
        {
        printf( "\nidarn v%1.1f\n", VERSION );
        printf( "\nUse: idarn infile.gds [bss_nm] [layers] [output file name] [angle] \n" );
        printf(   "     Separate doughnut holes from polygons. Output name defaults to cored.gds\n" );
        printf(   "     bss defaults to 1nm. Layers default to 'all'. Layer -1 also indicates 'all'.\n" );
        printf(   "     Layer numbers should be separated by commas, without spaces.\n" );
        printf(   "     Angle, in degrees, will spawn 'flatten' to rotate the pattern.\n" );
        printf(   "     Setting bss_nm to zero means 'no snapping'.\n\n" );
        exit(0);
        }

    we_must_snap = strcmp( argv[0], "idarn" );  // idarn by any other name means "do not snap"

    ifilename = argv[1]; 

    if ( argc > 2 )
        {
        sscanf( argv[2], "%lf", &bss_nm );
        we_must_snap = (bss_nm > 0.0);           // idarn with bss=0 also means 'no snap'
        }
    else
        {
        bail( "missing arguments to 'idarn'." );
        }

    allayers = TRUE;

    if ( argc > 3 )
        {
        layers = parse_layers( argv[3] );
        allayers = in_int_list( layers, minusone );
        }

    rotname    = (char *) malloc( 256 );
    rotempname = (char *) malloc( 256 );
    ofilename  = (char *) malloc( 256 );
    tempfile1  = (char *) malloc( 256 );
    tempfile2  = (char *) malloc( 256 );
    tempfile3  = (char *) malloc( 256 );
    tempfile4  = (char *) malloc( 256 );

    if ( argc > 4 )
        {
        strcat( ofilename, argv[4] );
        }
    else
        {
        strcat( ofilename, "cored.gds" );
        }

    noheader = FALSE;
    nofooter = FALSE;

    angle = 0.0;
    if ( argc > 5 ) sscanf( argv[5], "%lf", &angle );
    

    //----------------------------------------
    // If we must snap then we first call snap.
    // Currently this is turned off because it

    we_must_snap = FALSE;
    if ( we_must_snap )  // causes small overlap problems when edges are snapped horizontally and vertically
        {                // but now... are we going to cause problems by not using snap.c ?
                         // yes. we now get some corner fill-in errors in annezone.gds, 
                         // but not when we use convexify instead of poly2trap.

        sprintf( tempfile1, "tmp/temp1_%d.gds", pid );
        sprintf( tempfile2, "tmp/temp2_%d.gds", pid );
        sprintf( tempfile3, "tmp/temp3_%d.gds", pid );
        sprintf( tempfile4, "tmp/temp4_%d.gds", pid );

        sprintf( command,  "snap %s %s 1 nofix", ifilename, tempfile2 ); // snap to 1nm, do not fix crossings
        //printf( "%s\n", command );
        system(  command );

        //sprintf( command,  "snap %s %s %1.3lf fix", tempfile1, tempfile2, bss_nm ); // snap to bss, fix crossings
        //printf( "%s\n", command );
        //system(  command );

        sprintf( command,  "idarn %s 0 %s %s", tempfile2, argv[3], tempfile3 ); // bss=0 => no snap
        //printf( "%s\n", command );
        system(  command );

        sprintf( command,  "snap %s %s %1.3lf fix", tempfile3, tempfile4, bss_nm );  // snap and fix
        //printf( "%s\n", command );
        system(  command );

        sprintf( ifilename, "%s", tempfile4 );
        }

    //--------------------
    
    read_gds_units( ifilename, &dbu_uu, &dbu_m );   

    dbu_um = dbu_m * 1.0e6;

    if ( dbu_um > 0.001 ) 
	dbu_rescale = dbu_m / 1e-9;         // This could be 10 if the dbu is 10nm.
    else                                    // We are going to rescale only when 
        dbu_rescale = 1.0;                  // the gds has a stupidly large database unit.
                                            // Smaller database units are ok since
                                            // we might need high resolution pitch/placement control.
    dbu_m  = dbu_m / dbu_rescale;
    dbu_uu = dbu_m * 1.0e6;
    dbu_um = dbu_m * 1.0e6; 

    bss = (int) round( bss_nm / (dbu_m * 1.0e9));   // beam step size in bits 
    if ( bss == 0 ) bss = 1;

    // float nearness was 0.8   presque was 3.0

    nearness = 1; //2;
    presque  = 3; 
    tooclose = 2; //1;                     // was nearness/2

    fd = open( ifilename, 0 );
    if ( fd <= 0 ) bail( "unable to open input file" );
    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! item ) bail( "unable to allocate gds item" );
    nullitem = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! nullitem ) bail( "unable to allocate gds nullitem" );

    fdo = open( ofilename, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( fdo <= 0 ) bail( "unable to open output gds file" );
    dbu_um = dbu_m * 1.0e6;

    if ( ! noheader )
        {
        gds_create_lib_correctly( fdo, "cored", dbu_um, dbu_uu );
        gds_write_bgnstr( fdo );
        gds_write_strname( fdo, "cored" );
        }


    // Main polygon reading loop-----------------------------------
    //                          ... should be in a separate function, 
    //                          but the file descriptor gets confused. Why?

    list     = initialize_queue();    
    nice     = initialize_queue();    
    naughty  = initialize_queue();    
    dblist   = initialize_queue();    // debugging only

    n = 0;
    not_eof = read( fd, &count, 2 ) > 0;

    while ( not_eof )
        {
        gds_swap2bytes( (BYTE *) &count );
        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );

            switch( token )
                {
                case 0x2B02:  // printf( "\nPROPATTR means obese \n" ); => obese
                          SKIPOVER( fd, count );
                          break;
                case 0x0D02:  //printf( "LAYER \n" ) ;
                          gds_read_layer( fd, count, &item, 0 );
                          layer = item->layer;
                          break;
                case 0x0E02:  //printf( "DATATYPE \n" ) ;
                          gds_read_datatype( fd, count, &item, 0 );
                          dt = item->dt;
                          break;
                case 0x1003:  //printf( "XY \n" ) ;
                          gds_read_xy( fd, count, 0, &item, 0 );
                          if ( allayers || in_int_list( layers, layer ) )
                              {
                              item->layer = layer;
                              item->dt = dt;
                              n++;
                              len = item->n;
                              poly = initialize_polygon( len, layer, dt, FALSE );
                              copy_from_gds_polygon( item, poly, dbu_rescale );
                              poly->obese = FALSE;

                              append_queue( list, poly );

                              }
                          break;
                default:      //
                          SKIPOVER( fd, count );
                }    
            }      
        not_eof = read( fd, &count, 2 ) > 0;
        }          


    close(fd);

    pr = list->first;
    while ( pr )
        {
        gob = initialize_queue();    
        p = pr->p;
        append_queue( gob, p );

        
        gr = gob->first;
        while ( gr )
            {     
            g = gr->p;   
            p1p2 = look_for_self_crossing_polygon( g );

            if ( p1p2.b->len > 2 )
                {
                g->active = FALSE;
                set_shape_extents( p1p2.a );
                set_shape_extents( p1p2.b );
                append_queue( gob, p1p2.a );
                append_queue( gob, p1p2.b );
                }

            gr = gr->next;
            }
	

        if ( SHOW_HOLES ) 
	    {
            concatenate( gob, dblist );
	    }
        else
            {
            if ( there_are_holes_in( gob ) )
                {
                if ( ! eat_doughnut( gob, naughty ) ) 
                    {
                    v = vertex( gob->first->p, 0 );
                    printf( "...... bad hole somewhere near %ld, %ld\n", v.x, v.y );
                    printf( "...... bad gob written to %s\n", ofilename );
                    bail( "unable to remove hole" );
		    }
		}
            else
		{
                concatenate( gob, nice );
		}
            }
 
        free_queue( gob );
        pr = pr->next;
        }

    if ( SHOWPROGRESS ) { printf( "\n         writing..." ); fflush(stdout); }

    if ( SHOW_HOLES )
        write_queue( fdo, dblist );
    else
        {
        write_queue( fdo, naughty );
        write_queue( fdo, nice    );
        }

    if ( ! nofooter )
        {
        gds_write_endstr( fdo ); 
        gds_write_endlib( fdo );    
        }

    close( fdo );

    if ( angle != 0.0 )
        {  
        sprintf( rotname,     "%s_rot.gds",    ofilename );      
	sprintf( rotempname,  "%s_rotemp.gds", ofilename );
        //printf( "flatten %s flat -yesbox -nodump -silent -layer -1 %1.3lf %s\n", ofilename, angle, rotempname );
        sprintf( command, "flatten %s cored -yesbox -nodump -silent -layer -1 %1.3lf %s\n", ofilename, angle, rotempname );
        system(  command );
        //printf( "snap %s %s %1.2lf \n", rotempname, rotname, bss_nm );
        sprintf( command, "snap %s %s %1.2lf \n", rotempname, rotname, bss_nm );
        system(  command );
        //printf( "mv %s %s\n", rotname, ofilename );
        sprintf( command, "mv %s %s\n", rotname, ofilename );
        system(  command );
        }

    if ( SHOWPROGRESS ) 
        printf( "\nDone.\n" );

    if ( SHOW_HOLES ) bail( "not, holes are marked in output file" );

    return(1);
    }


