/* 

FLATTEN          Flatten the cell hierarchy and render all scaling, rotation, arrays, paths 
                 and boxes as simple polygons. A single cell will be written into the 
                 output file flat.gds. Written by M. Rooks, Yale University
                 

requires libgds

2.2  libgds now filters out degenerate vertices.
     dumps a text file, if requested

2.3  Upon encountering zero-width paths, flatten stops and asks
     if you would like to convert these to single-pixel rectangles.

2.4  Minor improvement to libgds when reading floating point

2.5  Use libgds 2.4 which now pads strings according to the silly gds spec

2.6  Fix the improper rotated arrays from the Layout CAD program,
     by replacing rotated arrays with arrays of rotated cells.
     Fix the improper scaled arrays from Layout by replacing
     scaled arrays with arrays of scaled cells. 
     If an array is both scaled and rotated, then that's crazy,
     so the program will still abort.

2.7  Create the header with a specified user unit. I previously thought
     that this number did not matter. Sorry. Now you have the function
     gds_create_lib_correctly. The old version works fine if you
     are using 1000 db units per user unit, which is almost always the case.

2.8  Polygons with fewer than 3 vertices will be filtered out,
     instead of causing a complete halt.

2.9  Creates fullstop.txt upon error

2.10 No questions option keeps it rolling without interruption

2.11 Reports the name of the calling cell when encountering strange
     aref properties

3.0  Fixed some serious problems with multiple rotations and scaling
     of both srefs and arefs. Those were bad bugs. Sorry.

3.1  Tiny issue with fgets resolved by using strtrim to chop off cr

3.2  Layers can be specified in the command line - this can save a lot of time

3.3  srefs of arefs sometimes are wrong. It's not clear what's going on.
     Did I fix it? Not sure. Look below for "holy tap dancing".

3.4  Paths with 180 degree turns are flagged as stupid.

3.5  Paths now use matching points to reduce rounding error problems.

3.6  Additional rotation can be specified in command line.

3.7  Now we can set the output file name, which is useful for
     parallel processing. Who cares, right?

3.8  If the cell name is '?' then the first cell will be used.

3.9  Are two floats ever equal? Yes, but also no. Define EQUAL to be less fussy.
     Use the new function sine_cosine to get rid of floating point
     errors when the angle is a multiple of 90 degrees.

3.10 When flattening zero-width paths, the program asks the user for
     the width of the resulting rectangle.

3.11 Write the pattern extents in tmp/pattern_extents.dat, if tmp exists.
     I noticed that if the file "1" exists then bash will turn "?" into "1"
     which is so bizarre! Why would bash do this? You can override this
     by using \?, so we should probably make freebeam.sh take this precaution.

3.12 libgds now warns about too many vertices

3.13 Mirror (reflection) operations now reflect sref and aref origins properly.
     Rotated arrays are still bad because Layout CAD does not
     follow the gds specification. No worries - you can use Layout itself
     to flatten a gds file. 

3.14 Command line option to shift the data in X and Y

3.15 Cosmetic changes to printing in -silent mode
 
*/

#define VERSION 3.15

#include "libgds.h"

#define NINETY     1.57079633
#define ONEIGHTY   3.14159265
#define TWOSEVENTY 4.71238898
#define THREESIXTY 6.28318531

#define EQUAL( a, b ) ((fabs((a)-(b)) < 1e-6))


struct flat_itemtype
    {                                // In this flattened form, an item can be a polygon only.                    
    int   n;                         // The number of verticies 
    int   layer;                     // layer                                                                     
    int   dt;                        // datatype                                                                  
    int  *x;                         // array of x coordinates 
    int  *y;                         // array of y coordinates 
    struct flat_itemtype *nextitem;  // mag is always 1, angle always 0, no reflection.    
    };                               

struct intersection_coordinate
    {
    int x;
    int y;
    BOOL exists;                     // false if it would be infinity
    BOOL crossing;                   // true if the line segments overlap
    };

typedef struct long_list
    {
    long len;
    long max;
    long *n;
    }                     long_list;

/*------------------------------------------------------------------------------------------*/

void
strtrim( char *s )
    {                    // chop \n off the end of a null-terminated string
    static char *sp;
    sp = s;
    while ( *sp )
        {
        if ( *sp == '\n' ) *sp = '\0';
        sp++;
        }
    }

/*------------------------------------------------------------------------------------------*/

float
trim( float x )   
    {
    // A number like 9.9999999e-12 becomes 1e-11.

    float e, a, f;
    if ( x <= 0.0 ) return( x );
    e = floor( log10( x ) ) - 1;
    a = x / pow( 10, e );
    f = a - floor( a );
    if ( fabs( f - 1.0 ) < 1e-4 ) f = 1.0;
    a = floor(a) + f;
    a = a * pow( 10, e );
    return( a );
    }

//--------------------------------------------------------------------------------------------------


float
angle_of_segment( int ix1, int iy1, int ix2, int iy2 )    
    {
                                  // angle (in radians) of the vector from point 1 to point 2
    static float                  // ranging from 0 to 2*pi. It's trickier than you might expect.
        a,
        x1, y1,
        x2, y2;

    static int ndls=0;

    if ( (ix2 == ix1) && (iy2 == iy1) ) { nag( "DEGENERATE LINE SEGMENT", &ndls ); return( 0.0 ); }

    if ( ix2 == ix1 )
        {
        if ( iy2 > iy1 )
            return( NINETY );   
        else
            return( TWOSEVENTY );
        }
    else if ( iy2 == iy1 )
        {
        if ( ix2 > ix1 )
            return( 0.0 );
        else
            return( ONEIGHTY ); 
        }
    else
        {
        x1 = ix1;
        x2 = ix2;
        y1 = iy1;
        y2 = iy2;

        a = atanf( (y2 - y1) / (x2 - x1) );

        if ( a < 0.0 )
            {
            if ( ix2 > ix1 ) 
                a = THREESIXTY + a;
            else
                a = ONEIGHTY + a;
            }
        else
            {
            if ( ix2 < ix1 ) a = ONEIGHTY + a;
            }
        }

    return( a );

    }  // angle_of_segment


/*------------------------------------------------------------------------------------------*/

BOOL
crossing( int x, int y, int a1x, int a1y, int a2x, int a2y, int b1x, int b1y, int b2x, int b2y )
    {
    // Is the point x,y somewhere on both line segments a1-a2 and b1-b2 ?

    static int t;

    if ( a2x < a1x ) { t = a1x ; a1x = a2x ; a2x = t ; }
    if ( a2y < a1y ) { t = a1y ; a1y = a2y ; a2y = t ; }
    if ( b2x < b1x ) { t = b1x ; b1x = b2x ; b2x = t ; }
    if ( b2y < b1y ) { t = b1y ; b1y = b2y ; b2y = t ; }

    return( x >= a1x && x <= a2x && x >= b1x && x <= b2x &&
            y >= a1y && y <= a2y && y >= b1y && y <= b2y     ); 


    }  // crossing

/*------------------------------------------------------------------------------------------*/

struct intersection_coordinate 
intersection( int a1x, int a1y, int a2x, int a2y, int b1x, int b1y, int b2x, int b2y )
    {
    // Find the intersection of the line a1-a2 with the line b1-b2.
    // Return exists = false if the lines are parallel.
    // Return crossing = true if the line segments intersect.
    // Return a2 if they are coincident. Really? Yes, because reasons.

    static BOOL
        a_is_vertical,
        b_is_vertical;
    static int
        ix, iy;       // intersection      
    static float
        ma, mb,       // slopes
        ba, bb,       // intercepts
        fa1y,
        fa2y,
        fa1x,
        fa2x,
        fb1y,
        fb2y,
        fb1x,
        fb2x;
    static struct intersection_coordinate
        intersect;

    fa1y = a1y;
    fa2y = a2y;
    fa1x = a1x;
    fa2x = a2x;
    fb1y = b1y;
    fb2y = b2y;
    fb1x = b1x;
    fb2x = b2x;

    a_is_vertical = (a1x == a2x);
    b_is_vertical = (b1x == b2x);

    intersect.crossing = FALSE;  

    if ( a_is_vertical && b_is_vertical )
        {
        // they might be coincident!

        if ( a1x == b1x )
            {
            intersect.x = a2x;
            intersect.y = a2y;
            intersect.exists = TRUE;
            intersect.crossing = TRUE;
            return( intersect );
            }

        // but if they are just two vertical lines...

        intersect.x = 0;
        intersect.y = 0;
        intersect.exists = FALSE;
        intersect.crossing = FALSE;
        return( intersect );
        }

    if ( a_is_vertical )
        {
        intersect.x = a1x;
        mb = (fb1y - fb2y) / (fb1x - fb2x);
        intersect.y = floor( mb * fa1x + fb1y - mb * fb1x );
        intersect.exists = TRUE;
        intersect.crossing = crossing( intersect.x, intersect.y, a1x, a1y, a2x, a2y, b1x, b1y, b2x, b2y );
        return( intersect );
        }

    if ( b_is_vertical )
        {
        intersect.x = b1x;
        ma = (fa1y - fa2y) / (fa1x - fa2x);
        intersect.y = floor( ma * fb1x + fa1y - ma * fa1x );
        intersect.exists = TRUE;
        intersect.crossing = crossing( intersect.x, intersect.y, a1x, a1y, a2x, a2y, b1x, b1y, b2x, b2y );
        return( intersect );
        }

    ma = (fa1y - fa2y) / (fa1x - fa2x);
    mb = (fb1y - fb2y) / (fb1x - fb2x);

    bb = fb1y - mb * fb1x;
    ba = fa1y - ma * fa1x;

    if ( (ma == mb) && (ba == bb) )  // coincident lines
        {
        intersect.x = a2x;
        intersect.y = a2y;
        intersect.exists = TRUE;
        intersect.crossing = TRUE;
        return( intersect );
        }

    if ( ma == mb )  // parallel lines, but not coincident
        {
        intersect.x = 0;
        intersect.y = 0;
        intersect.exists = FALSE;
        intersect.crossing = FALSE;
        return( intersect );
        }

    // Finally, just regular lines not on-axis, not parallel

    intersect.x = floor( (bb - ba) / (ma - mb) );
    intersect.y = floor( ma * intersect.x + ba );
    intersect.exists = TRUE;
    intersect.crossing = crossing( intersect.x, intersect.y, a1x, a1y, a2x, a2y, b1x, b1y, b2x, b2y );    

    return( intersect );

    }  // intersection

/*------------------------------------------------------------------------------------------*/

BOOL
right_turn( float a, float b )
    {

    if ( b < a ) 
        return( (a - b) < ONEIGHTY );
    else
        return( !( (b - a) < ONEIGHTY ) ); 

    }

/*------------------------------------------------------------------------------------------*/

BOOL
right_turn_degree( float a, float b )
    {

    if ( b < a ) 
        return( (a - b) < 180.0 );
    else
        return( !( (b - a) < 180.0 ) ); 

    }

/*------------------------------------------------------------------------------------------*/

void 
reverse_polygon( int *x, int *y, int n )
    {
    static BOOL even;
    static int c, t, i;

    even = (n % 2) == 0;

    if ( even )
        {
        c = n / 2 - 1;
        for ( i=0; i<=c; i++ )
            {
            t        = x[n-i-1];
            x[n-i-1] = x[i];
            x[i]     = t;
            t        = y[n-i-1];
            y[n-i-1] = y[i];
            y[i]     = t;
            }
        }
    else
        {
        c = (n+1)/2 - 2;
        for ( i=0; i<=c; i++ )
            {
            t        = x[n-i-1];
            x[n-i-1] = x[i];
            x[i]     = t;
            t        = y[n-i-1];
            y[n-i-1] = y[i];
            y[i]     = t;                
            }                   // for
        }                       // if even

    } // reverse_polygon

      
/*------------------------------------------------------------------------------------------*/


void 
make_poly_ccw( int *x, int *y, int n )
    {
    // Verticies will be put in ccw order.
    // The CAD program Layout always produces ccw polygons, but you never know
    // what other programs might do. The GDS specification does not specify
    // the directionality of polygon verticies. Counter clockwise ordering
    // is important for other libraries of shape processing functions.

    static int 
        i,
        t,
        c;
    static float 
        net,
        a,
        pa;
    static BOOL
        even;


    pa = angle_of_segment( x[0], y[0], x[1], y[1] );    
    net = 0.0;

    for ( i=1; i < n-1; i++ )
        {
        a = angle_of_segment( x[i], y[i], x[i+1], y[i+1] );

        if ( ! right_turn( pa, a ) )
            {
            if ( a > pa )
                net += a - pa;
            else
                net += a - (pa - THREESIXTY);
            }
        else
            {
            if ( pa > a )
                net += a - pa;
            else
                net += (a - THREESIXTY) - pa;
            }

        pa = a;
        } 

    if ( net < 0.0 ) reverse_polygon( x, y, n );


    } // make_poly_ccw

/*------------------------------------------------------------------------------------------*/

void 
write_flat_gds( struct flat_itemtype *flat, char *filename, char *cellname, float dbu_um, float dbu_uu, BOOL dump )
    {
    int i, fd;
    FILE *dumpfile, *extentfile;
    struct flat_itemtype *item;
    static int npf3=0;
    double llx, lly, urx, ury, x, y;
    BOOL first;
    
    fd = open( filename, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if( fd <= 0 ) BAILOUT( "UNABLE TO OPEN OUTPUT FILE" );

    if ( dump ) 
        {
        dumpfile = fopen( "flat.bdm", "w" );
        if ( ! dumpfile ) BAILOUT( "UNABLE TO OPEN DUMP FILE" );
        }

    extentfile = fopen( "tmp/pattern_extents.dat", "w" );
    
    gds_create_lib_correctly( fd, filename, dbu_um, dbu_uu );
    gds_write_bgnstr( fd );
    gds_write_strname( fd, cellname );
    // gds_write_angle( fd, 0 );          // angle should be iplicitly zero
 
    item = flat->nextitem;      // the first item is a dummy, so start with the second

    if ( item == NULL ) BAILOUT( "THERE ARE NO POLYGONS TO WRITE OUT" );

    first = TRUE;
    
    while ( item != NULL && item->layer != -1 )
        {
        if ( item->n > 2 )
            {
            make_poly_ccw( item->x, item->y, item->n );

            if ( dump )
                {
                fprintf( dumpfile, "%d\n", item->n - 1 );
                for ( i=0; i < item->n - 1; i++ )
                    {
                    fprintf( dumpfile, "%d %d\n", item->x[i], item->y[i] );
                    }
                }

            if ( extentfile )
                {
                for ( i=0; i < item->n; i++ )
                    {
                    x = item->x[i] * dbu_um;
                    y = item->y[i] * dbu_um;
                    if ( first )
                        {
                        llx = x;
                        lly = y;
                        urx = x;
                        ury = y;
                        first = FALSE;
                        }
                    else
                        {
                        if ( x < llx ) llx = x;
                        if ( y < lly ) lly = y;
                        if ( x > urx ) urx = x;
                        if ( y > ury ) ury = y;
                        }
                    }
                }
            
            gds_write_boundary( fd );
            gds_write_layer( fd, item->layer );
            gds_write_datatype( fd, item->dt );
            gds_write_xy( fd, item->x, item->y, item->n );
            gds_write_endel( fd );         
            }
        else
            nag( "POLYGON HAS FEWER THAN 3 VERTICIES", &npf3 );

        item  = item->nextitem;

        } 

    gds_write_endstr( fd );
    gds_write_endlib( fd );    
    close( fd );

    if ( dump ) fclose( dumpfile );

    if ( extentfile )
        {
        fprintf( extentfile, "%lf %lf %lf %lf\n", llx, lly, urx, ury );
        fclose( extentfile );
        }

    }  // write_flat_gds

/*------------------------------------------------------------------------------------------*/

void
sine_cosine( float angle, float *sin_a, float *cos_a )
    {
    // No more floating point bullshit when the angles are multiples of 90 degrees.

    static float rangle, sina, cosa, ssa, sca;


    if ( EQUAL( fabs( fmodf( angle, 90.0 ) ), 0.0 ) )    // if the angle is a multiple of 90
	{
	if ( EQUAL( angle, 0.0 ) ) 
	    {
	    *sin_a = 0.0;
	    *cos_a = 1.0;
	    }
	else
            {
            rangle = 0.01745329 * angle;
            sina = sinf( rangle );             
            cosa = cosf( rangle ); 
            ssa = sina > 0.0 ? 1 : -1;
	    sca = cosa > 0.0 ? 1 : -1;
	    sina = fabs( sina );
	    cosa = fabs( cosa );

	    if ( sina > 0.5 ) 
	        sina = 1.0;
	    else
	        sina = 0.0;

	    if ( cosa > 0.5 ) 
		cosa = 1.0;
	    else
		cosa = 0.0;

	    *sin_a =  ssa * sina;
	    *cos_a =  sca * cosa;
            }
	}
    else
	{
	rangle = 0.01745329 * angle;
	*sin_a = sinf( rangle );             
	*cos_a = cosf( rangle ); 
	}
    }

//------------------------------------------------------------------------------------------

struct flat_itemtype *
copy_boundary( struct gds_itemtype *item, float mag, float angle, BOOL reflect, int x0, int y0 )
    {
    static int i;
    static float s, x, y, xp, yp, sin_a, cos_a;
    static struct flat_itemtype *newthing;

    if ( item->n <= 0 ) BAILOUT( "POLYGON HAS NO VERTICES" );
    newthing = (struct flat_itemtype *) malloc( sizeof( struct flat_itemtype ) );
    if ( ! newthing ) BAILOUT( "UNABLE TO ALLOCATE NEWTHING IN COPY_BOUNDARY" );
    newthing->nextitem = NULL;
    
    newthing->n     = item->n;
    newthing->layer = item->layer;
    newthing->dt    = item->dt;

    newthing->x = (int *) malloc( newthing->n * sizeof( int ) );
    newthing->y = (int *) malloc( newthing->n * sizeof( int ) );

    if ( ! newthing->x || ! newthing->y ) BAILOUT( "UNABLE TO ALLOCTE NEW VERTICIES FOR POLYGON" );

    for ( i=0; i < newthing->n; i++ )
        {
        newthing->x[i] = item->x[i];
        if ( reflect )
            newthing->y[i] = -item->y[i];
        else
            newthing->y[i] =  item->y[i];
        }
        
    if ( ( ! EQUAL(mag, 1.0) ) || ( ! EQUAL(angle, 0.0) ) )  // then rotate and scale
        {

        sine_cosine( angle, &sin_a, &cos_a );  // angle in degrees, sin and cos snap properly

        for ( i=0; i < newthing->n; i++ )
            {
	    x = newthing->x[i] * mag;
	    y = newthing->y[i] * mag; 
            xp = x * cos_a - y * sin_a;
            yp = x * sin_a + y * cos_a;
            newthing->x[i] = floor( xp );
            newthing->y[i] = floor( yp );
            }
        }


    for ( i=0; i < newthing->n; i++ )      // shift after scaling
        {                                
        newthing->x[i] = x0 + newthing->x[i];
        newthing->y[i] = y0 + newthing->y[i];
        }

    return( newthing );

    }  // copy_boundary


/*------------------------------------------------------------------------------------------*/

void
massage_angles( float *a, float *b )
    {
    if ( *a >= 0.0 && *a < ONEIGHTY )
        {
        if ( right_turn( *a, *b ) && *b > ONEIGHTY ) *b = *b - THREESIXTY;
        }
    else
        {
        if ( ! right_turn( *a, *b ) && *b < ONEIGHTY ) *b = *b - THREESIXTY;
        }
    }

/*------------------------------------------------------------------------------------------*/

void
massage_degrees( float *a, float *b )
    {
    if ( *a >= 0.0 && *a < 180.0 )
        {
        if ( right_turn_degree( *a, *b ) && *b > 180.0 ) *b = *b - 360.0;
        }
    else
        {
        if ( ! right_turn_degree( *a, *b ) && *b < 180.0 ) *b = *b - 360.0;
        }
    }

/*------------------------------------------------------------------------------------------*/

BOOL
equal( float a, float b )
    {
    return( fabs(a - b) < 1e-6 );
    }


struct gds_itemtype *
path_start( struct gds_itemtype *item, struct gds_itemtype *p,
            struct intersection_coordinate *pI, struct intersection_coordinate *pC ) 
    {
    static int 
        i, n,
        w2sina,
        w2cosa,
        L1x, L1y,
        R1x, R1y,
        R2x, R2y,
        L2x, L2y,
        L3x, L3y,
        R3x, R3y,
        LL1x, LL1y,
        RR1x, RR1y,
        LL2x, LL2y,
        RR2x, RR2y,
        Cx, Cy,
        P1x, P1y,
        P2x, P2y,
        P3x, P3y;
    static BOOL 
        extended, 
        round;
    static float
        w2sina1,
        w2sina2,
        w2cosa1,
        w2cosa2, 
        a1, a2, 
        w2,
        dh,
        ca;
    static struct intersection_coordinate
        I;


    P1x = item->x[0];
    P1y = item->y[0];
    P2x = item->x[1];
    P2y = item->y[1];
    P3x = item->x[2];
    P3y = item->y[2];

    w2 = item->width / 2.0;

    a1 = angle_of_segment( P1x, P1y, P2x, P2y ); 
    a2 = angle_of_segment( P2x, P2y, P3x, P3y );

    extended = item->path_end == 2;
    round    = item->path_end == 1;

    if ( extended )
        {
        P1x = P1x + w2 * cosf( a1 - ONEIGHTY );
        P1y = P1y + w2 * sinf( a1 - ONEIGHTY );
        }
    
    w2sina1 = floor( w2 * sinf( a1 ) );
    w2cosa1 = floor( w2 * cosf( a1 ) );
    w2sina2 = floor( w2 * sinf( a2 ) );
    w2cosa2 = floor( w2 * cosf( a2 ) );

    L1x  = P1x - w2sina1;
    L1y  = P1y + w2cosa1;
    R1x  = P1x + w2sina1;
    R1y  = P1y - w2cosa1;

    if ( equal( fabs(a1), 0      ) || equal( fabs(a1), ONEIGHTY   ) ) R1x = L1x; 
    if ( equal( fabs(a1), NINETY ) || equal( fabs(a1), TWOSEVENTY ) ) R1y = L1y; 
  
    R2x  = P2x + w2sina1;
    R2y  = P2y - w2cosa1;
    L2x  = P2x - w2sina1;
    L2y  = P2y + w2cosa1;

    L3x  = P3x - w2sina2;
    L3y  = P3y + w2cosa2;
    R3x  = P3x + w2sina2;
    R3y  = P3y - w2cosa2;

    LL2x = P2x - w2sina2;
    LL2y = P2y + w2cosa2;
    RR2x = P2x + w2sina2;
    RR2y = P2y - w2cosa2;

    LL1x = P1x + floor( w2 * cosf( a1 - ONEIGHTY ) - w2sina1/2.0 );   
    LL1y = P1y + floor( w2 * sinf( a1 - ONEIGHTY ) + w2cosa1/2.0 ); 
    RR1x = P1x + floor( w2 * cosf( a1 - ONEIGHTY ) + w2sina1/2.0 ); 
    RR1y = P1y + floor( w2 * sinf( a1 - ONEIGHTY ) - w2cosa1/2.0 ); 

    if ( round )
        {
        p->x = (int *) malloc( 8 * sizeof( int ) );
        p->y = (int *) malloc( 8 * sizeof( int ) );
        }
    else
        {
        p->x = (int *) malloc( 6 * sizeof( int ) );
        p->y = (int *) malloc( 6 * sizeof( int ) );
        }


    if ( ! p->x || ! p->y ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR PATH CONVERSION" );

    if ( right_turn( a1, a2 ) )
        {
        I = intersection( R1x, R1y, R2x, R2y, R3x, R3y, RR2x, RR2y );
        if ( ! I.exists ) 
            {
            printf( "\nERROR: STRANGE PATH, SOMEWHERE AROUND %d, %d (database units) \n", P1x, P1y );
            printf(   "       This is probably a path that traces over itself, \n" );
            printf(   "       creating an invisible overlap and therefore a double exposure.\n" );
            printf(   "       Find this stupid path and turn it into a normal polygon.\n" );
            BAILOUT( "YOUR PATTERN IS STUPID" );
            }
        massage_angles( &a1, &a2 );
        ca = fmodf( (a1 + a2)/2.0 + NINETY, THREESIXTY );
        Cx = P2x + floor( w2 * cosf( ca ) );
        Cy = P2y + floor( w2 * sinf( ca ) );
        //printf( "a1 = %f  a2 = %f  ca = %f \n", a1*180.0/ONEIGHTY, a2*180.0/ONEIGHTY, ca*180.0/ONEIGHTY );
        p->type        = 0;  // polygon
        p->cell_number = -1;
        p->mag         = 1.0;      
        p->angle       = 0.0;    
        p->abs_angle   = FALSE;
        p->abs_mag     = FALSE;
        p->reflect     = FALSE;
        p->layer       = item->layer;
        p->dt          = item->dt;

        if ( round )
            {
            p->n = 8;
            p->x[0] = R1x;  p->y[0] = R1y;
            p->x[1] = I.x;  p->y[1] = I.y;
            p->x[2] = Cx;   p->y[2] = Cy;
            p->x[3] = L2x;  p->y[3] = L2y;
            p->x[4] = L1x;  p->y[4] = L1y;
            p->x[5] = LL1x; p->y[5] = LL1y;
            p->x[6] = RR1x; p->y[6] = RR1y;
            p->x[7] = R1x;  p->y[7] = R1y;
            }
        else
            {
            p->n = 6;
            p->x[0] = R1x;  p->y[0] = R1y;
            p->x[1] = I.x;  p->y[1] = I.y;
            p->x[2] = Cx;   p->y[2] = Cy;
            p->x[3] = L2x;  p->y[3] = L2y;
            p->x[4] = L1x;  p->y[4] = L1y;
            p->x[5] = R1x;  p->y[5] = R1y;
            }
        
        p->nextitem = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        if ( ! p->nextitem ) BAILOUT( "UNABLE TO ALLOCATE NEXT ITEM" );
        p = p->nextitem;
        }
    else
        {
        I = intersection( L1x, L1y, L2x, L2y, LL2x, LL2y, L3x, L3y );
        if ( ! I.exists ) 
            {
            printf( "\nERROR: STRANGE PATH, SOMEWHERE AROUND %d, %d (database units) \n", P1x, P1y );
            printf(   "       This is probably a path that traces over itself, \n" );
            printf(   "       creating an invisible overlap and therefore a double exposure.\n" );
            printf(   "       Find this stupid path and turn it into a normal polygon.\n" );
            BAILOUT( "YOUR PATTERN IS STUPID" );
            }
        massage_angles( &a1, &a2 );
        ca = fmodf( (a1 + a2)/2.0 - NINETY, THREESIXTY );
        //printf( "a1 = %f  a2 = %f  ca = %f \n", a1*180.0/ONEIGHTY, a2*180.0/ONEIGHTY, ca*180.0/ONEIGHTY );
        Cx = P2x + floor( w2 * cosf( ca ) );
        Cy = P2y + floor( w2 * sinf( ca ) );

        p->type        = 0;  // polygon
        p->cell_number = -1;
        p->mag         = 1.0;      
        p->angle       = 0.0;    
        p->abs_angle   = FALSE;
        p->abs_mag     = FALSE;
        p->reflect     = FALSE;
        p->layer       = item->layer;
        p->dt          = item->dt;

        if ( round )
            {
            p->n = 8;
            p->x[0] = L1x;  p->y[0] = L1y;
            p->x[1] = I.x;  p->y[1] = I.y;
            p->x[2] = Cx;   p->y[2] = Cy;
            p->x[3] = R2x;  p->y[3] = R2y;
            p->x[4] = R1x;  p->y[4] = R1y;
            p->x[5] = RR1x; p->y[5] = RR1y;
            p->x[6] = LL1x; p->y[6] = LL1y;
            p->x[7] = L1x;  p->y[7] = L1y;
            }
        else
            {
            p->n = 6;
            p->x[0] = L1x;  p->y[0] = L1y;
            p->x[1] = I.x;  p->y[1] = I.y;
            p->x[2] = Cx;   p->y[2] = Cy;
            p->x[3] = R2x;  p->y[3] = R2y;
            p->x[4] = R1x;  p->y[4] = R1y;
            p->x[5] = L1x;  p->y[5] = L1y;
            }
        
        p->nextitem = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        if ( ! p->nextitem ) BAILOUT( "UNABLE TO ALLOCATE NEXT ITEM" );
        p = p->nextitem;
        }

    pI->x = I.x;
    pI->y = I.y;
    pI->exists = TRUE;
    pC->x = Cx;
    pC->y = Cy;
    pC->exists = TRUE;

    return( p );

    }  // path_start

/*------------------------------------------------------------------------------------------*/

struct gds_itemtype *
path_end( struct gds_itemtype *item, struct gds_itemtype *p, struct intersection_coordinate I )
    {
    static int 
        i, n,
        w2sina,
        w2cosa,
        L1x, L1y,
        R1x, R1y,
        R2x, R2y,
        L2x, L2y,
        L3x, L3y,
        R3x, R3y,
        LL2x, LL2y,
        RR2x, RR2y,
        LLLx, LLLy,
        RRRx, RRRy,
        Cx, Cy,
        P1x, P1y,
        P2x, P2y,
        P3x, P3y;
    static BOOL 
        extended, 
        round;
    static float
        endwidth,
        w2sina1,
        w2sina2,
        w2cosa1,
        w2cosa2, 
        a1, a2, 
        w2,
        dh,
        ca;
    

    n = item->n - 1;
    if ( n < 2 ) BAILOUT( "TOO FEW POINTS ON PATH - LOGIC ERROR" );
    
    P1x = item->x[n-2];
    P1y = item->y[n-2];
    P2x = item->x[n-1];
    P2y = item->y[n-1];
    P3x = item->x[n];
    P3y = item->y[n];

    a1 = angle_of_segment( P1x, P1y, P2x, P2y ); 
    a2 = angle_of_segment( P2x, P2y, P3x, P3y );

    extended = item->path_end == 2;
    round    = item->path_end == 1;

    w2 = item->width / 2.0;

    if ( extended )
        {
        P3x = P3x + floor( w2 * cosf( a2 ) );
        P3y = P3y + floor( w2 * sinf( a2 ) );
        }

    w2sina1 = floor( w2 * sinf( a1 ) );
    w2cosa1 = floor( w2 * cosf( a1 ) );
    w2sina2 = floor( w2 * sinf( a2 ) );
    w2cosa2 = floor( w2 * cosf( a2 ) );

    L1x  = P1x - w2sina1;
    L1y  = P1y + w2cosa1;
    R1x  = P1x + w2sina1;
    R1y  = P1y - w2cosa1;
    R2x  = P2x + w2sina1;
    R2y  = P2y - w2cosa1;
    L2x  = P2x - w2sina1;
    L2y  = P2y + w2cosa1;
    L3x  = P3x - w2sina2;
    L3y  = P3y + w2cosa2;
    R3x  = P3x + w2sina2;
    R3y  = P3y - w2cosa2;
    LL2x = P2x - w2sina2;
    LL2y = P2y + w2cosa2;
    RR2x = P2x + w2sina2;
    RR2y = P2y - w2cosa2;

    if ( equal( fabs(a2), 0      ) || equal( fabs(a2), ONEIGHTY   ) ) R3x = L3x; 
    if ( equal( fabs(a2), NINETY ) || equal( fabs(a2), TWOSEVENTY ) ) R3y = L3y; 

    RRRx = P3x + floor( w2 * cosf( a2 ) + w2sina2/2.0 ); 
    RRRy = P3y + floor( w2 * sinf( a2 ) - w2cosa2/2.0 ); 
    LLLx = P3x + floor( w2 * cosf( a2 ) - w2sina2/2.0 );
    LLLy = P3y + floor( w2 * sinf( a2 ) + w2cosa2/2.0 );

    if ( round )
        {
        p->x = (int *) malloc( 8 * sizeof( int ) );
        p->y = (int *) malloc( 8 * sizeof( int ) );
        }
    else
        {
        p->x = (int *) malloc( 6 * sizeof( int ) );
        p->y = (int *) malloc( 6 * sizeof( int ) );
        }
    
    if ( ! p->x || ! p->y ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR PATH CONVERSION" );

    if ( right_turn( a1, a2 ) )
        {
        //printf( "right turn\n" );
        if ( ! I.exists ) I = intersection( R1x, R1y, R2x, R2y, R3x, R3y, RR2x, RR2y );
        if ( ! I.exists ) 
            {
            printf( "\nERROR: STRANGE PATH, SOMEWHERE AROUND %d, %d (database units) \n", P1x, P1y );
            printf(   "       This is probably a path that traces over itself, \n" );
            printf(   "       creating an invisible overlap and therefore a double exposure.\n" );
            printf(   "       Find this stupid path and turn it into a normal polygon.\n" );
            BAILOUT( "YOUR PATTERN IS STUPID" );
            }
        massage_angles( &a1, &a2 );
        ca = fmodf( (a1 + a2)/2.0 + NINETY, THREESIXTY );
        Cx = P2x + floor( w2 * cosf( ca ) );
        Cy = P2y + floor( w2 * sinf( ca ) );

        p->type        = 0;  // polygon
        p->cell_number = -1;
        p->mag         = 1.0;      
        p->angle       = 0.0;    
        p->abs_angle   = FALSE;
        p->abs_mag     = FALSE;
        p->reflect     = FALSE;
        p->layer       = item->layer;
        p->dt          = item->dt;

        if ( round )
            {
            p->n = 8;
            p->x[0] = I.x;  p->y[0] = I.y;
            p->x[1] = R3x;  p->y[1] = R3y;
            p->x[2] = RRRx; p->y[2] = RRRy;
            p->x[3] = LLLx; p->y[3] = LLLy;
            p->x[4] = L3x;  p->y[4] = L3y;
            p->x[5] = LL2x; p->y[5] = LL2y;
            p->x[6] = Cx;   p->y[6] = Cy;
            p->x[7] = I.x;  p->y[7] = I.y;
            }
        else
            {
            p->n = 6;
            p->x[0] = I.x;  p->y[0] = I.y;
            p->x[1] = R3x;  p->y[1] = R3y;
            p->x[2] = L3x;  p->y[2] = L3y;
            p->x[3] = LL2x; p->y[3] = LL2y;
            p->x[4] = Cx;   p->y[4] = Cy;
            p->x[5] = I.x;  p->y[5] = I.y;
            }
        
        }
    else
        {
        //printf( "left turn\n" );
        if ( ! I.exists ) I = intersection( L1x, L1y, L2x, L2y, LL2x, LL2y, L3x, L3y );
        if ( ! I.exists ) 
            {
            printf( "\nERROR: STRANGE PATH, SOMEWHERE AROUND %d, %d (database units) \n", P1x, P1y );
            printf(   "       This is probably a path that traces over itself, \n" );
            printf(   "       creating an invisible overlap and therefore a double exposure.\n" );
            printf(   "       Find this stupid path and turn it into a normal polygon.\n" );
            BAILOUT( "YOUR PATTERN IS STUPID" );
            }

        massage_angles( &a1, &a2 );
        ca = fmodf( (a1 + a2)/2.0 - NINETY, THREESIXTY );
        Cx = P2x + floor( w2 * cosf( ca ) );
        Cy = P2y + floor( w2 * sinf( ca ) );

        p->type        = 0;  // polygon
        p->cell_number = -1;
        p->mag         = 1.0;      
        p->angle       = 0.0;    
        p->abs_angle   = FALSE;
        p->abs_mag     = FALSE;
        p->reflect     = FALSE;
        p->layer       = item->layer;
        p->dt          = item->dt;

        if ( round )
            {
            p->n = 8;
            p->x[0] = I.x;  p->y[0] = I.y;
            p->x[1] = Cx;   p->y[1] = Cy;
            p->x[2] = RR2x; p->y[2] = RR2y;
            p->x[3] = R3x;  p->y[3] = R3y;
            p->x[4] = RRRx; p->y[4] = RRRy;
            p->x[5] = LLLx; p->y[5] = LLLy;
            p->x[6] = L3x;  p->y[6] = L3y;
            p->x[7] = I.x;  p->y[7] = I.y;
            }
        else
            {
            p->n = 6;
            p->x[0] = I.x;  p->y[0] = I.y;
            p->x[1] = Cx;   p->y[1] = Cy;
            p->x[2] = RR2x; p->y[2] = RR2y;
            p->x[3] = R3x;  p->y[3] = R3y;
            p->x[4] = L3x;  p->y[4] = L3y;
            p->x[5] = I.x;  p->y[5] = I.y;
            }
        
        }

    return( p );

    }  // path_end

/*------------------------------------------------------------------------------------------*/

struct gds_itemtype *
path_middle( struct gds_itemtype *item, struct gds_itemtype *p, int n, 
             struct intersection_coordinate *pI2,
             struct intersection_coordinate *pC  )
    {
    static int 
        i, 
        L1x, L1y,
        R1x, R1y,
        R2x, R2y,
        L2x, L2y,
        L3x, L3y,
        R3x, R3y,
        L4x, L4y,
        R4x, R4y,
        L5x, L5y,
        R5x, R5y,
        L6x, L6y,
        R6x, R6y,
        C1x, C1y,
        C2x, C2y,
        P1x, P1y,
        P2x, P2y,
        P3x, P3y,
        P4x, P4y;
    static BOOL 
        right_turn_1,
        right_turn_2,
        extended, 
        round;
    static float
        w2sina,
        w2cosa,
        w2sina1,
        w2sina2,
        w2cosa1,
        w2cosa2, 
        w2sina3,
        w2cosa3,
        a1, a2, a3,
        a, b,
        w2,
        dh,
        ca;
    static struct intersection_coordinate
        I1,
        I2;

    if ( (n+3) > item->n ) BAILOUT( "LOGIC ERROR IN PATH_MIDDLE" );
    if ( (n-1) < 0 )       BAILOUT( "LOGIC ERROR IN PATH_MIDDLE" );

    P1x = item->x[n-1];
    P1y = item->y[n-1];
    P2x = item->x[n];
    P2y = item->y[n];
    P3x = item->x[n+1];
    P3y = item->y[n+1];
    P4x = item->x[n+2];
    P4y = item->y[n+2];

    a1 = angle_of_segment( P1x, P1y, P2x, P2y ); 
    a2 = angle_of_segment( P2x, P2y, P3x, P3y );
    a3 = angle_of_segment( P3x, P3y, P4x, P4y );

    w2 = item->width / 2.0;

    w2sina1 = floor( w2 * sinf( a1 ) );
    w2cosa1 = floor( w2 * cosf( a1 ) );
    w2sina2 = floor( w2 * sinf( a2 ) );
    w2cosa2 = floor( w2 * cosf( a2 ) );
    w2sina3 = floor( w2 * sinf( a3 ) );
    w2cosa3 = floor( w2 * cosf( a3 ) );

    L1x  = P1x - w2sina1;  L1y  = P1y + w2cosa1;
    L2x  = P2x - w2sina1;  L2y  = P2y + w2cosa1;
    L3x  = P2x - w2sina2;  L3y  = P2y + w2cosa2;
    L4x  = P3x - w2sina2;  L4y  = P3y + w2cosa2;   
    L5x  = P3x - w2sina3;  L5y  = P3y + w2cosa3;
    L6x  = P4x - w2sina3;  L6y  = P4y + w2cosa3;
    R1x  = P1x + w2sina1;  R1y  = P1y - w2cosa1;
    R2x  = P2x + w2sina1;  R2y  = P2y - w2cosa1;
    R3x  = P2x + w2sina2;  R3y  = P2y - w2cosa2;
    R4x  = P3x + w2sina2;  R4y  = P3y - w2cosa2;
    R5x  = P3x + w2sina3;  R5y  = P3y - w2cosa3;
    R6x  = P4x + w2sina3;  R6y  = P4y - w2cosa3;

    p->n = 7;
    p->type        = 0;  // polygon
    p->cell_number = -1;
    p->mag         = 1.0;      
    p->angle       = 0.0;    
    p->abs_angle   = FALSE;
    p->abs_mag     = FALSE;
    p->reflect     = FALSE;
    p->layer       = item->layer;
    p->dt          = item->dt;
    p->x = (int *) malloc( 7 * sizeof( int ) );
    p->y = (int *) malloc( 7 * sizeof( int ) );
    if ( ! p->x || ! p->y ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR PATH CONVERSION" );

    right_turn_1 = right_turn( a1, a2 );
    right_turn_2 = right_turn( a2, a3 );

    if ( right_turn_1 )
        {
        I1 = intersection( R1x, R1y, R2x, R2y, R4x, R4y, R3x, R3y );
        }
    else
        {
        I1 = intersection( L1x, L1y, L2x, L2y, L3x, L3y, L4x, L4y );
        }

    if ( ! I1.exists ) 
        {
        printf( "\nERROR: STRANGE PATH, SOMEWHERE AROUND %d, %d (database units) \n", P1x, P1y );
        printf(   "       This is probably a path that traces over itself, \n" );
        printf(   "       creating an invisible overlap and therefore a double exposure.\n" );
        printf(   "       Find this stupid path and turn it into a normal polygon.\n" );
        BAILOUT( "YOUR PATTERN IS STUPID" );
        }

    if ( right_turn_2 )
        {
        I2 = intersection( R5x, R5y, R6x, R6y, R3x, R3y, R4x, R4y );
        }
    else
        {
        I2 = intersection( L5x, L5y, L6x, L6y, L3x, L3y, L4x, L4y );
        }

    if ( ! I2.exists ) 
        {
        printf( "\nERROR: STRANGE PATH, SOMEWHERE AROUND %d, %d (database units) \n", P1x, P1y );
        printf(   "       This is probably a path that traces over itself, \n" );
        printf(   "       creating an invisible overlap and therefore a double exposure.\n" );
        printf(   "       Find this stupid path and turn it into a normal polygon.\n" );
        BAILOUT( "YOUR PATTERN IS STUPID" );
        }

    a = a1;
    b = a2;
    massage_angles( &a, &b );

    if ( right_turn_1 )
        ca = fmodf( (a + b)/2.0 + NINETY, THREESIXTY );
    else
        ca = fmodf( (a + b)/2.0 - NINETY, THREESIXTY );

    if ( ! pC->exists )
        {
        C1x = P2x + floor( w2 * cosf( ca ) );
        C1y = P2y + floor( w2 * sinf( ca ) );
        }
    else
        {
        C1x = pC->x;
        C1y = pC->y;
        }

    a = a2;
    b = a3;
    massage_angles( &a, &b );

    if ( right_turn_2 )
        ca = fmodf( (a + b)/2.0 + NINETY, THREESIXTY );
    else
        ca = fmodf( (a + b)/2.0 - NINETY, THREESIXTY );

    C2x = P3x + floor( w2 * cosf( ca ) );
    C2y = P3y + floor( w2 * sinf( ca ) );

    pC->x = C2x;
    pC->y = C2y;
    pC->exists = TRUE;

    if ( right_turn_1 && right_turn_2 )
        {
        // I1 C1 L3 L4 C2 I2
        p->x[0] = I1.x;     p->y[0] = I1.y;
        p->x[1] = C1x;      p->y[1] = C1y;
        p->x[2] = L3x;      p->y[2] = L3y;
        p->x[3] = L4x;      p->y[3] = L4y;
        p->x[4] = C2x;      p->y[4] = C2y;
        p->x[5] = I2.x;     p->y[5] = I2.y;
        p->x[6] = p->x[0];  p->y[6] = p->y[0];
        }
    else if ( right_turn_1 && ! right_turn_2 )
        {
        // I1 C1 L3 I2 C2 R4 
        p->x[0] = I1.x;     p->y[0] = I1.y;
        p->x[1] = C1x;      p->y[1] = C1y;
        p->x[2] = L3x;      p->y[2] = L3y;
        p->x[3] = I2.x;     p->y[3] = I2.y;
        p->x[4] = C2x;      p->y[4] = C2y;
        p->x[5] = R4x;      p->y[5] = R4y;
        p->x[6] = p->x[0];  p->y[6] = p->y[0];
        }
    else if ( ! right_turn_1 && right_turn_2 )
        {
        // I1 C1 R3 I2 C2 L4
        p->x[0] = I1.x;     p->y[0] = I1.y;
        p->x[1] = C1x;      p->y[1] = C1y;
        p->x[2] = R3x;      p->y[2] = R3y;
        p->x[3] = I2.x;     p->y[3] = I2.y;
        p->x[4] = C2x;      p->y[4] = C2y;
        p->x[5] = L4x;      p->y[5] = L4y;
        p->x[6] = p->x[0];  p->y[6] = p->y[0];
        }
    else // left, left
        {
        // I1 C1 R3 R4 C2 I2
        p->x[0] = I1.x;     p->y[0] = I1.y;
        p->x[1] = C1x;      p->y[1] = C1y;
        p->x[2] = R3x;      p->y[2] = R3y;
        p->x[3] = R4x;      p->y[3] = R4y;
        p->x[4] = C2x;      p->y[4] = C2y;
        p->x[5] = I2.x;     p->y[5] = I2.y;
        p->x[6] = p->x[0];  p->y[6] = p->y[0];
        }

    p->nextitem = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! p->nextitem ) BAILOUT( "UNABLE TO ALLOCATE NEXT ITEM" );
    p = p->nextitem;

    pI2->x = I2.x;
    pI2->y = I2.y;
    pI2->exists = I2.exists;


    return( p );

    }  // path_middle

/*------------------------------------------------------------------------------------------*/

struct gds_itemtype *
path2poly( struct gds_itemtype *item )
    {
    // render paths as a set of polygons,
    // which is amazingly complicated

    static int 
        verticies,
        i, n,
        LL1x, LL1y,
        RR1x, RR1y,
        LL2x, LL2y,
        RR2x, RR2y,
        L1x, L1y,
        R1x, R1y,
        R2x, R2y,
        L2x, L2y,
        L3x, L3y,
        R3x, R3y,
        Cx, Cy,
        P1x, P1y,
        P2x, P2y,
        P3x, P3y;
    static BOOL 
        extended, 
        round;
    static float
        endwidth,
        w2sina,
        w2cosa,
        w2sina1,
        w2sina2,
        w2cosa1,
        w2cosa2, 
        a1, a2, 
        w2,
        dh,
        ca;
    static struct gds_itemtype 
        *polys,
        *first,
        *next,
        *p;
    static struct intersection_coordinate
        I, C;

    I.exists = FALSE;
    C.exists = FALSE;

    if ( item->type != 1 ) BAILOUT( "WRONG TYPE THING SENT TO PATH2POLY" );
    if ( item->n    <  2 ) BAILOUT( "TOO FEW VERTICIES ON A PATH" );

    polys = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! polys ) BAILOUT( "UNABLE TO ALLOCATE NEW POLYGON LIST" );

                              // after we turn the path into polygons,
                              // we will replace "item" with "polys"
    first = item;             // and then wrap it back to the orginal list
    next  = item->nextitem;   // by sticking this on the end 

    round    = item->path_end == 1;
    extended = item->path_end == 2;

    w2       = item->width / 2.0;
    n        = item->n;
    p        = polys;

    if ( n == 2 )    // no bends - just beginning and end
        {
        P1x = item->x[0];
        P1y = item->y[0];
        P2x = item->x[1];
        P2y = item->y[1];

        a1 = angle_of_segment( P1x, P1y, P2x, P2y ); 

        if ( extended )
            {
            P1x = P1x + floor( w2 * cosf( a1 - ONEIGHTY ) );
            P1y = P1y + floor( w2 * sinf( a1 - ONEIGHTY ) );
            P2x = P2x + floor( w2 * cosf( a1            ) );
            P2y = P2y + floor( w2 * sinf( a1            ) );
            }

        w2sina1 = floor( w2 * sinf( a1 ) );
        w2cosa1 = floor( w2 * cosf( a1 ) );

        L1x  = P1x - w2sina1;
        L1y  = P1y + w2cosa1;
        R1x  = P1x + w2sina1;
        R1y  = P1y - w2cosa1;
        R2x  = P2x + w2sina1;
        R2y  = P2y - w2cosa1;
        L2x  = P2x - w2sina1;
        L2y  = P2y + w2cosa1;

        LL1x = P1x + floor( w2 * cosf( a1 - ONEIGHTY ) - w2sina1/2.0 );   
        LL1y = P1y + floor( w2 * sinf( a1 - ONEIGHTY ) + w2cosa1/2.0 ); 
        RR1x = P1x + floor( w2 * cosf( a1 - ONEIGHTY ) + w2sina1/2.0 ); 
        RR1y = P1y + floor( w2 * sinf( a1 - ONEIGHTY ) - w2cosa1/2.0 ); 
        RR2x = P2x + floor( w2 * cosf( a1            ) + w2sina1/2.0 ); 
        RR2y = P2y + floor( w2 * sinf( a1            ) - w2cosa1/2.0 ); 
        LL2x = P2x + floor( w2 * cosf( a1            ) - w2sina1/2.0 );
        LL2y = P2y + floor( w2 * sinf( a1            ) + w2cosa1/2.0 );

        if ( round )
            verticies = 9;
        else
            verticies = 5;
        
        p->x = (int *) malloc( verticies * sizeof( int ) );
        p->y = (int *) malloc( verticies * sizeof( int ) );

        if ( ! p->x || ! p->y ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR PATH CONVERSION" );

        if ( round )
            {
	    p->n = 9;
	    p->type        = 0;  // polygon
	    p->cell_number = -1;
	    p->mag         = 1.0;      
	    p->angle       = 0.0;    
	    p->abs_angle   = FALSE;
	    p->abs_mag     = FALSE;
	    p->reflect     = FALSE;
	    p->layer       = item->layer;
	    p->dt          = item->dt;
	    p->x[0] = R1x;  p->y[0] = R1y;   
	    p->x[1] = R2x;  p->y[1] = R2y;   
            p->x[2] = RR2x; p->y[2] = RR2y;  
            p->x[3] = LL2x; p->y[3] = LL2y;  
	    p->x[4] = L2x;  p->y[4] = L2y;   
	    p->x[5] = L1x;  p->y[5] = L1y;   
	    p->x[6] = LL1x; p->y[6] = LL1y;  
            p->x[7] = RR1x; p->y[7] = RR1y;  
	    p->x[8] = R1x;  p->y[8] = R1y;   
            }
        else
            {
	    p->n = 5;
	    p->type        = 0;  // polygon
	    p->cell_number = -1;
	    p->mag         = 1.0;      
	    p->angle       = 0.0;    
	    p->abs_angle   = FALSE;
	    p->abs_mag     = FALSE;
	    p->reflect     = FALSE;
	    p->layer       = item->layer;
	    p->dt          = item->dt;
	    p->x[0] = R1x;  p->y[0] = R1y;
	    p->x[1] = R2x;  p->y[1] = R2y;
	    p->x[2] = L2x;  p->y[2] = L2y;
	    p->x[3] = L1x;  p->y[3] = L1y;
	    p->x[4] = R1x;  p->y[4] = R1y;
            }

        
        item = polys;                // put the list of polygons on the main list      
        polys->nextitem = next;      // return to the previous list

        }                            // end if n == 2
    else if ( n == 3 )
        {                                    // remember that p points to polys initially
        p = path_start( item, p, &I, &C );   // p is updated but is not yet wrapped back 
        p = path_end(   item, p, I );   
        item = polys;
        p->nextitem = next;
        }                            
    else         
        {
        p = path_start( item, p, &I, &C ); 
        for ( i=1; i<=(n-3); i++ )
            p = path_middle( item, p, i, &I, &C );
        p = path_end( item, p, I );   
        item = polys;
        p->nextitem = next;
        }

    return( item );

    } // path2poly

/*------------------------------------------------------------------------------------------*/

void
flatten_aref( struct gds_itemtype *item, float mag, float angle, BOOL mirror, int x0, int y0, char *callingcell )
    {

    // Turn an aref into a set of srefs, injected onto the list 
    // and replacing the current aref item.

    // If the array itself includes mag != 1 then this is ambiguous-- because 
    // the GDS specification does not say whether or not to scale the 
    // array pitch. Same deal with mirroring and rotation. 
    // The GDS specification actually _is_ clear
    // about rotation, but programs such as Layout do not get it right. 
    // We will try to fix these problems, while warning the user.


    static struct gds_itemtype
        *needle,
        *previous,
        *the_rest;

    static int
        nap=0,
        nap2=0,
        nra=0,
        nsa=0,
        nrfa=0,
        nrfa2=0,
        cellnum=0,
        cols=0,
        rows=0,
        dx=0,
        dy=0,
        dxy=0,
        dyx=0,
        r=0,
	c=0;

    static float
        x=0,
        y=0,
        sin_a=0,
        cos_a=0,
        new_angle=0,
        new_mag=1;

    if ( item->type    != 2 )   
        {
        printf( "\nFLATTEN_AREF CANNOT FLATTEN THINGS OTHER THAN AREFS" );
        if ( item->cell_number >= 0 ) printf(  "         The offending cell is [%s]\n", callingcell );
        BAILOUT( "YOU IDIOT" );
        }

    if ( (item->mag != 1.0) && (item->angle != 0.0) ) 
        {
	printf( "\nSCALED AND ROTATED ARRAY. THE GDS SPECIFICATION \n" );
        printf( "IS TOO AMBIGUOUS. YOU SHOULD USE SCALED AND ROTATED CELLS, \n" );
        printf( "NOT SCALED AND ROTATED ARRAYS.\n" );
        if ( item->cell_number >= 0 ) printf(  "         The offending cell is [%s]\n", callingcell );
        BAILOUT( "YOU DOPE" );
        }        

    if ( item->angle   != 0.0 ) 
        {
        nag( "ROTATED ARRAY. You should use arrays of rotated cells instead.\n   Attempting to fix this error assuming 'Layout' style. This fix might be wrong.", &nra );
        if ( item->cell_number >= 0 && nra < 10 ) 
            printf(  "         The offending cell is [%s]\n", callingcell );
        }

    if ( item->mag     != 1.0 ) 
        {
        nag( "SCALED ARRAY. You should use arrays of scaled cells instead.\n     Attempting to fix this error assuming 'Layout' style. This fix might be wrong.", &nsa );
        if ( item->cell_number >= 0 && nsa < 10) printf(  "         The offending cell is [%s]\n", callingcell );
        }

    if ( item->reflect )        
        {
        nag( "REFLECTED ARRAYS ARE NOT ALLOWED, BECAUSE THE GDS SPECIFICATION IS AMBIGUOUS.", &nrfa );
        nag( "INSTEAD, YOU SHOULD REFLECT THE CELL CONTAINING THE ARRAY, \nOR YOU SHOULD REFLECT THE ARRAYED CELL.", &nrfa2 );
        if ( item->cell_number >= 0 && nrfa < 10 ) printf(  "         The offending cell is [%s]\n", callingcell );
        }

    if ( mag <= 0.0 )           
        {
        printf( "\nMAGNIFICATION PASSED TO FLATTEN_AREF IS ZERO\n" );
        if ( item->cell_number >= 0 ) printf(  "         The offending cell is [%s]\n", callingcell );
        BAILOUT( "ABORT" );
        }        

    the_rest = item->nextitem;
    
    new_angle = item->angle;  // fixes that pesky bug from Layout
    new_mag   = item->mag;    // yup, another Layout bug

    cellnum = item->cell_number;                        
    cols    = item->cols;
    rows    = item->rows;
    dx      = mag * item->col_pitch;
    dy      = mag * item->row_pitch;
    dxy     = mag * item->col_pitchy;  // usually zero, unless you are demented
    dyx     = mag * item->row_pitchx;  // usually zero

    sine_cosine( angle, &sin_a, &cos_a );

    if ( rows <= 0 ) BAILOUT( "INVALID NUMBER OF ROWS SENT TO FLATTEN_AREF" );
    if ( cols <= 0 ) BAILOUT( "INVALID NUMBER OF COLUMNS SENT TO FLATTEN_AREF" );

    if ( dx < 0 )   
        { 
        nag( "ARRAY X PITCH < 0", &nap ); 
        if ( nap < 10 ) printf( "                             X pitch = %d    ...now set to zero\n", dx );
        dx = 0;
        }

    if ( dy < 0 )   
        { 
        nag( "ARRAY Y PITCH < 0", &nap2 ); 
        if ( nap2 < 10 ) printf( "                            Y pitch = %d    ...now set to zero\n", dy );
        dy = 0;
        }

    needle = item;

    for ( c=0; c < cols; c++ )
        {
        for ( r=0; r< rows; r++ )
            {
            needle->type        = 3;        // sref
            needle->cell_number = cellnum;
            needle->mag         = new_mag;      
            needle->angle       = new_angle; 
            needle->abs_angle   = FALSE;
            needle->abs_mag     = FALSE;
            needle->reflect     = FALSE;
            needle->cols        = 0;
            needle->rows        = 0;
            needle->col_pitch   = 0;
            needle->row_pitch   = 0;
            needle->col_pitchy  = 0;
            needle->row_pitchx  = 0;
            needle->layer       = 0;
            needle->dt          = 0;

            x = x0 + c * dx + r * dyx;
            y = y0 + r * dy + c * dxy;

            if ( mirror ) y = -y;
            
            x = x * new_mag;
            y = y * new_mag;

            // The cell origins should be rotated, but Layout screws this up,
            // so we have to do it wrong.

            // Right way:
            //needle->x[0] = x * cos_a - y * sin_a;  
            //needle->y[0] = x * sin_a + y * cos_a;
            // Wrong way:

            needle->x[0] = x;  
            needle->y[0] = y;

            needle->nextitem = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
            previous = needle;
            needle = needle->nextitem;
            needle->x = (int *) malloc( sizeof( int ) );
            needle->y = (int *) malloc( sizeof( int ) );
 
            }    
        } 

    previous->nextitem = the_rest;
    
    }  // flatten_aref


/*------------------------------------------------------------------------------------------*/


struct flat_itemtype *                         // flatten a list, return a pointer to end of list
flatten( struct gds_itemtype  *item,           // The cell to be flattened 
         struct gds_celltype **celltable,      // The cell table
         struct flat_itemtype *pflat,          // Pointer to end of list (input)
         float mag,                            // start with mag=1, angle=0
         float angle,                          // when first called
         BOOL  reflect,
         int   x0,                             // anchor of sref or aref
         int   y0,
         float dbu_m,
         BOOL  nobox, BOOL silent, int topcellnum     )
    {

    struct gds_itemtype
        *next;                                 // pointer into the hierarchical list

    float
        fsp_width, 
        angler,
        sina,
        cosa,
        total_mag, 
        xf, yf,
        total_a;

    static BOOL  
        zero_width_encountered = FALSE,
        fix_zero_width = FALSE,
        dont_increment,
        init = FALSE,
        total_reflect,
        first = TRUE,
        hush_now = FALSE;
    int
        sp_width,
        xp, yp,
        ix, iy,
        total_x,
        total_y;

    static int
        nzw=0,
        nam=0,
        naa=0;

    char
        *ans;

    static char 
        *callingcell;

    ans = (char *) malloc( 10 );

    if ( first )
        {
        first = FALSE;
        callingcell = (char *) malloc( 10 );
        callingcell = "";
        if ( topcellnum >= 0 )
            callingcell = celltable[ topcellnum ]->name;    
        }

    dont_increment = FALSE;


    while ( item != NULL )
        {
        if ( item->abs_mag   ) nag( "ABSOLUTE MAGNIFICATION FOUND", &nam );
        if ( item->abs_angle ) nag( "ABSOLUTE ANGLE FOUND", &naa );

        // What sort of idiot would use an aref instead of an sref? A special asshole.
        // So let's clean it up:

        if ( ( item->type == 2 && item->rows == 1 && item->cols == 1 ) ||
             ( item->type == 2 && item->rows == 0 && item->cols == 1 ) ||
             ( item->type == 2 && item->rows == 1 && item->cols == 0 ) )    item->type = 3; 

        switch( item->type )
            {
            case -1: // printf( "Item chain:\n" ); 
                     break;
            case 0:  
                     // printf( "Boundary [%4.2f] [%4.2f] [%d]\n", mag, angle, reflect );
                     if ( item->layer >= 0 )
                         {
                         total_x = x0;
                         total_y = y0;
                         total_a = angle;
                         pflat->nextitem = copy_boundary( item, mag, total_a, reflect, total_x, total_y );
                         pflat = pflat->nextitem;
                         }
                     break;
            case 5:
                     // Boxes are usually treated the same as polygons.  
                     if ( item->layer >= 0 )
                         {
                         total_x = x0;
                         total_y = y0;
                         total_a = angle;
                         if ( ! nobox ) 
                             {
                             pflat->nextitem = copy_boundary( item, mag, total_a, reflect, total_x, total_y );
                             pflat = pflat->nextitem;
                             }
                         }
                     break;
            case 1:  // printf( "Path     [%4.2f] [%4.2f] [%d]\n", mag, angle, reflect ); 
                     // Turn this path into a set of polygons, inserting them at the current
                     // "item" pointer. Supress incrementation of the list pointer (one time)
                     //  so that this loop will munch on the new polygons.
                     if ( item->layer == -1 ) break;
                     if ( (item->width == 0) && (! zero_width_encountered) ) 
                         {
                         zero_width_encountered = TRUE;
                         if ( ! silent )
                             {
                             printf( "\n                            ZERO WIDTH PATH" );
                             printf( "\n\nWould you like to convert zero-width paths into rectangles? [y/n] " );
                             fflush( stdout );
                             fgets( ans, 10, stdin );
                             strtrim( ans );
                             fix_zero_width = (strcmp( ans, "y" ) == 0) || (strcmp( ans, "Y" ) == 0);
                             if ( fix_zero_width )
                                 {
                                 printf( "How wide should the rectangles be (in nm) ? " );
				 fflush( stdout );
				 scanf( "%d", &sp_width );

                                 if ( sp_width == 0 )
				     {
				     fix_zero_width = FALSE;
				     printf( "\nReally? Ok, whatever. Zero width paths will disappear.\n\n" );
				     }
				 else if ( sp_width < 0 )
				     {
				     BAILOUT( "Line width is less than zero. Abort." );
				     }
				 else
				     {
                                     fix_zero_width = TRUE;
                                     printf( "\nVery good. Zero width paths will be converted into %d nm wide rectangles.\n\n", sp_width );
                                     fsp_width = sp_width;
				     sp_width = (int) round( fsp_width / (dbu_m * 1.0e9)); 
				     }
                                 }
                             else
                                 {
                                 printf( "\nOk then. Zero width paths will disappear.\n\n" );
                                 }
                             }
                         else
                             {
                             fix_zero_width = TRUE;
                             sp_width = (int) round( 10.0 / (dbu_m * 1.0e9)); 
                             nag( "Zero width paths will be converted into 10 nm wide rectangles.", &nzw );
                             }
                         }
                     if ( (item->width == 0) && fix_zero_width ) item->width = sp_width;
                     item = path2poly( item );
                     dont_increment = TRUE;
                     break;
            case 2:                                 // aref
                     if ( item->abs_angle )
                         total_a = item->angle;
                     else
                         {
                         massage_degrees( &angle, &item->angle );
                         total_a = angle + item->angle;
                         }

                     if ( item->abs_mag )
                         total_mag = item->mag;
                     else
                         total_mag   = mag * item->mag;

                     total_reflect = item->reflect;
                     if ( reflect ) total_reflect = ! total_reflect;

                     sine_cosine( angle, &sina, &cosa );

                     xf = mag * item->x[0];  // mag is the magnification from
                     yf = mag * item->y[0];  // the calling cell, not this one!

                     xp = floor( xf * cosa - yf * sina );
                     yp = floor( xf * sina + yf * cosa );

                     if ( reflect ) yp = -yp;  // v3.13    reflect the aref origin

                     total_x = xf;// xp;// + x0;  // holy tap dancing jesus. this is a mysterious thing 
                     total_y = yf;// yp;// + y0;  // I am not sure why this works


                     // printf( "Aref     [%4.2f] [%4.2f] [%d] [%d] [%d]\n", 
                     //          total_mag, total_a, total_reflect, item->x[0], item->y[0] );

                     // Turn this aref into a set of srefs, injected into the "item" list.
                     // Then supress incrementation of the list pointer, so that
                     // this loop will start on the newly injected srefs.

                     flatten_aref( item, total_mag, total_a, total_reflect, total_x, total_y, callingcell );

                     dont_increment = TRUE;

                     break;
            case 3:                                    // sref
                     if ( item->abs_angle )
                         total_a = item->angle;
                     else
                         {
                         massage_degrees( &angle, &item->angle );
                         total_a = angle + item->angle;
                         }

                     if ( item->abs_mag )
                         total_mag = item->mag;
                     else
                         total_mag = mag * item->mag;

                     total_reflect = item->reflect;
                     if ( reflect ) total_reflect = ! total_reflect;

                     sine_cosine( angle, &sina, &cosa );

                     xf = mag * item->x[0];  // mag is the magnification from
                     yf = mag * item->y[0];  // the calling cell, not this one!

                     xp = floor( xf * cosa - yf * sina );
                     yp = floor( xf * sina + yf * cosa );

                     if ( reflect ) yp = -yp;  // v3.13  reflect the sref origin
                     
                     total_x = xp + x0; 
                     total_y = yp + y0;

                     // printf( "Sref     [%4.2f] [%4.2f] [%d] [%d] [%d]\n", 
                     //         total_mag, total_a, total_reflect, item->x[0], item->y[0] );

                     if ( item->cell_number >= 0 )
                         callingcell = celltable[ item->cell_number ]->name;    

                     next = celltable[ item->cell_number ]->item;                    

                     pflat = flatten( next, 
                                      celltable,
                                      pflat, 
                                      total_mag, 
                                      total_a, 
                                      total_reflect, 
                                      total_x,
                                      total_y,
				      dbu_m,
                                      nobox, silent, topcellnum    );
 
                     break;
            case 4:  // printf( "Text     [%4.2f] [%4.2f] [%d]\n", mag, angle, reflect ); 
                     break;
            default: printf( "[error: type %d]\n", item->type );
            }

        if ( dont_increment )
            dont_increment = FALSE;
        else
            item  = item->nextitem;

        } // while
    
    return( pflat );

    }  // flatten


//--------------------------------------------------------------------------------------------------------------

long_list *
initialize_long_list( long n )
    {
    static long_list *p;
    static int i;
    p = (long_list *) malloc( sizeof(long_list) );
    if ( ! p ) BAILOUT( "Cannot allocate memory for list of longs." );
    p->len = 0;
    p->max = n;
    p->n = (long *) malloc( n * sizeof( long ) );
    if ( ! p->n ) BAILOUT( "Cannot allocate memory for array of longs." );
    return p;
    }

void
append_long_list( long_list *p, long x )
    {
    if ( ! p ) BAILOUT( "null pointer sent to append_long_list function" );
    if ( p->len == p->max ) BAILOUT( "Too many numbers added to a list of longs\n" );
    p->n[p->len] = x;
    p->len += 1;
    } 

BOOL
in_long_list( long_list *p, long x )
    {
    long i;
    if ( p == NULL ) return FALSE;
    for ( i=0; i < p->len; i++ )
        {
        if ( p->n[i] == x ) return TRUE;
        }
    return FALSE;
    }

void
print_long_list( long_list *p )
    {
    long i;
    if ( p == NULL ) return;
    for ( i=0; i < p->len; i++ )
        {
        printf( "[%ld]  ", p->n[i] );
        }
    printf( "\n" );
    }


long_list *
parse_layers( char *layers )
    {
    // Create a list of layers to include.
    // Layer -1 implies "all".

    char *p, *q;
    long layer, len, num, i;
    long_list *list;
    
    q = layers;
    len = strlen( q );
    num = 0;
    for ( i=0; i<len; i++ ) 
        {
        if ( *q == ',' ) num++;
        q++;
        }
    num++;
    
    list = initialize_long_list( num );

    p = strtok( layers, "," );
    while ( p != NULL ) 
        {
        sscanf( p, "%ld", &layer );
        append_long_list( list, layer );
        p = strtok( NULL, "," );
        }

    return list;

    }  

//-------------------end parse_layers------------------------------------------------------------

/*------------------------------------------------------------------------------------------*/

int 
main( int argc, char *argv[] )
    {
    
    /* Read in a GDSII file, returning a list of cells and their respective extents. */


    struct flat_itemtype
        *flat;

    struct gds_itemtype 
        *item,
        *current_item;

    struct gds_celltype 
        **cell,                          /* Pointers into the library.  */ 
        *current_cell,
        *library;                        /* The library is a linked list of cells.                  */
                                         /* A cell is a linked list of items.                       */
                                         /* An item usually contains a list of coordinate pairs.    */
                                         /* This list has a variable length, allocated at run time. */

    char  
        **forward_name;                  /* array of forward referenced cell names */

    float
        dbu_uu,                          // user units per database unit
        dbu_m,                           // meters per database unit (usually 1e-9)
        dbu_um,                          // size of a database unit in microns (usually 1e-3)
        mag,
        angle,
        shiftx_um,
        shifty_um,
        current_mag   = 1.0,
        extra_angle = 0.0;

    int 
        x, 
        y,
        i,
        j,
        shiftx,
        shifty,
        forward_num,                     // number of cells with forward references 
        cell_table_size,
        num_cells;                       // The number of cells in this library.              

    char 
        *top_cell_name,                  // to be flattened etc
        *filename,
        *outfile = "flat.gds",
        *libraryname;

    int fd;                              // file descriptor 

    unsigned short
        word,
        count,                           // number of bytes in a record 
        token;                           // record type                 

    struct gds_celltype 
        *top_cell;

    BOOL
        use_first_cell        = FALSE,
        allayers              = TRUE,
        dump                  = FALSE,
        nobox                 = FALSE,
        not_found,
        forward_refs          = FALSE,
        absolute_angle        = FALSE,
        absolute_mag          = FALSE,
        silent                = FALSE,
        reflect               = FALSE;

    struct intersection_coordinate
        p;

    long 
        llayer = -1,
        minusone = -1;

    long_list
          *layers;


    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "flatten v%1.2f\n", VERSION );
        exit(0);
        }

    
    if ( argc == 2 )
        {
        if ( strcmp( argv[1], "-h" ) == 0 || strcmp( argv[1], "-help" ) == 0 )
            {
            printf( "\n\nUse: flatten filename.gds top_cell_name [-nobox|-yesbox] [-dump|-nodump] \n" );
            printf(     "     [-silent|-loud] [-layer <list>] [rotation] [output file] [shift_x shift_y]\n\n" );
            printf(     "     where -nobox indicates that boxes should not be included.\n" );
            printf(     "     -silent will supress any questions, making reasonable assumptions instead.\n" );
            printf(     "     Layers can be a list of numbers separated by commas, without spaces\n" );
            printf(     "     Layer numbers are preserved, not merged together. Layer -1 means 'all'.\n" );
            printf(     "     Rotation (in degrees) will add the given angle to the pattern.\n" );
            printf(     "     Use a top_cell_name of '?' to use the first cell in the file.\n" );
            printf(     "     The output file name defaults to flat.gds, unless specified at the end.\n" );
            printf(     "     Optional shift X Y numbers are in microns. If the directory tmp exists \n" );
            printf(     "     then the data extents will be written to tmp/pattern_extents.dat\n\n" );
            exit(0);
            }
        }

    if ( argc < 2 )     
        {
	filename = (char *) malloc( 256 );
	printf( "\n\nGDS file to be flattened? " );  fflush( stdout );
        fgets( filename, 255, stdin );
        strtrim( filename );
        }
    else
        filename = argv[1];


    if ( argc < 3 )
        {
        top_cell_name = (char *) malloc( 256 );
        printf( "\n\nEnter the top level cell name\nEnter nothing to get a list of cells, \n");
        printf( "or enter ? to use the first cell     > " ); fflush( stdout );
        fgets( top_cell_name, 255, stdin );
        strtrim( top_cell_name );
        }
    else
        top_cell_name = argv[2];

    use_first_cell = (strcmp( top_cell_name, "?" ) == 0 );


    if ( argc > 4 )
        {
        nobox = (strcmp( argv[3], "-nobox" ) == 0) || (strcmp( argv[4], "-nobox" ) == 0) ;
        dump =  (strcmp( argv[3], "-dump" ) == 0)  || (strcmp( argv[4], "-dump"  ) == 0) ;
        silent =  (strcmp( argv[3], "-silent") == 0)  || (strcmp( argv[4], "-silent") == 0);
        }
    else if ( argc > 3 )
        {
        nobox = strcmp( argv[3], "-nobox" ) == 0 ;
        dump = strcmp( argv[3], "-dump" ) == 0 ;
        silent = strcmp( argv[3], "-silent" ) == 0 ;
        }

    allayers = TRUE;

    if ( argc > 4 ) 
        {
        allayers = strcmp( argv[3], "-layer" );
        if ( ! allayers ) layers = parse_layers( argv[4] );
        }    
    if ( allayers && argc > 5 ) 
        {
        allayers = strcmp( argv[4], "-layer" );
        if ( ! allayers ) layers = parse_layers( argv[5] );
        }    
    if ( allayers && argc > 6 ) 
        {
        allayers = strcmp( argv[5], "-layer" );
        if ( ! allayers ) layers = parse_layers( argv[6] );
        }    

    if ( argc > 5 )
        silent = silent || (strcmp( argv[5], "-silent") == 0);
            
    if ( allayers && argc > 7 ) 
        {
        allayers = strcmp( argv[6], "-layer" );
        if ( ! allayers ) layers = parse_layers( argv[7] );
        }    

    if ( ! allayers ) allayers = in_long_list( layers, minusone );

    if ( argc > 8 )
        {
	sscanf( argv[8], "%f", &extra_angle );
	}

    if ( argc > 9 )
        {
        outfile = argv[9];
        }

    shiftx_um = 0.0;
    shifty_um = 0.0;
    
    if ( argc > 10 ) sscanf( argv[10], "%f", &shiftx_um );
    if ( argc > 11 ) sscanf( argv[11], "%f", &shifty_um );


    //----------

    if ( ! silent )
        {
        printf( "\n\nGDS Flattening Utility v%1.2f\n\n", VERSION );
        printf( "\nInput file:  %s", filename );
        printf( "\nOutput:      %s", outfile );
        printf( "\nTop cell:    %s", top_cell_name );
        if ( use_first_cell ) printf( "\nFlattening the first cell found in this file.\n" );
        if ( nobox ) printf( "Boxes will be ignored.\n" );
        if ( dump ) printf( "A text dump will be created in flat.bdm\n" );
        if ( allayers )
            printf( "\nAll layers will be included.\n" );
        else
            {
            printf( "\nLayers: " );
            print_long_list( layers );
            }
        if ( extra_angle != 0.0 ) printf( "Rotation: %1.3f degrees\n", extra_angle );
        if ( shiftx_um != 0.0 && shifty_um != 0.0 ) printf( "Shift:  %1.3f, %1.3f um \n", shiftx_um, shifty_um );
        printf( "\nReading:\n" );
        }
    else
        {
        if ( extra_angle == 0.0 )
            printf( "\nFlatten %s ", filename ); 
        else
            printf( "\nFlatten %s with rotation %1.3f degrees. ", filename, extra_angle );
        
        if ( shiftx_um != 0.0 || shifty_um !=0.0 )
            printf( "Shift %1.3f %1.3f um \n\n", shiftx_um, shifty_um );
        else
            printf( "\n\n" );
        }
    
    fd = open( filename, 0 );
    if ( fd == -1 ) BAILOUT( "UNABLE TO OPEN THE INPUT FILE" );

    current_mag    = 1.0;
    absolute_mag   = TRUE;
    absolute_angle = TRUE;
    forward_refs   = FALSE;

    while ( read( fd, &count, 2 ) )
        {
        gds_swap2bytes( (BYTE *) &count );

        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );


            switch( token )
                {
                case 0x0002:  //printf( "HEADER \n" ) ;
                              gds_read_header( fd, count, TRUE );
                              break;
                case 0x0102:  //printf( "BGNLIB \n" );
                              gds_read_bgnlib( fd, count, &cell_table_size, &library, &cell, &current_cell, 0 );
                              break;
                case 0x0206:  //printf( "LIBNAME \n" ) ;
                              libraryname = gds_read_libname( fd, count, ! silent );
                              break;
                case 0x0305:  //printf( "UNITS \n" ) ;
                              gds_read_units( fd, &dbu_um, &dbu_uu, &dbu_m, ! silent );
                              shiftx = shiftx_um / dbu_um;
                              shifty = shifty_um / dbu_um;
                              break;
                case 0x0400:  //printf( "ENDLIB \n" ) ;
                              gds_read_endlib( fd, count, 0 );
                              break;
                case 0x0502:  //printf( "BGNSTR \n" ) ;
                              gds_read_bgnstr( fd, count, &cell_table_size, &num_cells, &cell, &current_cell, &current_item, 0 );
                              break;
                case 0x0606:  //printf( "STRNAME \n" ) ;
                              gds_read_strname( fd, count, num_cells, &cell, &current_cell, ! silent );
                              break;
                case 0x0700:  //printf( "ENDSTR \n" ) ;
                              gds_read_endstr( fd, count, 0 );
                              break;
                case 0x0800:  //printf( "BOUNDARY \n" ) ;
                              gds_read_boundary( fd, count, &current_item, 0 );
                              break;
                case 0x0900:  //printf( "PATH \n" ) ;
                              gds_read_path( fd, count, &current_item, 0 );
                              break;
                case 0x0A00:  //printf( "SREF \n" ) ;
                              gds_read_sref( fd, count, &current_item, ! silent );
                              break;
                case 0x0B00:  //printf( "AREF \n" ) ;
                              gds_read_aref( fd, count, &current_item, ! silent );
                              break;
                case 0x0C00:  //printf( "TEXT \n" ) ;  
                              gds_read_text( fd, count, &current_item, 0 );
                              break;
                case 0x0D02:  //printf( "LAYER \n" ) ;
                              gds_read_layer( fd, count, &current_item, 0 );
                              llayer = current_item->layer;
                              break;
                case 0x0E02:  //printf( "DATATYPE \n" ) ;
                              gds_read_datatype( fd, count, &current_item, 0 );
                              break;
                case 0x0F03:  //printf( "WIDTH \n" ) ;
                              gds_read_width( fd, count, &current_item, 0 );
                              break;
                case 0x1003:  //printf( "XY \n" ) ;
                              gds_read_xy( fd, count, dbu_um, &current_item, 0 );
                              if ( ! allayers && ! in_long_list( layers, llayer ) ) current_item->layer = -1;
                              break;
                case 0x1100:  //printf( "ENDEL \n" ) ;
                              gds_read_endel( fd, count, 0 );
                              break;
                case 0x1206:  //printf( "SNAME \n" ) ;
                              gds_read_sname( fd, count, num_cells, &forward_num, &forward_name, &cell, &current_item, 0 );
                              if ( current_item->cell_number == -1 ) forward_refs = TRUE;
                              break;
                case 0x1302:  //printf( "COLROW \n" ) ;
                              gds_read_colrow( fd, count, &current_item, 0 );
                              break;
                case 0x1602:  //printf( "TEXTTYPE \n" ) ; 
                              gds_read_texttype( fd, count, &current_item, 0 );
                              break;
                case 0x1701:  //printf( "PRESENTATION \n" ) ; 
                              gds_read_presentation( fd, count, &current_item, 0 );
                              break;
                case 0x1906:  //printf( "STRING \n" ) ; 
                              gds_read_string( fd, count, &current_item, 0  );
                              break;
                case 0x1A01:  //printf( "STRANS \n" ) ;
                              gds_read_strans( fd, count, &current_item, 0 );
                              break;
                case 0x1B05:  //printf( "MAG \n" ) ;
                              gds_read_mag( fd, count, &current_item, 0 );
                              break;
                case 0x1C05:  //printf( "ANGLE \n" ) ;
                              gds_read_angle( fd, count, &current_item, 0 );
                              break;
                case 0x2102:  //printf( "PATHTYPE \n" ) ;
                              gds_read_pathtype( fd, count, &current_item, 0 );
                              break;
                case 0x2202:  //printf( "GENERATIONS \n" ) ; 
                              gds_read_generations( fd, count, 0 );
                              break;
                case 0x2D00:  //printf( "BOX \n" ) ; convert to boundary
                              gds_read_box( fd, count, &current_item, 0 );
                              if ( ! allayers && ! in_long_list( layers, llayer ) ) current_item->layer = -1;
                              break;
                case 0x2E02:  //printf( "BOXTYPE \n" ) ;
                              gds_read_boxtype( fd, count, &current_item, 0 );
                              break;
                default:      printf( "                              UNSUPPORTED TOKEN: [%4x] \n", token ) ;
                              SKIPOVER( fd, count );
                }    
            }
        }

    close( fd );

    /*
    printf( "\n\nCELL TABLE:\n\n" ) ;
    for ( i=0; i <= num_cells; i++ )
        printf( "[%2d] %s\n", i, cell[i]->name ) ;
    printf( "\n\n" ) ;
    */

    /*-------------------------------------------------------------------------*/
    /* Now we must parse the data again to resolve forward references to cells */
    /*-------------------------------------------------------------------------*/

    if ( forward_refs )
        {
        printf( "Forward references:\n\n" );
        current_cell = library;
        j = -1;   

        while ( current_cell != NULL )
            {
            current_item = current_cell->item;
       
            while ( current_item != NULL )
                {
                if ( (current_item->type == 2 || current_item->type == 3) && current_item->cell_number == -1 )
                    {   
                    j++;
                    i = 0;
                    not_found = TRUE;
                    while ( not_found && (i <= num_cells) )
                        {
                        not_found = strcmp( forward_name[j], cell[i]->name ) != 0;
                        i++;
                        }
                    i--;

                    if ( not_found ) 
                        {
                        printf( "    WARNING: missing cell definition: %s\n", forward_name[j] ); 
                        }
                    else
                        {
                        if ( ! silent ) printf( "    Cell [%2d] matches %s\n", i, forward_name[j] );
                        current_item->cell_number = i;
                        }
                    }          // end if      

                current_item = current_item->nextitem;

                }  // end while

            current_cell = current_cell->nextcell;
        
            } // end while

        }  // end if


    // look for the cell to be digested

    if ( use_first_cell ) 
	{
        if ( ! silent ) printf( "\nCell [0] is %s\n", cell[0]->name );
        top_cell = cell[0];	
	}
    else
	{
        i = -1;
        not_found = TRUE;

        while ( not_found && (i < num_cells) )
            {
            i++;
            not_found = strcmp( cell[i]->name, top_cell_name ) != 0;
            }

        if ( not_found ) 
            {
            printf( "\nERROR: cell [%s] not found\n\n", top_cell_name ); 
            exit(-1);
            }
        else
            {
            if ( ! silent ) printf( "\nCell [%d] is %s\n", i, top_cell_name );
            top_cell = cell[i];
            }
        }

    mag     = 1.0;
    angle   = extra_angle;
    reflect = FALSE;

    // Create the flattened list

    flat = (struct flat_itemtype *) malloc( sizeof( struct flat_itemtype ) );
    if ( ! flat ) BAILOUT( "UNABLE TO ALLOCATE TOP OF FLAT LIST" );
    flat->n = -1;            // top of list is a dummy entry
    flat->nextitem = NULL;

    item  = top_cell->item;
    
    flatten( item, cell, flat, mag, angle, reflect, shiftx, shifty, dbu_m, nobox, silent, i );

    // Instead of using the original cell name, we will call the cell "flat" to
    // work around a bug in the python gds library. The python library crashes
    // if the cell name has an odd number of characters. 

    write_flat_gds( flat, outfile, "flat", dbu_um, dbu_uu, dump ); 

    if ( ! silent ) printf( "\nOutput file: %s\n\n", outfile );

    }
