// gpfout
//             Convert a gds file to gpf shapes for the ebpg.
//             The input file must be flat, with nothing but
//             convex polygons, chopped into fields and subfields.
//             Default output is gpf polygons that require the
//             UPG pattern generator. 
//
//             The command line "traps" will generate the older gpf 
//             trapezia; that is, triangles and parallelograms,
//             using the split-and-bury raster fill strategy.
//             When generating traps, the input shapes must be
//             4-sided on-axis paralellograms or 3-sided triangles 
//             with one side on-axis.

//             On systems without the Firebird option, trapezia
//             will expose significantly faster than polygons.

#include "gpf_pub.h"
#include "libgdsmin.h"
#include "polyq.h"
#include "libgpfree.h"

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>

#define PRES(  a, b ) ((fabs((a)-(b)) <= 1))
#define CLOSE( a, b ) ((fabs((a)-(b)) <= bss))

#define MAINDACBITS  (20)
#define SUBDACBITS   (14)
#define MAINDACMID   (1 << (MAINDACBITS-1))
#define MAXVERTEX    512
#define TOOCLOSE( a ) ((fabs((a)) < bss))
#define MAXSUBFIELD  4.2
#define ERRORLAYER   15

#define printv noop
#define VERBOSE FALSE
#define VERSION 5.2


// v1.1  Allow dataunits other than 1 nm, since the user might need
//       sub-pixel placement accuracy
//
// v1.2  Stop doing the weird offset bump for triangles. It doesn't 
//       work for small lsb values. 
//
// v2.0  Uses polyq instead of folyq.
//       Replace most xy types with dxy.
//
// v2.1  Output gpf polygons exclusively
//
// v2.2  Allow deflections a wee bit larger than the subfield
//       Hmm. That number 4.2 um should be scaled with the voltage?
//
// v2.3  Do not snap the mainfield and trap deflection to the
//       beam step size. Snap to the main field and subfield resolution.
//
// v2.4  Explicitly thow away any sliver smaller than bss_um in
//       either X or Y.
//
// v2.5  So many problems with arrays (large rectangles)
//
// v2.6  Long ints for type xy
//
// v2.7  fixed pixeltime bug
//
// v2.8  sort output blocks with a path file
//       which is a lot faster than using gpfsort
//
// v3.0  allow beam step size specification in the dose file
//       bss in the dose file must be a multiple of the base bss,
//       then the shapes can have different msf values
//
// v3.1  stop filtering out single-pixel polygons
//
// v3.2  more realistic patternsize for the header
//
// v3.3  fix patternsize for cview
//
// v4.0  optionally produce parallelograms in gpf
//       instead of polygons. GPF 1.43 allows trapezia
//       to have 1-bit placement accuracy, with split-and-bury
//       filling mode. This works a lot better than the old
//       traps that had to snap to the beam-step-size grid.
//
// v4.1  prints the path file name. hardly worth mentioning.
//
// v4.2  lots of small changes to make ubuntu gcc happy
//
// v5.0  gpf_write_union() replaced with gpfree_write_union()
//       gpf_write_header() replaced with gpfree_write_header()
//       gpf_write_indexf() replaced with pgfree_write_indexf()
//       gpf_file_open() and gpf_file_close() also replaced
//       so that we do not need the Beams library.
//
// v5.1  New version of gpf.h, no more beams.h
//
// v5.2  gpf.h replaced by gpf_pub.h


BOOL not_a_large_rectangle( polygon *p, dxy sfs, double dbu_um, BOOL verbose );
BOOL this_is_a_rectangle( polygon *p, dxy sfs, double dbu_um );

double bss;
int mfdefs=0, nout=0;
void noop() {}


//--------------------------------------------------------------------------------------------------------

typedef struct twoints
    {
    gpf_uint_t lsw;
    gpf_uint_t msw;
    }               twoints;

//--------------------------------------------------------------------------------------------------------

void warning( char *message, int *n )
    {
    if ( *n == 10 )
	{
        printf( "Repeated warning messages suppressed.\n" );
        fflush( stdout );
        *n = *n + 1;
        }
    else if ( *n < 10 )
        {
	printf( "%s\n", message );
        fflush( stdout );
        *n = *n + 1;
        }
    }

//-------------------------------------------------------------------------------------------

int **
i_matrix( long nrl, long nrh, long ncl, long nch )
    {
    // Allocate a matrix of pointers to int
    // A matrix element could be referenced with m[r][c],
    // since m[r] is a pointer to an array of int.
    
    long    i;
    int   **m;
      
    m=(int **) malloc( (nrh-nrl+1)*sizeof( int* ));

    if ( ! m ) 
        {
        bail("allocation failure 1 in gpfout/i_matrix()");
        }

    m -= nrl;

    for(i=nrl;i<=nrh;i++) 
        {
        m[i]=(int *) malloc((nch-ncl+1)*sizeof(int));
        if (! m[i] ) 
            {
            bail("allocation failure 2 in gpfout/i_matrix()");
            }
        m[i] -= ncl;
        }

    return m; 

    }


// end i_matrix ----------------------------------------------------------------


double
**doublematrix( int nrl, int nrh, int ncl, int nch )
    {
    // Allocate a matrix of doubles, forcing it to
    // occupy a solid block of memory. This makes it possible to       
    // pass the matrix to a fortran program, while retaining the       
    // more efficient double indirection of C.  

    int
        nrows,
        ncols,
	i;

    double
        *block,
        **m;

    nrows = nrh - nrl + 1;
    ncols = nch - ncl + 1;

    if ( nrows <= 0 ) 
        bail( "Illegal row index bounds sent to doublematrix()" );
    if ( ncols <= 0 ) 
        bail( "Illegal column index bounds sent to doublematrix()" );
   
    block = (double *) malloc((unsigned) 
	    nrows * ncols * sizeof(double) );

    if (! block) bail( "Error: unable to allocate memory block in dmoubleatrix()" );
	
    m = (double **) malloc( (unsigned) nrows * sizeof( double* ) );
    
    if ( !m ) bail( "Error: unable to allocate vector of pointers in doublematrix()" );

    m -= nrl;    // offset for arbitrary array bounds 
       
    for( i=0; i < nrows; i++ ) 
        {
	m[i+nrl] = block + (ncols * i); 
	m[i+nrl] -= ncl;                
	}

    return m;

    }


//------------------------------------------------------------------------------

double
alittlebigger( double x )
    {
    return( x + 0.01 * fabs(x) );
    }

//--------------------------------------------------------------------------------------------------------

double 
alittlesmaller( double x )
    {
    return( x - 0.01 * fabs(x) );
    }

//--------------------------------------------------------------------------------------------------------

void
snap_to_grid( polygon *p, double step )
    {
    static int    i; 
    static dxy    dv;
    static xy     v;
    static int    minx, maxx, miny, maxy;

    if ( ! p->active ) return;

    for ( i=0; i < p->len; i++ )
        {
        v = vertex( p, i );
        dv = dvertex( p, i );
        v.x = (int) step * round( dv.x / step );
        v.y = (int) step * round( dv.y / step );
        assign_vertex( p, i, v );
        if ( i == 0 ) 
	    { minx = v.x; miny = v.y; maxx = v.x; maxy = v.y; }
	else
	    {
            if ( v.x < minx ) { minx = v.x; p->ll.x = v.x; }
            if ( v.y < miny ) { miny = v.y; p->ll.y = v.y; }
            if ( v.x > maxx ) { maxx = v.x; p->ur.x = v.x; }
            if ( v.y > maxy ) { maxy = v.y; p->ur.y = v.y; }
	    }
        }
    }


void 
snapointo( dxy *v, double n )
    {
    v->x = n * floor( v->x / n );
    v->y = n * floor( v->y / n );
    }

xy
isnapointo( xy v, double n )
    {
    static xy vv;
    vv.x = n * floor( v.x / n );
    vv.y = n * floor( v.y / n );
    return( vv );
    }

double 
dsnap( double x, double n )
    {
    return( n * round( x / n ) );
    }

int
isnap( int x, int n )
    {
    static double dx, dn;
    return( x );           // disabled! see comments below
    dx = x;
    dn = n;
    return( (int) dn * round( dx / dn ) );
    }


//-------------------------------------------------------------------------------------------------


void
print_queue( queue *q )
    {
    static polygon_ref *pr;
    static polygon     *p;
    static int          i;
    static xy           v;
    printf( "\nqueue length: %d\n", q->len );
    pr = q->first;
    while ( pr )
	{
	p = pr->p;
        printf( "    [%d]    ", p->len );
        if ( ! p->active )
	    {
	    printf( "inactive" );
	    }
	else
	    {
	    for ( i=0; i < p->len; i++ )
	        {
	        v = vertex( p, i );
	        printf( "    %4ld, %4ld    ", v.x, v.y );
	        }
            }
        printf( "\n" );
	pr = pr->next;
	}
    }

//--------------------------------------------------------------------------------------------------------

void
print_2dq( array_2dq *aq )
    {
    static int r, c;
    static queue *q;

    for ( c=0; c < aq->len.x; c++ )
	{
	for ( r=0; r < aq->len.y; r++ )
	    {
	    q = get_field_list( aq, c, r );
	    if ( q->len > 0 ) 
		{
		printf( "\n[%d][%d]\n", c, r );
		print_queue( q );
		}
	    }
	}

    }

//--------------------------------------------------------------------------------------------------------


double
trim( double x )   // this does not work
    {
    // A number like 9.9999999e-12 becomes 1e-11.
    return( x );

    double e, a, f;
    if ( x <= 0.0 ) return( x );
    e = floor( log10( x ) ) - 1;
    a = x / pow( 10, e );
    f = a - floor( a );
    if ( fabs( f - 1.0 ) < 1e-4 ) f = 1.0;
    a = floor(a) + f;
    a = a * pow( 10, e );
    return( a );
    }

//--------------------------------------------------------------------------------------------------

twoints
dose2clock( double d )
    {
    static gpf_uint_t horloge;
    static twoints clock;

    if ( d <= 0.5 ) bail( "invalid dose <= 0.5" );
    horloge = (gpf_uint_t) floor( 2147483648.0 / d );    // 2^31 / dose 
    clock.msw = horloge >> 16;
    clock.lsw = horloge & 0xFFFF;
    return( clock );
    }

gpf_uint_t
dose2oneclock( double d )
    {
    if ( d <= 0.5 ) bail( "invalid dose <= 0.5" );
    return( (gpf_uint_t) floor( 2147483648.0 / d ) );    // 2^31 / dose 
    }

//----------------------------------------------------------------------------------------

void
init_prims( gpf_union_t *prims, gpf_union_t *prims_org, int sub_msf, double mindose, double maxdose )
    {
    // This is approach is confusing. Look at write_trap() for a better
    // way to initialize parameters. Why do we need an array of primitives?
    // We don't. It's stupid.
    
    static int i;
    static twoints minclock, maxclock;

    for (i=0; i < GPF_PRIM_COUNT; i++)
        {
        prims_org[i].data = 0;
        prims[i].data = 0;
        }

    // Put the identification codes (IC) into the element of prims
    // reserved for that identification code. Weirdly redundant.

    prims[GPF_PRIM_MOVESTAGE].prim24.IC            = GPF_PRIM_MOVESTAGE;
    prims[GPF_PRIM_FREQUENCY_LSW].prim16.IC        = GPF_PRIM_FREQUENCY_LSW;
    prims[GPF_PRIM_FREQUENCY_MSW].prim16.IC        = GPF_PRIM_FREQUENCY_MSW;
    prims[GPF_PRIM_MSF].prim16.IC                  = GPF_PRIM_MSF;
    prims[GPF_PRIM_MAIN_X_LSW].prim24.IC           = GPF_PRIM_MAIN_X_LSW;
    prims[GPF_PRIM_MAIN_Y_LSW].prim24.IC           = GPF_PRIM_MAIN_Y_LSW;
    prims[GPF_PRIM_OFFSET_X_LSW].prim24.IC         = GPF_PRIM_OFFSET_X_LSW;
    prims[GPF_PRIM_OFFSET_Y_LSW].prim24.IC         = GPF_PRIM_OFFSET_Y_LSW;

    //prims[GPF_PRIM_POLYGON].prim24.IC               = GPF_PRIM_POLYGON;
    //prims[GPF_PRIM_POLYGON_CENTER_X_FRAC].prim24.IC = GPF_PRIM_POLYGON_CENTER_X_FRAC;
    //prims[GPF_PRIM_POLYGON_CENTER_X_LSW].prim24.IC  = GPF_PRIM_POLYGON_CENTER_X_LSW;
    //prims[GPF_PRIM_POLYGON_CENTER_Y_FRAC].prim24.IC = GPF_PRIM_POLYGON_CENTER_Y_FRAC;
    //prims[GPF_PRIM_POLYGON_CENTER_Y_LSW].prim24.IC  = GPF_PRIM_POLYGON_CENTER_Y_LSW;
    //prims[GPF_PRIM_POLYGON_ANGLE_MSW].prim24.IC     = GPF_PRIM_POLYGON_ANGLE_MSW;
    //prims[GPF_PRIM_POLYGON_ANGLE_LSW].prim24.IC     = GPF_PRIM_POLYGON_ANGLE_LSW;

    minclock = dose2clock( maxdose );
    maxclock = dose2clock( mindose );

    prims[GPF_PRIM_FREQUENCY_LSW].prim16.val1      = minclock.lsw;
    prims[GPF_PRIM_FREQUENCY_LSW].prim16.val2      = maxclock.lsw;
  
    prims[GPF_PRIM_FREQUENCY_MSW].prim16.val1      = minclock.msw;
    prims[GPF_PRIM_FREQUENCY_MSW].prim16.val2      = maxclock.msw;
  
    prims[GPF_PRIM_MSF].prim16.val2                = sub_msf;

    return;
    }

//-----------------------------------------------------------------------------------------------


void
init_headers( char               *filename, 
              gpf_header_t       *header, 
              double              mainfieldsize_x, 
              double              mainfieldsize_y, 
	      double              main_res_x, 
	      double              main_res_y, 
	      double              sub_res_x, 
	      double              sub_res_y, 
              int                 main_msf,
              int                 sub_msf,
              int                 fieldmax,
              double              subfieldsize_xmax,
              double              subfieldsize_ymax,
              gpf_header_user_t  *user,
              gpf_ulong_t        *gpfpos,
              gpf_indexf_t       *field_index,
              int                 ht,
              BOOL                assign_doses,
              double              mindose,
              double              maxdose,
              int                 ncols,
              int                 nrows,
              BOOL                traps
            )
    {
    int   i;
    float gpf_version;
    char  freebeam_version[48];
    char  versionstr[8];
    FILE *vfile;
    
    struct tm *time_structure;
    time_t time_val;

    for (i = 0;i < 80;i++)
        header->Title[i] = 0;

    strcpy(header->Title, GPF_COPYRIGHT );

    for (i = 0; i < 8; i++)
        {
        header->Version[i] = 0;
        versionstr[i] = 0;
        }

    if ( traps )
        strncpy( header->Version, GPF_VERSIONSTR_GPG, 8 );
    else
        strncpy( header->Version, GPF_VERSIONSTR_UPG, 8 );

    strncpy( versionstr, header->Version, 4 );
    sscanf( versionstr, "%f", &gpf_version );

    if ( traps && gpf_version < 1.43 )
        {
        printf( "\n\nGPF version: %4.2f\n", gpf_version );
        printf( "The version of GPF on this machine is less than 1.43 and so does not support\n" );
        printf( "split-and-bury writing strategy. You need an updated Beams library.\n" );
        bail( "GPF version problem." );
        }

    if ( ! traps && gpf_version < 1.43 )
        {
        printf( "\n\nGPF version: %4.2f\n", gpf_version );
        printf( "The version of GPF on this machine is less than 1.43 and so does not support \n" );
        printf( "gpf polygons. You need an updated Beams library to generate polygons,\n" );
        bail( "GPF version problem." );
        }

    header->MainFieldResolutionX = main_res_x;
    header->MainFieldResolutionY = main_res_y;
    header->SubFieldResolutionX  = sub_res_x;
    header->SubFieldResolutionY  = sub_res_y;

    header->ResolutionX   = main_msf;
    header->ResolutionY   = main_msf;
    header->BeamStepSizeX = sub_msf;
    header->BeamStepSizeY = sub_msf;

    header->MainFieldSizeX = lround(mainfieldsize_x / (main_msf * main_res_x));
    header->MainFieldSizeY = lround(mainfieldsize_y / (main_msf * main_res_y));

    // Make patternsize a bit larger (+100), because a shape can go outside
    // the trap field due to the rotation and therfore beyond the main field and
    // pattern size. This prevents e.g. a redraw in cview.

    header->PatternSizeX = lround(ncols * header->MainFieldSizeX) + 100;
    header->PatternSizeY = lround(nrows * header->MainFieldSizeY) + 100;

    header->SubFieldSizeX = lround(subfieldsize_xmax / (sub_msf * sub_res_x));
    header->SubFieldSizeY = lround(subfieldsize_ymax / (sub_msf * sub_res_y));

    header->HighTension = ht*1000000;

    if ( traps )
        header->ShapeTypes = GPF_SHAPETYPE_TRAPEZIUM;
    else
        header->ShapeTypes = GPF_SHAPETYPE_POLYGON;
    
    header->SubFieldPlacement = GPF_SUBFIELDPLACEMENT_LOWERLEFT;
    header->NrMainFields = fieldmax;
    header->MainFieldPlacement = GPF_MAINFIELDPLACEMENT_FLOATING;
    
    header->FractureStyle = GPF_FRACTURESTYLE_SUBFIELD | GPF_FRACTURESTYLE_HIGHRESOLUTION |
                            GPF_FRACTURESTYLE_OUTLINE  | GPF_FRACTURESTYLE_FILLMODE_SMOOTH ;

    header->NrMainFieldBits = MAINDACBITS;
    header->NrSubFieldBits  = SUBDACBITS;
    //header->WorkLevel = GPF_WORKLEVEL_HIGH; relevant only on Z-stage machines or maybe this is undefined?

    if ( assign_doses )
	{
        header->MinFreqFactor = dose2oneclock( maxdose );
	header->MaxFreqFactor = dose2oneclock( mindose );
	}
    else
	{
        header->MinFreqFactor = NOMINAL_FREQUENCY;
        header->MaxFreqFactor = NOMINAL_FREQUENCY;
	}

    header->NrFreqFactors = 1;
    header->PixelTime = 0;            /* Normalised pixel expose time    */
    header->FileSize = 0;
    header->HeaderOffset = GPF_HEADER_SIZE;
    header->IndexOffset = 0;
    header->StructureOffset = 0;
    header->StructureSize = 0;
    header->MarkerOffset = 0;
    header->MarkerSize = 0;
    //header->RefFreqFactor = NOMINAL_FREQUENCY;

    (void) time( &time_val);
    time_structure = localtime( &time_val );

    for (i = 0; i < 80; i++)
        user->PatternName[i] = 0;

    strcpy(user->PatternName, filename);

    for (i = 0; i < 16; i++)
        user->CreateDate[i] = 0;

    (void) sprintf( user->CreateDate,
                    "%04d-%02d-%02d", 1900 + time_structure->tm_year,
                    time_structure->tm_mon + 1, time_structure->tm_mday);

    for (i = 0; i < 16; i++)
        user->CreateTime[i] = 0;

    (void) sprintf( user->CreateTime,
                    "%02d:%02d:%02d", time_structure->tm_hour,
                    time_structure->tm_min, time_structure->tm_sec);

    for (i = 0; i < 48; i++)
        user->ConverterName[i] = 0;

    strcpy( user->ConverterName, "Freebeam" );

    for (i = 0; i < 48; i++)
        {
        user->ConverterVersion[i] = 0;
        freebeam_version[i] = 0;
        }

    vfile = fopen( "tmp/version", "r" );

    if ( ! vfile )
        {
        strcpy( freebeam_version, "1.0" );
        }
    else
        {
        fscanf( vfile, "%s", freebeam_version );
        fclose( vfile );
        }

    strcpy( user->ConverterVersion, freebeam_version );

    for (i = 0; i < 48; i++)
        user->SourceFormat[i] = 0;

    strcpy( user->SourceFormat, "GDS" );

    *gpfpos = header->HeaderOffset + GPF_HEADER_SIZE;

    for (i=0; i < header->NrMainFields; i++)
        {
        field_index[i].FieldToPatternOffsetX = 0.0;  
        field_index[i].FieldToPatternOffsetY = 0.0;  
        // does not work: field_index[i].FieldToPatternOffsetZ = 0.0;
        field_index[i].StartData = 0; 
        field_index[i].EndData = 0;   
        }

    return;
    }

//--------------------------------------------------------------------------------------------


void
write_prims( FILE *fp, gpf_union_t *prims, gpf_union_t *prims_org, gpf_ulong_t *gpfpos )
    {
    // Do not follow this example. Writing the primitives in a separate function is stupid.
    // Someday we'll clean this up and purge this old nonsense.
    
    static int  i;
    static long count;

    for (i=0; i < GPF_PRIM_COUNT; i++)
        {
        if ((prims[i].prim24.IC != 0) && (i != GPF_PRIM_ACTIVATE))
            {
            if (prims_org[i].data != prims[i].data)
                {
                // if ( VERBOSE ) gpf_print_union( stdout, &prims[i], &count );
                gpfree_write_union( fp, gpfpos, &prims[i] );
                prims_org[i].data = prims[i].data;
                }
            }
        }

    // we do not print the 'activate' line here. 

    return;
    }

//-----------------------------------------------------------------------------------------------

void
movestage( FILE         *fp,           
           double        x, 
           double        y, 
           gpf_union_t  *prims,       
           gpf_union_t  *prims_org,  
           gpf_indexf_t *field_index, 
           int          *fieldnr, 
           gpf_ulong_t  *gpfpos       )
    {
    static long count;

    field_index[*fieldnr].FieldToPatternOffsetX = x; 
    field_index[*fieldnr].FieldToPatternOffsetY = y; 
    //field_index[*fieldnr].FieldToPatternOffsetZ = 0.0;
    field_index[*fieldnr].StartData = *gpfpos;

    // According to the gpf manual, this "movestage fieldnr" does nothing.
    // And yet we write it anyway. Perhaps it is used by cview to 
    // mark the fields. That would make it useful. We should test it!
 
    prims[GPF_PRIM_MOVESTAGE].prim24.val  = *fieldnr + 1;
    //if ( VERBOSE ) gpf_print_union( stdout, &prims[GPF_PRIM_MOVESTAGE], &count );
    gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MOVESTAGE] );
    prims_org[GPF_PRIM_MOVESTAGE].data = prims[GPF_PRIM_MOVESTAGE].data;

    // Remember we have to update field_index[].EndData with the new gpfpos
    // after putting shapes into the field. Do this with end_field()

    }

//--------------------------------------------------------------------------------------------------

void
end_field( gpf_indexf_t *field_index, int *fieldnr, gpf_ulong_t *gpfpos )
    {
    field_index[*fieldnr].EndData = *gpfpos;
    (*fieldnr)++;
    }


//---------------------------------------------------------------------------------------------------
// Triangle helper functions


BOOL
x_triangle( polygon *p )
    {
    static xy a, b, c;
    if ( p->len != 3 ) return( FALSE );
    a = vertex( p, 0 );
    b = vertex( p, 1 );
    c = vertex( p, 2 );
    if ( a.y == b.y || a.y == c.y || b.y == c.y ) return( TRUE );
    return( FALSE );
    }


BOOL
pointing_up( polygon *p )
    {
    static xy a, b, c;
    a = vertex( p, 0 );
    b = vertex( p, 1 );
    c = vertex( p, 2 );
    if ( a.y == b.y ) return( c.y > a.y );
    if ( a.y == c.y ) return( b.y > a.y );
    if ( b.y == c.y ) return( a.y > b.y );
    printf( "\npointing_up() expected one edge to be on the X axis.\n" );
    printf( "a:  %ld  %ld \n", a.x, a.y );
    printf( "b:  %ld  %ld \n", b.x, b.y );
    printf( "c:  %ld  %ld \n", c.x, c.y );
    bail( "gpfout/pointing_up is confused" );
    }


BOOL
pointing_right( polygon *p )
    {
    static xy a, b, c;
    a = vertex( p, 0 );
    b = vertex( p, 1 );
    c = vertex( p, 2 );
    if ( a.x == b.x ) return( c.x > a.x );
    if ( a.x == c.x ) return( b.x > a.x );
    if ( b.x == c.x ) return( a.x > b.x );
    printf( "\npointing_right() expected one edge to be on the Y axis.\n" );
    printf( "a:  %ld  %ld \n", a.x, a.y );
    printf( "b:  %ld  %ld \n", b.x, b.y );
    printf( "c:  %ld  %ld \n", c.x, c.y );
    bail( "gpfout/pointing_right is confused" );
    }


void
swap( xy *a, xy *b )
    {
    static xy t;
    t.x = a->x;
    t.y = a->y;
    a->x = b->x;
    a->y = b->y;
    b->x = t.x;
    b->y = t.y;
    return;
    }


void
put_c_at_top( polygon *p, xy *a, xy *b, xy *c )  // of an x triangle
    {                                            // also put a on the left, b on the right
    static xy aa, bb, cc, t;
    aa = vertex( p, 0 );
    bb = vertex( p, 1 );
    cc = vertex( p, 2 );
    if ( aa.y == bb.y )    // cc is already at the top or bottom
        {
        a->x = aa.x;
        a->y = aa.y;
        b->x = bb.x;
        b->y = bb.y;
        c->x = cc.x;
        c->y = cc.y;
        if ( a->x > b->x ) swap( a, b );
        }
    else if ( aa.y == cc.y )   // bb is at top or bottom
        {
        c->x = bb.x;
        c->y = bb.y;
        a->x = aa.x;
        a->y = aa.y;
        b->x = cc.x;
        b->y = cc.y;
        if ( a->x > b->x ) swap( a, b );
        }
    else if ( bb.y == cc.y )   // aa is at top or bottom
        {
        c->x = aa.x;
        c->y = aa.y;
        b->x = bb.x;
        b->y = bb.y;
        a->x = cc.x;
        a->y = cc.y;
        if ( a->x > b->x ) swap( a, b );
        }
    else
        bail( "logic error in gpfout/put_c_at_top" );
    return;
    }


void
put_c_at_bottom( polygon *p, xy *a, xy *b, xy *c )
    {
    put_c_at_top( p, a, b, c );
    return;
    }


void
put_c_on_right( polygon *p, xy *a, xy *b, xy *c )  // and also put 'a' below 'b'
    {
    static xy aa, bb, cc, t;
    aa = vertex( p, 0 );
    bb = vertex( p, 1 );
    cc = vertex( p, 2 );
    if ( aa.x == bb.x )  // then cc is already on the right or left
        {
        a->x = aa.x;
        a->y = aa.y;
        b->x = bb.x;
        b->y = bb.y;
        c->x = cc.x;
        c->y = cc.y;
        if ( a->y > b->y ) swap( a, b );
        }
    else if ( aa.x == cc.x )   // then bb is on the right or left
        {
        c->x = bb.x;
        c->y = bb.y;
        a->x = aa.x;
        a->y = aa.y;
        b->x = cc.x;
        b->y = cc.y;
        if ( a->y > b->y ) swap( a, b );
        }
    else if ( bb.x == cc.x )   // then aa is on the right or left
        {
        c->x = aa.x;
        c->y = aa.y;
        b->x = bb.x;
        b->y = bb.y;
        a->x = cc.x;
        a->y = cc.y;
        if ( a->y > b->y ) swap( a, b );
        }
    else
        {
        printf( "\na.x != b.x  and  a.x != c.x\n" );
        printf( "a:  %ld  %ld \n", aa.x, aa.y );
        printf( "b:  %ld  %ld \n", bb.x, bb.y );
        printf( "c:  %ld  %ld \n", cc.x, cc.y );
        bail( "logic error in gpfout/put_c_on_right" );
        }
    return;
    }


void
put_c_on_left( polygon *p, xy *a, xy *b, xy *c )
    {
    put_c_on_right( p, a, b, c );
    return;
    }


BOOL
right_triangle( polygon *p )
    {
    static xy a, b, c;
    if ( p->len != 3 ) bail( "triangle expected in gpfout/right_triangle" );
    a = vertex( p, 0 );
    b = vertex( p, 1 );
    c = vertex( p, 2 );
    if ( a.x == b.x && (b.y == c.y || a.y == c.y) ) return( TRUE );
    if ( a.x == c.x && (a.y == b.y || c.y == b.y) ) return( TRUE );
    if ( b.x == c.x && (b.y == a.y || c.y == a.y) ) return( TRUE );
    return( FALSE );
    }


void
reorder_right_triangle( polygon *p, xy *a, xy *b, xy *c, int *corner )   // put 'a' in the right-angle corner
    {                                                                    // put 'c' at the top,
    static BOOL a_is_90;                                                 // and return the corner number of the right angle
    
    if ( p->len != 3 ) bail( "triangle expected in gpfout/reorder_right_triangle" );

    *a = vertex( p, 0 );
    *b = vertex( p, 1 );
    *c = vertex( p, 2 );
    
    a_is_90 = (a->y == b->y && a->x == c->x) || (a->y == c->y && a->x == b->x);

    if ( ! a_is_90 )
        {
        swap( a, b );
        a_is_90 = (a->y == b->y && a->x == c->x) || (a->y == c->y && a->x == b->x);
        if ( ! a_is_90 )
            {
            swap( a, c );
            a_is_90 = (a->y == b->y && a->x == c->x) || (a->y == c->y && a->x == b->x);
            if ( ! a_is_90 )
                bail( "logic error 1 in gpfout/reorder_right_triangle" );
            }
        }

    // now 'a' is the right angle

    if ( b->y > c->y ) swap( b, c );

    // now 'c' is higher

    if ( a->y == c->y && a->x > c->x )
                                             *corner = 0;  // UR
    else if ( a->y == c->y && c->x > a->x )
                                             *corner = 1;  // UL
    else if ( a->y == b->y && b->x > a->x )
                                             *corner = 2;  // LL
    else if ( a->y == b->y && a->x > b->x )
                                             *corner = 3;  // LR
    else
        bail( "logic error 2 in gpfout/reorder_right_triangle" );

    return;
    }

//---------------------------------------------------------------------------------------------------
// trapezia helper functions

BOOL
x_trap( polygon *p )        // Are the parallel sides parallel to the X axis?
    {                       // If not, they better be parallel to the Y axis.
    static xy a, b, c, d;
    if ( p->len != 4 ) bail( "four-sided shape expected" );
    a = vertex( p, 0 );
    b = vertex( p, 1 );
    c = vertex( p, 2 );
    d = vertex( p, 3 );
    if ( a.y == b.y && c.y == d.y ) return( TRUE );
    if ( b.y == c.y && d.y == a.y ) return( TRUE );
    if ( a.x == b.x && c.x == d.x ) return( FALSE );
    if ( b.x == c.x && d.x == a.x ) return( FALSE );
    bail( "This trap is not a parallelogram." );
    }


void                                                        //    d   c     this should be    
reorder_x_trap( polygon *p, xy *a, xy *b, xy *c, xy *d )    //    a   b     the output order
    {
    if ( p->ll.x == p->ur.x || p->ll.y == p->ur.y ) bail( "gpfout/reorder_x_trap: degenerate trapezoid" );
    if ( p->len != 4 ) bail( "gpfout/reorder_x_trap: shape does not have four sides" );

    *a = vertex( p, 0 );
    *b = vertex( p, 1 );
    *c = vertex( p, 2 );
    *d = vertex( p, 3 );
    
    if ( a->y == b->y && c->y == d->y )        //  a  b      b  a      a  b      b  a      c  d      c  d      d  c      d  c
        {                                      //  c  d      c  d      d  c      d  c      a  b      b  a      a  b      b  a
        if ( a->x > b->x ) swap( a, b );       //
        if ( d->x > c->x ) swap( d, c );       //  a  b      a  b      a  b      a  b      d  c      d  c      d  c      d  c   
        if ( a->y > d->y )                     //  d  c      d  c      d  c      d  c      a  b      a  b      a  b      a  b
            {                                  //
            swap( a, d );                      //  d  c      d  c      d  c      d  c      d  c      d  c      d  c      d  c
            swap( b, c );                      //  a  b      a  b      a  b      a  b      a  b      a  b      a  b      a  b
            }                                  //
        }
    else if ( b->y == c->y && d->y == a->y )   //   b  c      c  b      b  c      c  b      d  a      d  a      a  d      a  d
        {                                      //   d  a      d  a      a  d      a  d      b  c      c  b      b  c      c  b
        if ( a->x > d->x ) swap( a, d );       //             
        if ( b->x > c->x ) swap( b, c );       //   b  c      b  c      b  c      b  c      a  d      a  d      a  d      a  d
        if ( a->y > b->y )                     //   a  d      a  d      a  d      a  d      b  c      b  c      b  c      b  c
            {                                  //
            swap( a, b );                      //   b  c      b  c      b  c      b  c      b  c      b  c      b  c      b  c
            swap( d, c );                      //   a  d      a  d      a  d      a  d      a  d      a  d      a  d      a  d
            }                                  //
        swap( b, d );                          //   d  c       
        }                                      //   a  b
    else
        bail( "gpfout/reorder_x_trap: not a parallelogram" );

    if ( ! ( a->y == b->y && d->y == c->y && a->x < b->x && d->x < c->x && a->y < d->y ) )
        {
        printf( "\n" );
        printf( "a:  %6ld  %6ld \n", a->x, a->y );
        printf( "b:  %6ld  %6ld \n", b->x, b->y );
        printf( "c:  %6ld  %6ld \n", c->x, c->y );
        printf( "d:  %6ld  %6ld \n", d->x, d->y );
        bail( "logic error in reorder_x_trap" );
        }

    return;
    }



void                                                        //    d   c     this should be    
reorder_y_trap( polygon *p, xy *a, xy *b, xy *c, xy *d )    //    a   b     the output order
    {
    if ( p->ll.x == p->ur.x || p->ll.y == p->ur.y ) bail( "gpfout/reorder_y_trap: degenerate trapezoid" );
    if ( p->len != 4 ) bail( "gpfout/reorder_y_trap: shape does not have four sides" );

    *a = vertex( p, 0 );
    *b = vertex( p, 1 );
    *c = vertex( p, 2 );
    *d = vertex( p, 3 );
    
    if ( a->x == b->x && c->x == d->x )        //  a  d      a  c      b  d      b  c      d  a      d  a      d  b      c  b
        {                                      //  b  c      b  d      a  c      a  d      c  b      c  b      c  a      d  a
        if ( a->y > b->y ) swap( a, b );       //
        if ( d->y > c->y ) swap( d, c );       //  b  c      b  c      b  c      b  c      c  b      c  b      c  b      c  b
        if ( a->x > d->x )                     //  a  d      a  d      a  d      a  d      d  a      d  a      d  a      d  a
            {                                  //
            swap( a, d );                      //                                          b  c      b  c      b  c      b  c
            swap( b, c );                      //                                          a  d      a  d      a  d      a  d
            }                                  //  d  c
        swap( b, d );                          //  a  b
        }
    else if ( b->x == c->x && d->x == a->x )   //  b  a      b  d      c  a      c  d      a  b      d  b      a  c      d  c
        {                                      //  c  d      c  a      b  d      b  a      d  c      a  c      d  b      a  b
        if ( a->y > d->y ) swap( a, d );       //             
        if ( b->y > c->y ) swap( b, c );       //  c  d      c  d      c  d      c  d      d  c      d  c      d  c      d  c
        if ( a->x > b->x )                     //  b  a      b  a      b  a      b  a      a  b      a  b      a  b      a  b
            {                                  //
            swap( a, b );                      //  d  c      d  c      d  c      d  c      
            swap( d, c );                      //  a  b      a  b      a  b      a  b
            }                                  //
        }                                      // 
    else
        bail( "gpfout/reorder_y_trap: not a parallelogram" );

    if ( ! ( a->x == d->x && b->x == c->x && a->x < b->x && d->x < c->x && a->y < d->y ) )
        bail( "logic error in reorder_y_trap" );

    return;
    }


//---------------------------------------------------------------------------------------------------

BOOL
check_trap( gpf_union_t *offsetxrecord,
            gpf_union_t *offsetyrecord,
            gpf_union_t *baselsw,
            gpf_union_t *heightlsw,
            gpf_union_t *shiftltlsw,
            gpf_union_t *shiftrblsw,
            gpf_union_t *activaterecord,
            int          sstep )
    {

    // If the Y deflection goes negative by less than a beam step,
    // then we should fix the rb shift value. If the X deflection
    // goes a little negative then we should fix the lt shift value.
    // What else can go wrong? Anything! The TOP line might actually
    // go below zero, or the RIGHT side might go negative. So crazy.

    // Polygons are so much simpler. But as long as we can use pss=lsb,
    // and then split/bury the glitches, the pattern comes out pretty nice.

    // We also check for shapes that are "too sharp," that is, when
    // the shift is too large. The gpf manual gives a weird equation for this.

    static int base, height, top_shift, bottom_shift, offsetx, offsety, subits;
    static int bottom_run, top_run, top, bottom, nag=0, max_shift;
    static int left_shift, right_shift, left_run, right_run, left, right;
    static BOOL init=FALSE, top_neg, bottom_neg, left_neg, right_neg, lt90, rb90;
    static float fsubits, fleft_shift, fright_shift, fheight;
    static float fbottom_shift, fbase, ftop_shift;
    static long count;

    if ( ! init )
        {
        init = TRUE;
        subits = pow( 2, SUBDACBITS );  // 2^14 = 16384
        fsubits = subits;
        // maximum shift, according to the gpf manual. Beyond this, the trap is too "sharp".
        max_shift = (((subits * subits / 4) - 1) / sstep) - (subits / 2);   // ye gods
        }
    
    
    base    = baselsw->prim24.val;
    height  = heightlsw->prim24.val;
    fbase   = base;
    fheight = height;
    lt90    = activaterecord->activate.lt90 == 1;
    rb90    = activaterecord->activate.rb90 == 1;

    if ( base == 0 && height == 0 ) return( FALSE );  

    // Look for narrow slivers.
    // Should we replace slivers with single-pixel lines?
    // For now, we simply skip them.

    top_shift     = shiftltlsw->prim24.val;   // or possibly the left shift
    bottom_shift  = shiftrblsw->prim24.val;   // or possibly the right shift

    if ( lt90 )
        {
        top_shift = 0;
        left_shift = 0;
        shiftltlsw->prim24.val = 0;
        }
    
    if ( rb90 )
        {
        bottom_shift = 0;
        right_shift = 0;
        shiftrblsw->prim24.val = 0;
        }
    
    ftop_shift    = top_shift;
    fbottom_shift = bottom_shift;
    
    
    if ( top_shift > max_shift || bottom_shift > max_shift ) return( FALSE );


    // Look for zero shifts, which should be flagged as 90 degrees

    if ( top_shift    == 0 ) { lt90 = 1; activaterecord->activate.lt90 = 1; }
    if ( bottom_shift == 0 ) { rb90 = 1; activaterecord->activate.rb90 = 1; }
     
    // look for negative deflections
    
    if ( activaterecord->activate.OrientationY == 1 )
        {
        top_shift    = shiftltlsw->prim24.val;
        bottom_shift = shiftrblsw->prim24.val;
        offsetx      = offsetxrecord->prim24.val;
        offsety      = offsetyrecord->prim24.val;
        bottom_neg   = (activaterecord->activate.rbNeg == 1);
        top_neg      = (activaterecord->activate.ltNeg == 1);

        fbottom_shift = bottom_shift;
        bottom_run = round( fbase * fbottom_shift / fsubits );

        if ( rb90 ) bottom_run = 0;  // redundant
        if ( bottom_neg ) bottom_run = -bottom_run;

        bottom = offsety - bottom_run;   // the bottom according to the bottom shift

        if ( bottom < 0 )
            {
            if ( bottom < -sstep )
                {
                warning( "trap bottom side has a large -Y deflection", &nag );
                }
            else
                {
                // bottom_run = offsety;
                bottom_shift = subits * offsety / base;
                
                if ( bottom_neg )
                    bottom = offsety + base * bottom_shift / subits;
                else
                    bottom = offsety - base * bottom_shift / subits;

                if ( bottom < 0 ) printf( "** trap y bottom shift fix unsuccessful **\n" );
                shiftrblsw->prim24.val = bottom_shift;
                if ( bottom_shift == 0 ) { rb90 = 1; activaterecord->activate.rb90 = 1; }
                }
            }

        // now check the top side, which might dip below zero
        
        ftop_shift = top_shift;
        top_run = round( fbase * ftop_shift / fsubits );
        if ( lt90 ) top_run = 0;
        if ( top_neg ) top_run = -top_run;
        bottom  = offsety + height - top_run;

        if ( bottom < 0 )
            {
            if ( bottom < -sstep )
                {
                warning( "trap top side has large -Y deflection" , &nag );
                }
            else
                {
                // top_run = offsety + height;
                top_shift = subits * (offsety + height) / base;

                if ( top_neg )
                    bottom = offsety + height + base * top_shift / subits;
                else
                    bottom = offsety + height - base * top_shift / subits;
                
                if ( bottom < 0 ) printf( "** trap y top shift fix unsuccessful **\n" );
                shiftltlsw->prim24.val = top_shift;
                if ( top_shift == 0 ) { lt90 = 1; activaterecord->activate.lt90 = 1; }
                }
            }
        }
    else          // X orientation
        {
        left_shift   = shiftltlsw->prim24.val;
        right_shift  = shiftrblsw->prim24.val;
        offsetx      = offsetxrecord->prim24.val;
        offsety      = offsetyrecord->prim24.val;
        left_neg     = (activaterecord->activate.ltNeg == 1);
        right_neg    = (activaterecord->activate.rbNeg == 1);
        
        fleft_shift = left_shift;
        left_run  = round( fheight * fleft_shift / fsubits );

        if ( lt90 ) left_run = 0;  // redundant

        if ( ! left_neg  ) left_run = -left_run;

        left = offsetx - left_run;

        if ( left < 0 )
            {
            if ( left < -sstep )
                {
                warning( "trap has large -X deflection", &nag );
                }
            else
                {
                // left_run = offsetx;
                left_shift = subits * offsetx / height;

                if ( left_neg )
                    left = offsetx - height * left_shift / subits;
                else
                    left = offsetx - height * left_shift / subits;
                
                if ( left <  0 ) printf( "**trap x left shift fix unsuccessful**" );
                shiftltlsw->prim24.val = left_shift;
                if ( left_shift == 0 ) { lt90 = 1; activaterecord->activate.lt90 = 1; }
                }
            }

        // now check the right side, which might go too far to the left

        fright_shift = right_shift;
        right_run = round( fheight * fright_shift / fsubits );
        if ( rb90 ) right_run = 0;
        if ( right_neg ) right_run = -right_run;
        right = offsetx + base + right_run;

        if ( right < 0 )
            {
            if ( right < -sstep )
                {
                warning( "trap has large -X deflection from right side", &nag );
                }
            else
                {
                // right_run = offsetx + base;
                right_shift = subits * (offsetx + base) / height;

                if ( right_neg )
                    right = offsetx + base - height * right_shift / subits;
                else
                    right = offsetx + base + height * right_shift / subits;

                if ( right < 0 ) printf( "** unable to fix negative trap x shift from right side **\n" );
                
                shiftrblsw->prim24.val = right_shift;
                if ( right_shift == 0 ) { rb90 = 1; activaterecord->activate.rb90 = 1; }
                }
            }
        }
    return( TRUE );
    }


//---------------------------------------------------------------------------------------------------

void
write_trap(     FILE *fp,
                BOOL         start_subfield,
                dxy          block,            // um   block size
                gpf_union_t *prims, 
                gpf_union_t *prims_org, 
                polygon     *poly,             // array of xy, mfr units (usually 1 dbu = 1nm)
                dxy          fieldcenter,      // um   gds units
                dxy          subfieldsize,     // um
		dxy          main_res,         // um   usually 1.0 nm
		dxy          sub_res,          // um   usually 0.5 nm
		double       bss_um,           // um
                double       dbu_um,           // um   database unit, usually 1 nm, sometimes 0.1 nm
                BOOL         assign_doses,     
                double     **dose,             // dose matrix, doubles
                int        **msfm,             // msf multipliers, int
		int          nx,               //      number of columns
                int          ny,               //      number of rows
                double       dx,               // um   pitch in x
                double       dy,               // um   pitch in y
                gpf_ulong_t *gpfpos       )
    {
    static BOOL         init=FALSE, horizontal, ok;
    static int          submsf, mstep, minarea, warnings=0, base, height, temp, sstep, sstep2;
    static int          left_shift, right_shift, left_slope_neg, right_slope_neg, corner;
    static int          subits;
    static long         count;
    static xy           maxsubfield, v, a, b, c, d;
    static dxy          block2, llblock, ll, ur, deflection, far_ur, dv, dmaxsub;
    static double       x0, y0, offset_x, offset_y, gds2subit, rise, run;
    static char         mess[256];
    static twoints      clock, prev_clock;

    static gpf_union_t  clockrecord_lsw, clockrecord_msw, msfrecord, activaterecord;
    static gpf_union_t  mainxrecord, mainyrecord, offsetxrecord, offsetyrecord;
    static gpf_union_t  baselsw, heightlsw, shiftltlsw, shiftrblsw;
    static gpf_union_t  repvect1x, repvect1y, repvect2x, repvect2y, repvect1c, repvect2c;
    static gpf_uint_t   this_msf, prev_msf, prev_mainx, prev_mainy, prev_offsetx, prev_offsety;

    // The header says "outline mode," which should resize the shapes. But unfortunately
    // outline mode works only for polygons. Trapezia must be resized manually. Drat.

    // Integer arithmetic might cause trapezia to have negative deflections. The secret sauce
    // is the function check_trap(), which fixes sub-pixel errors and filters out
    // sharp shapes.
    
    if ( ! init )
	{
	init = TRUE;
        subits = pow( 2, SUBDACBITS ); 
        gds2subit = dbu_um / sub_res.x;        // 0.001 / 0.0005 usually
        submsf = floor( bss_um / sub_res.x );
        maxsubfield.x = subfieldsize.x * 1.02;
        maxsubfield.y = subfieldsize.y * 1.02;
        dmaxsub.x = maxsubfield.x;
        dmaxsub.y = maxsubfield.y;
        if ( maxsubfield.x > MAXSUBFIELD ) maxsubfield.x = MAXSUBFIELD;
	if ( maxsubfield.y > MAXSUBFIELD ) maxsubfield.y = MAXSUBFIELD;
        mstep = floor( bss_um / main_res.x );  // bss in mainfield bits
        sstep = floor( bss_um / sub_res.x );   // bss in subfield bits ... same as submsf             
        sstep2 = sstep / 2;                    // bss better be an even number of bits
        minarea = mstep * mstep * 10;
	block2.x = alittlebigger( block.x / 2.0 );
        block2.y = alittlebigger( block.y / 2.0 );
        clockrecord_lsw.prim16.IC       = GPF_PRIM_FREQUENCY_LSW;
        clockrecord_msw.prim16.IC       = GPF_PRIM_FREQUENCY_MSW;
        msfrecord.prim16.IC             = GPF_PRIM_MSF;
        mainxrecord.prim24.IC           = GPF_PRIM_MAIN_X_LSW;
        mainyrecord.prim24.IC           = GPF_PRIM_MAIN_Y_LSW;
        offsetxrecord.prim24.IC         = GPF_PRIM_OFFSET_X_LSW;
        offsetyrecord.prim24.IC         = GPF_PRIM_OFFSET_Y_LSW;
        repvect1x.prim24.IC             = GPF_PRIM_MAINREPVECT1_X_LSW;
        repvect1y.prim24.IC             = GPF_PRIM_MAINREPVECT1_Y_LSW;
        repvect2x.prim24.IC             = GPF_PRIM_MAINREPVECT2_X_LSW;
        repvect2y.prim24.IC             = GPF_PRIM_MAINREPVECT2_Y_LSW;
        repvect1c.prim24.IC             = GPF_PRIM_MAINREPVECT1_CNT;
        repvect2c.prim24.IC             = GPF_PRIM_MAINREPVECT2_CNT;
        baselsw.prim24.IC               = GPF_PRIM_BASE_LSW;
        heightlsw.prim24.IC             = GPF_PRIM_HEIGHT_LSW;
        shiftltlsw.prim24.IC            = GPF_PRIM_SHIFT_LT_LSW;
        shiftrblsw.prim24.IC            = GPF_PRIM_SHIFT_RB_LSW;
        prev_clock.msw = 0;
        prev_clock.lsw = 0;
        prev_msf = 0;
        prev_mainx = 0;
        prev_mainy = 0;
        prev_offsetx = 0;
        prev_offsety = 0;
	}

    // return if degenerate

    if ( ! poly->active ) return;
    if ( poly->len > 4) bail( "gpfout/write_trap: too many vertices in a trapezoid" );
    if ( poly->len < 3) return;
    if ( poly->ll.x >= poly->ur.x || poly->ll.y >= poly->ur.y ) return;

    activaterecord.prim24.val = 0;                                   // erase all those trapezia options
    activaterecord.prim24.IC        = GPF_PRIM_ACTIVATE;
    activaterecord.activate.ShapeID = GPF_SHAPE_TRAPEZIUM;
    
    ll.x = poly->ll.x * dbu_um - fieldcenter.x;                      // LL corner of the shape, um, origin at field center
    ll.y = poly->ll.y * dbu_um - fieldcenter.y;

    llblock.x = fieldcenter.x - block.x / 2.0;                       // LL corner of this block, gds um
    llblock.y = fieldcenter.y - block.y / 2.0;
                                                                     // x0, y0 is the main field deflection to the
                                                                     // lower-left corner of the subfield (um).

    x0 = llblock.x - fieldcenter.x + subfieldsize.x * (floor(((poly->ll.x * dbu_um) - llblock.x) / subfieldsize.x));
    y0 = llblock.y - fieldcenter.y + subfieldsize.y * (floor(((poly->ll.y * dbu_um) - llblock.y) / subfieldsize.y));

    /*
    printf( "\n" );
    printf( "dbu_um       = %10.5f \n", dbu_um );
    printf( "llblock      = %8.3f %8.3f \n", llblock.x, llblock.y );
    printf( "fieldcenter  = %8.3f %8.3f \n", fieldcenter.x, fieldcenter.y );
    printf( "subfieldsize = %8.3f %8.3f \n", subfieldsize.x, subfieldsize.y );
    printf( "poly->ll     = %8d %8d \n", poly->ll.x, poly->ll.y );
    printf( "llblock      = %8.3f %8.3f \n", llblock.x, llblock.y );
    printf( "x0, y0       = %8.3f %8.3f \n", x0, y0 );
    printf( "\n" );
    */
    
    offset_x = ll.x - x0;                                    // um, to be converted to subfield db units
    offset_y = ll.y - y0;                                    // um

    // check trap extents
 
    ur.x = (poly->ur.x * dbu_um - fieldcenter.x);                     // um  UR corner of the shape, origin at field center
    ur.y = (poly->ur.y * dbu_um - fieldcenter.y);                     // um

    deflection.x = ur.x - x0;
    deflection.y = ur.y - y0;

    if ( deflection.x < 0.0 || deflection.y < 0.0 )
	{
	printf( "\n\nLL= %1.3lf, %1.3lf\n", ll.x, ll.y );
	printf(     "UR= %1.3lf, %1.3lf\n", ur.x, ur.y );
        printf(     "field center= %1.3lf, %1.3lf\n", fieldcenter.x, fieldcenter.y );
        printf(     "dbu_um = %1.6lf\n", dbu_um );
	bail( "(gpfout/write_trap) arithmetic confusion" );
	}
    
    // we can let the deflection exceed the subfield by a tiny bit, 
    // but the deflection has to be less than 4.2 um

    if ( (deflection.x > maxsubfield.x) || (deflection.y > maxsubfield.y) ||
         (offset_x     < 0.0)           || (offset_y     < 0.0)            )   //  then we need a new mfd
	{
        x0 = ll.x;       // set mainfield deflection to LL corner of this polygon
	y0 = ll.y;
        deflection.x = ur.x - x0;
        deflection.y = ur.y - y0;
        offset_x = ll.x - x0;
        offset_y = ll.y - y0;
        
        if ( offset_x < 0 || offset_y < 0 || deflection.x < 0 || deflection.y < 0 )
            {
	    printf( "\n (gpfout/write_trap) negative deflection error: offset= %1.2lf, %1.2lf  defl=  %1.2lf, %1.2lf um     ", 
			offset_x, offset_y, deflection.x, deflection.y );
	    printf( " x0,y0= %1.4lf, %1.4lf um  ", x0, y0 );
            }
	                                           // keep score of extra mfdefs because extra deflections take extra time
        if ( (nx == 1) && (ny == 1) ) mfdefs++;    // but the mfdef for an array doesn't count 
	}


    if ( (deflection.x > maxsubfield.x) || (deflection.y > maxsubfield.y) )
	{
	printf( "\n\nsubfield origin = %1.3lf, %1.3lf um\n", x0, y0 );
        printf(     "shape ll        = %ld, %ld dbu\n", poly->ll.x, poly->ll.y );
        printf(     "shape ur        = %ld, %ld dbu\n", poly->ur.x, poly->ur.y );
        printf(     "deflection      = %1.3lf, %1.3lf um\n", deflection.x, deflection.y );
        printf(     "nx, ny          = %d, %d\n", nx, ny );
        printf(     "field center    = %1.6lf, %1.6lf um\n\n", fieldcenter.x, fieldcenter.y );

        if ( not_a_large_rectangle( poly, dmaxsub, dbu_um, TRUE ) )
	    printf( "This is not an arrayed rectangle, but it is still bigger than the subfield?\n\n" );
	else
	    printf( "This is a large rectangle, so why is it in write_trap instead of write_rectangles?\n\n" );

        print_polygon( poly, 1 );
        printf( "trapezoid area: %lf\n", polygon_area( poly, mstep ) );

        bail( "trapezoid size exceeds subfield size" );
	}

    // check array extents
    // it doesn't seem to matter if the array goes a bit beyond the block boundary,
    // so this check is disabled.

    // ur and ll are the corners of the shape, in um, with origin at field center
 
    if ( FALSE ) // (nx > 1) || (ny > 1) )
	{
        far_ur.x = ur.x + (nx - 1) * dx;
        far_ur.y = ur.y + (ny - 1) * dy;
        
        if ( (far_ur.x > block2.x) || (far_ur.y > block2.y) )
	    {
	    v = vertex( poly, 0 );
            dv.x = v.x * dbu_um;
	    dv.y = v.y * dbu_um;
	    sprintf( mess, "array exceeds block size at %lf, %lf um  d= %lf, %lf um  n= %d, %d  urext= %lf, %lf um  block2= %lf, %lf", 
		     dv.x, dv.y, dx, dy, nx, ny, far_ur.x, far_ur.y, block2.x, block2.y );
	    warning( mess, &warnings ); 
            poly->layer = ERRORLAYER;
	    }
        }


    
    if ( assign_doses )
	{
        clock = dose2clock( fabs( dose[ poly->layer ][ poly->dt ] ) );               // fabs because we mark a dose 'hit' with the sign
        dose[ poly->layer ][ poly->dt ] = -fabs( dose[ poly->layer ][ poly->dt ] );  // just to keep track of the number of doses
        
        clockrecord_lsw.prim24.val = clock.lsw;
        clockrecord_msw.prim24.val = clock.msw;

        this_msf = submsf * msfm[ poly->layer ][ poly->dt ];
        sstep  = this_msf;                                         // used for filtering sharp shapes in check_trap()
        sstep2 = sstep / 2;
        msfrecord.prim16.val2 = this_msf;
        
        if ( start_subfield || (prev_clock.msw != clock.msw) || (prev_clock.lsw != clock.lsw) || (prev_msf != this_msf ) )
	    {
            gpfree_write_union( fp, gpfpos, &clockrecord_lsw ); 
            gpfree_write_union( fp, gpfpos, &clockrecord_msw );
            gpfree_write_union( fp, gpfpos, &msfrecord       ); 
            prev_clock.lsw = clock.lsw;
            prev_clock.msw = clock.msw;
            prev_msf       = this_msf; 
	    }
	}

    
    // mainfield deflection... should not be set if it is the same as the previous
    // unless of course we are starting a new subfield (or a new field)

    mainxrecord.prim24.val = MAINDACMID + x0 / main_res.x; 
    mainyrecord.prim24.val = MAINDACMID + y0 / main_res.y; 

    // Polygon outline mode shifts the pixels by 0.5 bss, but for trapezoids                              
    // we must apply manual butting corrections. sstep2 is half the beam step size.
    // Shift the shape sstep2 up and right, then later make the trapezoid smaller by sstep.

    if ( start_subfield || prev_mainx != mainxrecord.prim24.val || prev_mainy != mainyrecord.prim24.val )
        {
        gpfree_write_union( fp, gpfpos, &mainxrecord );                 // write main deflection x
        gpfree_write_union( fp, gpfpos, &mainyrecord );                 // write main deflection y
        prev_mainx = mainxrecord.prim24.val;
        prev_mainy = mainyrecord.prim24.val;
        }

    activaterecord.repeat.MainRepVect1 = 0;
    activaterecord.repeat.MainRepVect2 = 0;
    

    if ( nx > 1 || ny > 1 )
	{
        repvect1x.prim24.val = lround( dx / main_res.x );      // floor does not help close the gaps
        repvect1y.prim24.val = 0;
        repvect2x.prim24.val = 0;
        repvect2y.prim24.val = lround( dy / main_res.y );      // floor does not help

        // Here's a subtle point: the SIZE of a trapezoid is in subfield resolution units,
        // but the PITCH of an array is in mainfield resolution units. If the pitch is
        // an odd number, then the size cannot match. But it's really not bad enough
        // to worry about it. At worst the ends of arrays might have little sub-pixel errors.

        if ( nx > 1 )
	    {
            repvect1c.prim24.val = nx - 1;
            activaterecord.repeat.MainRepVect1 = 1;
            gpfree_write_union( fp, gpfpos, &repvect1x );
            gpfree_write_union( fp, gpfpos, &repvect1y );
            gpfree_write_union( fp, gpfpos, &repvect1c ); 
            }

        if ( ny > 1 )
	    {
            repvect2c.prim24.val = ny - 1;
            activaterecord.repeat.MainRepVect2 = 1;
            gpfree_write_union( fp, gpfpos, &repvect2x ); 
            gpfree_write_union( fp, gpfpos, &repvect2y ); 
            gpfree_write_union( fp, gpfpos, &repvect2c ); 
            }

        // We do not have to remember the previous repeat values, since the activate line indicates the 
	// use of repeats. That is, if MRV1 or MRV2 appear after ACTIVATE, then we have repeats. There is no
        // implication from the previous ACTIVATE.

        }

    // Write the trap primitive and the activate line
    // ha ha that's the hard part. good luck asshole.

    // poly is in gds database units, typically 1 nm or 1 angstrom.
    // gds2subit = dbu_um / sub_res is normally 1 nm / 0.5 nm = 2.0, or 0.1nm/0.5nm = 0.2

    // Reduce height and base by one beam step. Note that we set the offset
    // +X and +Y by half a beam step. This is equivalent to "outline" mode for polygons.
    // So we really want the beam step to be an even number of bits.

    // First consider rectangles, then triangles, then parallelograms.
    // So many cases! This is why polygons are way better.

    if ( this_is_a_rectangle( poly, subfieldsize, dbu_um ) )
        {
        offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );           // offset for a rectangle
        offsetyrecord.prim24.val = sstep2 + floor( offset_y / sub_res.y );           // is simply the LL corner.

        base   = floor( (poly->ur.x - poly->ll.x) * gds2subit ) - sstep;
        height = floor( (poly->ur.y - poly->ll.y) * gds2subit ) - sstep;

        if ( base <= 0   ) base   = 1;
        if ( height <= 0 ) height = 1;

        horizontal = base > height;

        // If the shape is vertical, 
        // one does *not* switch base and height.
        // The function of "OY" on the activate line
        // is to indicate the direction of the raster.
        // Yeesh.

        baselsw.prim24.val   = base;
        heightlsw.prim24.val = height;
    
        activaterecord.activate.lt90 = 1;
        activaterecord.activate.rb90 = 1;
    
        if ( horizontal )
            activaterecord.activate.OrientationY = 0;
        else
            activaterecord.activate.OrientationY = 1;

        // Before writing the gtx records, check for negative deflections
        // and correct any errors smaller than one beam step.
        
        ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

        // Write the gpf records - offset, base, height, activate

        if ( ok )
            {
            gpfree_write_union( fp, gpfpos, &offsetxrecord );                               // always write the offset
            gpfree_write_union( fp, gpfpos, &offsetyrecord );                               // for each shape
            gpfree_write_union( fp, gpfpos, &baselsw   );
            gpfree_write_union( fp, gpfpos, &heightlsw );
            gpfree_write_union( fp, gpfpos, &activaterecord );
            }
        }
    else if ( poly->len == 3 )                                                // Triangle
        {
        horizontal = (poly->ur.x - poly->ll.x) > (poly->ur.y - poly->ll.y);

        if ( right_triangle( poly ) )
            {
            reorder_right_triangle( poly, &a, &b, &c, &corner );   // 'a' is now the right angle
                                                                   // and 'c' is higher
            if ( corner == 0 )       // UR                         //       c    a
                {                                                  //            b
                offset_x = offset_x + (b.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (b.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val = -sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val =  sstep  + floor( offset_y / sub_res.y );
                base   = 0;
                height = floor( (a.y - b.y) * gds2subit ) - sstep - sstep2;
                if ( height <= 0 ) height = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;
                rise = (a.y - b.y) * gds2subit;
                run  = (a.x - c.x) * gds2subit - sstep;   // subtle difference here
                left_shift  = abs( floor( subits * run / rise ) );
                shiftltlsw.prim24.val = left_shift;
                activaterecord.activate.OrientationY = 0;
                activaterecord.activate.ltNeg = 1;
                activaterecord.activate.rb90  = 1;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );                
                    gpfree_write_union( fp, gpfpos, &shiftltlsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }                                                 //  a   c
            else if ( corner == 1 )  // UL                        //  b
                {
                offset_x = offset_x + (b.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (b.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val =  sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val =  sstep2 + floor( offset_y / sub_res.y );
                base   = floor( (c.x - a.x) * gds2subit ) - sstep;
                height = floor( (a.y - b.y) * gds2subit ) - sstep;
                if ( height <= 0 ) height = 1;
                if ( base   <= 0 ) base   = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;
                //rise = c.x - a.x;
                //run  = a.y - b.y;
                right_shift  = abs( floor( subits * height / base ) );
                //right_shift  = abs( floor( subits * run / rise ) );
                shiftrblsw.prim24.val = right_shift;
                activaterecord.activate.OrientationY = 1;      // OY
                activaterecord.activate.rbNeg = 1;
                activaterecord.activate.lt90  = 1;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );
                    gpfree_write_union( fp, gpfpos, &shiftrblsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }                                                //  c   
            else if ( corner == 2 )  // LL                       //  a   b
                {
                if ( horizontal )
                    {
                    offset_x = offset_x + (a.x - poly->ll.x) * dbu_um;        // um 
                    offset_y = offset_y + (a.y - poly->ll.y) * dbu_um;        // um
                    offsetxrecord.prim24.val =  sstep2 + floor( offset_x / sub_res.x );
                    offsetyrecord.prim24.val =  sstep2 + floor( offset_y / sub_res.y );
                    base   = floor( (b.x - a.x) * gds2subit ) - sstep;
                    height = floor( (c.y - a.y) * gds2subit ) - sstep;
                    if ( height <= 0 ) height = 1;
                    if ( base   <= 0 ) base   = 1;
                    baselsw.prim24.val   = base;
                    heightlsw.prim24.val = height;
                    right_shift  = abs( floor( subits * base / height ) );
                    shiftrblsw.prim24.val = right_shift;
                    activaterecord.activate.OrientationY = 0;
                    activaterecord.activate.rbNeg = 1;
                    activaterecord.activate.lt90  = 1;

                    ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                    if ( ok )
                        {
                        gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                        gpfree_write_union( fp, gpfpos, &offsetyrecord );
                        gpfree_write_union( fp, gpfpos, &baselsw   );
                        gpfree_write_union( fp, gpfpos, &heightlsw );
                        gpfree_write_union( fp, gpfpos, &shiftrblsw );
                        gpfree_write_union( fp, gpfpos, &activaterecord );
                        }
                    }
                else              // c
                    {             // a  b
                    offset_x = offset_x + (a.x - poly->ll.x) * dbu_um;        // um 
                    offset_y = offset_y + (a.y - poly->ll.y) * dbu_um;        // um
                    offsetxrecord.prim24.val =  sstep2 + floor( offset_x / sub_res.x );
                    offsetyrecord.prim24.val =  sstep2 + floor( offset_y / sub_res.y );
                    base   = floor( (b.x - a.x) * gds2subit ) - sstep;
                    height = floor( (c.y - a.y) * gds2subit ) - sstep;
                    if ( height <= 0 ) height = 1;
                    if ( base   <= 0 ) base   = 1;
                    baselsw.prim24.val   = base;
                    heightlsw.prim24.val = height;
                    left_shift  = abs( floor( subits * height / base ) );
                    shiftltlsw.prim24.val = left_shift;
                    activaterecord.activate.OrientationY = 1;
                    activaterecord.activate.ltNeg = 0;
                    activaterecord.activate.rb90  = 1;

                    ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                    if ( ok )
                        {
                        gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                        gpfree_write_union( fp, gpfpos, &offsetyrecord );
                        gpfree_write_union( fp, gpfpos, &baselsw   );
                        gpfree_write_union( fp, gpfpos, &heightlsw );
                        gpfree_write_union( fp, gpfpos, &shiftltlsw );
                        gpfree_write_union( fp, gpfpos, &activaterecord );
                        }
                    }
                }
            else                     // LR                         //      c
                {                                                  //  b   a 
                offset_x = offset_x + (b.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (b.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val =  sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val =  sstep2 + floor( offset_y / sub_res.y );
                base   = floor( (a.x - b.x) * gds2subit ) - sstep;
                height = floor( (c.y - a.y) * gds2subit ) - sstep;
                if ( height <= 0 ) height = 1;
                if ( base   <= 0 ) base   = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;
                left_shift  = abs( floor( subits * base / height ) );
                shiftltlsw.prim24.val = left_shift;
                activaterecord.activate.OrientationY = 0;
                activaterecord.activate.ltNeg = 0;
                activaterecord.activate.rb90  = 1;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );
                    gpfree_write_union( fp, gpfpos, &shiftltlsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }
            }
        else if ( x_triangle( poly ) )                          // X triangle, but not a right-triangle
            {
            if ( pointing_up( poly ) )                 
                {                                                                       //      c
                put_c_at_top( poly, &a, &b, &c );                                       //   a     b
                offset_x = offset_x + (a.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (a.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val = sstep2 + floor( offset_y / sub_res.y );
                
                base   = floor( (b.x - a.x) * gds2subit ) - sstep;
                height = floor( (c.y - a.y) * gds2subit ) - sstep;
                if ( base   <= 0 ) base   = 1;
                if ( height <= 0 ) height = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;
                
                rise = (c.y - a.y);
                run  = (c.x - a.x);
                left_slope_neg = 0;
                if ( c.x < a.x ) left_slope_neg = 1;
                left_shift  = abs( floor( subits * run / rise ) );    // subits = 16384
                run  = (b.x - c.x);
                right_slope_neg = 0;
                if ( c.x < b.x ) right_slope_neg = 1;
                right_shift = abs( floor( subits * run / rise ) );
                shiftltlsw.prim24.val = left_shift;
                shiftrblsw.prim24.val = right_shift;

                activaterecord.activate.OrientationY = 0;
                activaterecord.activate.ltNeg = left_slope_neg;
                activaterecord.activate.rbNeg = right_slope_neg;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );
                    gpfree_write_union( fp, gpfpos, &shiftltlsw );
                    gpfree_write_union( fp, gpfpos, &shiftrblsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }
            else                                       //   a     b
                {                                      //      c
                put_c_at_bottom( poly, &a, &b, &c );
                offset_x = offset_x + (c.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (c.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val = sstep2 + floor( offset_y / sub_res.y );
                
                base   = 0;
                height = floor( (a.y - c.y) * gds2subit ) - sstep;
                if ( height <= 0 ) height = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;

                // we can't shrink the base, so we have to
                // shrink the other end- the 'run'
                
                rise = a.y - c.y;
                run  = c.x - a.x - sstep2 ;
                left_slope_neg = 0;
                if ( c.x > a.x ) left_slope_neg = 1;
                left_shift  = abs( floor( subits * run / rise ) );
                run  = b.x - c.x - sstep2;
                right_slope_neg = 0;
                if ( c.x > b.x ) right_slope_neg = 1;
                right_shift = abs( floor( subits * run / rise ) );
                shiftltlsw.prim24.val = left_shift;
                shiftrblsw.prim24.val = right_shift;

                activaterecord.activate.OrientationY = 0;
                activaterecord.activate.ltNeg = left_slope_neg;
                activaterecord.activate.rbNeg = right_slope_neg;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );
                    gpfree_write_union( fp, gpfpos, &shiftltlsw );
                    gpfree_write_union( fp, gpfpos, &shiftrblsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }
            }
        else                                                              // Y triangle, but not a right-triangle
            {                                          //   b
            if ( pointing_right( poly ) )              //      c
                {                                      //   a
                put_c_on_right( poly, &a, &b, &c ); 
                offset_x = offset_x + (a.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (a.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val = sstep2 + floor( offset_y / sub_res.y );
                
                base   = floor( (c.x - a.x) * gds2subit ) - sstep;
                height = floor( (b.y - a.y) * gds2subit ) - sstep;
                if ( base   <= 0 ) base   = 1;
                if ( height <= 0 ) height = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;

                rise = abs(c.x - b.x) * gds2subit - sstep;
                run  = abs(b.y - c.y) * gds2subit - sstep2;
                left_slope_neg = 0;
                if ( c.y > b.y ) left_slope_neg = 1;
                left_shift  = abs( floor( subits * run / rise ) );    // subits = 16384
                run  = abs(c.y - a.y) * gds2subit - sstep2;
                right_slope_neg = 1;
                if ( c.y < a.y ) right_slope_neg = 0;
                right_shift = abs( floor( subits * run / rise ) );
                shiftltlsw.prim24.val = left_shift;
                shiftrblsw.prim24.val = right_shift;

                activaterecord.activate.OrientationY = 1;
                activaterecord.activate.ltNeg = left_slope_neg;
                activaterecord.activate.rbNeg = right_slope_neg;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );
                    gpfree_write_union( fp, gpfpos, &shiftltlsw );
                    gpfree_write_union( fp, gpfpos, &shiftrblsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }
            else                                       //      b               
                {                                      //   c
                put_c_on_left( poly, &a, &b, &c );     //      a
                offset_x = offset_x + (c.x - poly->ll.x) * dbu_um;        // um 
                offset_y = offset_y + (c.y - poly->ll.y) * dbu_um;        // um
                offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );
                offsetyrecord.prim24.val =          floor( offset_y / sub_res.y );
                
                base   = floor( abs(b.x - c.x) * gds2subit ) - sstep;
                height = 0;
                if ( base <= 0 ) base = 1;
                baselsw.prim24.val   = base;
                heightlsw.prim24.val = height;

                rise = abs(b.x - c.x) * gds2subit - sstep;
                run  = abs(b.y - c.y) * gds2subit - sstep2;
                left_slope_neg = 1;
                if ( c.y > b.y ) left_slope_neg = 0;
                left_shift  = abs( floor( subits * run / rise ) );    // subits = 16384
                run  = abs(c.y - a.y) * gds2subit - sstep2;
                right_slope_neg = 0;
                if ( c.y < a.y ) right_slope_neg = 1;
                right_shift = abs( floor( subits * run / rise ) );
                shiftltlsw.prim24.val = left_shift;
                shiftrblsw.prim24.val = right_shift;

                activaterecord.activate.OrientationY = 1;
                activaterecord.activate.ltNeg = left_slope_neg;
                activaterecord.activate.rbNeg = right_slope_neg;

                ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

                if ( ok )
                    {
                    gpfree_write_union( fp, gpfpos, &offsetxrecord );   
                    gpfree_write_union( fp, gpfpos, &offsetyrecord );
                    gpfree_write_union( fp, gpfpos, &baselsw   );
                    gpfree_write_union( fp, gpfpos, &heightlsw );
                    gpfree_write_union( fp, gpfpos, &shiftltlsw );
                    gpfree_write_union( fp, gpfpos, &shiftrblsw );
                    gpfree_write_union( fp, gpfpos, &activaterecord );
                    }
                }
            }
        }
    else if ( x_trap( poly ) )
        {
        reorder_x_trap( poly, &a, &b, &c, &d );     //   d  c
                                                    //  a    b
        
        offset_x = offset_x + (a.x - poly->ll.x) * dbu_um;        // um 
        offset_y = offset_y + (a.y - poly->ll.y) * dbu_um;        // um
        offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );
        offsetyrecord.prim24.val = sstep2 + floor( offset_y / sub_res.y );
                
        base   = floor( abs(b.x - a.x) * gds2subit ) - sstep;
        height = floor( abs(d.y - a.y) * gds2subit ) - sstep;
        if ( base   <= 0 ) base   = 1;
        if ( height <= 0 ) height = 1;
        baselsw.prim24.val   = base;
        heightlsw.prim24.val = height;

        if ( d.x == a.x )
            {
            activaterecord.activate.lt90 = 1;
            }
        else
            {
            rise = abs(d.y - a.y) * gds2subit - sstep;
            run  = abs(d.x - a.x) * gds2subit - sstep2;
            left_slope_neg = 1;
            if ( d.x > a.x ) left_slope_neg = 0;
            activaterecord.activate.ltNeg = left_slope_neg;
            left_shift  = abs( floor( subits * run / rise ) );    // subits = 16384
            shiftltlsw.prim24.val = left_shift;
            }

        if ( c.x == b.x )
            {
            activaterecord.activate.rb90 = 1;
            }
        else
            {
            rise = abs(d.y - a.y) * gds2subit - sstep;
            run  = abs(c.x - b.x) * gds2subit - sstep2;
            right_slope_neg = 1;
            if ( c.x > b.x ) right_slope_neg = 0;
            activaterecord.activate.rbNeg = right_slope_neg;
            right_shift = abs( floor( subits * run / rise ) );
            shiftrblsw.prim24.val = right_shift;
            }

        activaterecord.activate.OrientationY = 0;

        ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

        if ( ok )
            {
            gpfree_write_union( fp, gpfpos, &offsetxrecord );   
            gpfree_write_union( fp, gpfpos, &offsetyrecord );
            gpfree_write_union( fp, gpfpos, &baselsw   );
            gpfree_write_union( fp, gpfpos, &heightlsw );

            if ( activaterecord.activate.rb90 != 1 )
                gpfree_write_union( fp, gpfpos, &shiftrblsw );

            if ( activaterecord.activate.lt90 != 1 )
                gpfree_write_union( fp, gpfpos, &shiftltlsw );
        
            gpfree_write_union( fp, gpfpos, &activaterecord );
            }

        /*
        gpf_print_union( stdout, &baselsw,        &count );
        gpf_print_union( stdout, &heightlsw,      &count );
        gpf_print_union( stdout, &shiftltlsw,     &count );
        gpf_print_union( stdout, &shiftrblsw,     &count );
        gpf_print_union( stdout, &activaterecord, &count );
        */
        
        }
    else   // y trap
        {                                           //       c
        reorder_y_trap( poly, &a, &b, &c, &d );     //  d    
                                                    //  a    b
        
        offset_x = offset_x + (a.x - poly->ll.x) * dbu_um;        // um 
        offset_y = offset_y + (a.y - poly->ll.y) * dbu_um;        // um
        offsetxrecord.prim24.val = sstep2 + floor( offset_x / sub_res.x );
        offsetyrecord.prim24.val = sstep2 + floor( offset_y / sub_res.y );
                
        base   = floor( abs(b.x - a.x) * gds2subit ) - sstep;
        height = floor( abs(d.y - a.y) * gds2subit ) - sstep;
        if ( base   <= 0 ) base   = 1;
        if ( height <= 0 ) height = 1;
        baselsw.prim24.val   = base;
        heightlsw.prim24.val = height;

        if ( c.y == d.y )
            {
            activaterecord.activate.lt90 = 1;
            }
        else
            {
            rise = abs(c.x - d.x) * gds2subit - sstep;
            run  = abs(c.y - d.y) * gds2subit - sstep2;
            left_slope_neg = 1;
            if ( d.y > c.y ) left_slope_neg = 0;
            activaterecord.activate.ltNeg = left_slope_neg;
            left_shift  = abs( floor( subits * run / rise ) );    // subits = 16384
            shiftltlsw.prim24.val = left_shift;
            }

        if ( a.y == b.y )
            {
            activaterecord.activate.rb90 = 1;
            }
        else
            {
            rise = abs(b.x - a.x) * gds2subit - sstep;
            run  = abs(b.y - a.y) * gds2subit - sstep2;
            right_slope_neg = 1;
            if ( a.y > b.y ) right_slope_neg = 0;
            activaterecord.activate.rbNeg = right_slope_neg;
            right_shift = abs( floor( subits * run / rise ) );
            shiftrblsw.prim24.val = right_shift;
            }

        activaterecord.activate.OrientationY = 1;

        ok = check_trap( &offsetxrecord, &offsetyrecord, &baselsw, &heightlsw, &shiftltlsw, &shiftrblsw, &activaterecord, sstep );

        if ( ok )
            {
            gpfree_write_union( fp, gpfpos, &offsetxrecord );   
            gpfree_write_union( fp, gpfpos, &offsetyrecord );
            gpfree_write_union( fp, gpfpos, &baselsw   );
            gpfree_write_union( fp, gpfpos, &heightlsw );

            if ( activaterecord.activate.rb90 != 1 )
                gpfree_write_union( fp, gpfpos, &shiftrblsw );

            if ( activaterecord.activate.lt90 != 1 )
                gpfree_write_union( fp, gpfpos, &shiftltlsw );

            gpfree_write_union( fp, gpfpos, &activaterecord );
            }

        //gpf_print_union( stdout, &baselsw,        &count );
        //gpf_print_union( stdout, &heightlsw,      &count );
        //gpf_print_union( stdout, &shiftltlsw,     &count );
        //gpf_print_union( stdout, &shiftrblsw,     &count );
        //gpf_print_union( stdout, &activaterecord, &count );

        }
    
    nout = nout + nx * ny;
    
    }



//----------------------------------------------------------------------------------------------------------------
// Don't look here for an example of gpf output. write_polygon() uses the old style array of primitives.
// It's an ugly mess. Instead look at the function write_trap() above for a much better example.


void
write_polygon(  FILE *fp,
                BOOL         traps,            // generate parallelograms instead of polygons
                BOOL         start_subfield,
                dxy          block,            // um   block size
                gpf_union_t *prims, 
                gpf_union_t *prims_org, 
                polygon     *poly,             // array of xy, mfr units (usually 1 dbu = 1nm)
                dxy          fieldcenter,      // um   gds units
                dxy          subfieldsize,     // um
		dxy          main_res,         // um   usually 1.0 nm
		dxy          sub_res,          // um   usually 0.5 nm
		double       bss_um,           // um
                double       dbu_um,           // um   database unit, usually 1 nm, sometimes 0.1 nm
                BOOL         assign_doses,     
                double     **dose,             // dose matrix, doubles
                int        **msfm,             // msf multipliers, int
		int          nx,               //      number of columns
                int          ny,               //      number of rows
                double       dx,               // um   pitch in x
                double       dy,               // um   pitch in y
                gpf_ulong_t *gpfpos       )
    {

    // The subfield is calculated for each shape. If this main deflection is not the same as
    // the previous, then the shape will get a new mainfield deflection.

    // What happens if the shape extends into the adjacent field? It still works!
    // It definitely will not work if the shape goes beyond the maximum deflection.
    // This program will bail if the shape exceeds the subfield size, so the shape
    // cannot extend more than one subfield beyond the block. 

    static int         i, ix, iy, step, mstep, minarea, nwarnings=0, submsf;
    static gpf_uint_t  this_msf;
    static double      offset_x, offset_y;
    static dxy         ll, ur, maxsubfield, dv;
    static double      xum, yum, x0, y0;
    static gpf_union_t gpoly;
    static xy          v;
    static dxy         deflection, llblock, far_ur, block2;
    static polygon    *p;
    static BOOL        init=FALSE;
    static twoints     clock;
    static char        mess[256];

    if ( ! init )
	{
        submsf = floor( bss_um / sub_res.x );
	init = TRUE;
        p = initialize_polygon( MAXVERTEX, 0, 0, 0 );
        maxsubfield.x = subfieldsize.x * 1.02;
        maxsubfield.y = subfieldsize.y * 1.02;
        if ( maxsubfield.x > MAXSUBFIELD ) maxsubfield.x = MAXSUBFIELD;
	if ( maxsubfield.y > MAXSUBFIELD ) maxsubfield.y = MAXSUBFIELD;
        step = floor( bss_um / sub_res.x );
        mstep = floor( bss_um / main_res.x );
        minarea = mstep * mstep * 10;
	block2.x = alittlebigger( block.x / 2.0 );
        block2.y = alittlebigger( block.y / 2.0 );
	}

    if ( ! poly->active ) return;

    if ( poly->len < 3 ) { poly->active = FALSE; return; } 

    if ( traps && poly->len > 4 ) bail( "gpfout: trapezoid has more than 4 vertices" );
    
    if ( (poly->ll.x == poly->ur.x) || (poly->ll.y == poly->ur.y) )  { poly->active = FALSE; return; } 

    if ( nx < 1 || ny < 1 ) bail( "number of instances must be > 0.  nx and ny must both be at least 1." );

    if ( poly->len > MAXVERTEX ) bail( "polygon has too many vertices" );

    if ( traps && poly->len > 4 ) bail( "Polygon has too many sides for a trapezoid. Fracturing and slicing produced too many sides." ); 
    
    //if ( polygon_area( poly, mstep ) < minarea ) return; way too late for this!


    if ( traps )
        {
        write_trap( fp, start_subfield, block, prims, prims_org, poly, fieldcenter, subfieldsize, main_res,
                    sub_res, bss_um, dbu_um, assign_doses, dose, msfm, nx, ny, dx, dy, gpfpos );
        return;
        }
    

    ll.x = poly->ll.x * dbu_um - fieldcenter.x;                      // LL corner of polygon, um, origin at field center
    ll.y = poly->ll.y * dbu_um - fieldcenter.y;

    llblock.x = fieldcenter.x - block.x / 2.0;                       // LL corner of this block, gds um
    llblock.y = fieldcenter.y - block.y / 2.0;

                                                                     // x0, y0 is the main field deflection to the
                                                                     // lower-left corner of the subfield (um).
                                                                     // Subfields are tiled as in slice.c, starting at the 
                                                                     // lower-left. Block size is a multiple of the subfield size.

    x0 = llblock.x - fieldcenter.x + subfieldsize.x * (floor(((poly->ll.x * dbu_um) - llblock.x) / subfieldsize.x));
    y0 = llblock.y - fieldcenter.y + subfieldsize.y * (floor(((poly->ll.y * dbu_um) - llblock.y) / subfieldsize.y));

    offset_x = ll.x - x0;                                             // um, to be converted to subfield db units
    offset_y = ll.y - y0;                                             // um
    
    // check polygon extents
 
    ur.x = (poly->ur.x * dbu_um - fieldcenter.x);                     // um  UR corner of polygon, origin at field center
    ur.y = (poly->ur.y * dbu_um - fieldcenter.y);                     // um

    deflection.x = ur.x - x0;
    deflection.y = ur.y - y0;

    if ( deflection.x < 0.0 || deflection.y < 0.0 )
	{
	printf( "\n\nLL= %1.3lf, %1.3lf\n", ll.x, ll.y );
	printf(     "UR= %1.3lf, %1.3lf\n", ur.x, ur.y );
        printf(     "field center= %1.3lf, %1.3lf\n", fieldcenter.x, fieldcenter.y );
        printf(     "dbu_um = %1.6lf\n", dbu_um );
	bail( "(gpfout) arithmetic confusion in write_polygon" );
	}

    // we can let the deflection exceed the subfield by a tiny bit, 
    // but the deflection has to be less than 4.2 um

    if ( (deflection.x > maxsubfield.x) || (deflection.y > maxsubfield.y) ||
         (offset_x     < 0.0)           || (offset_y     < 0.0)            )
	{
        x0 = ll.x;       // set mainfield deflection to LL corner of this polygon
	y0 = ll.y;
        deflection.x = ur.x - x0;
        deflection.y = ur.y - y0;
        offset_x = ll.x - x0;
        offset_y = ll.y - y0;

        if ( offset_x < 0 || offset_y < 0 || deflection.x < 0 || deflection.y < 0 )
            {
	    printf( "\n (gpfout) negative deflection error: offset= %1.2lf, %1.2lf  defl=  %1.2lf, %1.2lf um     ", 
			offset_x, offset_y, deflection.x, deflection.y );
	    printf( " x0,y0= %1.4lf, %1.4lf um  ", x0, y0 );
            }
	                                           // keep score of extra mfdefs because extra deflections take extra time
        if ( (nx == 1) && (ny == 1) ) mfdefs++;    // but the mfdef for an array doesn't count 
	}

    if ( (deflection.x > maxsubfield.x) || (deflection.y > maxsubfield.y) )
	{
	printf( "\n\nsubfield origin = %1.3lf, %1.3lf um\n", x0, y0 );
        printf(     "shape ll        = %ld, %ld dbu\n", poly->ll.x, poly->ll.y );
        printf(     "shape ur        = %ld, %ld dbu\n", poly->ur.x, poly->ur.y );
        printf(     "deflection      = %1.3lf, %1.3lf um\n", deflection.x, deflection.y );
        printf(     "nx, ny          = %d, %d\n", nx, ny );
        printf(     "field center    = %1.6lf, %1.6lf um\n\n", fieldcenter.x, fieldcenter.y );

        if ( not_a_large_rectangle( poly, subfieldsize, dbu_um, TRUE ) )
	    printf( "This is not a large rectangle.\n\n" );
	else
	    printf( "This is a large rectangle, so why is it here?\n\n" );

        print_polygon( poly, 1 );
        printf( "polygon area: %lf\n", polygon_area( poly, mstep ) );

        bail( "polygon size exceeds subfield size" );
	}

    // check array extents

    // ur and ll are the corners of the shape, in um, with origin at field center
 
    if ( (nx > 1) || (ny > 1) )
	{
        far_ur.x = ur.x + (nx - 1) * dx;
        far_ur.y = ur.y + (ny - 1) * dy;
        if ( (far_ur.x > block2.x) || (far_ur.y > block2.y) )
	    {
	    v = vertex( poly, 0 );
            dv.x = v.x * dbu_um;
	    dv.y = v.y * dbu_um;
	    sprintf( mess, "array exceeds block size at %lf, %lf um  d= %lf, %lf um  n= %d, %d  urext= %lf, %lf um  block2= %lf, %lf", 
		     dv.x, dv.y, dx, dy, nx, ny, far_ur.x, far_ur.y, block2.x, block2.y );
	    warning( mess, &nwarnings ); 
            poly->layer = ERRORLAYER;
	    }
	}

    // check those extents extra hard

    // poly is in gds database units, typically 1 nm.
    // x * dbu_um / sub_res.x  converts poly to subfield bits
    // and fixes any weird gds database units (like angstroms)
    
    zero_polygon( p );

    for ( i=0; i < poly->len; i++ )
	{
        v = vertex( poly, i );
        xum = v.x * dbu_um - fieldcenter.x;                                    // dbu * pss
        yum = v.y * dbu_um - fieldcenter.y;                                    // we set pss = main field lsb, 1 nm
        xum = xum - x0 - offset_x;
        yum = yum - y0 - offset_y;
	ix = (int) round( xum / sub_res.x );                                   // poly coordinates are in subfield units
        iy = (int) round( yum / sub_res.y );
        append_polygon( p, ix, iy );                                           // polygon coord in gpf subfield unit
	}

    // Old versions of cview show weird misplaced single pixels,
    // but that was just a bug.
    
    //if ( (p->ur.x - p->ll.x) < step ) return;      // polygons smaller than a beam step used to be a problem
    //if ( (p->ur.y - p->ll.y) < step ) return;      // now we want to write single-pixel polygons
              
    
    if ( (p->ll.x >= p->ur.x) || (p->ll.y >= p->ur.y) ) return;

    // Don't look here for an example of writing gpf. This use of
    // an array of "prims" is awful and confusing. Blech.
    // Look above in the function write_trap() for a better example.
    
    if ( assign_doses )
	{
        clock = dose2clock( fabs( dose[ poly->layer ][ poly->dt ] ) );               // fabs because we mark a dose 'hit' with the sign
        dose[ poly->layer ][ poly->dt ] = -fabs( dose[ poly->layer ][ poly->dt ] );  // just to keep track of the number of doses

        //printf( "L %2d DT %2d dose %6.3lf\n", poly->layer, poly->dt, fabs( dose[poly->layer][poly->dt] ) );
        
        prims[GPF_PRIM_FREQUENCY_LSW].prim24.val = clock.lsw;
        prims[GPF_PRIM_FREQUENCY_MSW].prim24.val = clock.msw;

        this_msf = submsf * msfm[ poly->layer ][ poly->dt ];
        prims[GPF_PRIM_MSF].prim16.val2 = this_msf;
        
        if ( start_subfield || (prims_org[GPF_PRIM_FREQUENCY_MSW].prim24.val != clock.msw) ||
                               (prims_org[GPF_PRIM_FREQUENCY_LSW].prim24.val != clock.lsw) ||
                               (prims_org[GPF_PRIM_MSF].prim16.val2          != this_msf ) )
	    {
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_FREQUENCY_LSW] ); 
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_FREQUENCY_MSW] );
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MSF] );              // each shape can have its own msf! neat!
            prims_org[GPF_PRIM_FREQUENCY_LSW].prim24.val = clock.lsw;
            prims_org[GPF_PRIM_FREQUENCY_MSW].prim24.val = clock.msw;
            prims_org[GPF_PRIM_MSF].prim16.val2          = this_msf;
	    }
	}

    
    // mainfield deflection... should not be set if it is the same as the previous

    prims[GPF_PRIM_MAIN_X_LSW].prim24.val = MAINDACMID + isnap( x0 / main_res.x, mstep );   // we do not really need to snap
    prims[GPF_PRIM_MAIN_Y_LSW].prim24.val = MAINDACMID + isnap( y0 / main_res.y, mstep );   // to the beam step. that's just a
                                                                                            // remnant of the old days
    // outline mode already shifts the pixels by 0.5 bss.                                   // so isnap now does nothing
    // we do not have to apply any manual butting corrections.

    prims[GPF_PRIM_OFFSET_X_LSW].prim24.val = isnap( offset_x / sub_res.x, step );          // we do not really need to snap
    prims[GPF_PRIM_OFFSET_Y_LSW].prim24.val = isnap( offset_y / sub_res.y, step );          // the offset = trap deflection 
                                                                                            // to the subfield beam step
    /*
    if ( start_subfield || (prims[GPF_PRIM_MAIN_X_LSW].prim24.val != prims_org[GPF_PRIM_MAIN_X_LSW].prim24.val) )
        gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAIN_X_LSW] );                 // write main deflection x

    if ( start_subfield || (prims[GPF_PRIM_MAIN_Y_LSW].prim24.val != prims_org[GPF_PRIM_MAIN_Y_LSW].prim24.val) )
        gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAIN_Y_LSW] );                 // write main deflection y
    */

    
    if ( start_subfield || (prims[GPF_PRIM_MAIN_X_LSW].prim24.val != prims_org[GPF_PRIM_MAIN_X_LSW].prim24.val) 
    	                || (prims[GPF_PRIM_MAIN_Y_LSW].prim24.val != prims_org[GPF_PRIM_MAIN_Y_LSW].prim24.val) )
        {
        gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAIN_X_LSW] );                 // write main deflection x
        gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAIN_Y_LSW] );                 // write main deflection y
        }
    

    prims_org[GPF_PRIM_MAIN_X_LSW].prim24.val = prims[GPF_PRIM_MAIN_X_LSW].prim24.val;
    prims_org[GPF_PRIM_MAIN_Y_LSW].prim24.val = prims[GPF_PRIM_MAIN_Y_LSW].prim24.val;

    gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_OFFSET_X_LSW] );               // write offset x
    gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_OFFSET_Y_LSW] );               // write offset y

    prims[GPF_PRIM_ACTIVATE].repeat.MainRepVect1  = 0;
    prims[GPF_PRIM_ACTIVATE].repeat.MainRepVect2  = 0;

    if ( nx > 1 || ny > 1 )
	{
        prims[GPF_PRIM_MAINREPVECT1_X_LSW].prim24.IC  = GPF_PRIM_MAINREPVECT1_X_LSW;
        prims[GPF_PRIM_MAINREPVECT1_Y_LSW].prim24.IC  = GPF_PRIM_MAINREPVECT1_Y_LSW;
        prims[GPF_PRIM_MAINREPVECT2_X_LSW].prim24.IC  = GPF_PRIM_MAINREPVECT2_X_LSW;
        prims[GPF_PRIM_MAINREPVECT2_Y_LSW].prim24.IC  = GPF_PRIM_MAINREPVECT2_Y_LSW;
        prims[GPF_PRIM_MAINREPVECT1_CNT].prim24.IC    = GPF_PRIM_MAINREPVECT1_CNT; 
        prims[GPF_PRIM_MAINREPVECT2_CNT].prim24.IC    = GPF_PRIM_MAINREPVECT2_CNT; 

        prims[GPF_PRIM_MAINREPVECT1_X_LSW].prim24.val = lround( dx / main_res.x );      // floor does not help close the gaps
        prims[GPF_PRIM_MAINREPVECT1_Y_LSW].prim24.val = 0;
        prims[GPF_PRIM_MAINREPVECT2_X_LSW].prim24.val = 0;
        prims[GPF_PRIM_MAINREPVECT2_Y_LSW].prim24.val = lround( dy / main_res.y );      // floor does not help

        // Here's a subtle point: the SIZE of a polygon is in subfield resolution units,
        // but the PITCH of an array is in mainfield resolution units. If the pitch is
        // an odd number, then the size cannot match. But it's really not bad enough
        // to worry about it. The ends of arrays might have little sub-pixel errors.

        /*
        xy v0, v1;
	int idx, idy;
        v0 = vertex( p, 0 );
	v1 = vertex( p, 1 );
	idx = abs( v0.x - v1.x );
	idy = abs( v0.y - v1.y );
        printf( "\npitch: %5d  %5d", 2 * lround( dx / main_res.x ), 2 * lround( dy / main_res.y ));
        printf( "\nsize:  %5d  %5d\n", idx, idy ); 
	*/

        if ( nx > 1 )
	    {
            prims[GPF_PRIM_MAINREPVECT1_CNT].prim24.val   = nx - 1;
            prims[GPF_PRIM_ACTIVATE].repeat.MainRepVect1  = 1;
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAINREPVECT1_X_LSW] );
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAINREPVECT1_Y_LSW] );
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAINREPVECT1_CNT] );
            }

        if ( ny > 1 )
	    {
            prims[GPF_PRIM_MAINREPVECT2_CNT].prim24.val   = ny - 1;
            prims[GPF_PRIM_ACTIVATE].repeat.MainRepVect2  = 1;
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAINREPVECT2_X_LSW] );
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAINREPVECT2_Y_LSW] );
            gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_MAINREPVECT2_CNT] );
            }

        // We do not have to update prims_org since the activate line indicates the use of repeats.
	// That is, if MRV1 or MRV2 appear after ACTIVATE, then we have repeats. There is no
        // implication from the previous ACTIVATE.

        }

    //prims[GPF_PRIM_POLYGON_CENTER_X_LSW].prim24.val = 0;
    //prims[GPF_PRIM_POLYGON_CENTER_Y_LSW].prim24.val = 0;
    //prims[GPF_PRIM_POLYGON_ANGLE_MSW].prim24.val    = 0;
    //prims[GPF_PRIM_POLYGON_ANGLE_LSW].prim24.val    = 0;

    prims[GPF_PRIM_POLYGON].prim24.IC   = GPF_PRIM_POLYGON;
    prims[GPF_PRIM_POLYGON].prim16.val2 = poly->len;
    prims[GPF_PRIM_POLYGON].prim16.val1 = POLYGON_CLEAR;

    gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_POLYGON] );                    // polygon vector n clear
    prims_org[GPF_PRIM_POLYGON].data = prims[GPF_PRIM_POLYGON].data; 

    for ( i=0; i < p->len; i++ )
	{
        v = vertex( p, i ); 
	gpoly.coord.x = v.x;
        gpoly.coord.y = v.y;
        gpfree_write_union( fp, gpfpos, &gpoly );                                 // polygon coord x y
	}

    prims[GPF_PRIM_ACTIVATE].prim24.IC            = GPF_PRIM_ACTIVATE;
    prims[GPF_PRIM_ACTIVATE].activate.ShapeID     = GPF_SHAPE_POLYGON;
    prims[GPF_PRIM_ACTIVATE].polygon.outline      = 1;
    prims[GPF_PRIM_ACTIVATE].polygon.redistribute = 0;
    //prims[GPF_PRIM_ACTIVATE].polygon.exels = 0;  depricated, never used

    gpfree_write_union( fp, gpfpos, &prims[GPF_PRIM_ACTIVATE] );                   // activate polygon outline

    nout = nout + nx * ny;

    }     // write_polygon
    

//--------------------------------------------------------------------------------------------------

#define EQUAL( a, b ) ((a) == (b))

BOOL
not_a_large_rectangle( polygon *p, dxy sfs, double dbu_um, BOOL verbose )
    {
    // p    gds database units
    // sfs  subfield size in um
    // mfr  mainfield resolution in um, usually 0.001 um

    static xy zero, one, two, three;
    static double dx, dy;
    static BOOL init=FALSE;
    static dxy maxsfs;

    if ( ! init )
	{
	maxsfs.x = sfs.x; // * 1.01;
	maxsfs.y = sfs.y; // * 1.01;
	init = TRUE;
	}

    if ( p->len != 4 ) return( TRUE );  // it's not four-sided

    dx = (p->ur.x - p->ll.x) * dbu_um;
    dy = (p->ur.y - p->ll.y) * dbu_um;

    if ( (dx <= maxsfs.x) && ( dy <= maxsfs.y ) ) 
	{
	//printf( "Not a large rectangle because it's smaller than a subfield.\n" );
	return( TRUE ); 
	}

    zero  = vertex( p, 0 );
    one   = vertex( p, 1 );
    two   = vertex( p, 2 );
    three = vertex( p, 3 );

    // let's make sure it is a rectangle...

    if ( ! (EQUAL(zero.y, three.y) || EQUAL(zero.x, three.x))) return( TRUE ); 
    if ( ! (EQUAL(zero.y,   one.y) || EQUAL(zero.x,   one.x))) return( TRUE ); 
    if ( ! (EQUAL(one.y,    two.y) || EQUAL(one.x,    two.x))) return( TRUE ); 
    if ( ! (EQUAL(two.y,  three.y) || EQUAL(two.x,  three.x))) return( TRUE ); 
    return( FALSE );
    }

//-------------------------------------------------------------------------------------------------------------------

BOOL
this_is_a_rectangle( polygon *p, dxy sfs, double dbu_um )
    {
    // p    gds database units
    // sfs  subfield size in um

    static xy zero, one, two, three;
    static double dx, dy;
    static BOOL init=FALSE;
    static dxy maxsfs;

    if ( ! init )
	{
	maxsfs.x = sfs.x * 1.01;
	maxsfs.y = sfs.y * 1.01;
	init = TRUE;
	}

    if ( p->len != 4 ) return( FALSE );  // it's not four-sided

    dx = (p->ur.x - p->ll.x) * dbu_um;
    dy = (p->ur.y - p->ll.y) * dbu_um;

    if ( (dx > maxsfs.x) || ( dy > maxsfs.y ) )  bail( "Rectangle is larger than a subfield." );

    zero  = vertex( p, 0 );
    one   = vertex( p, 1 );
    two   = vertex( p, 2 );
    three = vertex( p, 3 );

    if ( ! (EQUAL(zero.y, three.y) || EQUAL(zero.x, three.x))) return( FALSE );
    if ( ! (EQUAL(zero.y,   one.y) || EQUAL(zero.x,   one.x))) return( FALSE );
    if ( ! (EQUAL(one.y,    two.y) || EQUAL(one.x,    two.x))) return( FALSE );
    if ( ! (EQUAL(two.y,  three.y) || EQUAL(two.x,  three.x))) return( FALSE );

    return( TRUE );
    }

//-------------------------------------------------------------------------------------------------------------------


gpf_ulong_t 
pixel_count( polygon *p, double dbu_um, double bss_um, double **dose, int **msfm )
    {
    static double area, bss;
    static gpf_ulong_t a;
    if ( ! p->active ) return( 0 );
    bss = bss_um * msfm[p->layer][p->dt];
    polygon_ccw( p, &area, bss );
    area = area * dbu_um * dbu_um / (bss * bss);
    area = area * fabs( dose[ p->layer ][ p->dt ] );
    a = (gpf_ulong_t) round( area );
    return( a ); 
    }

//-------------------------------------------------------------------------------------------------------------------

double
frac( double x )
    {
    static double f;
    f = fabs(x - floor( x ));
    //printf( "\nx = %g  floorx = %g  f = %g\n", x, floor(x), f );
    if ( fabs( f - 1.0 ) < 1.0e-7 ) f = 0.0;   // 1e-6 is sometimes too large!!
    return f;
    }

//--------------------------------------------------------------------------------------------------

BOOL
samepoint( xy a, xy b ) 
    {
    return( abs(a.x - b.x) < 2 && abs(a.y - b.y) < 2 );
    }

//---------------------------------------------------------------------------------------------------


void
remove_duplicate_vertices( polygon *p )
    {
    static int n, nm1, i;
    static xy   vi, vi1, v0, vnm1;
    
    n   = p->len;
    nm1 = n - 1;
    i   = 0;

    while ( i < nm1 )
        {
        vi  = vertex( p, i );
        vi1 = vertex( p, i+1 );
        while ( (i < nm1) && samepoint( vi, vi1 ) )
            {
            delete_vertex( p, i+1 );
            n = p->len;
            nm1 = n - 1;
            vi  = vertex( p, i );
            if ( i == nm1 )
                vi1 = vertex( p, 0 );
            else 
                vi1 = vertex( p, i+1 );
            }
        i += 1;
        }
    
    n = p->len;
    nm1 = n - 1;

    if ( n > 2 ) 
        {
        v0   = vertex( p, 0 );
        vnm1 = vertex( p, nm1 );
        if ( samepoint( v0, vnm1 ) ) delete_vertex( p, nm1 ); 
        }

    if ( n < 3 ) p->active = FALSE;

    }

//---------------------------------------------------------------------------------------------------------------------

void
copy_from_gds_polygon( struct gds_itemtype *item, polygon *poly )
    {
    // copy item to poly
    // poly has already been initialized

    static int i;
    static int x, y;

    // the append function checks for overflow

    for ( i=0; i < item->n; i++ )
        {
        x = item->x[i];
        y = item->y[i];
        append_polygon( poly, x, y );
        }

    remove_duplicate_vertices( poly );

    }  // copy_from_gds_polygon


//--------------------------------------------------------------------------------------------------

void
read_input_polys( char *iname, queue *shape, double *extent, double *dbu, int *nump, double bss_um )
    {
 
    int                  infile, layer, dt, i;
    BOOL                 not_eof, got_one, first;
    unsigned short       token;
    double               dbu_um, dbu_uu, dbu_m, step; 
    double               dminx, dmaxx, dminy, dmaxy;
    int                  minx, maxx, miny, maxy;
    int                  numpolys, len;
    struct gds_itemtype *item;
    polygon             *poly;
    short                count;
    xy                   v;
    
    first = TRUE;

    infile = open( iname, 0 );
    if ( infile <= 0 ) bail( "read_input_polys: unable to open input file." );

    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! item ) bail( "read_input_polys: unable to allocate gds item " );

    numpolys = -1;

    not_eof = read( infile, &count, 2 ) > 0;
     
    while ( not_eof )
        {
        got_one = FALSE;
        
        while ( not_eof && ! got_one )
            {
            gds_swap2bytes( (BYTE *) &count );
            if ( count >= 4 )
                {
                read( infile, &token, 2 );
                gds_swap2bytes( (BYTE *) &token );
            
                switch( token )
                    {
                    case 0x0305:  //printf( "UNITS \n" );  fflush(stdout);
                              gds_read_units( infile, &dbu_um, &dbu_uu, &dbu_m, 0 );
                              dbu_uu = trim( dbu_uu );                   // round off to sensible values
                              dbu_m  = trim( dbu_m );
                              dbu_um = dbu_m * 1.0e6;
                              dbu_um = trim( dbu_um );
                              step = 0.5 * bss_um / dbu_um;
                              break;
                    case 0x0D02:  //printf( "LAYER \n" );  fflush(stdout);
                              gds_read_layer( infile, count, &item, 0 );
                              layer = item->layer;
                              break;
                    case 0x0E02:  //printf( "DATATYPE \n" ); fflush(stdout);
                              gds_read_datatype( infile, count, &item, 0 );
                              dt = item->dt;
                              break;
                    case 0x1003:  //printf( "XY \n" ) ; fflush(stdout);
                              numpolys++;
                              if ( ! (numpolys % 10000) ) 
                                  {
                                  printf( "\015           %d       ", numpolys );
                                  fflush( stdout ); 
                                  }
                              gds_read_xy( infile, count, 0, &item, 0 );
                              item->layer = layer;
                              item->dt = dt;
                              got_one = TRUE;
                              len = item->n;
                              poly = initialize_polygon( len, layer, dt, FALSE );
                              copy_from_gds_polygon( item, poly );
                              //snap_to_grid( poly, step );

                              //------------------------------------
                              if ( poly->len > 2 && poly->ur.x > poly->ll.x && poly->ur.y > poly->ll.y  ) 
				  append_queue( shape, poly );
			      else
				  poly->active = FALSE;
                              //------------------------------------

                              if ( poly->active && first )
                                  {
                                  minx = poly->ll.x;
                                  maxx = poly->ur.x;
                                  miny = poly->ll.y;
                                  maxy = poly->ur.y;
                                  first = FALSE;
                                  }
                              else if ( poly->active )
                                  {
                                  if ( poly->ll.x < minx ) minx = poly->ll.x;
                                  if ( poly->ur.x > maxx ) maxx = poly->ur.x;
                                  if ( poly->ll.y < miny ) miny = poly->ll.y;
                                  if ( poly->ur.y > maxy ) maxy = poly->ur.y;
                                  }

                              break;
                    default:      //
                              SKIPOVER( infile, count );
                    }    
                }      
            not_eof = read( infile, &count, 2 ) > 0;
            }          

        }  // while not eof

    dminx = minx;
    dminy = miny;
    dmaxx = maxx;
    dmaxy = maxy;
    extent[0] = dminx * dbu_um;    // extent is now in um
    extent[1] = dminy * dbu_um;
    extent[2] = dmaxx * dbu_um;
    extent[3] = dmaxy * dbu_um;

    *nump = numpolys;
    *dbu = dbu_um;

    printf( "\n\n" );
    close( infile );

    }  // read_input_polys


//---------

int 
pardonne( double x1, double x2, double s )
    {
    
    // If a block assignment is just a little over, let it snap to the
    // larger side. Roundoff errors are such a bitch.

    static double low, high, xlow, xhigh;
    static int answer;
    
    x1 = dsnap( x1, s );
    x2 = dsnap( x2, s );

    answer = x1 < x2 ? floor(x1) : floor(x2);
    return( answer );

    // This.. does not work
    // low  = floor( xlow );
    // high = floor( xhigh ) + 1.0;
    // answer = low;
    // if ( abs( xlow - low ) > abs( xhigh - high ) ) answer = floor( xhigh );
    // return( answer );
    }

//--------------------------------------------------------------------------------------------------

array_2dq *
field_sort( queue *shape, dxy ll, dxy ur, dxy block, double dbu_um, int ny, int nx )    
    {
    dxy          ll_mfr, block_mfr;
    int          n, i, x, y, x2, y2;
    array_2dq   *field;
    polygon_ref *sr;
    polygon     *s;
    double       fx, fy, fx2, fy2, dx, dy;
    queue       *fieldq;

    if ( ll.x >= ur.x || ll.y >= ur.y ) bail( "field_sort: invalid bounds" );

    ll_mfr.x = ((double) ll.x) / dbu_um;
    ll_mfr.y = ((double) ll.y) / dbu_um; 
    block_mfr.x = block.x / dbu_um;
    block_mfr.y = block.y / dbu_um;

    n = shape->len;

    field = initialize_2dq( ny, nx );

    sr = shape->first;

    while ( sr )
        {
        s = sr->p;

        //s->ll = isnapointo( s->ll, bss );  does not help
	//s->ur = isnapointo( s->ur, bss );
        x  = (int) floor( (s->ll.x - ll_mfr.x) / block_mfr.x );
        y  = (int) floor( (s->ll.y - ll_mfr.y) / block_mfr.y );
        x2 = (int) floor( (s->ur.x - ll_mfr.x) / block_mfr.x );
        y2 = (int) floor( (s->ur.y - ll_mfr.y) / block_mfr.y );

      
        if ( x2 == x && y2 == y ) 
            {
	    //printf( "\n************** THIS SHAPE DEFINITELY GOES INTO FIELD %d, %d\n", x, y );
            fieldq = get_field_list( field, x, y );
            append_queue( fieldq, s );
            }
        else        // a shape spans two fields. sadly, pardonne() is very heuristic
            {
            dx = s->ll.x;
            fx  = (dx - ll_mfr.x) / block_mfr.x;
            dy = s->ll.y;
            fy  = (dy - ll_mfr.y) / block_mfr.y;
            dx = s->ur.x;
            fx2 = (dx - ll_mfr.x) / block_mfr.x;
            dy = s->ur.y;
            fy2 = (dy - ll_mfr.y) / block_mfr.y;

            fx  = trim( fx  ); 
            fx2 = trim( fx2 );
	    fy  = trim( fy  );
	    fy2 = trim( fy2 );

            x = pardonne( fx, fx2, 2e-4 );
            y = pardonne( fy, fy2, 2e-4 );
            
            //printf( "*************** x = %1.5lf or %1.5lf    y = %1.5lf or %1.5lf \n", fx, fx2, fy, fy2 );
            //printf( "*************** I THINK IT SHOULD GO INTO FIELD %d, %d\n", x, y );

            fieldq = get_field_list( field, x, y ); 
            append_queue( fieldq, s );
            }

        sr = sr->next;
        }

    return( field );

    }  // field_sort

//--------------------------------------------------------------------------------------------------

BOOL
bad_shape( polygon *p )
    {
    return( ((p->ur.x - p->ll.x) < 1) || ((p->ur.y - p->ll.y) < 1) );
    }

//--------------------------------------------------------------------------------------------------

void
write_subfield( FILE         *ofile,
                BOOL          traps,
                dxy           block, 
                dxy           subfield_origin, 
                dxy           fieldcenter, 
                queue        *subfieldq, 
                double        mfr, 
                double        sfr, 
                double        dbu_um, 
                double        dbupermfr, 
                double        centre, 
                double        sbss, 
                int           subits, 
                dxy           sfs_mfr,
                int          *fieldnum, 
                double        bss_um,
                BOOL          assign_doses,
                double      **dose,
                int         **msfm,
                gpf_union_t  *prims,
                gpf_union_t  *prims_org,
		gpf_ulong_t  *gpfpos   )
    {
    static int          lsign, rsign, x1, y2, dx, dy, msf, n;
    static BOOL         first=TRUE, start_subfield;
    static int          genus, h, b, ltshift, rbshift, ox, oy;
    static double       dbu2size, mfr2dbu, subits_per_mainbit;
    static dxy          mmfr, ssfr, sfs_um;
    static polygon     *p;
    static polygon_ref *pr;
    static double       previous_dose=0, thisdose; 
    static char        *clock, lsw[8], msw[8];

    start_subfield = TRUE;
    if ( ! subfieldq ) return;
    if ( subfieldq->len == 0 ) return;
    
    // subfield origin is mfr units, gds origin
    // fieldcenter is um units, gds origin 
    // centre is the DAC setting for field center, mfr units (bits), usually 524288

    if ( first ) 
        {
        first = FALSE;
        mmfr.x = mfr;
        mmfr.y = mfr;
        ssfr.x = sfr;
        ssfr.y = sfr;
        dbu2size = dbu_um / (sfr * sbss);
        mfr2dbu  = mfr / dbu_um;                          // usually 1
        subits_per_mainbit = mfr / sfr;                   // usually 2
        sfs_um.x  = sfs_mfr.x * mfr;                      // subfield size in um
        sfs_um.y  = sfs_mfr.y * mfr;                      // yeah I know this is stupid
        }

    pr = subfieldq->first;
    while ( pr && ! pr->p->active ) pr = pr->next; 

    //if ( pr )
    //    {
    //    p = pr->p;
    //    if ( bad_shape( p ) && subfieldq->len == 1 ) return;
    //    }

    // write_polygon will update the mainfield deflection if necessary
    // we don't have to do an explicit mainfield deflection here

    while ( pr ) 
        {
        p = pr->p;

	write_polygon( ofile,
                       traps,
                       start_subfield,
                       block,
                       prims, 
                       prims_org, 
                       p,                      // array of xy, mfr units (usually 1 dbu = 1nm)
		       fieldcenter,            // um, gds units
		       sfs_um,
		       mmfr,                   // um   main field resolution x, y
                       ssfr,                   // um   subfield resolution x, y
		       bss_um,
                       dbu_um,                 // um   gds database unit, usually 1nm
                       assign_doses,           // bool
                       dose,                   // dose matrix, doubles
                       msfm,                   // msf multipliers, int
		       1, 1,                   //      number of columns, rows
                       0.0, 0.0,               // um   pitch in x, y
                       gpfpos );

        start_subfield = FALSE;

        pr = pr->next;
	while ( pr && ! pr->p->active ) pr = pr->next; 

        }   // while pr 

    } // write_subfield


//----------------------------------------------------------------------------------------------------------

void
find_chunk( double base, double subfield, double mfr, double *b, int *n, double *extra )  
    {                                                       
    // Find the size 'b' of an element of an array, so that (n * b + extra) is nearly equal to 'base',
    // and b < subfield. Also b must be a multiple of the main field resolution (mfr). 
    // These units are all microns.

    double fn, fb;

    if ( base <= 0.0 )
	{
	*b = 0;
	*n = 0;
        *extra = 0;
	return;
	}

    if ( base <= subfield )
        {
        *n = 1;                               
        *b = base;
        *extra = 0;
        return;
        }

    fn = 1.0 + floor( base / subfield );
    fb = base / fn;
    fb = mfr * floor( fb / mfr );

    if ( fb < 0 ) fb = 0;
    *extra = base - fb * fn;

    if ( *extra > fb )            // this really happens!
        {
        //printf( "  extra = %4d  >  fb = %lf    fn = %4d  base = %lf\n", *extra, fb, fn, base );
        fn = fn + 1.0;
        fb = base / fn;
	fb = mfr * floor( fb / mfr );
	*extra = base - fb * fn;
	if ( *extra > fb ) 
            {
	    fn = fn + 1.0;
            fb = base / fn;
	    fb = mfr * floor( fb / mfr );
    	    *extra = base - fb * fn;
            if ( *extra > fb )
                {
		printf( ">>extra = %lf  >  fb = %lf    fn = %lf  base = %lf\n", *extra, fb, fn, base );
		bail( "gpfout: extra on end of array is still wrong!" );
		}
            }
	}
    *n = (int) floor( fn );
    *b = fb;

    //printf( "\n n= %d  b= %lf  when base= %lf  subf= %lf\n", *n, *b, base, subfield );

    return;
    }

//----------------------------------------------------------------------------------------------------------


void 
write_rectangles(  FILE        *fp,
                   BOOL         traps,
                   dxy          block,            // um   block size
                   gpf_union_t *prims, 
                   gpf_union_t *prims_org, 
                   queue       *rectangles,
                   dxy          fieldcenter,      // um   gds units
                   dxy          sfs,              // um   subfield size
                   dxy          mfr,              // um   mainfield resolution 0.001
                   dxy          sfr,              // um   subfield  resolution 0.0005
                   dxy          bss,              // um   beam step size
		   double       dbu_um,           // um   gds database unit
                   BOOL         assign_doses,     
                   double     **dose,             // dose matrix
                   int        **msfm,             // msf multipliers
                   gpf_ulong_t *gpfpos       )
 
    {
    // Use arrays of rectangular polygons to generate large rectangles.
    // This provides a very large compression to the output file, when the
    // pattern contains large features. Beams expands arrays before they
    // get to the pattern generator, so this is strictly compaction.
    
    static double       base, height, thisdose, previous_dose;
    static double       llx, lly, urx, ury, fny, fnx;
    static polygon_ref *pr;
    static polygon     *piece, *right, *top, *corner;
    static int          nx, ny, rightx, topy, ib, ih;
    static double       h, b;
    static dxy          chunk, extra;              
    static char        *clock, lsw[8], msw[8];
    static polygon     *p;
    static BOOL         first=TRUE, start_subfield;
    static xy           v, iextra;

    if ( first )
	{
	first      = FALSE;
	piece      = initialize_polygon( 4, 0, 0, 0 );
        right      = initialize_polygon( 4, 0, 0, 0 );
        top        = initialize_polygon( 4, 0, 0, 0 );
        corner     = initialize_polygon( 4, 0, 0, 0 );
	}
    
    if ( rectangles->len == 0 ) return;

    previous_dose = 0.0;
    start_subfield = TRUE;

    pr = rectangles->first;

    while ( pr )
        {
        p = pr->p;

	zero_polygon( piece  );

        piece->layer = p->layer;
        piece->dt    = p->dt;

	llx = p->ll.x * dbu_um;    // um
	lly = p->ll.y * dbu_um;
	urx = p->ur.x * dbu_um;
	ury = p->ur.y * dbu_um;

        base   = (urx - llx);     // um
        height = (ury - lly);

	if ( base   < 0.0 ) base   = 0.0;
	if ( height < 0.0 ) height = 0.0;

        find_chunk( height, sfs.y, mfr.y, &h, &ny, &extra.y ); 
        find_chunk( base,   sfs.x, mfr.x, &b, &nx, &extra.x );

        ib = round( b / dbu_um );   // dbu
        ih = round( h / dbu_um );

        append_polygon_xy( piece, p->ll );
        v.x = p->ll.x + ib;
        v.y = p->ll.y;
        append_polygon_xy( piece, v );
        v.y = p->ll.y + ih;
	append_polygon_xy( piece, v );
        v.x = p->ll.x;
        append_polygon_xy( piece, v );

        iextra.x = (int) floor( extra.x / dbu_um );   // dbu
        iextra.y = (int) floor( extra.y / dbu_um );

        rightx = p->ll.x + nx * ib;
        topy   = p->ll.y + ny * ih;

 	write_polygon( fp,
                       traps,
		       start_subfield,
                       block,
                       prims, 
                       prims_org, 
                       piece,                  // the polygon, mfr units, int 
	               fieldcenter,            // um, gds units
	               sfs,                    // um, subfield size
	               mfr,                    // um, mainfield resolution  0.001
                       sfr,                    // um, subfield resolution
		       bss.x,
		       dbu_um,                 // um, gds database resolution
                       assign_doses,           // bool
                       dose,                   // dose matrix, doubles
                       msfm,                   // msf multipliers, int
	               nx, ny,                 // number of columns, rows
                       b, h,                   // um, pitch in x, y
                       gpfpos );

        start_subfield = FALSE;
           
        if ( iextra.x > 0 )  // extra end-piece array on the right side
            {
	    zero_polygon( right );
            right->layer = p->layer;
            right->dt    = p->dt;

            if ( extra.x > sfs.x ) bail( "(write_polygon) extra x chunk is too big" );

	    v.x = rightx;
            v.y = p->ll.y;
	    append_polygon_xy( right, v );
            v.x = p->ur.x;
	    append_polygon_xy( right, v );
            v.y = p->ll.y + (int) floor( h / dbu_um );
	    append_polygon_xy( right, v );
            v.x = rightx;
	    append_polygon_xy( right, v );

 	    write_polygon( fp,
                           traps,
			   start_subfield,
                           block,
                           prims, 
                           prims_org, 
                           right,
	                   fieldcenter,            // um, gds units
	                   sfs,                    // um, subfield size
	                   mfr,                    // um, mainfield resolution
                           sfr,                    // um, subfield resolution
			   bss.x,
			   dbu_um,                 // um, gds database resolution
                           assign_doses,           // bool
                           dose,                   // dose matrix, doubles
                           msfm,                   // msf multipliers, int
	                   1, ny,                  // number of columns, rows
                           0, h,                   // um, pitch in x, y
                           gpfpos );            
            }

        if ( iextra.y > 0 )  // extra end-piece array on the top 
            {
            zero_polygon( top );
            top->layer = p->layer;
            top->dt    = p->dt;

            if ( extra.y > sfs.y ) bail( "(write_polygon) extra y chunk is too big" );

	    v.x = p->ll.x;
            v.y = topy;
	    append_polygon_xy( top, v );
            v.x = p->ll.x + (int) floor( b / dbu_um );
	    append_polygon_xy( top, v );
            v.y = p->ur.y;
	    append_polygon_xy( top, v );
            v.x = p->ll.x;
	    append_polygon_xy( top, v );

 	    write_polygon( fp,
                           traps,
			   start_subfield,
                           block,
                           prims, 
                           prims_org, 
                           top,
	                   fieldcenter,            // um, gds units
	                   sfs,                    // um, subfield size
	                   mfr,                    // um, mainfield resolution
                           sfr,                    // um, subfield resolution
			   bss.x,
			   dbu_um,                 // um, gds database resolution
                           assign_doses,           // bool
                           dose,                   // dose matrix, doubles
                           msfm,                   // msf multipliers, int
	                   nx, 1,                  // number of columns, rows
                           b, 0,                   // um, pitch in x, y
                           gpfpos );            
            } 

        if ( iextra.x > 0 && iextra.y > 0 )  // corner piece
	    {
	    zero_polygon( corner );
            corner->layer = p->layer;
	    corner->dt    = p->dt;

            v.x = rightx;
            v.y = topy;
            append_polygon_xy( corner, v );
            v.x = p->ur.x;
            append_polygon_xy( corner, v );
            v.y = p->ur.y;
            append_polygon_xy( corner, v );
            v.x = rightx;
            append_polygon_xy( corner, v );

 	    write_polygon( fp,
                           traps,
			   start_subfield,
                           block,
                           prims, 
                           prims_org, 
                           corner,
	                   fieldcenter,            // um, gds units
	                   sfs,                    // um, subfield size
	                   mfr,                    // um, mainfield resolution
                           sfr,                    // um, subfield resolution
			   bss.x,
			   dbu_um,                 // um, gds database resolution
                           assign_doses,           // bool
                           dose,                   // dose matrix, doubles
                           msfm,                   // msf multipliers, int
	                   1, 1,                   // number of columns, rows
                           0, 0,                   // um, pitch in x, y
                           gpfpos );            
	    }

        pr = pr->next;
        }


    }   // write_rectangles 

//-----------------------------------------------------------------------------------------------------------

void
write_field( FILE         *ofile,
             BOOL          traps,
             dxy           blockllmfr, 
	     dxy           fieldcenter, 
	     dxy           patcenter, 
	     int           col, 
	     int           row, 
             array_2dq    *field, 
	     double        sbss, 
	     dxy           sfs, 
	     dxy           block, 
	     double        dbu_um, 
	     double        mfr, 
             double        sfr, 
	     double        centre, 
	     int           subits, 
	     double        bss_um, 
	     BOOL          assign_doses, 
	     double      **dose,
             int         **msfm,
	     double        mindose,
	     double        maxdose,
             gpf_union_t  *prims,
             gpf_union_t  *prims_org,
             gpf_ulong_t  *gpfpos,
             gpf_indexf_t *field_index,
             int          *fieldnum,
             gpf_ulong_t  *pixeltime
           )
    {
    // ofile            output gtx file
    // fieldcenter (um) placement of this field, gds micron units
    // patcenter (um)   center of the pattern, gds micron units
    // (col,row)        index into the list of shapes in this field
    // field            list of shapes in this field (2D array of linked lists)
    // sbss             subfield beam step size, microns (e.g. 0.0005)
    // sfs              subfield size, microns (e.g. 4)
    // block            block size, microns (e.g. 500)
    // dbu_um           gds database unit, microns (e.g. 0.001)
    // mfr              main field resolution, microns (usually 0.001) = lsb of deflector
    // sfr              subfield resolution, microns (usually 0.0005)
    // centre           main field deflection to field center, usually 2^19 = 524288
    // subits           number of bits on the subfield DAC, 14
    // bss_um           beam step size in microns (e.g. 0.01)
    // dose             dose[layer][dt] relative dose matrix
    // msfm             msf[layer][dt] matrix of msf multipliers, mostly 1
    

    static dxy          llblock, deflection, size, maxsfs;
    static double       x0, y0, bss_um_2;
    static queue       *fieldq, *subfieldq;
    static int          n, nx, ny, x, y, x2, y2;
    static dxy          sfs_mfr, ll_mfr, ur_mfr, origin, mmfr, ssfr, bbss;
    static double       bss_mfr, dbupermfr, minx, miny, maxx, maxy, bss, left, bottom, fx, fx2, fy, fy2, mmsf;    
    static array_2dq   *subfield;
    static polygon     *s;
    static polygon_ref *pr;
    static int          sub_msf;   // subfield msf, eg 10 when the bss is 5 nm
    static int          numrows, numcols;
    static dxy          offset;
    static int          nsfs=0;
    
    numrows = block.x / sfs.x;
    numcols = block.y / sfs.y;

    bss_um_2 = bss_um * 0.5;  // throw away things smaller than this. not bss? nope.

    queue *rectangles;

    mmfr.x = mfr;  mmfr.y = mfr;
    ssfr.x = sfr;  ssfr.y = sfr;

    maxsfs.x = sfs.x * 1.02;
    maxsfs.y = sfs.y * 1.02;

    rectangles = initialize_queue();

    *pixeltime = 0;
    
    fieldq = get_field_list( field, col, row );
    n = fieldq->len;                               // number of shapes in this field
    if ( n == 0 ) return;

    dbupermfr = mfr / dbu_um;              // usually 1, but if gds is in angstroms, this might be 10

    mmsf = bss_um / mfr;                   // msf of main field  eg 5 = 5nm/1nm
    sub_msf = (int) floor( mmsf * 2.0 );   // msf of subfield    eg 10

    sfs_mfr.x = sfs.x / mfr;               // sfs in um, mfr in um/bit. So sfs_mfr units are mainfield bits
    sfs_mfr.y = sfs.y / mfr;      

    ll_mfr.x = blockllmfr.x;
    ll_mfr.y = blockllmfr.y;
    ur_mfr.x = blockllmfr.x + block.x / mfr + sfs_mfr.x;  // a shape can exceed the block size by 
    ur_mfr.y = blockllmfr.y + block.y / mfr + sfs_mfr.y;  // at most one subfield

    ur_mfr.x = alittlebigger( ur_mfr.x );
    ur_mfr.y = alittlebigger( ur_mfr.y );

    nx = (int) round( block.x / sfs.x ) + 2;    // number of subfields
    ny = (int) round( block.y / sfs.y ) + 2;    // upper limit

    subfield = initialize_2dq( ny, nx );

    pr = fieldq->first;

    bss_mfr = bss_um / mfr;   // same as mmsf

    // sort the shapes into subfields

    while ( pr )
        {
        s = pr->p;

        if ( s->active ) *pixeltime += pixel_count( s, dbu_um, bss_um, dose, msfm );  

        if ( s->active && not_a_large_rectangle( s, sfs, dbu_um, FALSE ) )  
            {
            minx = dsnap( ((double) s->ll.x) / dbupermfr, bss_mfr );
	    miny = dsnap( ((double) s->ll.y) / dbupermfr, bss_mfr );
            maxx = dsnap( ((double) s->ur.x) / dbupermfr, bss_mfr );
            maxy = dsnap( ((double) s->ur.y) / dbupermfr, bss_mfr );

            if ( ( minx < ll_mfr.x ) || ( miny < ll_mfr.y ) || ( maxx > ur_mfr.x ) || ( maxy > ur_mfr.y ) )
		{
		if ( ((maxx - minx) < bss_um_2) || ((maxy - miny) < bss_um_2) )
		    {
		    s->active = FALSE;
		    }
		else
		    {
		    /*
		    print_polygon( s, 1 );
                    printf( "shape ll (dbu):  %lf, %lf \n", minx, miny );
                    printf( "block ll (dbu):  %lf, %lf \n", ll_mfr.x, ll_mfr.y ); 
                    printf( "shape ur (dbu):  %lf, %lf \n", maxx, maxy );
                    printf( "block ur (dbu):  %lf, %lf \n", ur_mfr.x, ur_mfr.y ); 
                    if ( minx < ll_mfr.x ) bail( "shape x is less than block LL x" );
                    if ( miny < ll_mfr.y ) bail( "shape y is less than block LL y" );
		    if ( maxx > ur_mfr.x ) bail( "shape x is larger than block UR x" );
                    if ( maxy > ur_mfr.y ) bail( "shape y is larger than block UR y" );
		    */
		    }
		}

            llblock.x = fieldcenter.x - block.x / 2.0;
            llblock.y = fieldcenter.y - block.y / 2.0;
            // x0,y0 is the main field deflection from field center to the ll corner of the subfield, in microns
            x0 = llblock.x - fieldcenter.x + sfs.x * (floor(((s->ll.x * dbu_um) - llblock.x) / sfs.x));
            y0 = llblock.y - fieldcenter.y + sfs.y * (floor(((s->ll.y * dbu_um) - llblock.y) / sfs.y));
            // what is the index of the subfield, x,y ? The ll corner should be index 0,0
            x = floor( (x0 + block.x / 2.0) / sfs.x ); 
            y = floor( (y0 + block.y / 2.0) / sfs.y );
            if ( x < 0 ) x = 0;
	    if ( y < 0 ) y = 0;
            deflection.x = (s->ur.x * dbu_um - fieldcenter.x) - x0;   // trap deflection to ur of shape
            deflection.y = (s->ur.y * dbu_um - fieldcenter.y) - y0;   // um
            
            offset.x = (s->ll.x * dbu_um - fieldcenter.x) - x0;       // trap deflection to ll of shape
            offset.y = (s->ll.y * dbu_um - fieldcenter.y) - y0;       // um

            size.x = (s->ur.x - s->ll.x) * dbu_um;
            size.y = (s->ur.y - s->ll.y) * dbu_um;

            if ( size.x < bss_um_2 || size.y < bss_um_2 ) s->active = FALSE;    // throw away sub-pixel slivers

            if ( s->active && x < numcols && deflection.x > maxsfs.x )
		{
                while ( x < numcols && deflection.x > maxsfs.x )  // ridiculous but it works
		    {
                    x++;
                    x0 = x * sfs.x - block.x / 2.0; 
                    deflection.x = (s->ur.x * dbu_um - fieldcenter.x) - x0;
                    offset.x     = (s->ll.x * dbu_um - fieldcenter.x) - x0;
		    }
		}

            if ( s->active && y < numrows && deflection.y > maxsfs.y )
		{
                while ( y < numrows && deflection.y > maxsfs.y )  // ridiculous. how did we go so wrong above
		    {                                             // where we calculated x and y ?
                    y++;
                    y0 = y * sfs.y - block.y / 2.0;
		    deflection.y = (s->ur.y * dbu_um - fieldcenter.y) - y0;
                    offset.y     = (s->ll.y * dbu_um - fieldcenter.y) - y0;
		    }
		}

            if (0) // ( s->active && deflection.x > maxsfs.x && deflection.y > maxsfs.y ) 
		{
                warning( "X and Y deflection exceed subfield size", &nsfs );
		//printf( "\ndeflection = [%1.2lf], [%1.2lf] um   size= %1.2lf, %1.2lf um  ", deflection.x, deflection.y, size.x, size.y );
		//printf( " x0,y0= %1.4lf, %1.4lf um   x,y= %2d, %2d", x0, y0, x, y );
		}
            else if (0) // ( s->active && deflection.x > maxsfs.x )
		{
                warning( "X deflection exceeds subfield size", &nsfs );
		//printf( "\ndeflection = [%1.2lf],  %1.2lf  um   size= %1.2lf, %1.2lf um  ", deflection.x, deflection.y, size.x, size.y );
		//printf( " x0,y0= %1.4lf, %1.4lf um   x,y= %2d, %2d", x0, y0, x, y );
		}
            else if (0) // ( s->active && deflection.y > maxsfs.y )
		{
                warning( "Y deflection exceeds subfield size", &nsfs );
                if ( nsfs < 10 )
                    {
		    printf( "\ndeflection =  %1.2lf,  [%1.2lf] um   size= %1.2lf, %1.2lf um  ", deflection.x, deflection.y, size.x, size.y );
		    printf( " x0,y0= %1.4lf, %1.4lf um   x,y= %2d, %2d \n", x0, y0, x, y );
                    printf( "ll = %1.4f, %1.4f \n", (s->ll.x * dbu_um) - patcenter.x, (s->ll.y * dbu_um) - patcenter.y );
                    }
		}

	    if ( s->active && x > 0 && (deflection.x < 0 || offset.x < 0) )	// neg deflection? put x back to original value
		{                                                   
                x--;
                x0 = x * sfs.x - block.x / 2.0; 
                deflection.x = (s->ur.x * dbu_um - fieldcenter.x) - x0;   // um  trap defl to ur of shape
                offset.x     = (s->ll.x * dbu_um - fieldcenter.x) - x0;   // um  trap defl to ll of shape

                if ( (deflection.x < 0) && (fabs( deflection.x ) < bss_um) ) deflection.x = 0;
		if ( (offset.x     < 0) && (fabs( offset.x )     < bss_um) ) offset.x     = 0;    

                if ( deflection.x < 0 || offset.x < 0 )         // deflection will still be too high, so what can we do about that?
		    {                                           // ignore it and hope for the best?
		    printf( "\nnegative X offs at %lf, %lf \n     offs= %lf, %lf  defl=  %lf, %lf um   size= %1.2lf, %1.2lf um  ", 
		  	    s->ll.x * dbu_um, s->ll.y * dbu_um, offset.x, offset.y, deflection.x, deflection.y, size.x, size.y );
		    printf( " x0,y0= %1.4lf, %1.4lf um   x,y= %2d, %2d", x0, y0, x, y );
		    }
		}
	

	    if ( s->active && y > 0 && (deflection.y < 0 || offset.y < 0) )	
		{
                y--;
                y0 = y * sfs.y - block.y / 2.0; 
                deflection.y = (s->ur.y * dbu_um - fieldcenter.y) - y0;
                offset.y     = (s->ll.y * dbu_um - fieldcenter.y) - y0;

                if ( (deflection.y < 0) && (fabs( deflection.y ) < bss_um) ) deflection.y = 0;
		if ( (offset.y     < 0) && (fabs( offset.y )     < bss_um) ) offset.y     = 0;    

                if ( VERBOSE && (deflection.y < 0 || offset.y < 0) )  // this is not necessarily a fatal error
		    {
		    printf( "\nnegative Y at %lf, %lf \n     offs= %lf, %lf  defl=  %lf, %lf um   size= %1.2lf, %1.2lf um  ", 
		  	    s->ll.x * dbu_um, s->ll.y * dbu_um, offset.x, offset.y, deflection.x, deflection.y, size.x, size.y );
		    printf( " x0,y0= %1.4lf, %1.4lf um   x,y= %2d, %2d", x0, y0, x, y );
		    }
		}
	
            if ( s->active )
		{
                subfieldq = get_field_list( subfield, x, y ); 
                append_queue( subfieldq, s );
		}
            }
        else
            {
            if ( s->active ) append_queue( rectangles, s );
            }

        pr = pr->next;
        }

    // write stage move and initialize primitives

    init_prims( prims, prims_org, sub_msf, mindose, maxdose );

    // movestage actually just collects the stage move info in field_index.
    // we will write the stage move block at the end.

    movestage( ofile,
               fieldcenter.x - patcenter.x,
	       fieldcenter.y - patcenter.y,
               prims,
               prims_org,
               field_index,               
               fieldnum,
               gpfpos       );
   
    // It used to be that we would write_prims before each activate trap line,
    // because this would avoid repeating parameters that have not changed.
    // But when we are using polygons exclusively, there are no repeated
    // parameters to avoid. In fact, what's the point of having init_prims
    // and write_prims in separate functions? We should clean this up later.

    write_prims( ofile, prims, prims_org, gpfpos );

    origin.x = 0;
    origin.y = 0;

    bss = sbss * sfr / mfr;      
    bbss.x = bss_um;
    bbss.y = bss_um;
    left   = ll_mfr.x;  
    bottom = ll_mfr.y;  
    origin.x = left;             // origin of a subfield, lower-left, mfr units, gds origin
    origin.y = bottom;

    write_rectangles( ofile,
                      traps,
                      block,            // um   block size
                      prims, 
                      prims_org, 
                      rectangles,
                      fieldcenter,      // um   gds units
                      sfs,              // um   subfield size
                      mmfr,             // um   mainfield resolution 0.001
                      ssfr,             // um   subfield  resolution 0.0005
                      bbss,             // um   beam step size
		      dbu_um,           // um   gds database unit
                      assign_doses,     // bool
                      dose,             // dose matrix, doubles **
                      msfm,             // msf multipliers, int **
                      gpfpos  );


    // What if we did not sort into subfields? would that work?
    // Yes, but the subfields would not be writen in a serpentine order. 
    // The beam would deflect in the order of the input shapes. 

    // you could do this instead of looping over the subfields: 
    // write_subfield( ofile, block, origin, fieldcenter, fieldq, mfr, sfr, dbu_um, 
    //	  	       dbupermfr, centre, sbss, subits, sfs_mfr, fieldnum, bss_um, assign_doses, dose,
    //                 prims, prims_org, gpfpos );

    for ( y=0; y < ny; y++ )
        {
        for ( x=0; x < nx; x++ )
            {
            subfieldq = get_field_list( subfield, x, y );

            write_subfield( ofile, traps, block, origin, fieldcenter, subfieldq, mfr, sfr, dbu_um, 
			    dbupermfr, centre, sbss, subits, sfs_mfr, fieldnum, bss_um, assign_doses, dose, msfm,
                            prims, prims_org, gpfpos );

            origin.x = origin.x + sfs_mfr.x; 
            
            }
        origin.x = left;
        origin.y = origin.y + sfs_mfr.y;
        }
    
    end_field( field_index, fieldnum, gpfpos );    // increments *fieldnum

    }  // write_field

//--------------------------------------------------------------------------------------------------

char *
readln( FILE *fp )
    {
    static char ch, *line;
    static int i;
    static BOOL init=FALSE;
    if ( ! init )
        {
        init = TRUE;
        line = (char *) malloc( 512 * sizeof(char) );
        if ( ! line ) bail( "unable to allocate line in readln" );
        }
    i = 0;
    while ( ((ch=getc(fp)) != '\n' ) && (ch != EOF) ) 
        {
        if ( i > 510 ) bail( "line is too long in dose file" );
        line[i] = ch;
        i++;
        }
    line[i] = '\0';
    if ( ch == EOF && i == 0 ) line[0] = EOF;
    return line;
    }
    
//--------------------------------------------------------------------------------------------------

BOOL
read_dose_file( BOOL init_only, char *filename, double **dose, int **msfm, double *mindose, double *maxdose, double bss_base )
    {                                          
    FILE   *dosefile;
    char   ch, *line, *left, *right;
    int    layer, dt, i, j, nr, imsf; 
    double d, bss, frac, imsf2;
    BOOL   allayers, alldts;

    // This would be 1000 times easier in python
    
    for ( layer=0; layer < 256; layer++ )
        for ( dt=0; dt < 256; dt++ )
            {
            dose[layer][dt] = 1.0;
            msfm[layer][dt]  = 1;  // we will multiply the msf by this, later
            }

    if ( init_only ) return( FALSE );

    dosefile = fopen( filename, "r" );
    if ( ! dosefile ) 
        {
        printf( "\n>>>>>>>>>>>>>>>>>>>> Note lack of dose file.\n\n" );
	*mindose = 1.0;
	*maxdose = 1.0;
        return( FALSE );
        }

    printf( "\nDose assignments from %s \n[layer] [dt] [dose] [bss, nm]\n\n", filename );

    line = readln( dosefile );

    while ( line[0] == '#' ) 
        {
        printf( "%s\n", line );
        line = readln( dosefile );  // skip comments
        }

    while ( line[0] != EOF )
        {
        layer = 0;
        dt = 0;
        d  = 1.0;
        allayers = FALSE;
        alldts   = FALSE;

        if ( line[0] != '#' && strlen(line) > 4 )     // skip comments and blank lines
            {
            left  = strchr(  line, '*' );
            right = strrchr( line, '*' );
  
            if ( left && (left != right) )       // then we have two stars
                {
                allayers = TRUE;
                alldts   = TRUE;
                right++;
                nr = sscanf( right, "%lf %lf", &d, &bss );
                if ( nr < 2 ) bss = bss_base; 
                //printf( "\n***************** ALL LAYERS AND ALL DTS, DOSE = %g \n", d );
                } 
            else if ( left && (left == right) )  // then we have one star. which is it?
                {
                allayers = (left == line);       // star on left => layers
                alldts   = ! allayers;           // if it's not for layer then it's for dt
                left++;
                right++;
                if ( allayers )
                    {
                    nr = sscanf( left, "%d %lf %lf", &dt, &d, &bss );
                    if ( nr < 3 ) bss = bss_base;
                    //printf( "\n***************** ALL LAYERS, DT = %d  DOSE = %g \n", dt, d );
                    }
                else
                    {
                    sscanf( line, "%d", &layer );
                    nr = sscanf( right, "%lf %lf", &d, &bss );
                    if ( nr < 2 ) bss = bss_base;
                    //printf( "\n***************** ALL DTS, LAYER = %d  DOSE = %g \n", layer, d );
                    }
                }
            else
                {
                nr = sscanf( line, "%d %d %lf %lf", &layer, &dt, &d, &bss );
                if ( nr < 4 ) bss = bss_base;
                }

            if ( layer > 255 || layer < 0 ) bail( "invalid layer number in dose file" );
            if ( dt > 255 || dt < 0 ) bail( "invalid datatype number in dose file" );

            
            if ( bss < bss_base )
                {
                printf( "\nERROR: beam step size in dose file %1.3f nm is less than the base value %1.3f nm\n", bss, bss_base );
                bail( "illegal value of beam step size in dose file" );
                }

            frac = fabs( (bss/bss_base) - floor(bss/bss_base) );
            if ( frac > 0.001 )
                {
                printf( "\nWARNING: beam step in dose file %1.3f is not a multiple of the base value %1.3f\n", bss, bss_base );
                printf(   "         The beam step will be rounded down to the nearest multiple.\n\n" );
                }
            imsf = floor(bss / bss_base);  // we will multply the msf by this, later

            d = d * imsf * imsf; // relative dose is scaled up by the square of the beam step... 
                                 // well, we should say: the clock rate will be scaled down
                                 // by the square of the beam step multiplier. ok well that's confusing.
                                 // Think of it this way: this larger relative dose is appropriate
                                 // for the larger beam step we will set in the output gpf file.
            
            if ( d <= 0.5 && imsf > 1 )
                bail( "relative dose * ((beam step)/(base beam step))^2 cannot be <= 0.5" );
            else if ( d <= 0.5 )
                bail( "relative dose cannot be <= 0.5" );

            
            if ( allayers && alldts )
                {
                for ( i=0; i < 256; i++ ) 
                    for ( j=0; j < 256; j++ )
                        {
                        dose[i][j] = d;
                        msfm[i][j] = imsf;
                        }
                } 
            else if ( allayers )
                {
                for ( i=0; i < 256; i++ ) { dose[i][dt] = d; msfm[i][dt] = imsf; }
                }
            else if ( alldts )
                {
                for ( i=0; i < 256; i++ ) { dose[layer][i] = d; msfm[layer][i] = imsf; }
                }
            else
                {
                dose[layer][dt] = d;
                msfm[layer][dt] = imsf;
                }

            imsf2 = imsf * imsf;
            
            if ( allayers && alldts )
                printf( "  *   * %5.2f  %5.2f\n", dose[layer][dt]/imsf2, bss );
            else if ( allayers )
                printf( "  * %3d %5.2f  %5.2f\n", dt, dose[layer][dt]/imsf2, bss );
            else if ( alldts )
                printf( "%3d   * %5.2f  %5.2f\n", layer, dose[layer][dt]/imsf2, bss );
            else
                printf( "%3d %3d %5.2f  %5.2f\n", layer, dt, dose[layer][dt]/imsf2, bss );

            
            }
        line = readln( dosefile );
        }

    *mindose = dose[0][0];
    *maxdose = dose[0][0];

    for ( layer=0; layer < 256; layer++ )
        for ( dt=0; dt < 256; dt++ )
	    {
            if ( dose[layer][dt] < *mindose ) *mindose = dose[layer][dt];
	    if ( dose[layer][dt] > *maxdose ) *maxdose = dose[layer][dt];
	    }

    
    return( TRUE );
    }

//--------------------------------------------------------------------------------------------------

double
distance( xy p, xy q )
    {
    static double dx, dy, d;
    dx = p.x - q.x;
    dy = p.y - q.y;
    d = sqrt( dx * dx + dy * dy );
    return( d );
    }


double
pointoline_dist( xy p, xy a, xy b )
    {
    static double dx, dy, d, px, py, s, u, xx, yy;

    if ( samepoint( p, a ) )
        return( 0 );
    else if ( samepoint( p, b ) )
        return( 0 );
    else if ( samepoint( a, b ) ) 
        {
        dx = p.x - a.x;
        dy = p.y - a.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }
    else 
        {
        px = b.x - a.x;
        py = b.y - a.y;
        s = px * px + py * py;
        u = ((p.x - a.x) * px + (p.y - a.y) * py) / s;
        if ( u > 1.0 )
            u = 1.0;
        else if ( u < 0.0 )
            u = 0.0;
        xx = a.x + u * px;
        yy = a.y + u * py;
        dx = xx - p.x;
        dy = yy - p.y;
        d = sqrt( dx * dx + dy * dy );
        return( d );
        }

    }

//-------------------------------------------------------------------------------------------

void
plot_path( dxy *path, BOOL *jump, int n )
    {
    dxy    ll, ur;
    double low, high, extra;
    int    i;
    FILE   *gnufile, *tempath;

    ll.x = path[0].x;
    ll.y = path[0].y;
    ur.x = ll.x;
    ur.y = ll.y;
    
    for ( i=1; i < n; i++ )
        {
        if ( path[i].x < ll.x ) ll.x = path[i].x;
        if ( path[i].y < ll.y ) ll.y = path[i].y;
        if ( path[i].x > ur.x ) ur.x = path[i].x;
        if ( path[i].y > ur.y ) ur.y = path[i].y;
        }

    if ( ll.x == ur.x && ll.y == ur.y ) bail( "gpfout: path has no extent" );

    low = ll.x;
    if ( ll.y < low ) low = ll.y;
    high = ur.x;
    if ( ur.y > high ) high = ur.y;

    extra = fabs( high - low ) / 10.0;
    low   = low - extra;
    high  = high + extra;
    gnufile = fopen( "tmp/path.gnu", "w" );
    if ( ! gnufile )
        {
        printf( "\nUnable to open temporary file in tmp. Skipping path plot.\n" );
        return;
        }
    
    fprintf( gnufile, "set xlabel \"X um\"\n" );
    fprintf( gnufile, "set ylabel \"Y um\"\n" );
    fprintf( gnufile, "set size square\n" );
    fprintf( gnufile, "set xrange [%1.2f:%1.2f]\n", low, high );
    fprintf( gnufile, "set yrange [%1.2f:%1.2f]\n", low, high );
    fprintf( gnufile, "plot \"tmp/tempath.dat\" title 'writing path' with lines lt rgb \"blue\", " );
    fprintf( gnufile, "\"tmp/jump.dat\" title '' with lines lt rgb \"white\"\n" );
    fclose(  gnufile );

    tempath = fopen( "tmp/tempath.dat", "w" );
    if ( ! tempath )
        {
        printf( "\nUnable to open temporary file in tmp. Skipping path plot.\n" );
        return;
        }
    
    for ( i=0; i < n; i++ ) fprintf( tempath, "%lf %lf\n", path[i].x, path[i].y );
    fclose( tempath );

    tempath = fopen( "tmp/jump.dat", "w" );

    for ( i=0; i < n; i++ )
        {
        if ( i > 0 && jump[i] )
            {
            fprintf( tempath, "%lf %lf\n", path[i-1].x, path[i-1].y );
            fprintf( tempath, "%lf %lf\n", path[i].x,   path[i].y   );
            }
        else
            {
            fprintf( tempath, "%lf %lf\n", path[i].x, low );
            }
        }

    fclose( tempath );

    system( "gnuplot -persist tmp/path.gnu > /dev/null 2>&1 &" );

    }


//--------------------------------------------------------------------------------------------------

void
read_path( char *pathname, int *n, dxy **ppath, BOOL **pjump )   // this is so easy in python :(
    {
    // This path file might start with upper-right and lower-left lines,
    // but we do not need those coordinates. We can ignore them, if they exist.
    // gpfsort needs those lines because the input gpf file does not contain the gds origin.
    
    FILE *pf;                        
    dxy   p, *path;   
    char *line;
    BOOL *jump;
    int   i;
    
    // count the lines and allocate the array

    line = (char *) malloc( 256 );
    pf = fopen( pathname, "r" );
    if ( ! pf ) bail( "unable to open path file" );
    
    line = readln( pf );    // first coordinate or ll or ur
    *n = 0;    
    while ( line[0] != EOF )
        {
        line = readln( pf );
        *n = *n + 1;
        }
    fclose( pf );

    path = (dxy *)  malloc( (*n + 1) * sizeof(dxy)  );
    jump = (BOOL *) malloc( (*n + 1) * sizeof(BOOL) );
    
    pf = fopen( pathname, "r" );

    line = readln( pf );
    i = 0;
    
    while ( line[0] != EOF )
        {
        if ( strstr( line, "upper" ) || strstr( line, "lower" ) )   // if "upper" or "lower" in line
            {                                                       // then ignore the line
            *n = *n - 1;
            line = readln( pf );
            }
        else if ( strstr( line, "jump"  ) )   // if "jump" in line
            {
            jump[i] = TRUE;
            *n = *n - 1;
            line = readln( pf );
            }
        else
            {
            sscanf( line, "%lf %lf\n", &p.x, &p.y );
            path[i].x = p.x;
            path[i].y = p.y;
            i++;
            line = readln( pf );
            jump[i] = FALSE;
            }
        }

    plot_path( path, jump, *n );

    *ppath = path;
    *pjump = jump;
    
    return;
    }

//-------------------------------------------------------------------------------------------------

double
calc_y( dxy a, dxy b, double x )
    {
    double m, inter;
    
    if ( a.y == b.y ) return( a.y );
    if ( a.x == b.x ) return( (a.y + b.y) / 2.0 );  // unlikely
    m = (b.y - a.y) / (b.x - a.x);
    inter = a.y - m * a.x;
    return( m * x + inter );
    }


double
calc_x( dxy a, dxy b, double y )
    {
    double m, inter;
    
    if ( a.x == b.x ) return( a.x );
    if ( a.y == b.y ) return( (a.x + b.x) / 2.0 );  // unlikely
    m = (b.y - a.y) / (b.x - a.x);
    inter = a.y - m * a.x;
    return( (y - inter) / m );
    }


//-----------------------------------------------

dxy *
expand_path( int nin, int *nout, dxy *path, BOOL *jump, dxy block )
    {
    int i, n;
    dxy d, block4, extra, *chemin;

    block4.x = block.x / 4.0;
    block4.y = block.y / 4.0;
    
    *nout = 0;
    
    for ( i=1; i < nin; i++ )
        {
        d.x = fabs( path[i].x - path[i-1].x );
        d.y = fabs( path[i].y - path[i-1].y );

        if ( d.x > d.y )
            *nout = *nout + (int) floor( d.x / block4.x ) + 2;
        else
            *nout = *nout + (int) floor( d.y / block4.y ) + 2;
        }

    chemin = (dxy *) malloc( *nout * sizeof( dxy ) );

    // add points to the path in steps of block4 
    
    n = 0;
    for ( i=1; i < nin; i++ )
        {
        if ( ! jump[i] )
            {
            d.x = fabs( path[i].x - path[i-1].x );
            d.y = fabs( path[i].y - path[i-1].y );
        
            if ( d.x > block4.x || d.y > block4.y )
                {
                if ( d.x > d.y )
                    {
                    extra.x = path[i-1].x;
                    extra.y = path[i-1].y;
                    if ( path[i].x > path[i-1].x )
                        {
                        while ( extra.x < path[i].x )
                            {
                            chemin[n].x = extra.x;
                            chemin[n].y = extra.y;
                            extra.x = extra.x + block4.x;
                            extra.y = calc_y( path[i-1], path[i], extra.x );
                            n++;
                            }
                        chemin[n].x = path[i].x;
                        chemin[n].y = path[i].y;
                        n++;
                        }
                    else
                        {
                        while ( extra.x > path[i].x )
                            {
                            chemin[n].x = extra.x;
                            chemin[n].y = extra.y;
                            extra.x = extra.x - block4.x;
                            extra.y = calc_y( path[i-1], path[i], extra.x );
                            n++;
                            }
                        chemin[n].x = path[i].x;
                        chemin[n].y = path[i].y;
                        n++;
                        }
                    }
                else
                    {
                    extra.x = path[i-1].x;
                    extra.y = path[i-1].y;
                    if ( path[i].y > path[i-1].y )
                        {
                        while ( extra.y < path[i].y )
                            {
                            chemin[n].x = extra.x;
                            chemin[n].y = extra.y;
                            extra.y = extra.y + block4.y;
                            extra.x = calc_x( path[i-1], path[i], extra.y );
                            n++;
                            }
                        chemin[n].x = path[i].x;
                        chemin[n].y = path[i].y;
                        n++;
                        }
                    else
                        {
                        while ( extra.y > path[i].y )
                            {
                            chemin[n].x = extra.x;
                            chemin[n].y = extra.y;
                            extra.y = extra.y - block4.y;
                            extra.x = calc_x( path[i-1], path[i], extra.y );
                            n++;
                            }
                        chemin[n].x = path[i].x;
                        chemin[n].y = path[i].y;
                        n++;
                        }
                    }
                }
            else
                {
                chemin[n].x = path[i-1].x;
                chemin[n].y = path[i-1].y;
                n++;
                }

            } // end if
        }     // end for
    
    chemin[n].x = path[nin-1].x;
    chemin[n].y = path[nin-1].y;

    if ( (n+1) > *nout ) bail( "bad calculation of n in gpfout/expand_path" );
    
    *nout = n+1;

    return( chemin );
    }

//--------------------------------------------------------------------------------------------------

int
find_closest( dxy fieldcenter, dxy *chemin, int nchemin )
    {
    static int    i, besti;
    static double dx, dy, dist, smallest;

    dx = fieldcenter.x - chemin[0].x;
    dy = fieldcenter.y - chemin[0].y;
    smallest = sqrt( dx * dx + dy * dy );
    besti = 0;
    
    for ( i=1; i < nchemin; i++ )
        {
        dx = fieldcenter.x - chemin[i].x;
        dy = fieldcenter.y - chemin[i].y;
        dist = sqrt( dx * dx + dy * dy );
        if ( dist < smallest )
            {
            smallest = dist;
            besti = i;
            }
        }

    return( besti );
    }


//----------------------------

void
path_sort( int *theway, int numblocks, array_2dq *field, int *rows, int *cols, dxy *fieldcenters, dxy *chemin, int nchemin )
    {
    // Sort the blocks so they follow the path "chemin". Output is "theway".

    int    i, j, k, *closest;
    queue *fieldq;

    closest = (int *) malloc( numblocks * sizeof(int) );
    
    for ( i=0; i < numblocks; i++ )
        {
        fieldq = get_field_list( field, cols[i], rows[i] );
        
	if ( fieldq->len == 0 )
            {
            closest[i] = -1;
            }
        else
            {
            closest[i] = find_closest( fieldcenters[i], chemin, nchemin );
            }
        }

    // For each block n, closest[n] contains the index of the closest point on chemin.
    // We need to walk along the path, picking up blocks and putting them in theway[].
    // This is a dumb way of doing a sort, but, it's simple.
    
    k = 0;
    
    for ( i=0; i < nchemin; i++ )
        {
        for ( j=0; j < numblocks; j++ )   // we keep looking because there might
            {                             // be multiple blocks with the same 
            if ( closest[j] == i )        // closest point
                {
                theway[k] = j;
                k++;
                if ( k >= numblocks ) bail( "gpfout: path sort logic error" );
                }
            }
        }

    if ( k < numblocks ) theway[k] = -1;  // -1 marks the end. the leftover blocks are empty
    }

//==================================================================================================

int main( int argc, char *argv[] )
    {
    char
            *iname,
            *oname,
             pathname[256],
             gpfout[256],
             pattername[256];

    dxy      block,
             sfs,
             goff,
             center,
             fieldcenter,
            *fieldcenters,
             blockllmfr,
            *blockllmfrs,
	     blockmfr,
             ll, ur;

    double
             bss_nm,
             bss_um,
	     bss_mfr,
             mfr,
             sfr,
             sbss,
             subits,
             centre,
             extent[4],
             dbu_um,
             left,
             leftllmfr,
             dx, dy,
	   **dose,
	     mindose,
	     maxdose;

    int
	     layer,
             dt,
	     ht,
             i, j,
             mfbits,
             nrows,
             ncols,
	     nfields,
             nf,
             numpolys,
             r, c,
             row, col,
            *rows,
            *cols,
            *theway,  // block numbers following the path
             numblocks,
             sfbits;

    FILE
             *extents,
             *ofile;

    queue 
             *sub,
             *shape;

    array_2dq
	     *field;

    BOOL
             done,
             traps,
            *jump,
             assign_doses=FALSE,
             autoextent,
             sort_by_path=FALSE;

    queue
             *fieldq;

    polygon  *p;

    dxy      *path, *chemin;

    int      npath, nchemin,
             **msfm;             // msf multiplier



    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "gpfout v%1.1f\n", VERSION );
        exit(0);
        }

    
    
//--- new stuff for binary output

    subits = pow( 2, SUBDACBITS ); 

    gpf_header_t       *header;
    gpf_header_user_t  *user;
    gpf_indexf_t       *field_index;
    gpf_ulong_t         gpfpos;
    gpf_union_t         prims[GPF_PRIM_COUNT];
    gpf_union_t         prims_org[GPF_PRIM_COUNT];
    gpf_ulong_t         pixeltime, pixelcount;
    gpf_uint_t          NrFreqFactors;

    dxy main_res;   // um
    dxy sub_res;    // um

    dxy mainfieldsize;   // um
    dxy subfieldsize;    // um

    double subfieldsize_xmax = MAXSUBFIELD;    // in micron  
    double subfieldsize_ymax = MAXSUBFIELD;    // in micron

    int main_msf = 5;     // resolution in main resolution units
    int sub_msf  = 10;    // resolution in sub resolution units (bss !)

    int fieldmax = 30;     // maximum number of fields 
    int fieldnum = 0;

    header = (gpf_header_t *) malloc( sizeof( gpf_header_t ) );
    user   = (gpf_header_user_t *) malloc( sizeof( gpf_header_user_t ) );

//---

    //printf( "\n\nsize of ulong: %d\n\n", sizeof( gpf_ulong_t ) );
    //exit(0);
    
    if ( argc < 14 )
        {
        printf( "\ngpfout %1.1f\n", VERSION );
        printf( "\nUse: gpfout  infile.gds outfile                           \\ \n" );
        printf(   "             LLx  LLy  URx  URy                           \\ \n" );
        printf(   "             block_size_x                                 \\ \n" );
        printf(   "             block_size_y                                 \\ \n" );                  
        printf(   "             subfield_size_x                              \\ \n" );
        printf(   "             subfield_size_y                              \\ \n" );
        printf(   "             beam_step_size_nm                            \\ \n" );
        printf(   "             main_field_lsb_um                            \\ \n" );
        printf(   "             voltage_kv                                   \\ \n" );
        printf(   "             [file of relative doses | nodose]            \\ \n" );
        printf(   "             [path file name | nopath]                    \\ \n" );
        printf(   "             [traps]          \n\n"                              );
        printf(   "     All units are microns unless otherwise noted.\n" );
        printf(   "     This program should be used after 'slice'.\n" );
        printf(   "     The output file outfile.gpf will be created.\n\n" );
        printf(   "     GPF shapes will be polygons, unless the option 'traps' \n" );
        printf(   "     is used to generate gpf parallelograms.\n\n" );
        printf(   "     If a path file is specified then the blocks will\n" );
        printf(   "     be sorted by their proximity to this path. The \n" );
        printf(   "     format of the path file is the same that used by\n" );
        printf(   "     gpfsort. See 'extractpath'. Path file format:\n\n" );
        printf(   "     lower-left nnn nnn              (optional, ignored)\n" );
        printf(   "     upper-right nnn nnn             (optional, ignored)\n" );
        printf(   "     nnnn nnnn                       (coordinate from gds file)\n" );
        printf(   "     nnnn nnnn            \n" );
        printf(   "     jump                            (to ignore the path up to...)\n" );
        printf(   "     nnnn nnnn  \n" );
        printf(   "     nnnn nnnn  \n\n" );
        printf(   "     The file of relative doses has the format\n\n" );
        printf(   "     layer  datatype  relative-dose  [beam-step_nm]\n" );
        printf(   "     layer  datatype  relative-dose  [beam-step_nm]\n" );
        printf(   "     etc\n\n" );
        printf(   "     where 'layer' and 'dataype' could be numbers 0..255 or '*'. \n" );
        printf(   "     Beam-step values, in nm, must be multiples of the beam step\n" );
        printf(   "     given on the command line. These beam step values are optional.\n" );
        printf(   "     Relative dose values must normally be greater than 0.5, but note that\n" );
        printf(   "     *clock* values will be scaled by the square of (beam_step/base_beam_step).\n" );
        printf(   "     Therefore you can (for example) use a relative dose of 0.2 if the  \n" );
        printf(   "     beam step is 2 times larger than the base beam step (the beam step\n" );
        printf(   "     on the command line.)\n\n" );
        exit(0);
        }

    iname = argv[1];
    oname = argv[2];

    sscanf( argv[3],  "%lf", &ll.x );   
    sscanf( argv[4],  "%lf", &ll.y );   
    sscanf( argv[5],  "%lf", &ur.x );   
    sscanf( argv[6],  "%lf", &ur.y );   
    sscanf( argv[7],  "%lf", &block.x );    
    sscanf( argv[8],  "%lf", &block.y );    
    sscanf( argv[9],  "%lf", &sfs.x );    
    sscanf( argv[10], "%lf", &sfs.y );    
    sscanf( argv[11], "%lf", &bss_nm );
    sscanf( argv[12], "%lf", &mfr );
    sscanf( argv[13], "%d",  &ht );

    autoextent = ll.x == 0.0 && ll.y == 0.0 && ur.x == 0.0 && ur.y == 0.0;

    mindose = 1.0;
    maxdose = 1.0;

    dose = doublematrix( 0, 255, 0, 255 );
    msfm = i_matrix( 0, 255, 0, 255 );

    if ( argc > 14 && strcmp( argv[14], "nodose" ) != 0 ) 
        assign_doses = read_dose_file( FALSE, argv[14], dose, msfm, &mindose, &maxdose, bss_nm );
    else
        assign_doses = read_dose_file( TRUE, argv[14], dose, msfm, &mindose, &maxdose, bss_nm ); // init only

    pathname[0] = 0;
    
    if ( argc > 15 && ! (strcmp( argv[15], "nopath") == 0) )
        {
        strcpy( pathname, argv[15] );
        read_path( pathname, &npath, &path, &jump );
        sort_by_path = TRUE;
        
        for ( i=0; i < npath; i++ )
            {
            if ( jump[i] ) printf( "jump\n" );
            printf( "[%2d] %7.2lf  %7.2lf \n", i, path[i].x, path[i].y );
            }

        chemin = expand_path( npath, &nchemin, path, jump, block );

        //printf( "\nexpanded:\n\n" );
        //
        //for ( i=0; i < nchemin; i++ )
        //    {
        //    printf( "[%2d] %7.2lf  %7.2lf \n", i, chemin[i].x, chemin[i].y );
        //    }
            
        }

    traps = ( argc > 16 && (strcmp( argv[16], "traps") == 0 ) );

    mfdefs = 0;
    nout   = 0;

    sprintf( gpfout, "%s.gpf", oname );

    bss_um = bss_nm / 1000.0;

    if ( autoextent )
        {
        extents = fopen( "tmp/extents.dat", "r" );
        if ( ! extents ) bail( "unable to open tmp/extents.dat for reading" );
        fscanf( extents, "%lf %lf %lf %lf",  &ll.x, &ll.y, &ur.x, &ur.y );
        fclose( extents );
        }

    // we used to snap to 10nm, but that seems wrong.
    // what if we are using a very small lsb and a weird subfield size?
    // now we snape to angstroms.

    snapointo( &ll,    0.0001 );
    snapointo( &ur,    0.0001 );
    snapointo( &block, 0.0001 );
    snapointo( &sfs,   0.0001 );

    dx = ur.x - ll.x;
    dy = ur.y - ll.y;

    if ( block.x > dx ) block.x = dx;
    if ( block.y > dy ) block.y = dy;

    printf( "\n" );
    printf(   "Lower left:       (%1.2lf, %1.2lf) um\n", ll.x, ll.y );
    printf(   "Upper right:      (%1.2lf, %1.2lf) um\n", ur.x, ur.y );
    printf(   "Block size:       (%1.2lf, %1.2lf) um\n", block.x, block.y  );
    printf(   "Subfield:         (%1.3lf, %1.3lf) um\n", sfs.x, sfs.y  );
    printf(   "Beam step:        %1.1lf nm\n", bss_nm );
    printf(   "Beam voltage:     %1d kV\n", ht );

    strcpy( pattername, oname );

    mfbits = MAINDACBITS;                 // number of main field bits
    sfbits = SUBDACBITS;                  // number of subfield bits

    printf(   "Main field LSB:   %1.5lf um \n", mfr );

    sfr = ( mfr / 2.0 );      // equal to the pss

    printf(   "Subfield LSB:     %1.5lf um \n", sfr );

    // subfield beam step size, aka msf, in sfr units (e.g. 10 sfr = 5nm), usually 2*bss

    sbss = floor( bss_um / sfr ); 

    centre = pow( 2, MAINDACBITS-1 ); // mfbits-1 );   // 524288   field center in units of main field bits

    // beam step size in mainfield resolution units (aka 'msf') is usually half of the sbss

    bss = floor( sbss * sfr / mfr );      
    printf(   "Main field msf:   %1.0lf\n", bss );

    if ( pathname ) printf(   "Path file:        %s\n", pathname );

    // "number of fields" seems to matter

    ncols = 1 + (int) floor( fabs(ur.x - ll.x)/block.x );
    nrows = 1 + (int) floor( fabs(ur.y - ll.y)/block.y );

    nf = nrows * ncols;                // number of fields, upper limit

    printf( "\nConverting %s to %s...\n", iname, gpfout ); 

    // 'shape' is the master list of all trapezia in this pattern
    // 'extent' is a short list [minx, miny, maxx, maxy]

    shape = initialize_queue();

    printf( "\nReading...\n" ); fflush( stdout );

    read_input_polys( iname, shape, extent, &dbu_um, &numpolys, bss_um );      

    printf( "Pattern extent, um:  (%1.3lf,%1.3lf) to (%1.3lf,%1.3lf)\n", extent[0], extent[1], extent[2], extent[3] );

    printf( "Output boundary, um: (%1.3lf,%1.3lf) to (%1.3lf,%1.3lf)\n", ll.x, ll.y, ur.x, ur.y );

    printf( "Database unit, um:   %1.3e\n\n", dbu_um );

    //printf( "\nFULL QUEUE:\n" );
    //print_queue( shape );
    //printf( "\n-----------\n" );

    extent[0] += bss_um;
    extent[1] += bss_um;
    extent[2] -= bss_um;
    extent[3] -= bss_um;
 
    if ( ll.x > extent[0] || ll.y > extent[1] || ur.x < extent[2] || ur.y < extent[3] )
        bail( "The pattern falls outside the specified extents. Cropping cannot be done here." );


    //--- name changes for binary output

    main_res.x        = mfr;      // eg 0.001 um
    main_res.y        = mfr;
    sub_res.x         = sfr;      // eg 0.0005 um
    sub_res.y         = sfr;

    mainfieldsize.x   = block.x;  // eg 500.0 um
    mainfieldsize.y   = block.y;
    subfieldsize.x    = sfs.x;    // eg 4.0 um
    subfieldsize.y    = sfs.y;

    main_msf = (int) bss;         // resolution in main resolution units, eg 5
    sub_msf  = 2 * main_msf;      // resolution in sub resolution units, eg 10

    fieldmax = nf;                // maximum number of fields 
    fieldnum = 0;                 // will be incremented after each block

                                  // if fieldmax is wrong then this next malloc will
                                  // screw the program randomly. C is super-bad.

    field_index = (gpf_indexf_t *) malloc(fieldmax * sizeof(gpf_indexf_t));  

    // gtx header is written first
    // gpf header is written last, but at the top of the output file. so weird!

    
    init_headers(  pattername, 
                   header, 
                   mainfieldsize.x, 
                   mainfieldsize.y, 
  	           main_res.x, 
	           main_res.y, 
	           sub_res.x, 
	           sub_res.y, 
                   main_msf,
                   sub_msf,
                   fieldmax,
                   subfieldsize_xmax,
                   subfieldsize_ymax,
                   user,
		   &gpfpos,
                   field_index,
		   ht,
                   assign_doses,
                   mindose,
                   maxdose,
                   ncols,
                   nrows,
                   traps   );
    
    printf( "\nSorting into fields...\n" ); fflush(stdout);
    field = field_sort( shape, ll, ur, block, dbu_um, nrows, ncols );

    // count the number of populated fields

    nfields = 0;
    for ( row=0; row < nrows; row++ )
	{
	for ( col=0; col < ncols; col++ )
	    {
	    fieldq = get_field_list( field, col, row );
	    if ( fieldq->len > 0 ) nfields++;
	    }
	}

    // ofile = gpf_file_open( NULL, gpfout, "w+" );
    
    ofile = gpfree_file_open( gpfout, "w+" );

    if ( ofile == NULL) bail( "Can't open gpf file for writing\n" );

    center.x = (ur.x + ll.x) / 2.0;   // pattern center, um
    center.y = (ur.y + ll.y) / 2.0; 

    fieldcenter.x = ll.x + block.x / 2.0;  // lower-left field center, um  
    fieldcenter.y = ll.y + block.y / 2.0; 

    printf( "\ncenter:             %g, %g um\n", center.x, center.y );

    left         = fieldcenter.x;
    blockllmfr.x = round( ll.x / mfr ); 
    blockllmfr.y = round( ll.y / mfr );  
    leftllmfr    = blockllmfr.x;
    blockmfr.x   = round( block.x / mfr );
    blockmfr.y   = round( block.y / mfr );

    snapointo( &blockllmfr, 0.0001 );    // used to be 10nm
    snapointo( &blockmfr,   0.0001 );    // but these numbers are already rounded so this is silly 

    pixeltime = 0;
    printf( "lower left:         %g, %g um\n\n", blockllmfr.x * mfr, blockllmfr.y * mfr );

    numblocks = nrows * ncols;
    
    fieldcenters = (dxy *) malloc( numblocks * sizeof(dxy) );
    blockllmfrs  = (dxy *) malloc( numblocks * sizeof(dxy) );
    rows         = (int *) malloc( numblocks * sizeof(int) );
    cols         = (int *) malloc( numblocks * sizeof(int) );
    theway       = (int *) malloc( numblocks * sizeof(int) );
    
    //--------- first collect the field coordinates etc so that we can sort them later

    i = 0;
    
    for ( row=0; row < nrows; row++ )
        {
        for ( col=0; col < ncols; col++ )
            {
            rows[i] = row;
            cols[i] = col;
            fieldcenters[i].x = fieldcenter.x;
            fieldcenters[i].y = fieldcenter.y;
            blockllmfrs[i].x  = blockllmfr.x;
            blockllmfrs[i].y  = blockllmfr.y;
            theway[i] = i;                      // for now... maybe sort later
            i++;

            fieldcenter.x = fieldcenter.x + block.x; 
            blockllmfr.x  = blockllmfr.x  + blockmfr.x;
            }
        fieldcenter.x = left;
        fieldcenter.y = fieldcenter.y + block.y;
        blockllmfr.x  = leftllmfr;
        blockllmfr.y  = blockllmfr.y + blockmfr.y;
        }

    
    if ( sort_by_path ) path_sort( theway, numblocks, field, rows, cols, fieldcenters, chemin, nchemin );    

    //---------  write out the blocks in the order given in theway[]

    done = FALSE;

    for ( i=0; i < numblocks && ! done; i++ )
        {
        j = theway[i];    // this is the way
        done = (j < 0);

        if ( ! done )
            {
            row = rows[j];
            col = cols[j];

            fieldcenter.x = fieldcenters[j].x;
            fieldcenter.y = fieldcenters[j].y;
            blockllmfr.x  = blockllmfrs[j].x;
            blockllmfr.y  = blockllmfrs[j].y;
        
            fieldq = get_field_list( field, col, row );

            if ( fieldq->len > 1000 )
                {
                printf( "\015  Block (%4d,%4d)    Field  %d  of  %d        ", row, col, fieldnum+1, nfields );
                fflush( stdout );
                }

            write_field( ofile, traps, blockllmfr, fieldcenter, center, col, row, field, 
                         sbss, sfs, block, dbu_um, mfr, sfr, centre, subits, bss_um, 
                         assign_doses, dose, msfm, mindose, maxdose,
                         prims, prims_org, &gpfpos, field_index, &fieldnum, &pixelcount );
            }
        }


    header->IndexOffset = gpfpos;

    if ( assign_doses )     // the doses we actually used were marked by flipping the sign
	{                   // of dose[][]. cute :)
        NrFreqFactors = 0;  
        for ( layer=0; layer < 255; layer++ )
	    for ( dt=0; dt < 255; dt++ )
		if ( dose[layer][dt] < 0 ) NrFreqFactors++;
        if ( NrFreqFactors == 0 ) NrFreqFactors = 1;
	}
    else
	NrFreqFactors = 1;

    header->NrFreqFactors = NrFreqFactors;

    //printf( "\nNumber of fields: %d\n", fieldnum );

    for ( i = 0; i < fieldnum; i++)
        {
        // gpf_write_indexf( ofile, gpfpos, 0, &field_index[i] );
        gpfree_write_indexf( ofile, gpfpos, &field_index[i] );
        gpfpos += sizeof( gpf_indexf_t );
        }

    header->FileSize     = gpfpos;
    header->PixelTime    = pixeltime;
    
    header->NrMainFields = fieldnum;    // interesting refinement - might prevent crashing cview

    gpfree_write_header( ofile, header, user );

    gpfree_file_close( ofile );  


    float score;
    if ( shape->len > 0 )
	score = 100.0 * mfdefs / nout;
    else
	score = 0;

    printf( "\n\nNumber of shapes:     %d\n",  shape->len );
    //printf(     "Extra mfdef:          %1.1f%s\n", score, "%" ); 
    printf(     "Output file:          %s\n\n", gpfout );

    }

