#include <stdio.h>
#include "gpf_pub.h"
#include "polyq.h"
#include "libgds.h"
#include "libgpfree.h"

#define MAXVERTEX    512

#define TRUE  1
#define FALSE 0
#define BOOL  unsigned char
#define ERROR_LAYER 255
#define NUMDOSELAYERS 254

#define VERSION 2.2

int total=0;

void 
usage()
    {
    printf( "\ngpoly2gds v%1.1f\n", VERSION );
    printf( "\nUsage: gpoly2gds infile.gpf outfile.gds [lower_left x y (um)] [upper-right x y (um)] \n" );
    printf( "\nConvert gpf polygons to gds polygons. Just polygons, nothing else. \n");
    printf( "Include the lower-left and upper-right points if you want to recreate \n" );
    printf( "the original design coordinates. Doses are mapped onto layers 0..%d.\n", NUMDOSELAYERS-1 );
    printf( "Invalid polygons and field boundaries will appear on layer %d.\n", ERROR_LAYER );
    printf( "Output gds database unit will be equal to the input gpf subfield resolution.\n\n" ); 
    exit(0);
    }


//----------------------------------------------------------------------------------------

void
flag( struct gds_itemtype *g, dxy ll, dxy ur, dxy offset, char *message, double dbu_um )
    {
    static dxy d, size;
    d.x = g->x[0] * dbu_um; 
    d.y = g->y[0] * dbu_um;
    size.x = ur.x - ll.x;
    size.y = ur.y - ll.y;
    printf( "\n ERROR: %s near %1.3lf, %1.3lf um  size %1.4lf, %1.4lf um  offset %1.3lf, %1.3lf um", 
	    message, d.x, d.y, size.x, size.y, offset.x, offset.y );
    g->layer = ERROR_LAYER;
    }

//----------------------------------------------------------------------------------------

double
clock2dose( gpf_uint_t freq_msw, gpf_uint_t freq_lsw )
    {
    static unsigned long horloge;
    horloge = (freq_msw << 16) + freq_lsw;
    return( 2147483648.0 / horloge );
    }

double
oneclock2dose( gpf_uint_t freq )
    {
    return( 2147483648.0 / freq );
    }

int
dose2layer( double dose, double mindose, double maxdose )
    {
    return( (int) round( NUMDOSELAYERS * (dose - mindose) / (maxdose - mindose) ) );
    }


//----------------------------------------------------------------------------------------

void
write_gds_polygon( int out, struct gds_itemtype *poly, polygon *raw, dxy subfs, dxy offset, double dbu_um  )
    {
    static xy  ur, ll;
    static int i, oldlayer;
    static dxy dur, dll;

    oldlayer = poly->layer;
    ur.x = poly->x[0];
    ur.y = poly->y[0];
    ll.x = ur.x;
    ll.y = ur.y;

    for ( i=1; i < poly->n; i++ )
	{
	if ( poly->x[i] < ll.x ) ll.x = poly->x[i];
        if ( poly->y[i] < ll.y ) ll.y = poly->y[i];
	if ( poly->x[i] > ur.x ) ur.x = poly->x[i];
        if ( poly->y[i] > ur.y ) ur.y = poly->y[i];
	}
    
    dll.x = ll.x * dbu_um;
    dll.y = ll.y * dbu_um;
    dur.x = ur.x * dbu_um;
    dur.y = ur.y * dbu_um;
    
    // The polygon might become degenerate after all this floating point math,
    // so we should check the raw numbers instead.

    if ( raw->ll.x == raw->ur.x )   flag( poly, dll, dur, offset, "polygon is degenerate in x", dbu_um );
    if ( raw->ll.y == raw->ur.y )   flag( poly, dll, dur, offset, "polygon is degenerate in y", dbu_um );

    if ( offset.x < 0.0 ) flag( poly, dll, dur, offset, "offset x < 0", dbu_um );
    if ( offset.y < 0.0 ) flag( poly, dll, dur, offset, "offset y < 0", dbu_um );

    if ( (dur.x - dll.x + offset.x) > subfs.x ) flag( poly, dll, dur, offset, "polygon exceeds x subfield size", dbu_um );
    if ( (dur.y - dll.y + offset.y) > subfs.y ) flag( poly, dll, dur, offset, "polygon exceeds y subfield size", dbu_um );

    gds_write_boundary( out );
    gds_write_layer(    out, poly->layer );
    gds_write_datatype( out, poly->dt    );
    gds_write_xy(       out, poly->x, poly->y, poly->n );
    gds_write_endel(    out );
    poly->layer = oldlayer;                // because it might have been flagged
    total++;
    }

//----------------------------------------------------------------------------------------

void
write_block_outline( int out, dxy stage, dxy block, dxy shift, int fieldnum, double dbu_um )
    {
    static BOOL init=FALSE;
    static struct gds_itemtype *p;
    static int labelx, labely, size;
    static char sfieldnum[100];

    if ( ! init )
	{
	init = TRUE;
        p = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        p->x = (int *) malloc( 4 * sizeof( int ) ); 
        p->y = (int *) malloc( 4 * sizeof( int ) );
        p->layer = ERROR_LAYER;
        p->dt    = 0;
        p->n     = 4;
        }

    p->x[0] = (int) round((shift.x + stage.x - block.x / 2.0) / dbu_um );
    p->y[0] = (int) round((shift.y + stage.y - block.y / 2.0) / dbu_um );
    p->x[1] = (int) round((shift.x + stage.x + block.x / 2.0) / dbu_um );
    p->y[1] = (int) round((shift.y + stage.y - block.y / 2.0) / dbu_um );
    p->x[2] = (int) round((shift.x + stage.x + block.x / 2.0) / dbu_um );
    p->y[2] = (int) round((shift.y + stage.y + block.y / 2.0) / dbu_um );
    p->x[3] = (int) round((shift.x + stage.x - block.x / 2.0) / dbu_um );
    p->y[3] = (int) round((shift.y + stage.y + block.y / 2.0) / dbu_um );
    
    gds_write_boundary( out );
    gds_write_layer(    out, p->layer );
    gds_write_datatype( out, p->dt    );
    gds_write_xy(       out, p->x, p->y, p->n );
    gds_write_endel(    out );
    
    labelx = p->x[3] + (block.x / 20) / dbu_um; 
    labely = p->y[3] - (block.y / 20) / dbu_um;
    size = 0.05 * floor(block.x) / dbu_um;
    sprintf( sfieldnum, "%d", fieldnum );

    gds_create_text( out, sfieldnum, labelx, labely, ERROR_LAYER, size );  

    }

//----------------------------------------------------------------------------------------

void
write_gds_array( int out, struct gds_itemtype *poly, polygon *raw, dxy d1, dxy d2, int n1, int n2, double dbu_um, dxy subfs, dxy offset )
    {
    static int i, j, v, inst, max;
    static xy  id1, id2;
    static struct gds_itemtype *p;
    static BOOL init=FALSE;

    if ( ! init )
	{
	init = TRUE;
        p = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
        p->x = (int *) malloc( MAXVERTEX * sizeof( int ) ); 
        p->y = (int *) malloc( MAXVERTEX * sizeof( int ) );
	}

    p->layer = poly->layer;
    p->dt    = poly->dt;

    id1.x = (int) round( d1.x / dbu_um );
    id1.y = (int) round( d1.y / dbu_um );
    id2.x = (int) round( d2.x / dbu_um );
    id2.y = (int) round( d2.y / dbu_um );

    p->n = poly->n;

    // Diagonal arrays are so stupid
    // and have never been used, ever, 
    // since the beginning of time.
    // And yet here we are:

    max = (n1 + 1) * (n2 + 1);
    i = 0;
    j = 0;
    for ( inst=0; inst < max; inst++ )
	{
        for ( v=0; v < p->n; v++ )
	    {
	    p->x[v] = poly->x[v] + id1.x * i + id2.x * j ;
	    p->y[v] = poly->y[v] + id1.y * i + id2.y * j ;
	    }
        write_gds_polygon( out, p, raw, subfs, offset, dbu_um );	
        i++;
        if ( i > n1 ) { i=0; j++; }
        if ( j > n2 ) j = 0;
        }

    }    // end write_gds_array

//----------------------------------------------------------------------------------------

void
create_layer2dose( double mindose, double maxdose )
    {
    // Write a little python program that lets the user
    // convert layer numbers into relative doses.

    FILE *s;
    char percent[2], open[2], close[2];
    percent[0] = 0x25; percent[1] = 0x0;
    open[0]    = 0x28; open[1]    = 0x0;
    close[0]   = 0x29; close[1]   = 0x0;

    s = fopen( "layer2dose", "w" );
    fprintf( s, "#!/usr/bin/python3 \n" );
    fprintf( s, "import sys \n" );
    fprintf( s, "if len( sys.argv ) < 2 : \n" );
    fprintf( s, "    print( \"\\nuse: layer2dose layer\\n\" ) \n" );
    fprintf( s, "    sys.exit() \n" );
    fprintf( s, "layer = int( sys.argv[1] ) \n" );
    fprintf( s, "numlayers = %d \n", NUMDOSELAYERS );
    fprintf( s, "maxdose   = %lf \n", maxdose );
    fprintf( s, "mindose   = %lf \n", mindose );
    fprintf( s, "dose = mindose  + layer * ( maxdose - mindose ) / numlayers \n" );

    fprintf( s, "print( \"\\nLayer %sd = relative dose %s1.2f\\n\" %s %slayer, dose%s ) \n",
	    percent, percent, percent, open, close );

    fclose( s );
    system( "chmod a+x layer2dose" );
    }

//----------------------------------------------------------------------------------------

int main( int argc, char *argv[] )
    {

    FILE *in, *ext;
    int   out;

    gpf_header_t             *header;
    gpf_header_user_t        *user;
    gpf_ulong_t              *pos;
    gpf_uint_t                freq_msw, freq_lsw;
    gpf_union_t              *glob;
    gpf_indexf_t             *field_index;
    gpf_primitive_24_t       *thing;
    gpf_primitive_16_t       *thing2;
    gpf_primitive_coord_t    *vertex;
    gpf_primitive_activate_t *activate;
    struct gds_itemtype      *gdspoly;
    polygon                  *raw;
    double                    mainmid, dbu_um, dbu_uu, mindose, maxdose, dose;
    float                     dbu_um_f, dbu_uu_f;
    long                      shapecnt;
    dxy                       sfr, mfr, stage, main, offset, vert, d1, d2;
    dxy                       block, msf, smsf, subfs, ll, ur, shift;
    int                       n1, n2, blocknum;
    int                       i, v, p, numprims, main_bits, sub_bits, numv;
    char                     *infile, *outfile;


    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "gpoly2gds v%1.1f\n", VERSION );
        exit(0);
        }

    
    if ( argc < 3 ) usage();

    infile  = argv[1];
    outfile = argv[2];

    ll.x = 0;
    ll.y = 0;
    ur.x = 0;
    ur.y = 0;
    shift.x = 0;
    shift.y = 0;

    if ( argc == 7 )
	{
	sscanf( argv[3], "%lf", &ll.x );
	sscanf( argv[4], "%lf", &ll.y );
	sscanf( argv[5], "%lf", &ur.x );
	sscanf( argv[6], "%lf", &ur.y );
        shift.x = (ur.x + ll.x) / 2.0;
        shift.y = (ur.y + ll.y) / 2.0;
	}

    if ( ll.x == 0 && ll.y == 0 && ur.x == 0 && ur.y == 0 )
	{
	ext = fopen( "tmp/extents.dat", "r" );
	if ( ext )
	    {
	    fscanf( ext, "%lf %lf %lf %lf", &ll.x, &ll.y, &ur.x, &ur.y );
	    printf( "\nExtents from tmp/extents.dat: %1.3lf, %1.3lf  to  %1.3lf, %1.3lf \n", ll.x, ll.y, ur.x, ur.y );
	    }
	}

    raw = initialize_polygon( MAXVERTEX, 0, 0, 0 );   // used to store unprocessed ints, so we can look for degens

    in = gpfree_file_open( infile, "r" );
    

    printf( "\nInput  gpf file: %s\n", infile );
    printf(   "Output gds file: %s\n\n", outfile );

    header   = (gpf_header_t *)       malloc( sizeof( gpf_header_t ) );
    user     = (gpf_header_user_t *)  malloc( sizeof( gpf_header_user_t ) );
    pos      = (gpf_ulong_t *)        malloc( sizeof( gpf_ulong_t ) );
    glob     = (gpf_union_t *)        malloc( sizeof( gpf_union_t ) );

    thing    = (gpf_primitive_24_t *)       glob;
    thing2   = (gpf_primitive_16_t *)       glob;
    vertex   = (gpf_primitive_coord_t *)    glob;
    activate = (gpf_primitive_activate_t *) glob;

    gpfree_read_header( in, header, user );

    //gpf_print_header( stdout, header, user );

    sfr.x     = header->SubFieldResolutionX;      // usually 0.0005 um
    sfr.y     = header->SubFieldResolutionY;
    mfr.x     = header->MainFieldResolutionX;     // usually 0.001 um
    mfr.y     = header->MainFieldResolutionY;
    main_bits = header->NrMainFieldBits;
    sub_bits  = header->NrSubFieldBits;
    msf.x     = header->ResolutionX;              // eg 5
    msf.y     = header->ResolutionY;           
    smsf.x    = header->BeamStepSizeX;            // eg 10
    smsf.y    = header->BeamStepSizeY;
    block.x   = header->MainFieldSizeX * mfr.x * msf.x;   // um
    block.y   = header->MainFieldSizeY * mfr.y * msf.y;   // um
    subfs.x   = header->SubFieldSizeX  * sfr.x * smsf.x;  // um
    subfs.y   = header->SubFieldSizeY  * sfr.y * smsf.y;  // um

    dbu_um = sfr.x;     // might be 0.5 nm or even less
    dbu_uu = dbu_um;    // this makes the cad display um units 

    out = open( outfile, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( out <= 0 ) bail( "unable to open output gds file" );

    gds_create_lib_correctly( out, "gpf", dbu_um, dbu_uu );
    gds_write_bgnstr( out );
    gds_write_strname( out, "gpf" );
    gdspoly = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    gdspoly->x = (int *) malloc( MAXVERTEX * sizeof( int ) ); 
    gdspoly->y = (int *) malloc( MAXVERTEX * sizeof( int ) );
    gdspoly->layer = 0;
    gdspoly->dt    = 0;

    mindose = 1.0;
    maxdose = 1.0;

    if ( header->MaxFreqFactor > 0 )
        mindose   = oneclock2dose( header->MaxFreqFactor );
    if ( header->MinFreqFactor > 0 )	
        maxdose   = oneclock2dose( header->MinFreqFactor );

    mainmid   = mfr.x * pow( 2, main_bits - 1 );          // usually 524.288 um

    printf( "\nMain field size (um):       %1.3lf, %1.3lf", block.x, block.y );
    printf( "\nMax subfield size (um):     %1.3lf, %1.3lf", subfs.x, subfs.y );
    printf( "\nMain field resolution (um): %1.5lf", mfr.x );
    printf( "\nSub  field resolution (um): %1.5lf", sfr.x );
    printf( "\nMSF:                        %1.0lf", msf.x );
    printf( "\nNumber of main DAC bits:    %d", main_bits );
    printf( "\nMinimum relative dose:      %1.2lf", mindose );
    printf( "\nMaximum relative dose:      %1.2lf", maxdose );

    if ( argc == 7 )
	{
	printf( "\nLower left  (um):           %1.3lf, %1.3lf",   ll.x, ll.y );
	printf( "\nUpper right (um):           %1.3lf, %1.3lf\n", ur.x, ur.y );
	}
    else
	printf( "\n" );

    field_index = (gpf_indexf_t *) malloc( header->NrMainFields * sizeof(gpf_indexf_t));

    freq_msw = 0x8000;
    freq_lsw = 0;

    blocknum = 0; 
    printf( "\n" );

    for (i = 0; i < header->NrMainFields; i++)
        {
        printf( "\015 Field %d  ", i ); fflush( stdout );

        gpfree_read_indexf( in, header->IndexOffset, i, &field_index[i] );

        //gpf_print_indexf( stdout, i, &field_index[i] );

        stage.x = field_index[i].FieldToPatternOffsetX;  // um
        stage.y = field_index[i].FieldToPatternOffsetY;
    
        write_block_outline( out, stage, block, shift, i, dbu_um );

        *pos = field_index[i].StartData;
        numprims = (field_index[i].EndData - field_index[i].StartData) / 4;
        shapecnt = 0;

        main.x   = 0;
        main.y   = 0;
        offset.x = 0;
        offset.y = 0;
        d1.x     = 0;
        d1.y     = 0;
        d2.x     = 0;
        d2.y     = 0;
        n1       = 0;
        n2       = 0;

        for ( p=0; p < numprims; p++ )
	    {
            gpfree_read_union( in, pos, glob );   // glob is aka thing, thing2, vertex, and activate

            if      ( thing->IC == GPF_PRIM_MAIN_X_LSW )         main.x   = thing->val * mfr.x;  // um 
            else if ( thing->IC == GPF_PRIM_MAIN_Y_LSW )         main.y   = thing->val * mfr.y;
            else if ( thing->IC == GPF_PRIM_OFFSET_X_LSW )       offset.x = thing->val * sfr.x;
            else if ( thing->IC == GPF_PRIM_OFFSET_Y_LSW )       offset.y = thing->val * sfr.y;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT1_X_LSW ) d1.x     = thing->val * mfr.x; 
            else if ( thing->IC == GPF_PRIM_MAINREPVECT2_X_LSW ) d2.x     = thing->val * mfr.x;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT1_Y_LSW ) d1.y     = thing->val * mfr.x; 
            else if ( thing->IC == GPF_PRIM_MAINREPVECT2_Y_LSW ) d2.y     = thing->val * mfr.x;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT1_CNT )   n1       = thing->val;
            else if ( thing->IC == GPF_PRIM_MAINREPVECT2_CNT )   n2       = thing->val;
            else if ( thing->IC == GPF_PRIM_MOVESTAGE )          blocknum = blocknum + 1;
            else if ( thing->IC == GPF_PRIM_MSF )           {}
            else if ( thing->IC == GPF_PRIM_FREQUENCY_MSW )      freq_msw = thing->val; 
            else if ( thing->IC == GPF_PRIM_FREQUENCY_LSW )      freq_lsw = thing->val; 
            else if ( thing->IC == GPF_PRIM_POLYGON ) 
		{
		numv = thing2->val2;                         // number of vertices
                if ( numv > MAXVERTEX ) bail( "too many vertices" );
                gdspoly->n = numv;

                dose = clock2dose( freq_msw, freq_lsw );
                gdspoly->layer = dose2layer( dose, mindose, maxdose );

                zero_polygon( raw );                         // the unprocessed ints

                for ( v=0; v < numv; v++ )
		    {
		    gpfree_read_union( in, pos, glob );
		    append_polygon( raw, vertex->x, vertex->y );
		    vert.x = vertex->x * sfr.x;                     // remember, vertex is glob
		    vert.y = vertex->y * sfr.y;
		    vert.x = vert.x + main.x - mainmid + offset.x + stage.x + shift.x;   // um
                    vert.y = vert.y + main.y - mainmid + offset.y + stage.y + shift.y;
                    gdspoly->x[v] = (int) round( vert.x / dbu_um );
                    gdspoly->y[v] = (int) round( vert.y / dbu_um );
		    }
                
		gpfree_read_union( in, pos, glob );

		if ( thing->IC != GPF_PRIM_ACTIVATE ) bail( "bad gpf file: missing activate line" );
                if ( activate->ShapeID != GPF_SHAPE_POLYGON ) bail( "bad gpf file: this is not a polygon" );

                if ( ! (total % 5000) ) printf( "\015 Field %d      polygon %d  ", i, total ); fflush( stdout );

                if ( activate->MainRepVect1 || activate->MainRepVect2 ) 
		    {
		    if ( ! activate->MainRepVect1 ) n1 = 0;
		    if ( ! activate->MainRepVect2 ) n2 = 0;
                    write_gds_array( out, gdspoly, raw, d1, d2, n1, n2, dbu_um, subfs, offset );
		    }
		else
		    {
		    write_gds_polygon( out, gdspoly, raw, subfs, offset, dbu_um );
		    }

                p = p + numv + 1;
		}
            else
		{
		printf( "------------ Unknown primitive: %d\n", thing->IC );
		//gpf_print_union( stdout, glob, &shapecnt );
		}

	    }

        }

    gds_write_endstr( out ); 
    gds_write_endlib( out );    

    fclose( in );
    close( out );

    printf( "\n\nNumber of polygons: %d\n\n", total );

    if ( maxdose > mindose )
	{
        create_layer2dose( mindose, maxdose );
        printf( "Convert layer number to relative dose with ./layer2dose <n> \n\n" );
        }

    }

