//
// Copyright (c) 2024. Raith BV. All rights reserved.
//

#ifndef __GPF_H__
#define __GPF_H__

#ifdef GPG
typedef int          gpf_bool_t;
typedef int          gpf_int_t;
typedef unsigned int gpf_uint_t;
#else

#include <stdint.h>
#include <stdbool.h>

typedef double gpf_double_t;
typedef char gpf_char_t;
typedef bool gpf_bool_t;
typedef int32_t gpf_int_t;
typedef uint32_t gpf_uint_t;
typedef uint64_t gpf_ulong_t;
#endif

#define NOMINAL_FREQUENCY                       (0x80000000)

#define GPF_COPYRIGHT     "Generic Pattern Format, \xA9 RAITH nanofabrication"

// GPF version > 1.44 changes the subfield size to be in units of subfield bits.

#define GPF_VERSIONSTR_GPG                      "1.44"
#define GPF_VERSIONSTR_UPG                      "1.44UPG"

#define GPF_PRIM_MAX                             0xFF
#define GPF_PRIM_COUNT                           (GPF_PRIM_MAX + 1)
#define GPF_PRIM_NOT_USED                        0

#define GPF_PRIM_MAIN_X_LSW                      0x02
#define GPF_PRIM_MAIN_Y_LSW                      0x04
#define GPF_PRIM_BASE_LSW                        0x06
#define GPF_PRIM_HEIGHT_LSW                      0x08
#define GPF_PRIM_OFFSET_X_LSW                    0x0A
#define GPF_PRIM_OFFSET_Y_LSW                    0x0C
#define GPF_PRIM_FREQUENCY_MSW                   0x0D
#define GPF_PRIM_FREQUENCY_LSW                   0x0E
#define GPF_PRIM_MSF                             0x0F
#define GPF_PRIM_SHIFT_LT_LSW                    0x12
#define GPF_PRIM_SHIFT_RB_LSW                    0x14
#define GPF_PRIM_MAINREPVECT1_X_LSW              0x22
#define GPF_PRIM_MAINREPVECT1_Y_LSW              0x24
#define GPF_PRIM_MAINREPVECT2_X_LSW              0x26
#define GPF_PRIM_MAINREPVECT2_Y_LSW              0x28
#define GPF_PRIM_MAINREPVECT1_CNT                0x29
#define GPF_PRIM_MAINREPVECT2_CNT                0x30
#define GPF_PRIM_PATH                            0x4A
#define GPF_PRIM_PATH_CENTER_X_FRAC              0x2A
#define GPF_PRIM_PATH_CENTER_X_LSW               0x2B
#define GPF_PRIM_PATH_CENTER_Y_FRAC              0x2C
#define GPF_PRIM_PATH_CENTER_Y_LSW               0x2D
#define GPF_PRIM_PATH_ANGLE_MSW                  0x4B
#define GPF_PRIM_PATH_ANGLE_LSW                  0x4C
#define PATH_CLEAR                               0x01
#define PATH_FRACTION                            0x02
#define PATH_RELATIVE                            0x04
#define PATH_BBON                                0x08
#define PATH_FREQUENCY                           0x10
#define PATH_EXELCOUNT                           0x20
#define GPF_PRIM_POLYGON                         0x4D
#define GPF_PRIM_POLYGON_CENTER_X_FRAC           0x2E
#define GPF_PRIM_POLYGON_CENTER_X_LSW            0x2F
#define GPF_PRIM_POLYGON_CENTER_Y_FRAC           0x3E
#define GPF_PRIM_POLYGON_CENTER_Y_LSW            0x3F
#define GPF_PRIM_POLYGON_ANGLE_MSW               0x4E
#define GPF_PRIM_POLYGON_ANGLE_LSW               0x4F
#define POLYGON_CLEAR                            0x01
#define POLYGON_FRACTION                         0x02
#define POLYGON_RELATIVE                         0x04
#define POLYGON_REDISTRIBUTE                     0x08
#define POLYGON_OUTLINE                          0x10
#define GPF_PRIM_ACTIVATE                        0xF0
#define GPF_PRIM_MOVESTAGE                       0xF2

#define GPF_SHAPE_ITEMS(ITEM) \
  ITEM(UNDEFINED, 0, false) \
  ITEM(TRAPEZIUM, 1 << 0, true) \
  ITEM(BITMAP, 1 << 1, true) \
  ITEM(SEGMENT, 1 << 2, true) \
  ITEM(CIRCLE, 1 << 3, false) \
  ITEM(CUBE, 1 << 4, false) \
  ITEM(REPEAT, 1 << 6, true) \
  ITEM(SEQUENCE, 1 << 5, true) \
  ITEM(STRUCTURE, 1 << 7, true) \
  ITEM(LINE, 1 << 8, false) \
  ITEM(ELLIPSE, 1 << 9, false) \
  ITEM(POLYGON, 1 << 10, false) \
  ITEM(PATH, 1 << 11, false) \
  ITEM(MERGE, 1 << 12, true) \
  ITEM(BIGTRAPEZIUM, 1 << 13, true) \
  ITEM(SPIRALTRAPEZIUM, 1 << 14, false) \
  ITEM(SPIRALSEGMENT, 1 << 15, false) \
  ITEM(RECTANGLE, -1 , false) \
  ITEM(SPIRALRECTANGLE, -1 , false) \
  ITEM(GRADIENTCUBE, 1 << 16, false) \
  ITEM(GRADIENTELLIPSE, 1 << 17, false) \

enum GPF_SHAPES {
#define GPF_SHAPES_ITEM(name, typeVal, gpg) GPF_SHAPE_##name,
  GPF_SHAPE_ITEMS(GPF_SHAPES_ITEM)
};

#define GPF_SHAPE_STRING_ITEM(name, typeVal, gpg) #name,
#define GPF_SHAPE_STRINGS { GPF_SHAPE_ITEMS(GPF_SHAPE_STRING_ITEM) }

enum GPF_SHAPETYPES {
#define GPF_SHAPETYPES_ITEM(name, typeVal, gpg) GPF_SHAPETYPE_##name = typeVal,
  GPF_SHAPE_ITEMS(GPF_SHAPETYPES_ITEM)
};

typedef enum {
  GPF_MAINFIELDPLACEMENT_FLOATING = 1,
} gpf_mainFieldPlacement_t;

#define GPF_SUBFIELDPLACEMENT_LOWERLEFT         2
#define GPF_FRACTURESTYLE_SUBFIELD              0x00000003
#define GPF_FRACTURESTYLE_FILLMODE_SMOOTH  0x2
#define GPF_FRACTURESTYLE_HIGHRESOLUTION        0x10000000
#define GPF_FRACTURESTYLE_OUTLINE               0x80000000

#ifndef GPG


#define GPF_HEADER_SIZE 512
#pragma pack(push, 1)
typedef struct _gpf_header_t {
  gpf_char_t Title[80];
  gpf_char_t Version[8];
  gpf_double_t MainFieldResolutionX;
  gpf_double_t MainFieldResolutionY;
  gpf_double_t SubFieldResolutionX;
  gpf_double_t SubFieldResolutionY;
  gpf_uint_t ResolutionX;
  gpf_uint_t ResolutionY;
  gpf_uint_t BeamStepSizeX;
  gpf_uint_t BeamStepSizeY;
  gpf_ulong_t PatternSizeX;
  gpf_ulong_t PatternSizeY;
  gpf_uint_t MainFieldSizeX;
  gpf_uint_t MainFieldSizeY;
  gpf_uint_t SubFieldSizeX;
  gpf_uint_t SubFieldSizeY;
  gpf_uint_t HighTension;
  gpf_uint_t ShapeTypes;
  gpf_uint_t SubFieldPlacement;
  gpf_uint_t NrMainFields;
  gpf_uint_t MainFieldPlacement;
  gpf_uint_t FractureStyle;
  gpf_uint_t NrMainFieldBits;
  gpf_uint_t NrSubFieldBits;
  gpf_uint_t WorkLevel;
  gpf_uint_t MinFreqFactor;
  gpf_uint_t MaxFreqFactor;
  gpf_uint_t NrFreqFactors;
  gpf_ulong_t PixelTime;
  gpf_ulong_t FileSize;
  gpf_ulong_t HeaderOffset;
  gpf_ulong_t IndexOffset;
  gpf_ulong_t StructureOffset;
  gpf_ulong_t SequenceSize;
  gpf_ulong_t MarkerOffset;
  gpf_ulong_t StructureSize;
  gpf_ulong_t MarkerSize;
  gpf_uint_t RefFreqFactor;
  gpf_ulong_t LargestZipBlockSize;
  gpf_ulong_t DoseMappingOffset;
} gpf_header_t;
#pragma pack(pop)

#define GPF_HEADER_USER_SIZE 512
typedef struct _gpf_header_user_t {
  gpf_char_t PatternName[80];
  gpf_char_t CreateDate[16];
  gpf_char_t CreateTime[16];
  gpf_char_t ConverterName[48];
  gpf_char_t ConverterVersion[48];
  gpf_char_t SourceFormat[48];

} gpf_header_user_t;

typedef struct _gpf_indexf_t {
  gpf_double_t FieldToPatternOffsetX;
  gpf_double_t FieldToPatternOffsetY;
  gpf_double_t FieldToPatternOffsetZ;
  gpf_ulong_t StartData;
  gpf_ulong_t EndData;
  gpf_ulong_t Flags;
} gpf_indexf_t;

#endif

typedef struct _gpf_primitive_24_t {
  gpf_uint_t val: 24;
  gpf_uint_t IC: 8;
} gpf_primitive_24_t;

typedef struct _gpf_primitive_24i_t {
  gpf_int_t val: 24;
  gpf_uint_t IC: 8;
} gpf_primitive_24i_t;

typedef struct _gpf_primitive_16_t {
  gpf_uint_t val2: 16;
  gpf_uint_t val1: 8;
  gpf_uint_t IC: 8;
} gpf_primitive_16_t;

typedef struct _gpf_primitive_8_t {
  gpf_uint_t val3: 8;
  gpf_uint_t val2: 8;
  gpf_uint_t val1: 8;
  gpf_uint_t IC: 8;
} gpf_primitive_8_t;

typedef struct _gpf_primitive_step_t {
  gpf_uint_t len: 16;
  gpf_uint_t inc: 1;
  gpf_uint_t dirx: 1;
  gpf_uint_t beamonsteps: 1;
  gpf_uint_t jmp: 1;
  gpf_uint_t beamonjmp: 1;
  gpf_uint_t updfreq: 1;
  gpf_uint_t maindef: 1;
  gpf_uint_t reljmp: 1;
  gpf_uint_t id: 8;
} gpf_primitive_step_t;

typedef struct _gpf_primitive_step_smooth_t {
  gpf_uint_t len: 16;
  gpf_uint_t diry: 1;
  gpf_uint_t dirx: 1;
  gpf_uint_t beamonsteps: 1;
  gpf_uint_t jmp: 1;
  gpf_uint_t beamonjmp: 1;
  gpf_uint_t updfreq: 1;
  gpf_uint_t maindef: 1;
  gpf_uint_t reljmp: 1;
  gpf_uint_t id: 8;
} gpf_primitive_step_smooth_t;


typedef struct _gpf_primitive_coord_t {
  gpf_int_t x: 16;
  gpf_int_t y: 16;
} gpf_primitive_coord_t;


typedef struct _gpf_primitive_jump_t {
  gpf_int_t x: 15;
  gpf_uint_t beamonjmp: 1;
  gpf_int_t y: 15;
  gpf_uint_t jmp: 1;
} gpf_primitive_jump_t;


typedef struct _gpf_primitive_freq_t {
  gpf_uint_t lsw: 16;
  gpf_uint_t msw: 16;
} gpf_primitive_freq_t;

typedef struct _gpf_primitive_activate_t {
  gpf_uint_t hb: 1;
  gpf_uint_t rblt: 1;
  gpf_uint_t lt90: 1;
  gpf_uint_t rb90: 1;
  gpf_uint_t lt45: 1;
  gpf_uint_t rb45: 1;
  gpf_uint_t ltNeg: 1;
  gpf_uint_t rbNeg: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t StepXB: 1;
  gpf_uint_t StepYH: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t rbClip: 1;
  gpf_uint_t ltClip: 1;
  gpf_uint_t OrientationY: 1;
  gpf_uint_t Merge: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_activate_t;

typedef struct _gpf_primitive_cube_t {
  gpf_uint_t hb: 1;
  gpf_uint_t invert: 1;
  gpf_uint_t hbcube: 1;
  gpf_uint_t insideOut: 1;
  gpf_uint_t SubRepVectLong: 1;
  gpf_uint_t SubRepVectTrans: 1;
  gpf_uint_t MainRepVectLong: 1;
  gpf_uint_t MainRepVectTrans: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t StepXB: 1;
  gpf_uint_t StepYH: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t StepAL: 1;
  gpf_uint_t StepAT: 1;
  gpf_uint_t OrientationY: 1;
  gpf_uint_t Merge: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_cube_t;

typedef struct _gpf_primitive_polygon_t {
  gpf_uint_t outline: 1;
  gpf_uint_t redistribute: 1;
  gpf_uint_t reserved4: 1;
  gpf_uint_t internal: 1;
  gpf_uint_t v2: 1;
  gpf_uint_t reserved6: 1;
  gpf_uint_t reserved7: 1;
  gpf_uint_t reserved8: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t reserved9: 1;
  gpf_uint_t reserved10: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t reserved11: 1;
  gpf_uint_t reserved12: 1;
  gpf_uint_t reserved13: 1;
  gpf_uint_t Merge: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_polygon_t;

typedef struct _gpf_primitive_path_t {
  gpf_uint_t close: 1;
  gpf_uint_t exels: 1;
  gpf_uint_t reserved3: 1;
  gpf_uint_t reserved4: 1;
  gpf_uint_t reserved5: 1;
  gpf_uint_t reserved6: 1;
  gpf_uint_t reserved7: 1;
  gpf_uint_t reserved8: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t reserved9: 1;
  gpf_uint_t reserved10: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t reserved11: 1;
  gpf_uint_t reserved12: 1;
  gpf_uint_t reserved13: 1;
  gpf_uint_t Merge: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_path_t;

typedef struct _gpf_primitive_sequence_t {
  gpf_uint_t reserved1: 1;
  gpf_uint_t reserved2: 1;
  gpf_uint_t reserved3: 1;
  gpf_uint_t reserved4: 1;
  gpf_uint_t reserved5: 1;
  gpf_uint_t reserved6: 1;
  gpf_uint_t reserved7: 1;
  gpf_uint_t reserved8: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t reserved9: 1;
  gpf_uint_t reserved10: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t reserved11: 1;
  gpf_uint_t reserved12: 1;
  gpf_uint_t reserved13: 1;
  gpf_uint_t Merge: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_sequence_t;

typedef struct _gpf_primitive_repeat_t {
  gpf_uint_t reserved1: 1;
  gpf_uint_t reserved2: 1;
  gpf_uint_t reserved3: 1;
  gpf_uint_t reserved4: 1;
  gpf_uint_t reserved5: 1;
  gpf_uint_t reserved6: 1;
  gpf_uint_t reserved7: 1;
  gpf_uint_t reserved8: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t reserved9: 1;
  gpf_uint_t reserved10: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t reserved11: 1;
  gpf_uint_t reserved12: 1;
  gpf_uint_t reserved13: 1;
  gpf_uint_t Merge: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_repeat_t;

typedef struct _gpf_primitive_merge_t {
  gpf_uint_t reserved1: 1;
  gpf_uint_t reserved2: 1;
  gpf_uint_t reserved3: 1;
  gpf_uint_t reserved4: 1;
  gpf_uint_t reserved5: 1;
  gpf_uint_t reserved6: 1;
  gpf_uint_t reserved7: 1;
  gpf_uint_t reserved8: 1;
  gpf_uint_t SubRepVect1: 1;
  gpf_uint_t SubRepVect2: 1;
  gpf_uint_t MainRepVect1: 1;
  gpf_uint_t MainRepVect2: 1;
  gpf_uint_t reserved9: 1;
  gpf_uint_t reserved10: 1;
  gpf_uint_t BeamBlanker: 1;
  gpf_uint_t reserved11: 1;
  gpf_uint_t reserved12: 1;
  gpf_uint_t reserved13: 1;
  gpf_uint_t reserved14: 1;
  gpf_uint_t ShapeID: 5;
  gpf_uint_t IC: 8;
} gpf_primitive_merge_t;

typedef union _gpf_union_t {
  gpf_uint_t data;
  gpf_primitive_24_t prim24;
  gpf_primitive_24i_t prim24i;
  gpf_primitive_16_t prim16;
  gpf_primitive_8_t prim8;
  gpf_primitive_activate_t activate;
  gpf_primitive_polygon_t polygon;
  gpf_primitive_path_t path;
  gpf_primitive_coord_t coord;
  gpf_primitive_sequence_t sequence;
  gpf_primitive_repeat_t repeat;
  gpf_primitive_merge_t merge;
  gpf_primitive_step_t step;
  gpf_primitive_step_smooth_t step_smooth;
  gpf_primitive_jump_t jump;
  gpf_primitive_freq_t freq;
} gpf_union_t;


#endif
