#!/bin/bash
#               Freebeam installation script v2
#
# This script will check that you have 
# gcc, pyqt5, python3, and Layout.
# The Freebeam program set will be
# compiled from source and tested.

# For installation on an ebpg console,
# I suggest that you unpack the tar file in /home/pg/bin,
# creating the directory /home/pg/bin/freebeam.

# For installation on a different RHEL system,
# I suggest that you unpack the tar file in /usr/local/bin,
# creating the directory /usr/local/bin/freebeam.

echo
echo
echo "Installation of Freebeam on a RHEL system."
echo

RHEL=`uname -r | grep -Po ".el\K[^.]"`

echo
if [ $RHEL == "7" ]; then
    echo "RHEL 7"
    RHEL7=1
elif [ $RHEL == "8" ]; then
    echo "RHEL 8"
    RHEL7=0
elif [ $RHEL == "9" ]; then
    echo "RHEL 9"
    RHEL7=0
else
    echo "Is this a Red Hat linux system? I can't tell."
    echo "This installation script works only for "
    echo "RHEL 7 or 8 on an EBPG. Maybe 9. If you are trying "
    echo "to use some other system, you should follow"
    echo "the instructions for manual installation"
    echo "found in the INSTALL_RHEL or INSTALL_UBUNTU file."
    echo "Freebeam installation failed."
    echo 
    exit
fi

echo  


if ! command -v gcc > /dev/null ; then
    echo "The C compiler gcc is not installed."
    echo "You must be using some crazy sort of"
    echo "linux system without gcc. Wow."
    echo "Freebeam installation failed."
    echo
    exit
fi

if ! command -v python3 > /dev/null ; then
    echo "Freebeam requires python3, which is"
    echo "normally installed on an EBPG. "
    echo "If your system is registered"
    echo "with a linux repository, you can "
    echo "install python3 with..."
    echo
    if [ $RHEL7 == "1" ]; then
	echo "yum install -y python3-devel"
        echo "yum install -y python39-scipy"
        echo "pip3 install scipy"
        echo "yum install -y PyQt5"
    else
        echo "dnf install -y python3-devel"
        echo "dnf install -y python39-scipy"
        echo "pip3 install scipy"
        echo "dnf install -y PyQt5"
    fi
    echo
    echo "Freebeam installation failed."
    echo
    exit
fi


if [ -f /home/pg/pg_local  ]; then
    pushd . > /dev/null
    . /home/pg/pg_local
    popd
    echo
fi

echo "Checking python modules..."
echo

if [ `python3 -c "import PyQt5" 2>&1 | grep -c module` != "0" ]; then
    echo "        PyQt5 module is missing."
    echo
    if [ `python3 -c "import PyQt6" 2>&1 | grep -c module` == "0" ]; then
	echo "        You do have PyQt6 but those QT wankers did not make"
	echo "        version 6 compatible with version 5."
	echo
    fi
    echo "        If your system is registered"
    echo "        with a linux repository, you can "
    echo "        install Qt5 with..."
    echo
    if [ $RHEL7 == "1" ]; then
        echo "        yum install -y PyQt5"
    else
        echo "        dnf install -y PyQt5"
    fi
    echo
    echo "        Qt5 is required for freeview (used for viewing gpf patterns),"
    echo "        but this is not essential. You can use cview instead."
    echo
    read -p "Do you want to continue installing Freebeam? (y/n) " answer
    if [ $answer != 'y' ]; then 
	echo
        echo "Ok, exiting..."
        echo "Please install Qt5 and try again."
        echo
        exit
    fi
fi


if [ `python3 -c "import LayoutScript" 2>&1 | grep -c module` != "0" ]; then
    echo "        LayoutScript module is missing."
    echo
    echo "        You can download the free version of Layout from"
    echo "        layouteditor.com. After downloading the rpm file "
    echo "        from layouteditor.com, you can install it with"
    echo
    if [ $RHEL7 == "1" ]; then
	echo "        yum install layout-20230925-RHEL7.x86_64.rpm"
    else
	echo "        dnf install layout-20230925-RHEL8.x86_64.rpm"
    fi
    echo
    echo "        But of course the file name will be different."
    echo "        This presumes that your system is registered with"
    echo "        a linux repository."
    echo 
    echo "        Be sure to insert the line"
    echo
    echo "        export PYTHONPATH=\"\$PYTHONPATH:/opt/layout/python\" "
    echo 
    echo "        at the end of /home/pg/pg_local or in /etc/bashrc "
    echo "        and then set that parameter by 'sourcing' the file with"
    echo 
    echo "        . /home/pg/pg_local     "
    echo "  or    . /etc/bashrc           (the dot and space are important)"
    echo
    echo "        Layout is required for overlap removal,"
    echo "        but this is not essential. Instead, you can "
    echo "        design patterns properly, without overlaps."
    echo
    read -p "Do you want to continue with this installation? (y/n) " answer
    if [ $answer != 'y' ]; then 
	echo
        echo "Ok, exiting..."
        echo "Please install Layout and try again."
        echo
        exit
    fi
fi


here=`pwd`
echo "Current directory: $here"
echo

if [ $here == "/home/pg/bin" ]; then
    cd freebeam
    here="/home/pg/bin/freebeam"
fi

if [ $here == "/usr/local/bin" ]; then
    cd freebeam
    here="/usr/local/bin/freebeam"
fi

if [ $here == "/home/pg/bin/freebeam" ] || [ $here == "/usr/local/bin/freebeam" ]; then
    echo "We are already working in $here "
    echo "which is a good place."
else
    echo 
    echo "This directory, $here"
    echo "is not a good place for the installation."
    echo
    echo "I suggest that you start over. Move the "
    echo "tar file to /home/pg/bin or to /usr/local/bin"
    echo "then unpack it with"
    echo
    echo "        tar xvzf freebeam_distro_blabla.tgz"
    echo "        cd freebeam "
    echo "        ./install.sh"
    echo
    read -p "Do you really want to install Freebeam in $here ? (y/n) > " ans
    if [ $ans == 'y' ] || [ $ans == 'Y' ]; then
        echo
        echo "Ok you weirdo, here we go."
	echo
    else
        echo
	echo "Ok- please move the tar file and try again."
	echo
	exit
    fi
    echo 
fi

if [ ! -f freebeam.sh ] || [ ! -f include/polyq.h ] ; then
    echo
    echo "Some or all of the Freebeam files seem to be missing."
    echo "I suggest that you start over, by moving the "
    echo "tar file to /home/pg/bin or /usr/local/bin, then unpack it with"
    echo "tar xvzf freebeam_distro_blabla.tgz"
    echo
    echo "Or, you could follow the instructions for "
    echo "manual installation found in the file INSTALL_RHEL (or INSTALL_UBUNTU)."
    echo
    echo "Freebeam installation failed."
    echo 
    exit
fi

echo
echo "Checking the library path... "

if [ `echo $LD_LIBRARY_PATH | grep -c $here/lib` != "1" ]; then
    echo
    echo "        adding $here/lib to LD_LIBRARY_PATH..."
    echo
    if [ -f /home/pg/pg_local ]; then
        echo -e "\nexport LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:$here/lib\n" >> /home/pg/pg_local
        echo "        updated pg_local"
    elif [ -f /etc/bashrc ]; then
        echo -e "\nexport LD_LIBRARY_PATH=\$LD_LIBRARY_PATH:$here/lib\n" >> /etc/bashrc
	echo "        updated /etc/bashrc"
    else
	echo
	echo "I can't find either /etc/bashrc or /home/pg/pg_local"
	echo "So this is probably not a Red Hat linux box."
	echo "I'm going to stop now. You need to use a different installation script."
	echo
	exit
    fi
else
    echo "        $here/lib is ok"
fi
echo

echo "Checking the command path... "

if [ `echo $PATH | grep -c $here` != "1" ]; then
    echo
    echo "        adding /home/pg/bin/freebeam to path... "
    echo
    if [ -f /home/pg/pg_local ]; then
        echo -e "\nexport PATH=$here:\$PATH\n" >> /home/pg/pg_local
    elif [ -f /etc/bashrc ]; then
        echo -e "\nexport PATH=$here:\$PATH\n" >> /etc/bashrc
    else
	echo
	echo "I can't find either /etc/bashrc or /home/pg/pg_local"
	echo "So this is probably not a Red Hat linux box."
	echo "I'm going to stop now. You need to use a different installation script."
	echo
	exit
    fi
else
    echo "        $here is ok... "
fi

echo

echo "Compiling C programs..."
echo
echo
echo "-----------------------------------------------------------"

make clean
make
make install

echo "-----------------------------------------------------------"
echo

unset fail

if [ ! -f lib/libgds.a    ]; then echo "lib/libgds.c    failed to compile" ; fail=1 ; fi
if [ ! -f lib/libgdsmin.a ]; then echo "lib/libgdsmin.c failed to compile" ; fail=1 ; fi
if [ ! -f lib/libgds.o    ]; then echo "lib/libgdsmin.c failed to compile" ; fail=1 ; fi
if [ ! -f convexify       ]; then echo "convexify.c     failed to compile" ; fail=1 ; fi
if [ ! -f flatten         ]; then echo "flatten.c       failed to compile" ; fail=1 ; fi
if [ ! -f gpoly2gds       ]; then echo "gpoly2gds.c     failed to compile" ; fail=1 ; fi
if [ ! -f gpoly2txt       ]; then echo "gpoly2txt.c     failed to compile" ; fail=1 ; fi
if [ ! -f slice           ]; then echo "slice.c         failed to compile" ; fail=1 ; fi
if [ ! -f grid_snap       ]; then echo "grid_snap.c     failed to compile" ; fail=1 ; fi
if [ ! -f splitter        ]; then echo "splitter.c      failed to compile" ; fail=1 ; fi
if [ ! -f darn            ]; then echo "darn.c          failed to compile" ; fail=1 ; fi
if [ ! -f gpfout          ]; then echo "gpfout.c        failed to compile" ; fail=1 ; fi

if [ $fail ]; then
    echo
    echo "Compilation was not successful."
    echo "Freebeam installation failed."
    echo
    exit
else
    echo
    echo "Compilation was successful. Hooray!"
    echo
fi

unset fail
echo -n "Checking python programs... "

if [ ! -f check_extents       ]; then echo "check_extents       is missing" ; fail=1 ; fi
if [ ! -f check_subfield_size ]; then echo "check_subfield_size is missing" ; fail=1 ; fi
if [ ! -f flatten_with_layout ]; then echo "flatten_with_layout is missing" ; fail=1 ; fi
if [ ! -f fracture_feeder.py  ]; then echo "fracture_feeder.py  is missing" ; fail=1 ; fi
if [ ! -f freeview            ]; then echo "freeview            is missing" ; fail=1 ; fi
if [ ! -f kill.py             ]; then echo "kill.py             is missing" ; fail=1 ; fi
if [ ! -f remove_overlaps     ]; then echo "remove_overlaps     is missing" ; fail=1 ; fi
if [ ! -f subfield_size       ]; then echo "subfield_size       is missing" ; fail=1 ; fi

if [ $fail ]; then
    echo
    echo "        Some python files are missing."
    echo "        Freebeam installation failed."
    echo 
    exit
else
    echo "ok"
    echo
fi

if [ $here == /home/pg/bin/freebeam ]; then
    chown -R pg /home/pg/bin/freebeam
fi

mail -s "freebeam installed" ebpgfunpak@gmail.com < /dev/null

echo
echo "Freebeam installation complete."

echo
echo
echo "        *    When you convert a pattern, be sure to include the line"
echo "        *    "
echo "        *    polygons=\"no\" "
echo "        *    "
echo "        *    in the parameter script that runs Freebeam,"
echo "        *    if you have the older GPG pattern generator. Or, use the line "
echo "        *    "
echo "        *    polygons=\"yes\" "
echo "        *    "
echo "        *    if you have the newer UPG pattern generator with the Firebird option."
echo "        *    Your system needs to have gpf version 1.43 or higher."
echo
echo
read -p "Press [enter] to continue  " ans
echo

echo
echo
echo "Now I will test Freebeam by becoming the user 'pg'"
echo "and then running freebeam_example.sh"
echo
echo "You could do the same thing by logging in as 'pg' (on an ebpg), then"
echo
echo "        cd "
echo "        . pg_local           (or log out and in again)"
echo "        cd bin/freebeam"
echo "        freebeam_example.sh"
echo
echo "        If successful, this will end by using cview"
echo "        to display a pattern in the shape of the letters \"Yale\""
echo 
echo "Please report bugs to ebpgfunpak@gmail.com"
echo

if ! command -v cview > /dev/null ; then
    echo
    echo "*** cview is not installed. "
    echo "*** We can convert a pattern, but we cannot view it with cview."

    if [ `python3 -c "import PyQt5" 2>&1 | grep -c module` != "0" ]; then
        echo "***"
        echo "*** PyQt5 is also missing, so you cannot use freeview to view patterns."
    else
        echo "***"
        echo "*** You can use freeview instead."
        echo "*** When the test is done, view the pattern with"
        echo "*** freeview yale.gpf"
    fi
    echo
fi

read -p "Are you ready to proceed with the test? (y/n) " answer

if [ $answer != 'y' ]; then
    echo
    echo "Very well, please test Freebeam yourself."
    echo
else
    if [ -f /home/pg/pg_local ]; then
        bash -c ". /home/pg/pg_local ; cd $here ; freebeam_example.sh"
    else
	bash -c ". /etc/bashrc ; cd $here ; freebeam_example.sh"
    fi
    echo
    echo "*******************************************************"
    echo "If cview displayed \"Yale\" then the test was successful."
    echo "*******************************************************"
    echo
fi






 








	
