/* 

GDSLIBMIN 1.3x

    This version does not use realloc for allocating polygon space, since it causes crashes.

    A minimal library of C functions for reading and writing a subset of GDSII Stream format,
    One cell, flat, no arrays, paths, or strings. Polygons only, retaining only layer and 
    datatype. Mag is always 1, and angle is always 0.

    1.1  Add "hit" array to gds_itemtype, to indicate which vertices have been processed.
    1.2  Pad lib or struct name to force even-numbered length strings. 
    1.3  realloc causes problems, so go back to using malloc

Supported tokens:

    HEADER              version number (6)
    BGNLIB              creation date
    LIBNAME             library name
    UNITS               database units
    ENDLIB              end of library
    BGNSTR              begin structure (cell) 
    STRNAME             structure (cell) name
    ENDSTR              end structure (cell)
    BOUNDARY            polygon
    LAYER               layer number
    DATATYPE            datatype number (often used for dose assignment)
    XY                  coordinates for cell placement or polygon verticies
    ENDEL               end of element
    SNAME               structure (cell) name

Token order:

    Library:       Structure:   Elements:

    HEADER         BGNSTR       BOUNDARY     
    BGNLIB         STRNAME      LAYER        
    LIBNAME        <element>    DATATYPE     
    UNITS          ENDSTR       XY           
    <structures>                ENDEL        
    ENDLIB                                   

    

*/

#include "libgdsmin.h"


/*------------------------------------------------------------------------------------------*/
//    UTILITY FUNCTIONS
/*------------------------------------------------------------------------------------------*/


void
gds_swap4bytes( BYTE *four  )
    {
    static BYTE temp;
    #if BYTESWAP
        temp    = four[0];
        four[0] = four[3];
        four[3] = temp;
        temp    = four[1];
        four[1] = four[2];
        four[2] = temp;
    #endif
    }

/*------------------------------------------------------------------------------------------*/

void
gds_swap2bytes( BYTE *two )
    {
    static BYTE temp;
    #if BYTESWAP 
        temp   = two[0];
        two[0] = two[1];
        two[1] = temp;
    #endif
    }

/*------------------------------------------------------------------------------------------*/

void
gds_make_next_item( struct gds_itemtype **ci )
    {

    static struct gds_itemtype
        *current_item;


    current_item = *ci;    
    current_item->nextitem = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );

    if ( ! current_item->nextitem ) BAILOUT( "UNABLE TO ALLOCATE NEXT ITEM" );

    current_item = current_item->nextitem;

    // Each cell has a rotation, magnification and strans, but we do not need
    // to keep a running tally unless we are flattening the pattern.

    current_item->n             = 0;
    current_item->layer         = 0; 
    current_item->dt            = 0;
    current_item->nextitem = NULL;

    *ci = current_item;

    }  // make_next_item


/*------------------------------------------------------------------------------------------*/

double
gds_read_double( int fd )

    /* Real numbers are not represented in IEEE format. A floating point number is    */
    /* made up of three parts: the sign, the exponent, and the mantissa. The value    */
    /* of the number is defined to be (mantissa) (16) (exponent) . If "S" is the      */
    /* sign bit, "E" are exponent bits, and "M" are mantissa bits then an 8-byte      */
    /* real number has the format                                                     */
    /*                                                                                */
    /* SEEEEEEE MMMMMMMM MMMMMMMM MMMMMMMM                                            */
    /* MMMMMMMM MMMMMMMM MMMMMMMM MMMMMMMM                                            */
    /*                                                                                */
    /* The exponent is in "excess 64" notation; that is, the 7-bit field shows a      */
    /* number that is 64 greater than the actual exponent. The mantissa is always     */
    /* a positive fraction greater than or equal to 1/16 and less than 1. For an      */
    /* 8-byte real, the mantissa is in bits 8 to 63. The decimal point of the         */
    /* binary mantissa is just to the left of bit 8. Bit 8 represents the value       */
    /* 1/2, bit 9 represents 1/4, and so on.                                          */
    /*                                                                                */
    /* In order to keep the mantissa in the range of 1/16 to 1, the results of        */
    /* floating point arithmetic are normalized. Normalization is a process whereby   */
    /* the mantissa is shifted left one hex digit at a time until its left four       */
    /* bits represent a non-zero quantity. For every hex digit shifted, the           */
    /* exponent is decreased by one. Since the mantssa is shifted four bits at a      */
    /* time, it is possible for the left three bits of a normalized mantissa to be    */
    /* zero. A zero value is represented by a number with all bits zero. The          */
    /* representation of negative numbers is the same as that of positive numbers,    */
    /* except that the highest order bit is 1, not 0.                                 */

    /* Contributed by Hans Romijn (Raith/Vistec)                                      */

    {
    BYTE
        e,
        m[8];

    int
        i,
        b,
        bit,
        sign,
        bitmask,
        exponent;

    long mant = 0;
    long bitm = 0x80000000000000;

    double
        mantissa,
        number;

    read( fd, &m, 8 );

    if ( m[0] >> 7 )
        sign = -1;
    else
        sign = 1;

    exponent = (m[0] % 128) - 64;

    for ( b = 1; b <8; b++ )
        {
        bitmask   = 0x80;
        for ( bit = 7; bit >= 0; bit-- )
            {
            i = ( m[b] & bitmask ) >> bit;
            if (i)
                mant += bitm ;

            // mantissa = (double)mant/(double)(0x100000000000000);
            // number = sign * mantissa * pow( 16, exponent );
            // printf("%d 0x%16lX: %.24lf %.24lf\n",i,bitm,number,(float)number);

            bitmask = bitmask >> 1 ;
            bitm = bitm >> 1;
            }
        }

    mantissa = (double)mant/(double)(0x100000000000000);

    number = sign * mantissa * pow( 16, exponent );

    return( number );

    }  // read_double


/*------------------------------------------------------------------------------------------*/

double
gds_read_float( int fd )

    /* Read 8 bytes and interpret the number as a wacky GDS float. The format must be */
    /* the legacy of some very ancient General Electric computer, built before the    */
    /* establishment of IEEE floating point format.                                   */

    {
    BYTE 
        e,
        m[8];

    int
        i,
        b,
        bit,
        sign,
        bitmask,
        exponent;

    double
        value,
        mantissa, 
        number;

        
    read( fd, &e, 1 );

    if ( e >> 7 )
        sign = -1;
    else
        sign = 1; 
    
    exponent = (e % 128) - 64;

    read( fd, m, 7 );

    value     = 2.0;                  
    mantissa  = 0.0;

    for ( b = 0; b <= 6; b++ )
        {
        bitmask   = 0x80;
        for ( bit = 7; bit >= 0; bit-- )
            {
            i = ( m[b] & bitmask ) >> bit ;  
            mantissa += i / value ;
            value = value + value ;
            bitmask = bitmask >> 1 ;
            }
        }


    number = sign * mantissa * pow( 16, exponent );

    return( number );

    }  // read_float

/*------------------------------------------------------------------------------------------*/

void
gds_bindump( BYTE x )            // dump one byte in binary format
    {                            // way too clever hack from a forum post, sorry
    int z;
    static char b[9];

    b[0] = '\0';

    for ( z=128; z>0; z >>=1 )
        strcat( b, ((x & z) == z) ? "1" : "0" );

    printf( "%s ", b );        

    }

/*------------------------------------------------------------------------------------------*/

void
gds_write_float( int fd, double x )

    /* Write 8 bytes after converting back to wacky GDS float format. */
    
    {

    unsigned int
        b,
        sign,
        e64,
        bit[56];

    int            // important
        exponent,
        i;

    double
        fexponent,
        mantissa,
        mantita;    

    BYTE
        by,
        stupid[8];


    if ( ! BYTESWAP )
        ATTENTION( "WRITING OF FLOATING POINT ON LITTLE-ENDIAN MACHINES HAS NOT BEEN TESTED" );

    // printf( "\nencoding %g\n", x ); fflush( stdout );

    for ( i=0; i<8; i++ ) stupid[i] = 0;

    if ( x != 0.0 ) 
        {
        
        if ( x < 0.0 )
            sign = 1;
        else
            sign = 0;

        x = fabsf( x ); 

        exponent = 1 + floor( log( x ) / log( 16 ) );
        if ( exponent < -64 ) BAILOUT( "(gds_write_float) A NUMBER IS TOO SMALL TO ENCODE AS A GDS FLOAT" );
        fexponent = exponent;
        e64 = exponent + 64;
        mantissa = x / powf( 16.0, fexponent );
        mantita = mantissa;

        for( i=0; i<=56; i++ )
            {
            bit[i] = floor( pow(2,i+1) * mantita );
            mantita = mantita - bit[i] / pow(2,i+1);
            }

        stupid[0] = pow(2,7) * sign + e64;

        for ( i=0;  i<8;  i++ ) stupid[1] = stupid[1] + bit[i] * pow(2, 7-i);
        for ( i=8;  i<16; i++ ) stupid[2] = stupid[2] + bit[i] * pow(2,15-i);
        for ( i=16; i<24; i++ ) stupid[3] = stupid[3] + bit[i] * pow(2,23-i);
        for ( i=24; i<32; i++ ) stupid[4] = stupid[4] + bit[i] * pow(2,31-i);
        for ( i=32; i<40; i++ ) stupid[5] = stupid[5] + bit[i] * pow(2,39-i);
        for ( i=40; i<48; i++ ) stupid[6] = stupid[6] + bit[i] * pow(2,47-i);
        for ( i=48; i<56; i++ ) stupid[7] = stupid[7] + bit[i] * pow(2,55-i);

        }

    // printf( "\n" );
    for ( i=0; i<8; i++ ) 
        {
        by = stupid[i];
        write( fd, &by, 1 );
        // gds_bindump( by );    
        }            
    // printf( "\n" ); fflush( stdout );


    } // write_float

/*------------------------------------------------------------------------------------------*/
//    TOKEN I/O FUNCTIONS
/*------------------------------------------------------------------------------------------*/

				/* Contains GDS version number- we don't care */

void gds_read_header( int fd, int count, BOOL verbose )
    {
    int i;
    SKIPOVER( fd, count );
    }

/*------------------------------------------------------------------------------------------*/

void
gds_write_header( int fd )
    {
    unsigned short int 
        count,
        si,
        token;

    ssize_t w;

    if ( ! fd ) BAILOUT( "gds_write_header: GDS FILE HAS NOT BEEN OPENED" );

    count = 6;
    gds_swap2bytes( (BYTE *) &count );
    w = write( fd, &count, 2 );
    
    //if ( w <= 0 ) BAILOUT( "UNABLE TO WRITE TO OUTPUT FILE - CHECK OPEN() CALL" );

    token = 0x0002;                // header
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    si = 600;                      // GDSII version 6 
    gds_swap2bytes( (BYTE *) &si );
    write( fd, &si, 2 );          

    }  // write_header

/*------------------------------------------------------------------------------------------*/
                                /* Beginning of library. Allocate the first cell. */
                                /* Ignore the creation date.                      */
void
gds_read_bgnlib( int  fd,                        // file descriptor
                 int  count,                     // bytes in this record (always 4)
                 int *cell_table_size,           // cell table size
                 struct gds_celltype **lib,      // pointer to library (main anchor point)
                 struct gds_celltype ***c,       // pointer to the cell table
                 struct gds_celltype **cc,       // pointer to the current cell 
                 BOOL verbose )
    
    {
    struct gds_celltype *library, **cell, *current_cell;
    int i;

    SKIPOVER( fd, count );
    library = (struct gds_celltype *) malloc( sizeof( struct gds_celltype ) );   /* start of the list */
    current_cell = library;                                              /* top of the list   */
    library->nextcell = NULL;

    if ( ! library ) BAILOUT( "UNABLE TO ALLOCATE LIBRARY POINTER. THAT'S STRANGE." );

    cell = (struct gds_celltype **) malloc( 1025 * sizeof( struct gds_celltype * ) ); /* table of pointers to cells */
    *cell_table_size = 1024;

    if ( ! cell ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR CELL TABLE." );

    *lib = library;
    *c   = cell;
    *cc  = current_cell;

    }  // read_bgnlib

/*------------------------------------------------------------------------------------------*/

void
gds_write_bgnlib( int fd )
    {
    time_t *now;
    struct tm *date;

    short int 
        year,
        month,
        day,
        hour,
        minute,
        second,
        token,
        count;

    now  = (time_t *)    malloc( sizeof( time_t ) );
    date = (struct tm *) malloc( sizeof( struct tm ) );

    time( now );
    date = localtime( now );
    
    count = 0x1C;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0102;                 // BGNLIB
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );         
    year = 1900 + date->tm_year;    // last modification time
    gds_swap2bytes( (BYTE *) &year );
    write( fd, &year, 2 );
    month = 1 + date->tm_mon;
    gds_swap2bytes( (BYTE *) &month );
    write( fd, &month, 2 );
    day = date->tm_mday;
    gds_swap2bytes( (BYTE *) &day );
    write( fd, &day, 2 );
    hour = date->tm_hour;
    gds_swap2bytes( (BYTE *) &hour );
    write( fd, &hour, 2 );
    minute = date->tm_min;
    gds_swap2bytes( (BYTE *) &minute );
    write( fd, &minute, 2 );
    second = date->tm_sec;
    gds_swap2bytes( (BYTE *) &second );
    write( fd, &second, 2 );

    write( fd, &year, 2 );          // last access time (same)
    write( fd, &month, 2 );
    write( fd, &day, 2 );
    write( fd, &hour, 2 );
    write( fd, &minute, 2 );
    write( fd, &second, 2 );

    }  // write_bgnlib


/*------------------------------------------------------------------------------------------*/
                                /* Beginning of a cell. If this is after the first cell */
                                /* then allocate it.  Ignore the creation date.         */
void
gds_read_bgnstr( int fd, int count, 
                 int *cell_table_size, 
                 int *num_cells, 
                 struct gds_celltype ***c,          // cell table
                 struct gds_celltype **cc,          // current cell
                 struct gds_itemtype **ci,          // current item
                 BOOL verbose )
    {
    static BOOL 
        first_cell = TRUE;

    int i;

    static struct gds_celltype
        **cell,
        *current_cell;

    static struct gds_itemtype
        *current_item;

    current_item = *ci;
    cell = *c;
    current_cell = *cc;

    SKIPOVER( fd, count );

    if ( first_cell )  /* Then we already created the first cell in bgnlib */
        {
        first_cell = FALSE;
        *num_cells = 0;
        }
    else                       /* allocate a new cell and move up the pointer */
        {
        current_cell->nextcell = (struct gds_celltype *) malloc( sizeof( struct gds_celltype ) );  

        if ( ! current_cell->nextcell ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR THE NEXT CELL." );
  
        current_cell = current_cell->nextcell;
        current_cell->nextcell = NULL;

        *num_cells = *num_cells + 1;

        if ( *num_cells > *cell_table_size )
            {
            *cell_table_size *= 2;
            cell = (struct gds_celltype **) realloc( cell, (1 + *cell_table_size) * sizeof( struct gds_celltype * ) );
            if ( ! cell ) BAILOUT( "UNABLE TO REALLOCATE CELL TABLE. SORRY DUDE." );
            }

        }


    current_cell->item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    current_item = current_cell->item;
    current_item->nextitem = NULL;
    
    // Now let's be careful to not use the first item. 
    // Seems like a waste, but it makes the logic simpler.

    *c   = cell;
    *cc  = current_cell;
    *ci  = current_item;

    }  // read_bgnstr

/*------------------------------------------------------------------------------------------*/

void
gds_write_bgnstr( int fd )
    {
    static time_t *now;
    static struct tm *date;

    static short int 
        year,
        month,
        day,
        hour,
        minute,
        second,
        token,
        count;

    now  = (time_t *)    malloc( sizeof( time_t ) );
    date = (struct tm *) malloc( sizeof( struct tm ) );

    time( now );
    date = localtime( now );

    year   = 1900 + date->tm_year;   
    month  = 1 + date->tm_mon;
    day    = date->tm_mday;
    hour   = date->tm_hour;
    minute = date->tm_min;
    second = date->tm_sec;
    
    gds_swap2bytes( (BYTE *) &year   );
    gds_swap2bytes( (BYTE *) &month  );
    gds_swap2bytes( (BYTE *) &day    );
    gds_swap2bytes( (BYTE *) &hour   );
    gds_swap2bytes( (BYTE *) &minute );
    gds_swap2bytes( (BYTE *) &second );

    count = 28;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0502;                 // BGNSTR
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    write( fd, &year, 2 );          // creation time
    write( fd, &month, 2 );
    write( fd, &day, 2 );
    write( fd, &hour, 2 );
    write( fd, &minute, 2 );
    write( fd, &second, 2 );
    write( fd, &year, 2 );          // access time
    write( fd, &month, 2 );
    write( fd, &day, 2 );
    write( fd, &hour, 2 );
    write( fd, &minute, 2 );
    write( fd, &second, 2 );

    }  // write_bgnstr

/*------------------------------------------------------------------------------------------*/
                                /* Marks the end of a library */
void
gds_read_endlib( int fd, int count, BOOL verbose )
    {
    int i;
    SKIPOVER( fd, count );
    }

/*------------------------------------------------------------------------------------------*/

void
gds_write_endlib( int fd )
    {
    static short int
        count,
        token;

    count = 4;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0400;                 // ENDLIB
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    }  // write_endlib

/*------------------------------------------------------------------------------------------*/

void
gds_write_node( int fd )
    {
    static short int
        count,
        token;

    count = 4;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x1500;                 // NODE
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    }  // write_node

/*------------------------------------------------------------------------------------------*/
                                /* Marks the end of a structure */
void
gds_read_endstr( int fd, int count, BOOL verbose )
    {
    static int i;
    SKIPOVER( fd, count );
    }

/*------------------------------------------------------------------------------------------*/

void 
gds_write_endstr( int fd )
    {
    static short int
        count,
        token;

    count = 4;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0700;                 // ENDSTR
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    }  // write_endstr

/*------------------------------------------------------------------------------------------*/
                                /* Read the library name. */
char *
gds_read_libname( int fd, int count, BOOL verbose )
    {
    char *name; 
    name = (char *) malloc( count - 3 ); 

    read( fd, name, count-4 );
    name[count-4] = '\0';
    if ( verbose ) printf( "             %s\n", name );
    return( name );

    }  // read_libname

/*------------------------------------------------------------------------------------------*/

void
gds_write_libname( int fd, char *s )
    {
    short int                       // name should be null-terminated
        count,
        len, i,
        token;

    static BOOL pad;                // padding for even-numbered length!
    static char *ps;
    
    len = strlen( s );
    pad = (len % 2) != 0;
    if ( pad ) len++;

    ps = (char *) malloc( len );
    if ( pad )
        {
        for ( i=0; i < len-1; i++ ) ps[i] = s[i];
        ps[len-1] = 0;
        }
    else
        {
        for ( i=0; i < len; i++ ) ps[i] = s[i];
        }


    count = 4 + len;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0206;                 // LIBNAME
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    write( fd, ps, len );

    }  // write_libname


/*------------------------------------------------------------------------------------------*/
                                /* Contains a structure name.  */
void
gds_read_strname( int fd, 
                  int count, 
                  int num_cells,
                  struct gds_celltype ***c,
                  struct gds_celltype **cc,
                  BOOL verbose   )
    {
    static struct gds_celltype
        **cell,
        *current_cell;

    cell = *c;
    current_cell = *cc;

    current_cell->name = (char *) malloc( count - 3 );

    if ( ! current_cell->name ) BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR NEXT CELL NAME" );

    read( fd, current_cell->name, count-4 );
    current_cell->name[count-4] = '\0';
    if ( verbose ) printf( "             %s\n", current_cell->name );

    cell[ num_cells ] = current_cell;    /* save the pointer to this cell */

    *c   = cell;
    *cc  = current_cell;

    }  // read_strname

/*------------------------------------------------------------------------------------------*/

void
gds_write_strname( int fd, char *s )
    {
    static short int         // name should be null-terminated
        count,
        len, i,
        token;

    static BOOL pad;
    static char *ps;
    
    len = strlen( s );
    pad = (len % 2) != 0;
    if ( pad ) len++;

    ps = (char *) malloc( len );
    if ( pad )
        {
        for ( i=0; i < len-1; i++ ) ps[i] = s[i];
        ps[len-1] = 0;
        }
    else
        {
        for ( i=0; i < len; i++ ) ps[i] = s[i];
        }

    count = 4 + len;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0606;                 // STRNAME
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    write( fd, ps, len );

    }  // write_strname


/*------------------------------------------------------------------------------------------*/

void
gds_read_sname( int fd, int count )
    {
    static int i;
    SKIPOVER( fd, count );
    }

/*------------------------------------------------------------------------------------------*/

void 
gds_write_sname( int fd, char *s )
    {
    static short int                // s should be null-terminated
        count,
        token;

    count = 4 + strlen( s );
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x1206;                 // SNAME
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    write( fd, s, strlen(s) );

    }  // write_sname


/*------------------------------------------------------------------------------------------*/
                                /* Marks the beginning of a boundary (polygon)  */
void
gds_read_boundary( int fd, int count, struct gds_itemtype **ci, BOOL verbose )
    {
    static struct gds_itemtype
        *current_item;
    int i;

    current_item = *ci;
    SKIPOVER( fd, count );
    gds_make_next_item( &current_item );

    *ci = current_item;

    }  // read_boundary

/*------------------------------------------------------------------------------------------*/

void
gds_write_boundary( int fd )
    {                             // just the token here. "xy" writes the actual polygon.
    static short int
        count,
        token;

    count = 4;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0800;                 // BOUNDARY
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    } // write_boundary


/*------------------------------------------------------------------------------------------*/
                                /* Marks the end of an element, pointlessly. */
void
gds_read_endel( int fd, int count, BOOL verbose )
    {
    int i;
    SKIPOVER( fd, count );
    }

/*------------------------------------------------------------------------------------------*/
                                /* Indicates that the next shape is too big. */
void
gds_write_obese( int fd )
    {
    static short int
        count,
        token;

    count = 6;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x2B02;                 // PROPATTR
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    token = 0;
    write( fd, &token, 2 ); 
    }

/*------------------------------------------------------------------------------------------*/

void
gds_write_endel( int fd )
    {
    static short int
        count,
        token;

    count = 4;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x1100;                 // ENDEL
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    }  // write_endel


/*------------------------------------------------------------------------------------------*/
                                /* Contains the layer number 0..255 */
void
gds_read_layer( int fd, int count, struct gds_itemtype **ci, BOOL verbose )
    {
    static short num;
    static struct gds_itemtype *current_item;
    current_item = *ci;
    read( fd, &num, 2 );
    gds_swap2bytes( (BYTE *) &num );
    if ( verbose ) printf( "             %d\n", num );

    if ( num < 0 )
        ATTENTION( "NEGATIVE LAYER NUMBER" );

    if ( num > 255 )
        ATTENTION( "LAYER > 255 " );

    current_item->layer = num;
    *ci = current_item;

    }  // read_layer

/*------------------------------------------------------------------------------------------*/

void
gds_write_layer( int fd, short int layer )
    {
    static short int
        count,
        token;

    // Layers are actually supposed to go up to 64 only, but most programs allow
    // values up to 255, because 64 is stupid. Well 255 is also stupid, but less so.

    if ( layer < 0 || layer > 255 ) BAILOUT( "INVALID LAYER NUMBER" );
    count = 6;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0D02;                 // LAYER
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    gds_swap2bytes( (BYTE *) &layer );
    write( fd, &layer, 2 );

    }  // write_layer


/*------------------------------------------------------------------------------------------*/
                                /* Contains the datatype number 0..255 */
void
gds_read_datatype( int fd, int count, struct gds_itemtype **ci, BOOL verbose )
    {
    static short num;
    static struct gds_itemtype *current_item;
    current_item = *ci;

    read( fd, &num, 2 );
    gds_swap2bytes( (BYTE *) &num );

    if ( verbose ) printf( "             %d\n", num );

    if ( num < 0 )
        ATTENTION( "NEGATIVE DATATYPE NUMBER" ); 

    if ( num > 255 )
        ATTENTION( "DATATYPE > 255 " ); 

    current_item->dt = num;
    *ci = current_item;

    }  // read_datatype

/*------------------------------------------------------------------------------------------*/

void
gds_write_datatype( int fd, short int dt )
    {
    static short int
        count,
        token;

    if ( dt < 0 )
        ATTENTION( "NEGATIVE DATATYPE NUMBER" ); 

    if ( dt > 255 )
        ATTENTION( "DATATYPE > 255 " ); 

    count = 6;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0E02;                 // DATATYPE
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    gds_swap2bytes( (BYTE *) &dt );
    write( fd, &dt, 2 );

    }  // write_datatype


/*------------------------------------------------------------------------------------------*/
void
gds_read_xy( int fd, int count, double dbu_um, struct gds_itemtype **ci, BOOL verbose )
    {
                                // Read X,Y coordinate pairs for polygons only. 
                                // We can throw away the last vertex since it is a 
                                // repeat of the first. Now a triangle has 3 verticies, 
                                // not 4. That's a divergence from libgds.
    static int 
        i, 
        num,
        x, y,
        px, py;

    static struct gds_itemtype 
        *current_item;

    current_item = *ci;

    num = (count - 4) / 8;
    num--;

    // realloc causes a crash with large patterns
    //current_item->x   = (int *)  realloc( current_item->x,   (num+1) * sizeof( int  ) );  
    //current_item->y   = (int *)  realloc( current_item->y,   (num+1) * sizeof( int  ) );
    //current_item->hit = (BOOL *) realloc( current_item->hit, (num+1) * sizeof( BOOL ) );

    current_item->x   = (int *)  malloc( (num+1) * sizeof( int  ) );  
    current_item->y   = (int *)  malloc( (num+1) * sizeof( int  ) );
    current_item->hit = (BOOL *) malloc( (num+1) * sizeof( BOOL ) );
    
    if ( ! current_item->x || ! current_item->y || ! current_item->hit )
        BAILOUT( "UNABLE TO ALLOCATE MEMORY FOR COORDINATES." );


    for ( i=0; i < num; i++ )  
        {
        read( fd, &x, 4 );
        gds_swap4bytes( (BYTE *) &x );
        read( fd, &y, 4 );
        gds_swap4bytes( (BYTE *) &y );
        if ( verbose ) printf( "             (%8d, %8d) = (%9.3f, %9.3f) um \n", x, y, x * dbu_um, y * dbu_um ) ;
        current_item->x[i]   = x;
        current_item->y[i]   = y;
        current_item->hit[i] = FALSE;
        }

    read( fd, &x, 4 );  // skip over the repeat
    read( fd, &y, 4 );

    current_item->n = num;      // which means a triangle has n=3

    *ci = current_item;

    }  // read_xy

/*------------------------------------------------------------------------------------------*/

void
gds_write_xy( int fd, int *x, int *y, int n )
    {
    static short int            // Here in libgdsmin we have polygons only.
        count,                  // This function repeats the first vertex
        token;                  // to conform to GDS format.
                                // In libgdsmin, the datastructure does not 
                                // contain the repeated vertex.
    static int i, xx, yy;

    if ( n > 8200 ) BAILOUT( "WAY TOO MANY VERTICIES" );

    count = 4 + 8 * (n+1);
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x1003;                      // XY
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    for ( i=0; i<n; i++ )
        {
        xx = x[i];
        yy = y[i];
        gds_swap4bytes( (BYTE *) &xx );
        gds_swap4bytes( (BYTE *) &yy );
        write( fd, &xx, 4 );
        write( fd, &yy, 4 );
        }

    xx = x[0];                 // repeating the first vertex 
    yy = y[0];                 // is so stupid
    gds_swap4bytes( (BYTE *) &xx );
    gds_swap4bytes( (BYTE *) &yy );
    write( fd, &xx, 4 );       // but we have to do it
    write( fd, &yy, 4 );       // because that's gds format

    }  // write_xy

/*------------------------------------------------------------------------------------------*/

void
gds_write_long_xy( int fd, long *x, long *y, long n )
    {
    static short int            // Here in libgdsmin we have polygons only.
        count,                  // This function repeats the first vertex
        token;                  // to conform to GDS format.
                                // In libgdsmin, the datastructure does not 
                                // contain the repeated vertex.
    static int i, xx, yy;

    if ( n > 8200 ) BAILOUT( "WAY TOO MANY VERTICIES" );

    count = 4 + 8 * (n+1);
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x1003;                      // XY
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );

    for ( i=0; i<n; i++ )
        {
        xx = x[i];
        yy = y[i];
        gds_swap4bytes( (BYTE *) &xx );
        gds_swap4bytes( (BYTE *) &yy );
        write( fd, &xx, 4 );
        write( fd, &yy, 4 );
        }

    xx = x[0];                 // repeating the first vertex 
    yy = y[0];                 // is so stupid
    gds_swap4bytes( (BYTE *) &xx );
    gds_swap4bytes( (BYTE *) &yy );
    write( fd, &xx, 4 );       // but we have to do it
    write( fd, &yy, 4 );       // because that's gds format

    }  // write_long_xy

/*------------------------------------------------------------------------------------------*/

void
gds_read_units( int fd, double *Pdbu_um, double *Pdbu_uu, double *Pdbu_m, BOOL verbose )

    /* Read the database UNIT in "user units" (typically 0.001), then read the database unit  */
    /* in units of meters. The first number doesn't really matter, unless you want a CAD      */
    /* program to display the data in convenient units (such as microns).                     */
    /* The imporant number is the second one: meters per bit.                                 */


    {
    double 
        dbu_uu,        /* database unit in user units, usually 0.001 */
        dbu_m;         /* database unit in meters, usually 1e-9      */

    dbu_uu = gds_read_double( fd ); 

    dbu_m  = gds_read_double( fd ); 

    if ( dbu_uu <= 0.0 ) BAILOUT( "INVALID DATABASE USER UNIT, < 0" );

    if ( dbu_m <= 0.0 )  BAILOUT( "INVALID DATABASE UNIT, < 0" );


    if ( verbose ) printf( "             Database units in microns: %f\n", dbu_m * 1e6 ) ;

    *Pdbu_um = dbu_m * 1e6;     // microns per database unit 
    *Pdbu_uu = dbu_uu;          // user units per database unit
    *Pdbu_m  = dbu_m;           // meters per database unit

    }  // read_units

/*------------------------------------------------------------------------------------------*/

void
gds_write_units( int fd, double dbu_uu, double dbu_m )
    {
    short int
        count,
        token;

    count = 20;
    gds_swap2bytes( (BYTE *) &count );
    write( fd, &count, 2 );
    token = 0x0305;                 // UNITS
    gds_swap2bytes( (BYTE *) &token );
    write( fd, &token, 2 );
    gds_write_float( fd, dbu_uu );
    gds_write_float( fd, dbu_m );

    }  // write_units



/*------------------------------------------------------------------------------------------*/
//  HIGHER LEVEL FUNCTIONS
/*------------------------------------------------------------------------------------------*/


void
gds_create_lib( int fd, char *libname, double dbu_um )
    {
    // Write HEADER, BGNLIB, LIBNAME, and UNITS.
    
    
    double
        dbu_uu,        // database user units, 1 nm per bit
        dbu_m;         // database unit in meters, usually 1e-9m


    dbu_uu = 0.001;
    dbu_m  = dbu_um * 1.0e-6; 

    gds_write_header( fd );
    gds_write_bgnlib( fd );
    gds_write_libname( fd, libname );
    gds_write_units( fd, dbu_uu, dbu_m );

    }  // create_lib


/*------------------------------------------------------------------------------------------*/

void
gds_create_lib_correctly( int fd, char *libname, double dbu_um, double dbu_uu )
    {
    // Write HEADER, BGNLIB, LIBNAME, and UNITS.
    // OMG this is so confusing! 
    
    double
        dbu_m;        


    dbu_m  = 1.0e-6 * dbu_uu;

    gds_write_header( fd );
    gds_write_bgnlib( fd );
    gds_write_libname( fd, libname );
    gds_write_units( fd, dbu_uu, dbu_m );

    }  // create_lib



