// polyq.c    polygons and queues of polygons that can be
//            addressed with index checking. Lists of int also!
//            This version uses longs for vertex coordinates.
// v2.1       now with mv_polygon
// v2.2       minor printf formatting corrections for ubuntu

#include "polyq.h"

void
bail( char *message )
    {
    FILE *arret;
    printf( "\n-------------------------------------------" );
    printf( "\nERROR: %s ", message );
    printf( "\n-------------------------------------------\n\n" );
    fflush( stdout );
    arret = fopen( "fullstop.txt", "w" );
    fprintf( arret, "%s\n", message );
    fclose( arret );
    abort();   // force a crash, allowing the use of 'bt' in gdb
    }

xy 
vertex( polygon *p, int i )
    {
    if ( ! p ) bail( "null polygon sent to vertex function" );
    if ( i < 0 ) i = p->len + i;
    if ( p->len == 0 ) bail( "vertex() was sent a zero-length polygon" );
    if ( i >= p->len ) i = i % p->len;
    return( p->v[i] );
    }


xy 
next_vertex( polygon *p, int i )
    {
    static int ii;

    if ( ! p ) bail( "null polygon sent to next_vertex function" );

    ii = (i+1) % p->len;

    if ( ii < 0 || ii >= p->len ) 
        {
        printf( "\nlen: %d   i: %d", p->len, ii );
        bail( "next_vertex: index out of bounds" );
        }
    return( p->v[ii] );
    }

xy 
previous_vertex( polygon *p, int i )
    {
    static int ii;

    if ( ! p ) bail( "null polygon sent to previous_vertex function" );

    i = i % p->len;

    if ( i == 0 )
	ii = p->len - 1;
    else
	ii = i - 1;

    if ( ii < 0 || ii >= p->len ) 
        {
        printf( "\nlen: %d   i: %d", p->len, ii );
        bail( "previous_vertex: index out of bounds" );
        }
    return( p->v[ii] );
    }


dxy 
dvertex( polygon *p, int i )
    {
    static dxy v;
    if ( ! p ) bail( "null polygon sent to dvertex function" );
    if ( ! p->active ) bail( "(dvertex) attempt to use inactive polygon" );
    if ( i < 0 || i >= p->len ) 
        {
        printf( "\nlen: %d   i: %d", p->len, i );
        bail( "vertex: index out of bounds" );
        }
    v.x = p->v[i].x;
    v.y = p->v[i].y;
    return( v );
    }


BOOL
equalpoints( xy a, xy b )
    {
    return( (a.x == b.x) && (a.y == b.y) );
    }


int
insert_vertex( xy v, /*into*/ polygon_ref *pr, /*after index*/ int jj )
    {
    // ...but not if the vertex is redundant
    // Return the actual insertion index or the redundant one.

    polygon *p, *newp;
    static int i, n, nm1;

    if ( ! pr ) bail( "null polygon ref sent to insert_vertex" );

    p = pr->p;

    if ( equalpoints( v,      vertex( p, jj ) ) ) return( jj );
    if ( equalpoints( v, next_vertex( p, jj ) ) ) return( (jj+1) % p->len );

    n = p->len;
    nm1 = n - 1;

    if ( jj < 0 || jj >= p->len ) bail( "insert_vertex illegal index" );

    if ( p->len < p->max )
        {
        p->len = p->len + 1;
        for ( i=nm1; i > jj; i-- ) assign_vertex( p, i+1, vertex( p, i ) );
        assign_vertex( p, jj+1, v );
        set_extents( p );
        }
    else
        {
        newp = initialize_polygon( p->max + 4, p->layer, p->dt, p->obese );
        for ( i=0; i <= jj; i++ )  append_polygon_xy( newp, vertex( p, i ) );
        append_polygon_xy( newp, v );
        for ( i=jj+1; i < n; i++ ) append_polygon_xy( newp, vertex( p, i ) );
	set_extents( newp );
        pr->p = newp;
        }

    return( (jj+1) % p->len );
    }



BOOL
nearlyequalpoints( xy a, xy b, int bss )
    {
    return( (abs( a.x - b.x ) < bss) && (abs( a.y - b.y ) < bss) );
    }


BOOL
insert_vertex_near( xy v, /*into*/ polygon_ref *pr, /*after index*/ int jj, int bss )
    {
    // ...but not if the vertex is closer than bss to an adjacent vertex.
    // Return true if this creates a new vertex, or return
    // false if it was a duplicate (closer than bss).

    polygon *p, *newp;
    static int i, n, nm1;
    static xy w;

    if ( ! pr ) bail( "null polygon ref sent to insert_vertex" );

    p = pr->p;

    if ( nearlyequalpoints( v, vertex( p, jj ), bss ) ) 
        { 
        //w = vertex( p, jj );
        //printf( "%ld %ld is too close to %ld %ld \n", v.x, v.y, w.x, w.y );
        assign_vertex( p, jj, v ); 
        return( FALSE ); 
        }
    if ( nearlyequalpoints( v, next_vertex( p, jj ), bss ) ) 
        {
        //w = next_vertex( p, jj );
        //printf( "%ld %ld is too close to the next vertex %ld %ld\n", v.x, v.y, w.x, w.y  ); 
        assign_vertex( p, (jj+1) % p->len, v );
        return( FALSE ); 
        }

    n = p->len;
    nm1 = n - 1;

    if ( jj < 0 || jj >= p->len ) bail( "insert_vertex illegal index" );

    if ( p->len < p->max )
        {
        p->len = p->len + 1;
        for ( i=nm1; i > jj; i-- ) assign_vertex( p, i+1, vertex( p, i ) );
        assign_vertex( p, jj+1, v );
        set_extents( p );
        }
    else
        {
        newp = initialize_polygon( p->max + 4, p->layer, p->dt, p->obese );
        for ( i=0; i <= jj; i++ )  append_polygon_xy( newp, vertex( p, i ) );
        append_polygon_xy( newp, v );
        for ( i=jj+1; i < n; i++ ) append_polygon_xy( newp, vertex( p, i ) );
	set_extents( newp );
        pr->p = newp;
        }

    return( TRUE );
    }



void 
set_vertex( polygon *p, int i, xy v )
    {
    if ( ! p ) bail( "null polygon sent to set_vertex function" );
    if ( ! p->active ) bail( "(set_vertex) attempt to use inactive polygon" );
    i = i % p->len;
    if ( i < 0 ) 
        {
        printf( "\nlen: %d   i: %d", p->len, i );
        bail( "set_vertex: negative index" );
        }
    p->v[i].x = v.x;
    p->v[i].y = v.y;
    if ( v.x < p->ll.x ) p->ll.x = v.x;
    if ( v.y < p->ll.y ) p->ll.y = v.y;
    if ( v.x > p->ur.x ) p->ur.x = v.x;
    if ( v.y > p->ur.y ) p->ur.y = v.y; 
    return;
    }

void 
set_dvertex( polygon *p, int i, dxy v )
    {
    if ( ! p ) bail( "null polygon sent to set_dvertex function" );
    if ( ! p->active ) bail( "(set_dvertex) attempt to use inactive polygon" );
    if ( i < 0 || i >= p->len ) 
        {
        printf( "\nlen: %d   i: %d", p->len, i );
        bail( "set_dvertex: index out of bounds" );
        }
    p->v[i].x = (long) floor( v.x );
    p->v[i].y = (long) floor( v.y );
    return;
    }


void
print_polygon( polygon *p, float scale )
    {
    static int   i;
    static xy    v;
    static float fx, fy;

    if ( ! p ) bail( "attempt to print null polygon" );
    printf( "\n" );
    if ( ! p->active ) printf( "INACTIVE:\n" );
    for ( i=0; i < p->len; i++ ) 
        {
        v = vertex( p, i );
        fx = v.x;
        fy = v.y;
        fx = fx / scale;
        fy = fy / scale;
        printf( "[%2d] [%6ld] [%6ld] = (%3.3f, %3.3f) ", i, v.x, v.y, fx, fy );
        if ( p->hit[i] )
            printf( "H " );
        else
            printf( "  " );
        if ( p->precious[i] ) 
            printf( "P\n" ); 
        else 
            printf( "\n" );
        }
    printf( "\n" );
    }


void
printn_polygon( polygon *p, float scale, int n )
    {
    static int   i;
    static xy    v;
    static float fx, fy;

    if ( ! p ) bail( "attempt to print null polygon" );
    printf( "\n" );
    if ( ! p->active ) printf( "INACTIVE:\n" );
    for ( i=0; i < n; i++ ) 
        {
        v = vertex( p, i );
        fx = v.x;
        fy = v.y;
        fx = fx / scale;
        fy = fy / scale;
        printf( "[%2d] [%6ld] [%6ld] = (%3.3f, %3.3f) ", i, v.x, v.y, fx, fy );
        if ( p->hit[i] )
            printf( "H " );
        else
            printf( "  " );
        if ( p->precious[i] ) 
            printf( "P\n" ); 
        else 
            printf( "\n" );
        }
    printf( "\n" );
    }


void
assign_vertex( polygon *p, int i, xy v )
    {
    set_vertex( p, i, v );
    }


BOOL
delete_vertex( polygon *p, int gone )
    {
    static xy v;
    static int nm1, j;

    if ( ! p ) bail( "null polygon sent to delete_vertex function" );
    if ( ! p->active ) bail( "(delete_vertex) attempt to use inactive polygon" );

    nm1 = p->len - 1;

    gone = gone % p->len;
    
    if ( gone < 0 ) bail( "delete_vertex: index out of bounds" );

    if ( p->precious[gone] )
        {
	//v = vertex( p, gone );
	//printf( "vertex %ld,%ld is precious\n", v.x, v.y ); 
        return( TRUE );   // true => break out of the loop that looks for consecutive points  
        }
    
    // If we delete the last vertex, then we need only decrement p->len.
    
    for ( j=gone; j < nm1; j++ )
        {
        p->v[j].x = p->v[j+1].x;
        p->v[j].y = p->v[j+1].y;
        p->precious[j] = p->precious[j+1];
        p->hit[j] = p->hit[j+1];
        }

    p->len -= 1;
    set_extents( p );
    return( FALSE );     // false => keep looking if you like
    }


BOOL
delete_precious_vertex( polygon *p, int gone )
    {
    static xy v;
    static int nm1, j;

    if ( ! p ) bail( "null polygon sent to delete_precious_vertex function" );

    nm1 = p->len - 1;

    gone = gone % p->len;
    
    if ( gone < 0 ) bail( "delete_precious_vertex: index out of bounds" );

    for ( j=gone; j < nm1; j++ )
        {
        p->v[j].x = p->v[j+1].x;
        p->v[j].y = p->v[j+1].y;
        p->precious[j] = p->precious[j+1];
        }

    p->len -= 1;
    set_extents( p );
    return( FALSE ); 
    }

void
nag_too_many_vertices()
    {
    printf( "\nERROR: unable to allocate polygon memory.\n" );
    printf(   "       You can reduce the number of vertices by using Layout CAD.\n" );
    printf(   "       In Layout, use Utilities > Setup > File Format > GDSII \n" );
    printf(   "       then set the 'maximal number of points' to 512. \n" );
    printf(   "       Save the GDS file and try again.\n" );
    }

void
nag_too_many_polygons()
    {
    printf( "\nERROR: this pattern will have too many shapes in the output gpf file.\n" );
    printf(   "       You might try setting array=\"yes\" in the script, \n" );
    printf(   "       or you could delete any outrageously large shapes.\n" );
    }

polygon *
initialize_polygon( int n, int layer, int dt, BOOL obese )
    {
    static polygon *p;
    static int i;
    p = (polygon *) malloc( sizeof(polygon) );
    if ( ! p )
        {
        nag_too_many_vertices();
        bail( "Cannot allocate memory for polygon." );
        }
    if ( n < 4 ) n = 4;  // helps us fix single-pixel triangles
    n = n + 2;           // so we can insert extra vertices quickly
    p->len   = 0;
    p->max   = n;
    p->layer = layer;
    p->dt    = dt;
    p->v = (xy *) malloc( n * sizeof( xy ) );
    if ( ! p->v )
        {
        nag_too_many_vertices();
        bail( "Cannot allocate memory for polygon vertices." );
        }
    p->precious = (BOOL *) malloc( n * sizeof( BOOL ) );
    if ( ! p->precious )
        {
        nag_too_many_vertices();
        bail( "Cannot allocate memory for polygon preciousness." );
        }
    for ( i=0; i < n; i++ ) p->precious[i] = FALSE;
    p->hit  = (BOOL *) malloc( n * sizeof( BOOL ) );
    p->hit2 = (BOOL *) malloc( n * sizeof( BOOL ) );
    if ( ! p->hit || ! p->hit2 )
        {
        nag_too_many_vertices();
        bail( "Cannot allocate polygon hit array" );
        }
    for ( i=0; i < n; i++ ) { p->hit2[i] = FALSE; p->hit[i] = FALSE; }
    p->active = TRUE;
    p->obese  = obese;
    p->ll.x   = 0;
    p->ll.y   = 0;
    p->ur.x   = 0;
    p->ur.y   = 0;
    return p;
    }


void
unhit_all( polygon *p )
    {                   
    static int i;
    if ( ! p ) bail( "null polygon sent to unhit_all" );
    for ( i=0; i < p->len; i++ ) 
	{ 
	p->hit[i]  = FALSE;
	p->hit2[i] = FALSE;
	}
    }

void
unhit_polygon_array( polygon_array *pa )
    {
    static int i;
    for ( i=0; i < pa->len; i++ ) unhit_all( pa->par[i] );
    }

void
no_precious( polygon *p )
    {
    static int i;
    if ( ! p ) bail( "null polygon sent to no_precious" );
    for ( i=0; i < p->len; i++ ) 
	{ 
	p->precious[i]  = FALSE;
	}    
    }

void 
hit_vertex( polygon *p, int i )
    {
    if ( ! p ) bail( "null polygon sent to hit_vertex" );
    if ( i < 0 ) i = p->len + i;
    i = i % p->len;
    if ( i < 0 || i >= p->len ) bail( "spooky illegal index in hit_vertex" );
    p->hit[i] = TRUE;
    }
void 

unhit_vertex( polygon *p, int i )
    {
    if ( ! p ) bail( "null polygon sent to unhit_vertex" );
    if ( i < 0 ) i = p->len + i;
    i = i % p->len;
    if ( i < 0 || i >= p->len ) bail( "spooky illegal index in unhit_vertex" );
    p->hit[i] = FALSE;
    }

BOOL
is_hit( polygon *p, int i )
    {
    if ( i < 0 ) i = p->len + i;
    i = i % p->len;
    return( p->hit[i] );    
    }


void
zero_polygon( polygon *p )
    {
    if ( ! p ) bail( "attempt to zero a polygon before initialization." );
    p->len   = 0;
    p->active = TRUE;
    p->obese = FALSE;
    p->layer = 0;
    p->dt = 0;
    // notice that p->max is unchanged. The space is still there.
    }

polygon_ref *
initialize_polygon_ref()
    {
    polygon_ref *p;
    p = (polygon_ref *) malloc( sizeof( polygon_ref ) );
    if ( ! p )
        {
        nag_too_many_polygons();
        bail( "unable to allocate polygon_ref" );
        }
    p->p = NULL;
    p->next = NULL;
    p->prev = NULL;
    return( p );
    }

queue_ref *
initialize_queue_ref()
    {
    queue_ref *q;
    q = (queue_ref *) malloc( sizeof( queue_ref ) );
    if ( ! q ) bail( "unable to allocate queue_ref" );
    q->q = NULL;
    q->active = FALSE;
    q->next = NULL;
    q->prev = NULL;
    return( q );
    }

int_list *
initialize_int_list( int n )
    {
    static int_list *p;
    static int i;
    p = (int_list *) malloc( sizeof(int_list) );
    if ( ! p ) bail( "Cannot allocate memory for list of ints." );
    p->len = 0;
    p->max = n;
    p->n = (int *) malloc( n * sizeof( int ) );
    if ( ! p->n ) bail( "Cannot allocate memory for array of ints." );
    return p;
    }

void
free_int_list( int_list *p )
    {
    if ( ! p ) bail( "null pointer sent to free_int_list function" );
    free( p->n );
    free( p );
    }

queue *
initialize_queue()
    {
    queue *q;
    q = (queue *) malloc( sizeof( queue ) );
    if ( ! q )
        {
        nag_too_many_polygons();
        bail( "Cannot allocate memory for queue." );
        }
    q->len   = 0;
    q->first = NULL;
    q->last  = NULL;
    return q;
    }

void
zero_queue( queue *q )
    {
    q->len = 0;
    q->first = NULL;
    q->last  = NULL;
    }

queue_queue *
initialize_queue_queue()
    {
    queue_queue *qq;
    qq = (queue_queue *) malloc( sizeof( queue_queue ) );
    if ( ! qq ) bail( "Cannot allocate memory for queue_queue." );
    qq->first = NULL;
    qq->last  = NULL; 
    return qq;
    }

void
free_queue( queue *q )   // why does this not work?
    {
    /*
    polygon_ref *pr, *beer;
    if ( ! q ) return;
    pr = q->first;
    while ( pr )
        {
        if ( pr->p->v ) free( pr->p->v );
        if ( pr->p )    free( pr->p );
        beer = pr;
        pr = pr->next;
        if ( beer ) free( beer );
        }
    */
    if ( q ) free( q );
    }


void
append_polygon( polygon *p, long x, long y )
    {
    if ( ! p ) bail( "null polygon sent to append_polygon function" );
    if ( p->len == p->max ) 
        {
        print_polygon( p, 1 );
        bail( "append_polygon: too many vertices added to polygon" );
        }
    p->v[p->len].x = x;
    p->v[p->len].y = y;
    p->precious[p->len] = FALSE;
    p->len += 1;
    if ( p->len == 1 )
	{
	p->ll.x = x;
	p->ll.y = y;
	p->ur.x = x;
	p->ur.y = y;
	}
    else
	{
        if ( x < p->ll.x ) p->ll.x = x;
        if ( x > p->ur.x ) p->ur.x = x;
        if ( y < p->ll.y ) p->ll.y = y;
        if ( y > p->ur.y ) p->ur.y = y;
	}
    } 

void
append_polygon_xy( polygon *p, xy v )
    {
    if ( ! p ) bail( "null polygon sent to append_polygon_xy function" );
    if ( ! p->active ) bail( "(append_polygon_xy) attempt to use inactive polygon" );
    if ( p->len == p->max ) 
        {
        print_polygon( p, 1 );
        printf( "\nlen: %d ", p->len );
        bail( "append_polygon_xy: too many vertices added to polygon" );
        }
    p->v[p->len].x = v.x;
    p->v[p->len].y = v.y;
    p->precious[p->len] = FALSE;
    p->len += 1;
    if ( p->len == 1 )
	{
	p->ll.x = v.x;
	p->ll.y = v.y;
	p->ur.x = v.x;
	p->ur.y = v.y;
	}
    else
        {
        if ( v.x < p->ll.x ) p->ll.x = v.x;
        if ( v.x > p->ur.x ) p->ur.x = v.x;
        if ( v.y < p->ll.y ) p->ll.y = v.y;
        if ( v.y > p->ur.y ) p->ur.y = v.y;
        }
    } 

void
append_polygon_dxy( polygon *p, dxy dv )
    {
    static xy v;
    if ( ! p ) bail( "null polygon sent to append_polygon_dxy function" );
    if ( ! p->active ) bail( "(append_polygon_dxy) attempt to use inactive polygon" );
    if ( p->len == p->max ) 
        {
        printf( "\nlen: %d ", p->len );
        bail( "append_polygon_dxy: too many vertices added to polygon\n" );
        }
    v.x = (long) floor( dv.x );
    v.y = (long) floor( dv.y );
    p->v[p->len].x = v.x;
    p->v[p->len].y = v.y;
    p->precious[p->len] = FALSE;
    p->len += 1;
    if ( p->len == 1 )
	{
	p->ll.x = v.x;
	p->ll.y = v.y;
	p->ur.x = v.x;
	p->ur.y = v.y;
	}
    else
        {
        if ( v.x < p->ll.x ) p->ll.x = v.x;
        if ( v.x > p->ur.x ) p->ur.x = v.x;
        if ( v.y < p->ll.y ) p->ll.y = v.y;
        if ( v.y > p->ur.y ) p->ur.y = v.y;
        }
    } 


void
append_int_list( int_list *p, int x )
    {
    if ( ! p ) bail( "null pointer sent to append_int_list function" );
    if ( p->len == p->max ) bail( "Too many numbers added to a list of ints\n" );
    p->n[p->len] = x;
    p->len += 1;
    } 

BOOL
in_int_list( int_list *p, int x )
    {
    int i;
    if ( p == NULL ) return FALSE;
    for ( i=0; i < p->len; i++ )
        {
        if ( p->n[i] == x ) return TRUE;
        }
    return FALSE;
    }

void
print_int_list( int_list *p )
    {
    int i;
    if ( p == NULL ) return;
    for ( i=0; i < p->len; i++ )
        {
        printf( "[%d] ", p->n[i] );
        }
    printf( "\n" );
    }

polygon *
pop_queue_fifo( queue *q )  // FIFO
    {
    polygon_ref *ret;
    if ( ! q ) bail( "queue has not been initialized - pop_queue_fifo" );
    if ( q->len == 0 ) return NULL;    
    ret = q->first;
    q->first = q->first->next;
    if ( q->first != NULL ) q->first->prev = NULL;
    q->len -= 1;
    return ret->p;
    }


int 
int_item( int_list *p, int i )
    {
    if ( ! p ) bail( "null pointer sent to int_item function" );
    if ( i < 0 || i >= p->len ) bail( "illegal index to list of ints" );
    return( p->n[i] );
    }


polygon_ref *
append_queue( queue *q, polygon *p )
    {
    polygon_ref *prev, *newp;
    if ( ! p ) bail( "(append_queue) null polygon" );
    if ( ! q ) bail( "(append_queue) queue has not been initialized." );
    if ( ! p->active ) return( NULL ); // bail( "(append_queue) inactive polygon" );
    newp = initialize_polygon_ref();
    newp->p = p;

    if ( q->len == 0 )
        {
        q->len   = 1;
        q->first = newp;
        q->last  = newp;
        newp->next = NULL;
        newp->prev = NULL;
        }
    else
        {
        q->len += 1;
        q->last->next = newp;
        prev = q->last;
        q->last = newp;
        q->last->prev = prev;
        q->last->next = NULL;
        }

    return( newp );
    }


void
append_queue_queue( queue_queue *qq, queue *q )
    {
    queue_ref *prevq, *newq;
    if ( ! q )  bail( "null queue sent to append_queue_queue" );
    if ( ! qq ) bail( "attempt to append a null queue_queue." );

    newq = initialize_queue_ref();
    newq->q = q;
    newq->active = TRUE;

    if ( qq->last ) 
        {
        qq->last->next = newq;
        prevq = qq->last;
        qq->last = newq;
        qq->last->prev = prevq;
        qq->last->next = NULL;
        }
    else   // first entry
        {
        qq->first = newq;
        qq->last  = newq;
        qq->last->prev = NULL;
        qq->last->next = NULL;
        }

    }


polygon *
copy_polygon( polygon *from )
    {                             // init and then copy
    polygon *to;
    static int i;

    if ( ! from ) bail( "null polygon sent to copy_polygon" );
    if ( ! from->active ) bail( "(copy_polygon) attempt to use inactive polygon" );

    to = initialize_polygon( from->len, from->layer, from->dt, from->obese );

    to->layer  = from->layer;
    to->dt     = from->dt;
    to->len    = from->len;
    to->active = from->active;
    to->obese  = from->obese; 
    to->ll.x   = from->ll.x;
    to->ll.y   = from->ll.y;
    to->ur.x   = from->ur.x;
    to->ur.y   = from->ur.y;


    for ( i=0; i < from->len; i++ ) 
        {
        to->v[i].x = from->v[i].x;
        to->v[i].y = from->v[i].y;
        to->precious[i] = from->precious[i];
        }
    return( to );
    }


void
mv_polygon( polygon *from, polygon *to )
    {                                        // copy, no malloc
    static int i;
    
    if ( ! from || ! to ) bail( "Null polygon sent to mv_polygon. Target and source must be allocated." );
    if ( to->max < from->len ) bail( "target polygon is too small for mv_polygon" );

    to->layer  = from->layer;
    to->dt     = from->dt;
    to->len    = from->len;
    to->active = from->active;
    to->obese  = from->obese;
    to->ll.x   = from->ll.x;
    to->ll.y   = from->ll.y;
    to->ur.x   = from->ur.x;
    to->ur.y   = from->ur.y;

    for ( i=0; i < from->len; i++ ) 
        {
        to->v[i].x = from->v[i].x;
        to->v[i].y = from->v[i].y;
        to->precious[i] = from->precious[i];
        }

    }


void
just_copy_polygon( polygon *from, polygon_ref *to )   // copy a poly to a poly reference
    {                                                 // just copy, no malloc

    if ( ! from ) bail( "null polygon sent to just_copy_polygon" );
    if ( ! to   ) bail( "null polygon ref sent to just_copy_polygon" );
    if ( ! from->active ) bail( "(just_copy_polygon) attempt to use inactive polygon" );

    if ( to->p->max < from->len ) bail( "(just_copy_polygon) target polygon is not big enough" );

    mv_polygon( from, to->p );   // it's a polygon reference for some reason

    return;
    }


void
reverse_polygon( polygon *p )
    {
    static int i, n, n2, n21, ni1;
    static xy t;
    static BOOL ring;

    if ( ! p ) bail( "null polygon sent to reverse_polygon" );
    if ( ! p->active ) bail( "(reverse_polygon) attempt to use inactive polygon" );
    if ( p->len < 2 ) return;
    n = p->len;
    
    n2 = 1 + p->len / 2; 
    n21 = n2 - 1;

    for ( i=0; i < n21; i++ ) 
        {
        t = vertex( p, i );
        ring = p->precious[i];
        ni1 = n - i - 1;
        assign_vertex( p, i, vertex( p, ni1 ) );
        p->precious[i] = p->precious[ni1];
        assign_vertex( p, ni1, t ); 
        p->precious[ni1] = ring;
        }
    }



polygon *
copy_flip_polygon( polygon *from )
    {                              // init and then copy, flipping x and y
    static int i;
    polygon *to;

    if ( ! from ) bail( "null polygon sent to copy_flip_polygon" );
    if ( ! from->active ) bail( "(copy_flip_polygon) attempt to use inactive polygon" );
    to = initialize_polygon( from->len, from->layer, from->dt, from->obese );
    to->layer  = from->layer;
    to->dt     = from->dt;
    to->len    = from->len;
    to->active = from->active;
    to->obese  = from->obese; 

    for ( i=0; i < from->len; i++ ) 
        {
        to->v[i].x = from->v[i].y;
        to->v[i].y = from->v[i].x;
        to->precious[i] = from->precious[i];
        }
    return to;
    }
    

void
multiply_polygon( polygon *p, float f )
    {
    static int i;
    if ( ! p ) bail( "null polygon sent to multiply_polygon" );
    if ( ! p->active ) bail( "(multiply_polygon) attempt to use inactive polygon" );
    if ( f == 1.0 ) return;
    for ( i=0; i < p->len; i++ )
        {
        p->v[i].x = floor( f * p->v[i].x );
        p->v[i].y = floor( f * p->v[i].y );
        }
    set_extents( p );
    }


array2d_ill *
initialize_array2d_ill( int nx, int ny )
    {
    int x, y;
    array2d_ill *all;

    all = (array2d_ill *) malloc( sizeof( array2d_ill ) );
    if ( ! all ) bail( "unable to allocate 2d int list array" );

    all->lenx = nx;
    all->leny = ny;
    
    all->element = (p_list **) malloc( nx * sizeof( p_list * ) );

    if ( ! all->element ) bail( "Cannot allocate memory for 2D list of ints" );

    for ( x=0; x < nx; x++ ) 
        {
        all->element[x] = (p_list *) malloc( ny * sizeof( p_list ) ); 
        if ( ! all->element[x] ) bail( "Cannot allocate memory for list of ints!" );

        for ( y=0; y < ny; y++ )
            {  
            all->element[x][y].top = (ill *) malloc( sizeof( ill ) );
            if ( ! all->element[x][y].top ) bail( "Cannot allocated memory for list of ints" );
            all->element[x][y].bottom = all->element[x][y].top;
            all->element[x][y].top->n = 0; 
            all->element[x][y].top->prev = (ill *) 1; // not_initialized = true
            all->element[x][y].top->next = NULL; 
            }
        }
    return( all );
    }


ill *
initialize_ill()
    {
    ill *p;
    p = (ill *) malloc( sizeof( ill ) );
    if ( ! p ) bail( "unable to initialize int linked list" );
    p->n = -1; 
    // Set the 'previous' pointer to a valid address, to indicate
    // that the list has not yet been used. Normally, the top of
    // the list has the 'previous' pointer set to Null.
    p->prev = (ill *) malloc( 1 ); 
    if ( ! p->prev ) bail( "unable to initialize int linked list!" );
    p->next = NULL;
    return( p );
    }


ill *
remove_from_ill( ill *p, ill *q )
    {
    // delete item q from the list p, by just snipping it out

    ill *prev, *next;

    if ( ! p || ! q ) return( p );

    prev = q->prev;
    next = q->next;

    if ( prev ) 
        prev->next = next;
    else
        p = next;

    if ( next ) 
        next->prev = prev;
    else
        q->prev->next = NULL;

    return( p );
    }



void
append_array2d( array2d_ill *list, int x, int y, int n )
    {
    static ill *new, *p;

    if ( x < 0 || x >= list->lenx ) bail( "illegal x index in append_array2d" );
    if ( y < 0 || y >= list->leny ) bail( "illegal y index in append_array2d" );
    
    p = list->element[x][y].top;

    if ( p->prev )  // then this is the first entry
        {
        p->prev = NULL;    // not yet used = false
        p->n = n;
        }
    else
        {
        new = (ill *) malloc( sizeof( ill ) );
        if ( ! new ) bail( "unable to allocate memory in append_array2d" );

        new->n = n;                          
        new->prev = list->element[x][y].bottom;
        new->next = NULL;
        list->element[x][y].bottom->next = new;
        list->element[x][y].bottom = new;
        }
    }


ill *
append_ill( ill *p, int n )
    {
    ill *new, *pp, *last;

    if ( p->prev )  // then this is the first entry
        {
        p->prev = NULL;    // not yet used = false
        p->n = n;
        }
    else
        {
        new = (ill *) malloc( sizeof( ill ) );
        if ( ! new ) bail( "unable to append int to linked list" );
        last = p;
        pp = p->next;
        while ( pp )     // find the end
            {
            last = pp;
            pp = pp->next;   
            }
        last->next = new;
        new->n = n; 
        new->prev = last;
        new->next = NULL;  
        }    
    return( p );
    }


ill *
element_of_array2d( array2d_ill *list, int x, int y )
    {
    static ill *p;

    if ( x < 0 || x >= list->lenx ) bail( "illegal x index to element_of_array2d" );
    if ( y < 0 || y >= list->leny ) bail( "illegal y index to element_of_array2d" );
    
    p = list->element[x][y].top;

    return( p );
    }


void
print_array2d( array2d_ill *list, int x, int y )
    {
    ill *p;
        
    p = element_of_array2d( list, x, y );

    if ( p->prev ) return;

    while ( p )
        {
        printf( "%d", p->n );
        p = p->next;
        if ( p ) printf( ", " );
        }

    printf( "\n" );
    }


void
print_ill( ill *p )
    {
    if ( p->prev ) return;
    while ( p )
        {
        printf( "%d", p->n );
        p = p->next;
        if ( p ) printf( ", " );
        }

    printf( "\n" );
    }


BOOL **
initialize_bool( int r, int c )
    {
    BOOL **b;
    int i, j;

    b = (BOOL **) malloc( r * sizeof( BOOL * ) );

    if ( ! b ) bail( "Cannot allocate memory for 2D boolean array" );

    for ( i=0; i < r; i++ ) 
        {
        b[i] = (BOOL *) malloc( c * sizeof( BOOL ) ); 
        if ( ! b[i] ) bail( "Cannot allocate memory for boolean array!" );
        }

    for ( i=0; i < r; i++ )
        for ( j=0; j < c; j++ )
            b[i][j] = FALSE;

    return( b );
    }

void
set_all_false( BOOL **b, int r, int c )
    {
    int i, j;
    for ( i=0; i < r; i++ )
        for ( j=0; j < c; j++ )
            b[i][j] = FALSE;
    return;
    }


array_2dq *
initialize_2dq( int nr, int nc )       // 2D array of linked lists. Holy cow.
    {
    array_2dq *a;
    int r, c;

    a = (array_2dq *) malloc( sizeof( array_2dq ) );
    if ( ! a ) bail( "Unable to allocate array of queues" );

    a->len.x = nc;
    a->len.y = nr;

    a->q = (queue ***) malloc( nr * sizeof( queue ** ) );
    if ( ! a->q ) bail( "Unable to allocate array of queues (y)" );

    for ( r=0; r < nr; r++ )
        {
        a->q[r] = (queue **) malloc( nc * sizeof( queue * ) );
        if ( ! a->q[r] ) bail( "Unable to allocate array of queues (x)" );
        }

    for ( r=0; r < nr; r++ )
        for ( c=0; c < nc; c++ )
            a->q[r][c] = initialize_queue();

    return a;           
    }
    

queue *
get_field_list( array_2dq *a, int x, int y )
    {
    if ( x < 0 || x >= a->len.x || y < 0 || y >= a->len.y )
        {
        printf( "ERROR: in get_field_list( x, y )\n" ); 
        printf( "x = %d is not in the range (0, %ld)  and/or\n", x, a->len.x );
        printf( "y = %d is not in the range (0, %ld)\n", y, a->len.y );
        bail( "(polyq) Illegal index in field list" );
        }
    
    return a->q[y][x];
    }

queue *
get_field_list_db( array_2dq *a, int x, int y, int location )
    {
    if ( x < 0 || x >= a->len.x || y < 0 || y >= a->len.y )
        {
        printf( "\n\nx: %d >= %ld  y: %d >= %ld [location %d]", x, a->len.x, y, a->len.y, location ); 
        printf( "\nWARNING (polyq) Illegal index in field list\n" );
	if ( x < 0 ) x = 0;
	if ( y < 0 ) y = 0;
	if ( x >= a->len.x ); x = a->len.x - 1;
	if ( y >= a->len.y ); y = a->len.y - 1;
        }
    
    return a->q[y][x];
    }

polygon_pair *
initialize_polygon_pair()
    {
    polygon_pair *pp;
    pp = (polygon_pair *) malloc( sizeof( polygon_pair ) );
    pp->a = (polygon *) NULL;
    pp->b = (polygon *) NULL;
    return( pp );
    }


BOOL
polygon_ccw( polygon *p, double *area, int bss )
    {
    static int    n;
    static double c, toosmall, s, dx, dy, miny;
    static int    i, j;
    static xy     vi, vj;

    toosmall = bss * bss; // yikes

    n = p->len;
    c = 0.0;

    for ( i=0; i < n; i++ )
        {
        j = (i + 1) % n;
        vi = vertex( p, i );
        vj = vertex( p, j );
	vi.x = vi.x - p->ll.x;    
	vi.y = vi.y - p->ll.y;
        vj.x = vj.x - p->ll.x;
	vj.y = vj.y - p->ll.y;
	s = (vj.x > vi.x) ? 1 : -1;   
        miny = (vj.y < vi.y) ? vj.y : vi.y;
        dx = abs( vj.x - vi.x );
        dy = abs( vj.y - vi.y );
	c = c + s * ( miny * dx + (dx * dy) / 2.0 );
        }

    if ( fabs(c) < toosmall )
	*area = 0;
    else
        *area = fabs( c );
    
    return( c > 0.0 );
    }



double
polygon_area( polygon *p, int bss )
    {
    static double area;
    polygon_ccw( p, &area, bss );
    return( area );
    }


void 
delete_queue_member( queue *q, polygon_ref *pr )
    {
    q->len = q->len - 1;
    if(  pr->prev )
        (pr->prev)->next = pr->next;
    else
	q->first = pr->next;
    }

void
set_extents( polygon *p )
    {
    static int minx, maxx, miny, maxy;
    static xy v;
    static int i;
    if ( ! p->active ) return;
    if ( p->len == 0 ) return;
    v = vertex( p, 0 );
    minx = v.x; 
    maxx = v.x;
    miny = v.y;
    maxy = v.y;
    for ( i=1; i < p->len; i++ )
	{
	v = vertex( p, i );
	if ( v.x < minx ) minx = v.x;
	if ( v.x > maxx ) maxx = v.x;
	if ( v.y < miny ) miny = v.y;
	if ( v.y > maxy ) maxy = v.y;
	}
    p->ll.x = minx;
    p->ll.y = miny;
    p->ur.x = maxx;
    p->ur.y = maxy;
    }

void
free_polygon( polygon *p )
    {
    //free( p->hit );  something is
    //free( p->hit2 ); really wrong
    //free( p->v );    with c
    //free( p );
    }

polygon_array *
initialize_polygon_array( int np, int n )
    {                               
    // np = number of polygons
    // n  = maximum number of vertices in each polygon

    static int i;
    static polygon_array *pa;

    pa = (polygon_array *) malloc( sizeof( polygon_array ) );
    if ( ! pa ) bail( "unable to allocate pointer to polygon array" );
    pa->len = np;
    pa->par = (polygon **) malloc( np * sizeof( polygon * ) );
    if ( ! pa->par ) bail( "unable to allocate polygon array" );

    for ( i=0; i < np; i++ )
        pa->par[i] = initialize_polygon( n, 0, 0, 0 );

    if ( ! pa->par[np-1] ) bail( "unable to allocate polygons in an array" );
    
    return( pa );
    }
