// splitter.c
//                 Split a pattern file into n parts, then
//                 generate n scripts that convert each part
//                 with darn, convexify, and slice.

//                 Now checks the pattern extent, making sure
//                 it is less than MAXEXTENT = 2 meters.
//                 What if the gds file uses angstrom units?
//                 The data is not rescaled until it gets to darn.c,
//                 so we need to consider here the dataunit and scale MAXEXTENT.


#include "libgdsmin.h"
#include "polyq.h"
#include <string.h>
#include <stdlib.h>

#define MAXSUBFIELD 4.0
#define CENTER_AT_ZERO 0

#define MAXEXTENT 2e9   // database units

#define VERSION 4.5

// 3.2  compiled with long int version of polyq. Nothing changed in this file.

// 4.0  No longer prints out a sequence file, and no longer uses display_progress.
//      We also do not need a launcher.sh because piece_*.sh will be lauched with fracture_feeder.py. 

// 4.1  bss is now on the end of the command line, and the name 'pss' is used properly.
//      We need bss in case the pattern is rotated. Rotated patterns must be snapped to the bss grid.

// 4.2  The 'traps' option will force convexify to generate triangles and parallelograms.

// 4.3  Slice might create > 4 sides, so we must run the pattern through convexify again
//      if trapezia are required. 

// 4.4  When generating traps, use bss in darn, not 1.0

// 4.5  Fixed problem with auto-extents

//---------------------------------------------------------------------------------------------------------

void
count_polygons( char *ifilename, double *dbu_uu, double *dbu_m, double *dbu_rescale, int *numpolys )   
    {
    // and also return the database unit

    double
        dbu_um;

    int i, fd;

    long layer;

    unsigned short
        count,           
        token;           

    FILE
        *thetotal;

    thetotal = fopen( "tmp/darntotal", "w" );
    if ( ! thetotal ) bail( "Unable to open/write darntotal file" );

    struct gds_itemtype *item;

    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    
    fd = open( ifilename, 0 );
    if ( fd <= 0 ) bail( "unable to open input file" );

    *numpolys = 0;
    layer = -1;

    while ( read( fd, &count, 2 ) )
        {
        gds_swap2bytes( (BYTE *) &count );

        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );
            
            if ( token == 0x0305 )  // UNITS
                {
                gds_read_units( fd, &dbu_um, dbu_uu, dbu_m, 0 );
                *dbu_rescale = 1.0e-9 / *dbu_m ; 
                }
            else if ( token == 0x0D02 )  // LAYER
                {
                gds_read_layer( fd, count, &item, 0 );
                layer = item->layer;
                }
            else if ( token == 0x1003 )  // XY
                {
                SKIPOVER( fd, count );
                *numpolys = *numpolys + 1;
                }
            else
                {
                SKIPOVER( fd, count );
                }
            }
        }
    fprintf( thetotal, "%d\n", *numpolys * 2 );  // * 2 because darn needs two passes. so... not accurate
    fclose( thetotal );
    } 

//-----------------end count_polygons-----------------------------------------------------------------------


void write_queue( int fdo, struct gds_itemtype *list ) 
    {
    static struct gds_itemtype *item;

    item = list;
    while ( item )
        {
        gds_write_boundary( fdo );
        gds_write_layer(    fdo, item->layer );
        gds_write_datatype( fdo, item->dt    );
        gds_write_xy(       fdo, item->x, item->y, item->n );
        gds_write_endel(    fdo );      
        item = item->nextitem;
        }

    // note that this function does not close the gds structure 
    } // write_queue

//----------------------------------------------------------------------------------------------------------

double                                            
subfield_size( double block, double bss_nm, double lsb )
    { 
    static char command[256];
    static FILE *fp;
    static double subfield;

    sprintf( command, "subfield_size %1.2lf %1.2lf %1.6lf > tmp/subfield.txt", block, bss_nm, lsb );
    system( command );
    fp = fopen( "tmp/subfield.txt", "r" );
    fscanf( fp, "%lf", &subfield );
    fclose( fp );

    if ( subfield < 0.0 ) 
        {
        printf( "block = %1.2lf um\n", block );
        printf( "bss   = %1.2lf nm\n", bss_nm );
        printf( "lsb   = %1.2lf nm\n", lsb );
	system( "cat tmp/subfield.txt" );
	bail( "unable to find an appropriate subfield size" );
        }

    return( subfield ); 
    }

//----------------------------------------------------------------------------------------------------------

void
check_pattern_extent( struct gds_itemtype *item, dxy *ll, dxy *ur )
    {
    static BOOL init=FALSE;
    static int i;

    if ( ! init ) 
        {
        init = TRUE;
        ll->x = item->x[0];
        ll->y = item->y[0];
        ur->x = ll->x;
        ur->y = ll->y;
        }

    for ( i=0; i < item->n; i++ )
        {
        if ( item->x[i] > ur->x ) ur->x = item->x[i];
        if ( item->y[i] > ur->y ) ur->y = item->y[i];
        if ( item->x[i] < ll->x ) ll->x = item->x[i];
        if ( item->y[i] < ll->y ) ll->y = item->y[i];
        }

    return;
    }
  
//----------------------------------------------------------------------------------------------------------

void
print_sequence( FILE *fp, int *sequence, char *message )
    {
    // The script will create an empty file tmp/step_nnn before each major step.
    // The program display_progress will count the step_* files to show our progress.

    //static char seqname[40];    
    //sprintf( seqname, "tmp/step_%d", *sequence );
    fprintf( fp, "if [ -f fullstop.txt ]; then rm -f tmp/wait_for_piece_* ; echo \"%s\" ; exit; fi \n", message );
    //fprintf( fp, "touch %s \n", seqname );
    //fprintf( fp, "display_progress \n" );
    //(*sequence)++;
    }

//==========================================================================================================

int main( int argc, char *argv[] )
    {
    FILE 
	 *piece,
	 *extent,
         *fp;

    char *ifilename,
         *ofilename,
         *tfilename,
	 *command, 
  	 *waitname,
         *piecename;

    double
	  lsb,
 	  maxextent,
	  dx, dy,
          xblock,
          yblock, 
  	  angle,
          xsub,
          ysub,
          s,
	  llx,
	  lly,
	  urx,
          ury,
          crllx,
          crlly,
          crurx,
          crury,
          progress,
          bss_nm,
          pss_nm,
          dbu_uu,
          dbu_nm,
          dbu_m,
          dbu_um,
 	  dbu_rescale,
          maxx,
          maxy;

    int   layer,
          dt,
	  sequence,
          fd, 
          fdo,
          chunk,
          nchunks=0,
          nom,
          n,
          numpolys,
          i,
	  part, 
          CHUNKS=200;

    long  
          len;

    BOOL  first=TRUE,
	  first_piece=TRUE,
          not_eof,
          noarray,
          nosplit,
          traps;

    char
          *layers;

    unsigned short 
          count,           
          token;           

    struct gds_itemtype 
          *top,
          *previous,
          *item;

    polygon 
          *p,
          *poly;

    polygon_ref 
          *pr;

    queue
          *list;

    dxy   ll, ur;


    if ( argc == 2 && strcmp( argv[1], "-v" ) == 0 )
        {
        printf( "splitter v%1.1f\n", VERSION );
        exit(0);
        }

    command   = (char *) malloc( 256 );
    ofilename = (char *) malloc( 256 );
    tfilename = (char *) malloc( 256 );
    piecename = (char *) malloc( 256 );
    waitname  = (char *) malloc( 256 );

    //printf( "\nsplitter\n" );

    if ( argc < 2 )
        {
        printf( "\nsplitter v%1.1f\n", VERSION );
        printf( "\nUse: splitter infile.gds [pss_nm] [layers] [xblock] [yblock] \n" );
        printf(   "     [llx] [lly] [urx] [ury] [crllx] [crlly] [crurx] [crury] [xsub] [ysub] [angle] \n" );
        printf(   "     [nosplit] [mf lsb um] [#splits] [noarray] [bss_nm] [traps] \n\n" );
        printf(   "     Split a gds file into n parts, then generate shells which run each part \n" );
        printf(   "     through 'darn', 'convexify', and 'slice', in parallel. \n" );
        printf(   "     The 'traps' option forces convexify to generate trapezia instead of polygons. \n" );
        printf(   "     The pss and bss default to 1nm. Layers default to 'all'.\n" );
        printf(   "     Layer numbers should be separated by commas, without spaces.\n" );
        printf(   "     LL-UR is the bounding box for field slicing (um). \n" );
        printf(   "     CRLL-CRUR is the optional cropping box (um). Optionally use the\n" );
        printf(   "     string 'nosplit' to prevent parallel execution." );
        printf(   "     Rotation angle is in degrees.\n\n" );
        exit(0);
        }

    ifilename = argv[1]; 

    if ( argc > 2 )
        sscanf( argv[2], "%lf", &pss_nm ); 
    else
        pss_nm = 1.0;

    if ( argc > 3 )
        layers = argv[3];
    else
        {
        layers = "-1";
        }

    xblock = 0;
    yblock = 0;
    xsub   = 0;
    ysub   = 0;
    llx    = 0;
    lly    = 0;
    urx    = 0;
    ury    = 0;
    crllx  = 0;
    crlly  = 0;
    crurx  = 0;
    crury  = 0;

    if ( argc > 13 )
        {
        sscanf( argv[4],  "%lf", &xblock );
        sscanf( argv[5],  "%lf", &yblock );
        sscanf( argv[6],  "%lf", &llx );
        sscanf( argv[7],  "%lf", &lly );
        sscanf( argv[8],  "%lf", &urx );
        sscanf( argv[9],  "%lf", &ury );
        sscanf( argv[10], "%lf", &crllx );
        sscanf( argv[11], "%lf", &crlly );
        sscanf( argv[12], "%lf", &crurx );
        sscanf( argv[13], "%lf", &crury );
        }

    if ( argc > 15 )
	{
        sscanf( argv[14], "%lf", &xsub );
        sscanf( argv[15], "%lf", &ysub );
        //printf( "subfield = %1.3lf, %1.3lf\n", xsub, ysub );
	}

    angle = 0;
    if ( argc > 16 ) 
        {
        sscanf( argv[16], "%lf", &angle );
        }
    //printf( "\nangle = %1.3lf\n", angle );

    nosplit = FALSE;
    if ( argc > 17 ) nosplit = ( strcmp( argv[17], "nosplit" ) == 0 );

    lsb = 0.001;
    if ( argc > 18 ) sscanf( argv[18], "%lf", &lsb );  // um        

    if ( argc > 19 ) sscanf( argv[19], "%d", &nchunks );

    if ( argc > 20 ) noarray = (strcmp( argv[20], "noarray" ) == 0 );

    bss_nm = 1.0;
    
    if ( argc > 21 ) sscanf( argv[21], "%lf", &bss_nm );

    traps = FALSE;    // default is polygon
    
    if ( argc > 22 ) traps = ( strcmp( argv[22], "traps" ) == 0 );



    
    if ( xsub == 0.0 || ysub == 0.0 )    // not likely
	{
        xsub = subfield_size( xblock, pss_nm, lsb );
        ysub = subfield_size( yblock, pss_nm, lsb );
        //printf( "\nsubfield size: %1.3lf, %1.3lf\n", xsub, ysub );
	}

    piece = fopen( "tmp/piece_0.sh", "w" );        
    if ( ! piece ) bail( "Unable to create piece_0.sh" );


    printf( "\nInput file:     %s\n", ifilename );
    printf(   "Beam step, nm:  %1.1lf\n\n", bss_nm );

    count_polygons( ifilename, &dbu_uu, &dbu_m, &dbu_rescale, &numpolys );

    dbu_um = dbu_m * 1.0e6; 
    dbu_uu = dbu_m * 1.0e6;

    printf( "Number of polygons:  %d\n", numpolys );

    if ( numpolys <= 0 ) bail( "NO INPUT DATA" );

    fd = open( ifilename, 0 );
    if ( fd <= 0 ) bail( "unable to open input file" );

    item = (struct gds_itemtype *) malloc( sizeof( struct gds_itemtype ) );
    if ( ! item ) bail( "unable to allocate gds item" );
    item->n        = 0;
    item->layer    = 0; 
    item->dt       = 0;
    item->nextitem = NULL;
    top = item;


    sprintf( ofilename, "part_0.gds" );
    sprintf( tfilename, "tmp/part_0.gds" );
    fdo = open( tfilename, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
    if ( fdo <= 0 ) ( "unable to open output gds file" );

    gds_create_lib_correctly( fdo, "pcored", dbu_um, dbu_uu );
    gds_write_bgnstr( fdo );
    gds_write_strname( fdo, "pcored" );


    // Main polygon reading loop-----------------------------------


    if ( nosplit ) printf( "\n----------- No pattern splitting.\n" );

    if ( nosplit )
        chunk = numpolys;
    else 
        {
	if ( numpolys > 1000 * CHUNKS )
	    chunk = numpolys / 200;
        else if ( numpolys > 100 * CHUNKS )
            chunk = numpolys / (CHUNKS-1);
        else if ( numpolys > 64 * CHUNKS )
            chunk = numpolys / 64;
        else if ( numpolys > CHUNKS )
            chunk = numpolys / 10;
        else if ( numpolys > 64 )
            chunk = numpolys / 5;
        else
            chunk = numpolys;
        }

    if ( nchunks == 0 )
	{
        if ( ! (numpolys % chunk) )
            nchunks = numpolys / chunk;
        else
            nchunks = 1 + numpolys / chunk;
        }
    else
	{
	chunk = numpolys / nchunks;
	}

    if ( chunk > 0 )
	{
        printf( "\nSplitting the pattern into %d files of ~ %d shapes each...\n", nchunks, chunk );
        printf( "\n\010   Creating chunk 0 \015" );
        fflush( stdout );
	}

    sequence = 0;

    list = initialize_queue();    
    n = 0;
    nom = 0;
    not_eof = read( fd, &count, 2 ) > 0;

    if ( angle > 0.0 ) pss_nm = bss_nm;  // because if we rotate the pattern, it must be snapped back to the bss grid.
                                         // well, I don't think that's true anymore. we should test it.

    while ( not_eof )
        {
        gds_swap2bytes( (BYTE *) &count );
        if ( count >= 4 )
            {
            read( fd, &token, 2 );
            gds_swap2bytes( (BYTE *) &token );

            switch( token )
                {
                case 0x2B02:  // printf( "\nPROPATTR means obese \n" ); => obese
                          SKIPOVER( fd, count );
                          break;
                case 0x0D02:  //printf( "LAYER \n" ) ;
                          gds_make_next_item( &item );
                          gds_read_layer( fd, count, &item, 0 );
                          layer = item->layer;
                          break;
                case 0x0E02:  //printf( "DATATYPE \n" ) ;
                          gds_read_datatype( fd, count, &item, 0 );
                          dt = item->dt;
                          break;
                case 0x1003:  //printf( "XY \n" ) ;
                          progress = 100 * n / numpolys;

                          if ( first )
                              {
                              first = FALSE;
                              item = top;
                              item->layer = layer;
                              item->dt = dt;
                              gds_read_xy( fd, count, 0, &item, 0 );
                              item->nextitem = NULL;
                              }
                          else
                              {
                              gds_read_xy( fd, count, 0, &item, 0 );
                              }

                          n++;
                          check_pattern_extent( item, &ll, &ur );

                          if ( n == chunk )
                              {
    	                      write_queue( fdo, top );
                              gds_write_endstr( fdo ); 
                              gds_write_endlib( fdo );    
                              close( fdo );
                              first = TRUE;
			      sprintf( waitname, "tmp/wait_for_piece_%d", nom );
                              
                              fprintf( piece, "#!/bin/bash\n" );
                              fprintf( piece, "touch %s\n", waitname );
                              print_sequence( piece, &sequence, "failure in splitter" );
                              fprintf( piece, "darn tmp/%s %1.3f %s tmp/precored_%s %1.3lf \n", ofilename, pss_nm, layers, ofilename, angle  );
                              print_sequence( piece, &sequence, "failure in first darn" );
                              if ( traps )
                                  {
                                  fprintf( piece, "darn tmp/precored_%s %1.3f %s tmp/cored_%s 0.0 \n", ofilename, bss_nm, layers, ofilename  );
                                  print_sequence( piece, &sequence, "failure in second darn" );
                                  fprintf( piece, "grid_snap tmp/cored_%s tmp/snapped_%s %1.1lf fix\n", ofilename, ofilename, bss_nm );
                                  fprintf( piece, "convexify tmp/snapped_%s tmp/trapped_%s %1.1lf header traps \n", ofilename, ofilename, pss_nm );
                                  }
                              else
                                  {
                                  fprintf( piece, "darn tmp/precored_%s %1.3f %s tmp/cored_%s 0.0 \n", ofilename, pss_nm, layers, ofilename  );
                                  print_sequence( piece, &sequence, "failure in second darn" );
                                  fprintf( piece, "convexify tmp/cored_%s tmp/trapped_%s %1.1lf header notraps \n", ofilename, ofilename, pss_nm );
                                  }
                              print_sequence( piece, &sequence, "failure in convexify" );

                              if ( first_piece )
				  {
				  first_piece = FALSE;
                                  fprintf( piece, "slice tmp/trapped_%s tmp/sliced_%s ", ofilename, ofilename );
                                  fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf ", llx, lly, urx, ury );
                                  fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf %1.3lf ", xblock, yblock, xsub, ysub, pss_nm ); 
                                  fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf ", crllx, crlly, crurx, crury );
                                  if ( noarray ) fprintf( piece, "yesheader noarray" );
                                  fprintf( piece, "\n" );
                                  if ( traps )
                                      {
                                      fprintf( piece, "convexify tmp/sliced_%s tmp/temp_%s %1.1lf header traps\n",
                                                        ofilename, ofilename, pss_nm );
                                      fprintf( piece, "mv tmp/temp_%s tmp/sliced_%s\n", ofilename, ofilename );
                                      }
				  }
			      else
				  {
                                  fprintf( piece, "slice tmp/trapped_%s tmp/sliced_%s ", ofilename, ofilename );
                                  fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf ", llx, lly, urx, ury );
                                  fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf %1.3lf ", xblock, yblock, xsub, ysub, pss_nm );
                                  if ( traps )             // then slice produces a header, which is stripped off with convexify
                                      {
                                      fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf header ", crllx, crlly, crurx, crury ); 
                                      if ( noarray ) fprintf( piece, "noarray" );
                                      fprintf( piece, "\n" );
                                      fprintf( piece, "convexify tmp/sliced_%s tmp/temp_%s %1.1lf noheader traps\n",
                                                        ofilename, ofilename, pss_nm );
                                      fprintf( piece, "mv tmp/temp_%s tmp/sliced_%s\n", ofilename, ofilename );
                                      }
                                  else
                                      {
                                      fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf noheader ", crllx, crlly, crurx, crury ); 
                                      if ( noarray ) fprintf( piece, "noarray" );
                                      fprintf( piece, "\n" );
                                      }
				  }

                              print_sequence( piece, &sequence, "failure in slice" );

                              fprintf( piece, "rm -f %s\n", waitname );

                              n = 0;
                              nom++;
                              printf( "\010   Creating chunk %d \015", nom );
                              fflush( stdout );
                              sprintf( tfilename, "tmp/part_%d.gds", nom );
                              sprintf( ofilename, "part_%d.gds", nom );
                              fdo = open( tfilename, O_CREAT | O_TRUNC | O_WRONLY, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
                              if ( fdo <= 0 ) ( "unable to open output gds file tmp/part_%d.gds", nom );
                              gds_create_lib_correctly( fdo, "cored", dbu_um, dbu_uu );
                              gds_write_bgnstr( fdo );
                              gds_write_strname( fdo, "cored" );

                              fclose( piece );
			      sprintf( piecename, "tmp/piece_%1d.sh", nom );
			      piece = fopen( piecename, "w" );
			      if ( ! piece ) bail( "splitter: unable to create script file" );
                              }
                              
                          break;
                default:      //
                          SKIPOVER( fd, count );
                }    
            }      
        not_eof = read( fd, &count, 2 ) > 0;
        }          

    
    if ( n > 0 )
        {
        write_queue( fdo, top );
        gds_write_endstr( fdo ); 
        gds_write_endlib( fdo );    
        close( fdo );
        sprintf( waitname, "tmp/wait_for_piece_%d", nom );
        fprintf( piece, "#!/bin/bash\n" );
        fprintf( piece, "touch %s\n", waitname );
        print_sequence( piece, &sequence, "fail in splitter" );
        fprintf( piece, "darn tmp/%s %1.3f %s tmp/precored_%s %1.3lf \n", ofilename, pss_nm, layers, ofilename, angle );
        print_sequence( piece, &sequence, "fail in first darn" );
        if ( traps )
            {
            fprintf( piece, "darn tmp/precored_%s %1.3f %s tmp/cored_%s 0.0 \n", ofilename, bss_nm, layers, ofilename );
            print_sequence( piece, &sequence, "fail in second darn" );
            fprintf( piece, "grid_snap tmp/cored_%s tmp/snapped_%s %1.1lf fix\n", ofilename, ofilename, bss_nm );
            fprintf( piece, "convexify tmp/snapped_%s tmp/trapped_%s %1.1lf header traps \n", ofilename, ofilename, pss_nm );
            }
        else
            {
            fprintf( piece, "darn tmp/precored_%s %1.3f %s tmp/cored_%s 0.0 \n", ofilename, pss_nm, layers, ofilename );
            print_sequence( piece, &sequence, "fail in second darn" );
            fprintf( piece, "convexify tmp/cored_%s tmp/trapped_%s %1.1lf header notraps \n", ofilename, ofilename, pss_nm );
            }
            
        print_sequence( piece, &sequence, "fail in convexify" );

        if ( first_piece )
            {
	    first_piece = FALSE;
            fprintf( piece, "slice tmp/trapped_%s tmp/sliced_%s ", ofilename, ofilename );
            fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf ", llx, lly, urx, ury );
            fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf %1.3lf ", xblock, yblock, xsub, ysub, pss_nm ); 
            fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf ", crllx, crlly, crurx, crury ); 
            if ( noarray ) fprintf( piece, "yesheader noarray" );
            fprintf( piece, "\n" );
            if ( traps )
                {
                fprintf( piece, "convexify tmp/sliced_%s tmp/temp_%s %1.1lf header traps\n",
                         ofilename, ofilename, pss_nm );
                fprintf( piece, "mv tmp/temp_%s tmp/sliced_%s\n", ofilename, ofilename );
                }
            
	    }
        else
	    {
            fprintf( piece, "slice tmp/trapped_%s tmp/sliced_%s ", ofilename, ofilename );
            fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf ", llx, lly, urx, ury );
            fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf %1.3lf ", xblock, yblock, xsub, ysub, pss_nm );
            
            if ( traps )  // then slice spits out a header, and convexify strips the header
                {
                fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf header ", crllx, crlly, crurx, crury ); 
                if ( noarray ) fprintf( piece, "noarray" );
                fprintf( piece, "\n" );
                fprintf( piece, "convexify tmp/sliced_%s tmp/temp_%s %1.1lf noheader traps\n",
                         ofilename, ofilename, pss_nm );
                fprintf( piece, "mv tmp/temp_%s tmp/sliced_%s\n", ofilename, ofilename );
                }
            else
                {
                fprintf( piece, "%1.3lf %1.3lf %1.3lf %1.3lf noheader ", crllx, crlly, crurx, crury ); 
                if ( noarray ) fprintf( piece, "noarray" );
                fprintf( piece, "\n" );
                }
	    }

        print_sequence( piece, &sequence, "fail in slice" );

        fprintf( piece, "rm -f %s\n", waitname );

        }
    else
        {
        gds_write_endstr( fdo ); 
        gds_write_endlib( fdo );    
        close( fdo );
        sprintf( command, "rm -f %s", tfilename );
        system( command );
        }        

    printf( "\n\n" );
    close( fd );
    fclose( piece );


    if ( llx == 0 && lly == 0 && urx == 0 && ury == 0 )
        {
        dx = abs( ll.x - ur.x );
        dy = abs( ll.y - ur.y );
        
        ll.x = (ll.x - 0.2 * dx) * dbu_um;
        ur.x = (ur.x + 0.2 * dx) * dbu_um;
        ll.y = (ll.y - 0.2 * dy) * dbu_um;
        ur.y = (ur.y + 0.2 * dy) * dbu_um;

        // snap the chip extents to the block size grid,
	// so we are less likely to generate out-of-range shapes
        
        dx = fabs( ll.x - ur.x );
        dy = fabs( ll.y - ur.y );

        if ( dx < xblock )
	    {
	    ll.x = (ll.x + ur.x)/2.0 - xblock/2.0;
	    ur.x = ll.x + xblock;
	    }
	else
	    {
            ll.x = (round( ll.x / xblock ) - 1.0) * xblock;
	    ur.x = ll.x + xblock * (1.0 + round( fabs(ur.x - ll.x) / xblock ));
	    }

        if ( dy < yblock )
	    {
	    ll.y = (ll.y + ur.y)/2.0 - yblock/2.0;
	    ur.y = ll.y + yblock;
	    }
	else
	    {
            ll.y = (round( ll.y / yblock ) - 1.0) * yblock;
            ur.y = ll.y + yblock * (1.0 + round( fabs(ur.y - ll.y) / yblock ));
	    }

        if ( CENTER_AT_ZERO )
            {
            maxx = fabs( ur.x );
            maxy = fabs( ur.y );
            if ( fabs(ll.x) > maxx ) maxx = fabs(ll.x);
            if ( fabs(ll.y) > maxy ) maxy = fabs(ll.y);
            if ( maxx < xblock/2.0 ) maxx = xblock/2.0;
	    if ( maxy < yblock/2.0 ) maxy = yblock/2.0;
            ll.x = -maxx;
            ur.x =  maxx;
            ll.y = -maxy;
            ur.y =  maxy;
            }
        }
    else
	{
        //ll.x = ll.x * dbu_um;
        //ur.x = ur.x * dbu_um;
        //ll.y = ll.y * dbu_um;
        //ur.y = ur.y * dbu_um;
        ll.x = llx;
        ur.x = urx;
        ll.y = lly;
        ur.y = ury;
	}

    extent = fopen( "tmp/extents.dat", "w" );
    if ( ! extent ) bail( "Unable to open tmp/extents.dat" );
    fprintf( extent, "%1.3lf %1.3lf %1.3lf %1.3lf\n", ll.x, ll.y, ur.x, ur.y );
    fclose( extent );
    
    llx = ll.x;
    lly = ll.y;
    urx = ur.x;
    ury = ur.y;
    dx = fabs( llx - urx );
    dy = fabs( lly - ury );

    maxextent = MAXEXTENT * dbu_rescale;   

    if ( dx > maxextent || dy > maxextent || isnan( dx ) || isnan( dy ) ) 
        {
        printf( "\nERROR: pattern extent is too large. \n" );
        printf(   "       dbu_rescale = %g   is usually 1.0 \n",  dbu_rescale );
	printf(   "       dbu_m       = %g   is usually 1e-9 \n", dbu_m );
        dx = 1e-6 * dx / dbu_rescale;
        dy = 1e-6 * dy / dbu_rescale;
        printf(   "       dx = %g mm   dy = %g mm \n", dx, dy );
        bail( "Pattern extent is too large." );
        }
    else
	{
        printf( "\ndbu_rescale = %g   dbu_m = %g\n", dbu_rescale, dbu_m );
        printf( "Pattern extent:  %g, %g um\n", dx, dy );
	}

    system( "chmod a+x tmp/piece_*.sh" );

    fp = fopen( "tmp/step_total", "w" );
    fprintf( fp, "%d\n", sequence );
    fclose( fp );

    //printf( "\nsplitter done.\n" );
    }


