#!/bin/bash
#               Freebeam update script v3.0
#               No longer requires the Beams libraries
#               and so this can be used on any RHEL system.


echo
echo
echo "Update Freebeam..."
echo


if ! command -v gpfout > /dev/null ; then
    echo
    echo "It looks like Freebeam is not installed on this system."
    echo "You should use the install.sh script instead."
    echo "Update failed."
    echo
    exit
fi


g=`which gpfout`
path=${g:0:-7}
echo "Installation location: $path"
echo

# do we have write permission here?

touch $path/write_test.123

if [ ! -f $path/write_test.123 ]; then
    echo
    echo "You do not have write permission in $path"
    echo "Try becoming root with 'su' first."
    echo "Update failed."
    echo
    exit
fi

rm -f $path/write_test.123

# if this is an ebpg then pg_local contains local definitions
# such as the ld_library_path. sourcing this is probably not necessary.

pushd . > /dev/null
if [ -f /home/pg/pg_local ]; then
    . /home/pg/pg_local
fi
popd > /dev/null


echo
echo "Checking the library path... "

if [ `echo $LD_LIBRARY_PATH | grep -c $path/lib` != "1" ]; then
    echo
    echo "        PROBLEM:"
    echo "        LD_LIBRARY_PATH does not include $path/lib"
    echo "        which is weird because this had to be right for the"
    echo "        initial installation of Freebeam. I suggest that"
    echo "        you go through the full installation."
    echo
    echo "        Freebeam update failed."
    echo
    exit
else
    echo "        $path/lib ok"
fi
echo


echo "Checking the command path... "

if [ `echo $PATH | grep -c $path` != "1" ]; then
    echo
    echo "        PROBLEM:"
    echo "        The command PATH does not include $path"
    echo "        which is weird because this had to be right for the"
    echo "        initial installation of Freebeam. I suggest that you"
    echo "        go through the complete installation using install.sh"
    echo
    echo "        Freebeam update failed."
    echo
    exit
else
    echo "        $path is in the PATH... "
fi

echo
echo "Pulling the version file..."
echo

rm -f latest_freebeam_version

wget https://github.com/ebpgfunpak/freebeam/blob/main/version?raw=true -O latest_freebeam_version

read version < latest_freebeam_version
rm -f latest_freebeam_version

if [ "$version" == "" ]; then
    echo
    echo "Unable to connect to github.com to check for updates."
    echo "Freebeam update failed."
    echo
    exit
fi


pushd . > /dev/null
cd $path/..
rm -rf old_freebeam
mv freebeam old_freebeam

echo
echo "Downloading the new Freebeam version..."
echo

wget https://github.com/ebpgfunpak/freebeam/blob/main/freebeam_$version.tgz?raw=true -O freebeam_$version.tgz

echo
echo "The previous installation has been moved to $path/old_freebeam"
echo
echo "Unpacking the updates..."
echo

tar xvzf freebeam_$version.tgz

if [ ! -f freebeam/install.sh ]; then
    echo
    echo "        PROBLEM:"
    echo "        The archive (tar) file did not unpack properly."
    echo "        I suggest that you delete this mess with"
    echo "        rm -rf freebeam"
    echo "        then restore the previous version with"
    echo "        mv old_freebeam freebeam"
    echo
    echo "        Freebeam update failed."
    echo
    exit
fi

cd freebeam

echo "Compiling C programs..."
echo
echo
echo "-----------------------------------------------------------"

make clean
make
make install

echo "-----------------------------------------------------------"
echo

unset fail

if [ ! -f lib/libgds.a    ]; then echo "lib/libgds.c    failed to compile" ; fail=1 ; fi
if [ ! -f lib/libgdsmin.a ]; then echo "lib/libgdsmin.c failed to compile" ; fail=1 ; fi
if [ ! -f lib/libgds.o    ]; then echo "lib/libgdsmin.c failed to compile" ; fail=1 ; fi
if [ ! -f convexify       ]; then echo "convexify.c     failed to compile" ; fail=1 ; fi
if [ ! -f flatten         ]; then echo "flatten.c       failed to compile" ; fail=1 ; fi
if [ ! -f gpoly2gds       ]; then echo "gpoly2gds.c     failed to compile" ; fail=1 ; fi
if [ ! -f gpoly2txt       ]; then echo "gpoly2txt.c     failed to compile" ; fail=1 ; fi
if [ ! -f slice           ]; then echo "slice.c         failed to compile" ; fail=1 ; fi
if [ ! -f grid_snap       ]; then echo "grid_snap.c     failed to compile" ; fail=1 ; fi
if [ ! -f splitter        ]; then echo "splitter.c      failed to compile" ; fail=1 ; fi
if [ ! -f darn            ]; then echo "darn.c          failed to compile" ; fail=1 ; fi
if [ ! -f gpfout          ]; then echo "gpfout.c        failed to compile" ; fail=1 ; fi

if [ $fail ]; then
    echo
    echo "Compilation was not successful."
    echo "I suggest that you delete this mess with"
    echo "    cd /home/pg/bin"
    echo "    rm -rf freebeam"
    echo "then restore the previous version with"
    echo "    mv old_freebeam freebeam"
    echo "Freebeam update failed."
    echo
    exit
else
    echo
    echo "Compilation was successful. Hooray!"
    echo
fi

mail -s "freebeam updated" ebpgfunpak@gmail.com < /dev/null

echo
echo "Freebeam update complete. Version $version installed."
echo
echo
echo "Now I will test Freebeam by running freebeam_example.sh"
echo
echo "You could do the same thing with..."
echo
echo "        cd "
echo "        cd bin/freebeam"
echo "        freebeam_example.sh"
echo
echo "        If successful, this will end by using cview"
echo "        to display a pattern in the shape of the letters \"Yale\""
echo 
echo "Please report bugs to ebpgfunpak@gmail.com"
echo

read -p "Are you ready to proceed with the test? (y/n) " answer

if [ $answer != 'y' ]; then
    echo
    echo "Very well, please test Freebeam yourself."
    echo
else
    cd /home/pg/bin/freebeam
    freebeam_example.sh

    echo
    echo "*******************************************************"
    echo "If cview displayed \"Yale\" then the test was successful."
    echo "*******************************************************"
    echo
fi

popd > /dev/null

echo
echo "Freebeam version $version installed."
echo
echo 
echo "You can revert to the previous version by doing this:"
echo
echo "        cd /home/pg/bin"
echo "        rm -rf freebeam"
echo "        mv old_freebeam freebeam"
echo
echo









	
